/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#pragma once

#include "stdafx.h"

#include "JpArray.h"
#include "JpRandom.h"
#include "JpUtils.h"



template< class TYPE >
class JpArray
{
	public:

		//-----------------------
		// public class declarations
		class Iterator
		{
			public:
				Iterator( JpArray< TYPE >* pArray ) 
				{
					_pArray = pArray;
					Reset();
				}

				void Reset()
				{
					_iCurrentIndex = 0;
					_pData = NULL;
				}

				bool First() 
				{ 
					_iCurrentIndex = 0;
					return Next();
				}

				bool Last() 
				{ 
					_iCurrentIndex = ( _pArray->GetNumEntries() - 1 );
					return Prev();
				}

				bool Next() 
				{
					if ( _iCurrentIndex < _pArray->GetNumEntries() )
					{
						_pData = _pArray->GetAt( _iCurrentIndex );
						_iCurrentIndex ++;
						return true;
					}

					_pData = NULL;
					return false;
				}

				bool Prev() 
				{
					if ( _iCurrentIndex >= 0 )
					{
						_pData = _pArray->GetAt( _iCurrentIndex );
						_iCurrentIndex --;
						return true;
					}

					_pData = NULL;
					return false;
				}

				TYPE* GetData() { return _pData; }

			private:

				Iterator() {} // default constructor is private so it can't be used.

				JpArray< TYPE >* _pArray;
				int _iCurrentIndex;
				TYPE* _pData;
		};


		//-----------------------
		// function declarations

		JpArray( int iArraySize )
		{
			Init( iArraySize );
		}

		JpArray( JpArray& Copy )
		{
			ASSERTF( 0, ( "\nCannot copy instances of this class" ) );
		}

		~JpArray()
		{
			Deinit();
		}

		void Init( int iArraySize );

		void Deinit();

		TYPE* GetAt( unsigned int i );
		TYPE* GetRandom( JpRandom* pRng );

		void AddLast( TYPE* p );
		TYPE* GetLast();
		TYPE* RemoveLast();

		void Clear();

		int GetMaxEntries() { return _iMaxEntries; }
		int GetNumEntries() { return _iNumEntries; }

	private:

		//-----------------------
		// member variables

		TYPE** _papData;
		int _iMaxEntries;
		int _iNumEntries;


		//-----------------------
		// member functions
		void SetAt( unsigned int i, TYPE* p );
		JpArray() {}; // can't use default constructor
};




/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
template< class TYPE > 
void JpArray< TYPE >::Init( int iArraySize )
{
	_papData = (TYPE**)malloc( iArraySize * sizeof( TYPE* ) );

	memset( _papData, 0, iArraySize * sizeof( TYPE* ) );

	_iMaxEntries = iArraySize;
	_iNumEntries = 0;
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
template< class TYPE > 
void JpArray< TYPE >::Deinit()
{
	free( _papData );
	_papData = NULL;
}



/******************************************************************************
 FUNCTION:  GetAt
******************************************************************************/
template <class TYPE> 
inline TYPE* JpArray<TYPE>::GetAt( unsigned int i )
{ 
	if ( i >= 0 && i < (unsigned int)_iMaxEntries )
		return _papData[ i ]; 

	return NULL;
}




/******************************************************************************
 FUNCTION:  SetAt
******************************************************************************/
template <class TYPE> 
inline void JpArray<TYPE>::SetAt( unsigned int i, TYPE* p )
{ 
	if ( i >= 0 && i < (unsigned int)_iMaxEntries )
		_papData[ i ] = p;
}



/******************************************************************************
 FUNCTION:  AddLast
******************************************************************************/
template <class TYPE> 
inline void JpArray<TYPE>::AddLast( TYPE* p )
{
	ASSERTF( _iNumEntries < _iMaxEntries, ( "\nNot enough space in JpArray" ) );

	if ( _iNumEntries < _iMaxEntries )
	{
		_papData[ _iNumEntries ] = p;
		_iNumEntries ++;
	}
}



/******************************************************************************
 FUNCTION:  GetLast
******************************************************************************/
template <class TYPE> 
inline TYPE* JpArray<TYPE>::GetLast()
{
	if ( _iNumEntries > 0 )
		return _papData[ _iNumEntries - 1 ];

	return NULL;
}



/******************************************************************************
 FUNCTION:  RemoveLast
******************************************************************************/
template <class TYPE> 
inline TYPE* JpArray<TYPE>::RemoveLast()
{
	if ( _iNumEntries > 0 )
	{
		TYPE* pType = _papData[ _iNumEntries - 1 ];
		_papData[ _iNumEntries - 1 ] = NULL;
		_iNumEntries --;
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  Clear
******************************************************************************/
template <class TYPE> 
inline void JpArray<TYPE>::Clear()
{
	memset( _papData, 0, _iMaxEntries * sizeof( TYPE* ) );
	 _iNumEntries = 0;
}



/******************************************************************************
 FUNCTION:  GetRandom
******************************************************************************/
template< class TYPE > 
inline TYPE* JpArray<TYPE>::GetRandom( JpRandom* pRng )
{
	if ( _iNumEntries > 0 )
	{
		int iRandomIndex = pRng->GetInt() % _iNumEntries;
		return GetAt( iRandomIndex );
	}

	return NULL;
}




