/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#ifndef __JpHashTable_h
#define __JpHashTable_h

#include "stdafx.h"

#include "JpDebug.h"
#include "JpUtils.h"

/*
	//-----------------------
	NOTE: The SIZE you pass in should be 2x the actual number of
	entires you expect the hash table to hold.  So if you want
	a hash table that keeps 1024 actor objects, then pass in
	2048 as the SIZE when the hash table is created.

	The reason for this is because when the hash table grows
	to more than 50% capacity, it begins to slow.  It's a result
	of the way this hash table was implemented - When there is
	a hash collision, the Store routine does a linear walk down
	the storage array to find the first available empty slot.
	When the storage array is less than half full, the probability
	that the next slot is available is 50% or greater.  But
	if the storage array is more than half full, this probability
	is less than 50%.

	This storage mechanism thus affects Find, Remove and other
	operations.



	//-----------------------
	// This is code I used to test the JpHashTable::Iterator class.
	// mdomowicz 2009_07_03

	char acString00[] = "This";
	char acString01[] = "Is";
	char acString02[] = "A";
	char acString03[] = "Test";

	JpHashTable< char, 128> _StringTable;
	_StringTable.Store( JpHash( acString00 ), acString00 );
	_StringTable.Store( JpHash( acString01 ), acString01 );
	_StringTable.Store( JpHash( acString02 ), acString02 );
	_StringTable.Store( JpHash( acString03 ), acString03 );

	JpHashTable< char, 128 >::Iterator Iter( &_StringTable );
	while ( Iter.Next() )
	{
		char* pString = Iter.GetData();
	}
	
	JpHashTable< char, 128 >::Iterator Iter2( &_StringTable );
	for ( char* pString = Iter2.First(); Iter2.HasMore(); pString = Iter2.Next() )
	{
		int i = 0;
	}

*/

typedef unsigned int hash;
typedef unsigned int guid;

extern unsigned int JpHash( const char* pacText );
extern unsigned int JpHashWithCase( const char* pacText );


template< class TYPE, int SIZE >
class JpHashTable
{
	public:

		//-----------------------
		// public enums and classes


		// This allows for linear walk over the data using WHILE and FOR
		// loop constructions.  
		//
		//    // WHILE example
		//    JpHashTable< TYPE, SIZE >::Iterator Iter( &_ItemTable );
		//    while ( Iter.Next() )
		//        Item* pItem = Iter.GetData();
		//
		//    // FOR example 
		//    JpList< Item >::Iterator Iter( &_ItemList );
		//    for ( Item* pItem = Iter.First(); Iter.MoreExist(); pItem = Iter.Next() )
		//        pItem->Update( dt );
		//
		// NOTE: Unlike with JpList Iterators, this Iterator does NOT support
		// data removal during iteration.  If you Remove() data while iterating
		// you will get unexpected results.
		//
		// NOTE: As of this writing, this Iterator has not been well tested.  beware!
		// 2009_07_03 mdomowicz
		class Iterator
		{
			public:
				Iterator( JpHashTable< TYPE, SIZE >* pTable ) 
				{
					_pTable = pTable;
					_iMaxKeys = SIZE;
					Reset();
				}

				void Reset() // supports WHILE construction
				{
					_iCurrentKey = -1;
					_iNumEntriesFound = 0;
				}

				TYPE* First() // supports FOR construction
				{ 
					_iCurrentKey = -1;
					return Next();
				}

				TYPE* Next() 
				{
					for ( int i = _iCurrentKey + 1; i < _iMaxKeys; i ++ )
					{
						if ( _pTable->_uiKeys[ i ] != 0 )
						{
							_iCurrentKey = i;
							_iNumEntriesFound ++;
							return _pTable->Find( _pTable->_uiKeys[ i ] );
						}
					}
					
					return NULL;
				}

				bool HasMore() { return ( _iNumEntriesFound < _pTable->_iNumEntries ); }

				TYPE* GetData() { return _pTable->Find( _pTable->_uiKeys[ _iCurrentKey ] ); }

				unsigned int GetKey() { return _pTable->_uiKeys[ _iCurrentKey ]; }

			private:

				Iterator() {} // default constructor is private so it can't be used.

				JpHashTable< TYPE, SIZE >* _pTable;
				int _iCurrentKey;
				int _iMaxKeys;
				int _iNumEntriesFound;
		};

		friend class Iterator;


		//-----------------------
		// public functions
		JpHashTable()
		{
			ASSERTF( IS_POWER_OF_TWO( SIZE ), ( "Size of hash table must be a power of 2" ) );
			Init();
		}

		void Init();

		TYPE* Find( unsigned int iKey );

		bool Exists( unsigned int iKey );

		bool Store( unsigned int iKey, TYPE* pValue );

		bool StoreUniquely( unsigned int iKey, TYPE* pValue );

		void Remove( unsigned int iKey );

		void RemoveAll();

		int GetNumEntries() 
		{ 
			return _iNumEntries; 
		}

		int GetMaxEntries() 
		{ 
			return SIZE; 
		}

		void DeleteAll();

		void ArrayDeleteAll();

		void FreeAll();


	private:


		//-----------------------
		// private variables

		int	_iNumEntries;
		TYPE* _pValues[ SIZE ];
		unsigned int _uiKeys[ SIZE ];

		int GetIndexFromKey( unsigned int iKey );
};




template< class TYPE, int SIZE > 
void JpHashTable< TYPE, SIZE >::Init()
{
	for ( int i = 0; i < SIZE; i ++ )
	{
		_pValues[ i ] = NULL;
		_uiKeys[ i ] = 0;
	}

	_iNumEntries = 0;
}



template< class TYPE, int SIZE > 
TYPE* JpHashTable< TYPE, SIZE >::Find( unsigned int uiKey )
{
	int iIndex = GetIndexFromKey( uiKey );
	if ( iIndex >= 0 )
		return _pValues[ iIndex ];

	return NULL;
}




template< class TYPE, int SIZE > 
bool JpHashTable< TYPE, SIZE >::Exists( unsigned int uiKey )
{
	int iIndex = GetIndexFromKey( uiKey );
	if ( iIndex >= 0 )
		return true;

	return false;
}




template< class TYPE, int SIZE > 
int JpHashTable< TYPE, SIZE >::GetIndexFromKey( unsigned int uiKey )
{
	// SIZE is always a power of two, so SIZE-1 is always
	// a bunch of ones.  For example, if size is 256, SIZE-1
	// is 255 = 0000 0000 0000 0000 0000 0000 1111 1111.
	// 
	// Using SIZE-1 as an AND bitmask forces all values 
	// to be within numerical range of [0, SIZE-1]

	unsigned int uiBitMask = SIZE - 1;

	int iIndex = uiKey & uiBitMask;
	while ( _uiKeys[ iIndex ] != 0 )
	{
		if ( _uiKeys[ iIndex ] == uiKey )
			return iIndex;

		iIndex = ( iIndex + 1 ) & uiBitMask;
	}

	return -1;
}




template< class TYPE, int SIZE > 
bool JpHashTable< TYPE, SIZE >::Store( unsigned int uiKey, TYPE* pValue )
{
	ASSERTF( uiKey != 0, ( "\nkeys must be non 0" ) );
	ASSERTF( _iNumEntries < SIZE / 2, ("\nhash table is more than half full. This is slow. Comment at top of file explain why.") );
	ASSERTF( _iNumEntries < SIZE - 1, ("\nhash table full") );

	if ( uiKey != 0 && _iNumEntries < SIZE - 1 )
	{
		unsigned int uiBitMask = SIZE - 1;
		int iIndex = uiKey & uiBitMask;

		while ( _uiKeys[ iIndex ] != 0 )
		{
			ASSERTF( _uiKeys[ iIndex ] != uiKey, ( "duplicate ID (0x%x).", uiKey ) );
			if ( _uiKeys[ iIndex ] == uiKey )
				return false;
		
			iIndex = ( iIndex + 1 ) & uiBitMask;
		}

		_uiKeys[ iIndex ] = uiKey;
		_pValues[ iIndex ] = pValue;
		_iNumEntries += 1;
		return true;
	}

	return false;
}




template< class TYPE, int SIZE > 
bool JpHashTable< TYPE, SIZE >::StoreUniquely( unsigned int uiKey, TYPE* pValue )
{
	if ( !Exists( uiKey ) )
		return Store( uiKey, pValue );

	return false;
}




template< class TYPE, int SIZE > 
void JpHashTable< TYPE, SIZE >::Remove( unsigned int uiKey )
{
	if ( uiKey )
	{
		int iIndex = GetIndexFromKey( uiKey );
		if ( iIndex >= 0 )
		{
			_uiKeys[ iIndex ] = 0;
			_pValues[ iIndex ] = NULL;
			_iNumEntries -= 1;

			// Re-insert any entries that follow.
			unsigned int uiBitMask = SIZE - 1;
			iIndex = ( iIndex + 1 ) & uiBitMask;

			while ( _uiKeys[ iIndex ] != 0 )
			{
				uiKey = _uiKeys[ iIndex ];
				TYPE* pValue = _pValues[ iIndex ];

				_uiKeys[ iIndex ] = 0;
				_pValues[ iIndex ] = 0;

				_iNumEntries -= 1;
				Store( uiKey, pValue );

				iIndex = ( iIndex + 1 ) & uiBitMask;
			}
		}
	}

}


template< class TYPE, int SIZE > 
void JpHashTable< TYPE, SIZE >::RemoveAll()
{
	for ( int i = 0; i < SIZE; i ++ )
	{
		_pValues[ i ] = NULL;
		_uiKeys[ i ] = 0;
	}

	_iNumEntries = 0;
}



template< class TYPE, int SIZE > 
void JpHashTable< TYPE, SIZE >::DeleteAll()
{
	for ( int i = 0; i < SIZE; i ++ )
	{
		if ( _pValues[ i ] != NULL )
			delete _pValues[ i ];

		_pValues[ i ] = NULL;
		_uiKeys[ i ] = 0;
	}

	_iNumEntries = 0;
}



template< class TYPE, int SIZE > 
void JpHashTable< TYPE, SIZE >::ArrayDeleteAll()
{
	for ( int i = 0; i < SIZE; i ++ )
	{
		if ( _pValues[ i ] != NULL )
			delete[] _pValues[ i ];

		_pValues[ i ] = NULL;
		_uiKeys[ i ] = 0;
	}

	_iNumEntries = 0;
}



template< class TYPE, int SIZE > 
void JpHashTable< TYPE, SIZE >::FreeAll()
{
	for ( int i = 0; i < SIZE; i ++ )
	{
		if ( _pValues[ i ] != NULL )
			free( _pValues[ i ] );

		_pValues[ i ] = NULL;
		_uiKeys[ i ] = 0;
	}

	_iNumEntries = 0;
}



#endif //__JpHashTable_h



