/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#ifndef __JpList_h
#define __JpList_h

#include "stdafx.h"

#include "JpList.h"
#include "JpRandom.h"
#include "JpUtils.h"

/*
	//-----------------------
	// The following is sample code
	// that shows how to use this
	// list class.

	JpList< int > TestList;
	int iData = 6;
	int iData2 = 7;
	int iData3 = 10;

	TestList.Add( &iData, 1);
	TestList.Add( &iData2, 2 );
	TestList.Add( &iData2, 3 );
	TestList.Add( &iData, 4 );
	TestList.Add( &iData3, 5 );

	JpList< int >::Node* pCurrent = TestList.GetFirst();
	for ( ; pCurrent != TestList.GetHead(); pCurrent = pCurrent->GetNext() )
	{
		int iTest = *(pCurrent->GetData());
	}


	//-----------------------
	// Here is another example of how
	// to walk through a list using
	// the new iterator class.
	// As of the time of this writing
	// these iterators have not been
	// well tested.  Use at your own
	// risk.
	// mdomowicz 2008_12_08

	JpList< SoAction >::Iterator Iter( &_ActionList );
	for ( SoAction* pAction = Iter.First(); Iter.HasMore(); pAction = Iter.Next() )
	{
		pAction->Update( dt );
	}


	//-----------------------
	// The following While loop should
	// also work, but again it's not
	// yet extremely well tested.
	// mdomowicz 2008_12_08

	JpList< SoAction >::Iterator Iter( &_ActionList );
	while ( Iter.Next() )
	{
		SoAction* pAction = Iter.GetNodeData();
		pAction->Update( dt );
	}
*/



template< class TYPE >
class JpList
{
	public:

		//-----------------------
		// class declarations

		class Node
		{
			public:
				Node() 
				{
					_pNext = NULL;
					_pPrev = NULL;
					_pData = NULL;
					_iSortKey = 0;
				}

				
				Node* GetNext() { return _pNext; }
				Node* GetPrev() { return _pPrev; }
				TYPE* GetData() { return _pData; }

				int GetSortKey() { return _iSortKey; }
				void SetSortKey( int iSortKey ) { _iSortKey = iSortKey; }
				

			private:
				Node* _pNext;
				Node* _pPrev;
				TYPE* _pData;
				int _iSortKey;


				void SetNext( Node* pNext ) { _pNext = pNext; }
				void SetPrev( Node* pPrev ) { _pPrev = pPrev; }
				void SetData( TYPE* pData ) { _pData = pData; }
				
				friend class JpList;
		};


		class Iterator
		{
			public:
				Iterator( JpList< TYPE >* pList ) 
				{
					_pList = pList;
					Reset();
				}


				void Reset()
				{
					// Set the current node to the head 
					// to support WHILE construction like:
					// 
					//    JpList< Item >::Iterator Iter( &_ItemList );
					//    while ( Iter.Next() )
					//        Item* pAction = Iter.GetNodeData();
					//
					_pCurrentNode = _pList->GetHead();
					_pNextNode = _pCurrentNode->GetNext();
					_pPrevNode = _pCurrentNode->GetPrev();
				}


				TYPE* First() 
				{ 
					// Set the current node to the first item
					// to support FOR construction like:
					// 
					//    JpList< Item >::Iterator Iter( &_ItemList );
					//    for ( Item* pItem = Iter.First(); Iter.HasMore(); pItem = Iter.Next() )
					//        pItem->Update( dt );
					//
					_pCurrentNode = _pList->GetFirst();
					_pNextNode = _pCurrentNode->GetNext();
					return _pCurrentNode->GetData();
				}

				TYPE* Last() 
				{ 
					// Set the current node to the last item
					// to support FOR construction like:
					// 
					//    JpList< Item >::Iterator Iter( &_ItemList );
					//    for ( Item* pItem = Iter.Last(); Iter.HasMore(); pItem = Iter.Prev() )
					//        pItem->Update( dt );
					//
					_pCurrentNode = _pList->GetLast();
					_pPrevNode = _pCurrentNode->GetPrev();
					return _pCurrentNode->GetData();
				}

				TYPE* Next() 
				{
					_pCurrentNode = _pNextNode;

					// Many loops delete the Current node.
					// Thus we save the Next node immediately.
					_pNextNode = _pCurrentNode->GetNext();
					return _pCurrentNode->GetData();
				}

				TYPE* Prev() 
				{
					_pCurrentNode = _pPrevNode;

					// Many loops delete the Current node.
					// Thus we save the Prev node immediately.
					_pPrevNode = _pCurrentNode->GetPrev();
					return _pCurrentNode->GetData();
				}

				bool HasMore()
				{
					return ( _pCurrentNode != _pList->GetHead() );
				}

				Node* GetNode() { return _pCurrentNode; }

				TYPE* GetNodeData() { return _pCurrentNode->GetData(); }

			private:

				Iterator() {} // default constructor is private so it can't be used.

				Node* _pCurrentNode;
				Node* _pNextNode;
				Node* _pPrevNode;
				JpList< TYPE >* _pList;
		};


		//-----------------------
		// function declarations

		JpList()
		{
			Init();
		}

		JpList( JpList& Copy )
		{
			ASSERTF( 0, ( "\nCannot copy instances of this class" ) );
		}

		~JpList()
		{
			RemoveAll();
		}

		void Init();

		int GetNumEntries() 
		{ 
			return _iNumEntries; 
		}

		bool Exists( TYPE* pData );

		Node* Add( TYPE* pData, int iSortKey );
		
		Node* AddUniquely( TYPE* pData, int iSortKey );

		Node* AddFirst( TYPE* pData );
		
		Node* AddLast( TYPE* pData );

		Node* AddBefore( TYPE* pData, Node* pReferenceNode );

		Node* AddAfter( TYPE* pData, Node* pReferenceNode );

		Node* Find( TYPE* pData );

		Node* GetFirst() { return _Head.GetNext(); }

		Node* GetLast() { return _Head.GetPrev(); }
		
		Node* GetHead() { return &_Head; }

		Node* GetAtIndex( int iIndex ); // O(n), use sparingly!

		Node* GetRandom( JpRandom* pRng ); // Also O(n), use sparingly!

		// Remove just removes the Node, but the data is unaffected.
		void Remove( TYPE* pData );
		TYPE* RemoveFirst();
		TYPE* RemoveLast();
		void RemoveAll();

		// Delete removes the node and deallocates the data using delete
		void Delete( TYPE* pData );
		void DeleteFirst();
		void DeleteLast();
		void DeleteAll();

		void QuickSort();

		void Combine( JpList< TYPE >* pList );
	
		void Copy( JpList< TYPE >* pList );
		void CopyOnly( JpList< TYPE >* pList );


	private:

		//-----------------------
		// member variables.

		Node _Head;
		int	_iNumEntries;


		//-----------------------
		// member function.
		void RemoveNode( Node* pNode );

		void QuickSortArray( Node** papNodePointers, int Left, int Right );

		int PartitionArray( Node** papNodePointers, int Left, int Right );
};




/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::Init()
{
	_iNumEntries = 0;
	_Head.SetNext( &_Head );
	_Head.SetPrev( &_Head );
	_Head.SetData( NULL );
	_Head.SetSortKey ( 0 );
}



/******************************************************************************
 FUNCTION:  Exists
 NOTES:     walks the list to see if there is a match.
******************************************************************************/
template< class TYPE > 
bool JpList< TYPE >::Exists( TYPE* pData )
{
	if ( Find( pData ) != NULL )
		return true;
	
	return false;
}



/******************************************************************************
 FUNCTION:  Add
 NOTES:     If two nodes have the same iSortKey, then their relative position 
            in the list will match the order in which they were added.  The
            first added will be in front of the second added, etc.
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::Add( TYPE* pData, int iSortKey )
{
	_iNumEntries ++;

	ASSERTF( _Head.GetNext() != NULL, ("\nHead node has null next pointer.  Has this list been initialized?") );

	//-----------------------
	// create the new node.
	typename JpList< TYPE >::Node* pNewNode = new typename JpList< TYPE >::Node;
	pNewNode->SetData( pData );
	pNewNode->SetSortKey( iSortKey );


	//-----------------------
	// find the place in the list
	// to insert the node.
	Node* pCurrent = &_Head;
	while ( pCurrent->GetPrev() != &_Head )
	{
		if ( iSortKey >= pCurrent->GetPrev()->GetSortKey() )
			break;
		else
			pCurrent = pCurrent->GetPrev();
	}


	//-----------------------
	// insert the node.
	pNewNode->SetNext( pCurrent );
	pNewNode->SetPrev( pCurrent->GetPrev() );
	pCurrent->GetPrev()->SetNext( pNewNode );
	pCurrent->SetPrev( pNewNode );

	return pNewNode;
}



/******************************************************************************
 FUNCTION:  AddUniquely
 NOTES:     Only adds to list if item is not already in list.
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::AddUniquely( TYPE* pData, int iSortKey )
{
	if ( Exists( pData ) == false )
		return Add( pData, iSortKey );

	return NULL;
}



/******************************************************************************
 FUNCTION:  AddFirst
 NOTES:     This adds to the front of the list, regardless of the SortKeys.
            Thus using this will likely make your list unsorted.
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::AddFirst( TYPE* pData )
{
	ASSERTF( _Head.GetNext() != NULL, ("\nHead node has null next pointer.  Has this list been initialized?") );

	Node* pNewNode = AddAfter( pData, GetHead() );
	return pNewNode;
}



/******************************************************************************
 FUNCTION:  AddLast
 NOTES:     This adds to the back of the list, regardless of the SortKeys.
            Thus using this will likely make your list unsorted.
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::AddLast( TYPE* pData )
{
	ASSERTF( _Head.GetNext() != NULL, ("\nHead node has null next pointer.  Has this list been initialized?") );

	Node* pNewNode = AddBefore( pData, GetHead() );
	return pNewNode;
}




/******************************************************************************
 FUNCTION:  AddBefore
 NOTES:     This creates a new node and inserts it in front of the given node, 
            regardless of the SortKeys.  Using this will likely make your list 
            unsorted.

            If your reference object is of type TYPE* and not of type Node*,
            you can use the Find function to get the Node* from the TYPE*

            For speed reasons, this function does not verify that the 
            reference node exists in the list.  
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::AddBefore( TYPE* pData, Node* pReferenceNode )
{
	_iNumEntries ++;

	//-----------------------
	// create the new node.
	typename JpList< TYPE >::Node* pNewNode = new typename JpList< TYPE >::Node;
	pNewNode->SetData( pData );
	pNewNode->SetSortKey( 0 );


	//-----------------------
	// insert the node.
	pNewNode->SetNext( pReferenceNode );
	pNewNode->SetPrev( pReferenceNode->GetPrev() );

	pReferenceNode->GetPrev()->SetNext( pNewNode );
	pReferenceNode->SetPrev( pNewNode );

	return pNewNode;
}




/******************************************************************************
 FUNCTION:  AddAfter
 NOTES:     This creates a new node and inserts it after the given node, 
            regardless of the SortKeys.  Using this will likely make your list 
            unsorted.

            If your reference object is of type TYPE* and not of type Node*,
            you can use the Find function to get the Node* from the TYPE*

            For speed reasons, this function does not verify that the 
            reference node exists in the list.  
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::AddAfter( TYPE* pData, Node* pReferenceNode )
{
	_iNumEntries ++;

	//-----------------------
	// create the new node.
	typename JpList< TYPE >::Node* pNewNode = new typename JpList< TYPE >::Node;
	pNewNode->SetData( pData );
	pNewNode->SetSortKey( 0 );


	//-----------------------
	// insert the node.
	pNewNode->SetNext( pReferenceNode->GetNext() );
	pNewNode->SetPrev( pReferenceNode );
	pReferenceNode->GetNext()->SetPrev( pNewNode );
	pReferenceNode->SetNext( pNewNode );

	return pNewNode;
}




/******************************************************************************
 FUNCTION:  Find
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::Find( TYPE* pData )
{
	Node* pCurrent = _Head.GetNext();
	while ( pCurrent != &_Head )
	{
		if ( pCurrent->GetData() == pData )
			return pCurrent;
		else
			pCurrent = pCurrent->GetNext();
	}

	return NULL;
}




/******************************************************************************
 FUNCTION:  RemoveNode
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::RemoveNode( Node* pNode )
{
	if ( pNode != NULL )
	{
		_iNumEntries --;

		pNode->GetNext()->SetPrev( pNode->GetPrev() );
		pNode->GetPrev()->SetNext( pNode->GetNext() );
		delete pNode;
	}
}



/******************************************************************************
 FUNCTION:  Remove
 NOTE:      Remove just removes the Node, but the data is unaffected.
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::Remove( TYPE* pData )
{
	Node* pNode = Find( pData );
	RemoveNode( pNode );
}



/******************************************************************************
 FUNCTION:  RemoveFirst
******************************************************************************/
template< class TYPE > 
TYPE* JpList< TYPE >::RemoveFirst()
{
	TYPE* pData = NULL;
	Node* pFirst = GetFirst();
	if( pFirst )
	{
		pData = pFirst->GetData();
		RemoveNode( pFirst );
	}

	return pData;
}




/******************************************************************************
 FUNCTION:  RemoveLast
******************************************************************************/
template< class TYPE > 
TYPE* JpList< TYPE >::RemoveLast()
{
	TYPE* pData = NULL;
	Node* pLast = GetLast();
	if( pLast )
	{
		pData = pLast->GetData();
		RemoveNode( pLast );
	}

	return pData;
}




/******************************************************************************
 FUNCTION:  RemoveAll
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::RemoveAll()
{
	typename JpList< TYPE >::Node* pCurrent = GetFirst();
	for ( ; pCurrent != GetHead();  )
	{
		typename JpList< TYPE >::Node* pNext = pCurrent->GetNext();
		delete pCurrent;
		pCurrent = pNext;
	}

	_iNumEntries = 0;
	_Head.SetNext( &_Head );
	_Head.SetPrev( &_Head );
	_Head.SetData( NULL );
}




/******************************************************************************
 FUNCTION:  Delete
 NOTE:      Delete removes the node and deallocates the data using delete
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::Delete( TYPE* pData )
{
	Node* pNode = Find( pData );
	RemoveNode( pNode );
	delete pData;
}



/******************************************************************************
 FUNCTION:  DeleteFirst
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::DeleteFirst()
{
	Node* pFirst = GetFirst();
	if( pFirst )
	{
		TYPE* pData = pFirst->GetData();
		RemoveNode( pFirst );
		delete pFirst;
	}
}




/******************************************************************************
 FUNCTION:  DeleteLast
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::DeleteLast()
{
	Node* pLast = GetLast();
	if( pLast )
	{
		TYPE* pData = pLast->GetData();
		RemoveNode( pLast );
		delete pLast;
	}
}




/******************************************************************************
 FUNCTION:  DeleteAll
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::DeleteAll()
{
	typename JpList< TYPE >::Node* pCurrent = GetFirst();
	for ( ; pCurrent != GetHead();  )
	{
		typename JpList< TYPE >::Node* pNext = pCurrent->GetNext();
		delete pCurrent->GetData();
		delete pCurrent;
		pCurrent = pNext;
	}

	_iNumEntries = 0;
	_Head.SetNext( &_Head );
	_Head.SetPrev( &_Head );
	_Head.SetData( NULL );
}




/******************************************************************************
 FUNCTION:  GetAtIndex
 NOTE:      This is SUPER SLOW!!!  Using sparingly.
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::GetAtIndex( int iIndex )
{
	int iCurrent = 0;

	Node* pCurrent = _Head.GetNext();
	while ( pCurrent != &_Head )
	{
		if ( iCurrent == iIndex )
			return pCurrent;
		
		pCurrent = pCurrent->GetNext();
		iCurrent ++;
	}

	return NULL;
}




/******************************************************************************
 FUNCTION:  GetRandom
 NOTE:      Because this uses GetAtIndex, this is also SUPER SLOW!!!  
            Using sparingly.
******************************************************************************/
template< class TYPE > 
typename JpList< TYPE >::Node* JpList< TYPE >::GetRandom( JpRandom* pRng )
{
	int iNumEntries = GetNumEntries();

	if ( iNumEntries > 0 )
	{
		int iRandomIndex = pRng->GetInt() % iNumEntries;
		return GetAtIndex( iRandomIndex );
	}

	return NULL;
}




/******************************************************************************
 FUNCTION:  QuickSort
 INFO:      This will sort a linked list by the SortKey of each element or 
            "node" in the list.

            This implementation uses a version of QuickSort that works on 
            fixed sized arrays.  To do this, the linked list is broken up and 
            reorganzied as an array.  This array is fed to the QuickSortArray 
            routine.  When finished, the array is then reorganized into a 
            linked list, ordered by sort key from lowest to highest.

 NOTES:

	//-----------------------
	// The following is a unit test
	// for the QuickSort
	// don't erase.  Maybe useful later.
	// 2007_07_13 mdomowicz

	JpList< SoActor > ActorList;
	for ( int i = 0; i < 100; i ++ )
	{
		SoActor* pNewActor = new SoActor;
		ActorList.Add( pNewActor, 0 );
	}


	JpList< SoActor >::Node* pCurrent = ActorList.GetFirst();
	for ( ; pCurrent != ActorList.GetHead(); pCurrent = pCurrent->GetNext() )
	{
		int iRandomSortKey = JpRandom::G.GetInt()%100;
		pCurrent->SetSortKey( iRandomSortKey );
	}

	ActorList.QuickSort();


******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::QuickSort()
{
	//-----------------------
	// create a pointer to an array of
	// pointers
	typename JpList< TYPE >::Node** papNodePointers = new typename JpList< TYPE >::Node* [ _iNumEntries ];


	//-----------------------
	// set every pointer in the array
	// to point at each node in the
	// list.
	Node* pCurrent = &_Head;
	pCurrent = pCurrent->GetNext();

	int iIndex= 0;
	while ( pCurrent != &_Head )
	{
		papNodePointers[ iIndex ] = pCurrent;
		pCurrent = pCurrent->GetNext();
		iIndex++;
	}


	//-----------------------
	// Disconnect everything.
	// There is no more list.
	_Head.SetNext( NULL );
	_Head.SetPrev( NULL );

	for ( int i = 0; i < _iNumEntries; i ++ )
	{
		Node* pNode = papNodePointers[ i ];
		pNode->SetNext( NULL );
		pNode->SetPrev( NULL );
	}


	//-----------------------
	// quick sort the pointer array
	QuickSortArray( papNodePointers, 0, _iNumEntries - 1 );


	//-----------------------
	// now that the array is sorted
	// we recreate the link list
	pCurrent = &_Head;
	for ( int j = 0; j < _iNumEntries; j ++ )
	{
		Node* pNode = papNodePointers[ j ];

		pCurrent->SetNext( pNode );
		pNode->SetPrev( pCurrent );
		pCurrent = pNode;
	}

	pCurrent->SetNext( &_Head );
	_Head.SetPrev( pCurrent );


	//-----------------------
	// done!
	delete[] papNodePointers;
}




/******************************************************************************
 FUNCTION: QuickSortArray
 PURPOSE:  A fast, recursive, in place, sorting routine
 ARGS:     iLeft is the left most index of the array that needs to be sorted.
           iRight is the right most index
           Values is the array of values to be sorted
 RETURN:   NA
 USAGE:    note that if Values is an array of 100 elements, 0 should be passed
           in for iLeft, and 99 should be passed in for iRight           
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::QuickSortArray( Node** papNodePointers, int Left, int Right )
{
	int Middle; 

	if (Left < Right)
	{
		Middle = PartitionArray( papNodePointers, Left, Right );
		QuickSortArray( papNodePointers, Left, Middle - 1 );
		QuickSortArray( papNodePointers, Middle + 1, Right );
	}
}



/******************************************************************************
 FUNCTION: PartitionArray
 PURPOSE:  Used by QuickSortArray to do the dirty work of the sort
 ARGS:     iLeft and iRight are parameters sent by (and only by) QuickSort 
 RETURN:   NA
 USAGE:    This function should only be called by QuickSortArray
******************************************************************************/
template< class TYPE > 
int JpList< TYPE >::PartitionArray( Node** papNodePointers, int Left, int Right )
{
	double Pivot = papNodePointers[ Left ]->GetSortKey();
	int L = Left + 1;
	int R = Right;
 
	while (L <= R)
	{
		while (L <= Right && papNodePointers[ L ]->GetSortKey() <= Pivot)
			L ++;

		while (R >= Left && papNodePointers[ R ]->GetSortKey() > Pivot)
			R --;

		if (L < R)
		{
			Node* Temp = papNodePointers[L];
			papNodePointers[L] = papNodePointers[R];
			papNodePointers[R] = Temp;
			L ++;
			R --;
		}
	}


	Node* Temp2 = papNodePointers[Left];
	papNodePointers[Left] = papNodePointers[R];
	papNodePointers[R] = Temp2;

	return R;
}




/******************************************************************************
 FUNCTION:  Combine
 PURPOSE:   Take every node off the given list and AddUniquely to this list
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::Combine( JpList< TYPE >* pList )
{
	typename JpList< TYPE >::Node* pCurrent = pList->GetFirst();
	for ( ; pCurrent != pList->GetHead(); pCurrent = pCurrent->GetNext() )
	{
		TYPE* pData = pCurrent->GetData();
		int iSortKey = pCurrent->GetSortKey();
		AddUniquely( pData, iSortKey );
	}

	pList->RemoveAll();
}




/******************************************************************************
 FUNCTION:  Copy
 PURPOSE:   Copy all source list nodes to this list, leaving source intact
            Note:  'This' list is zeroed out before adding the source.
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::Copy( JpList< TYPE >* pList )
{
	RemoveAll();

	typename JpList< TYPE >::Node* pCurrent = pList->GetFirst();
	for ( ; pCurrent != pList->GetHead(); pCurrent = pCurrent->GetNext() )
	{
		TYPE* pData = pCurrent->GetData();
		int iSortKey = pCurrent->GetSortKey();

		Add( pData, iSortKey );
	}
}




/******************************************************************************
 FUNCTION:  CopyOnly
 PURPOSE:   Copy all source list nodes to this list, leaving source intact
            Note:  'This' list is NOT zeroed out before adding the source.
******************************************************************************/
template< class TYPE > 
void JpList< TYPE >::CopyOnly( JpList< TYPE >* pList )
{
	typename JpList< TYPE >::Node* pCurrent = pList->GetFirst();
	for ( ; pCurrent != pList->GetHead(); pCurrent = pCurrent->GetNext() )
	{
		TYPE* pData = pCurrent->GetData();
		int iSortKey = pCurrent->GetSortKey();

		Add( pData, iSortKey );
	}
}




#endif //__JpList_h



