/*-----------------------------------------------------------------------------
 |  JpMatrix.h
 |  Copyright (c) 2010, Mark F. Domowicz    All Rights Reserved
 +----------------------------------------------------------------------------*/



#ifndef __JpMatrix_H
#define __JpMatrix_H

#include "stdafx.h"

#include <math.h>
#include "JpDebug.h"
#include "JpVector.h"



#define M00   _vX.X
#define M10   _vX.Y
#define M20   _vX.Z
#define M30   _vX.W

#define M01   _vY.X
#define M11   _vY.Y
#define M21   _vY.Z
#define M31   _vY.W

#define M02   _vZ.X
#define M12   _vZ.Y
#define M22   _vZ.Z
#define M32   _vZ.W

#define M03   _vW.X
#define M13   _vW.Y
#define M23   _vW.Z
#define M33   _vW.W


class JpMatrix
{
	public:

		//-----------------------
		// public variables
		JpVector4 _vX;
		JpVector4 _vY;
		JpVector4 _vZ;
		JpVector4 _vW;


		//-----------------------
		// public functions
		JpMatrix() { Identity(); };

		// copy constructor
		JpMatrix( const JpMatrix& mSrc ) { Init( mSrc._vX, mSrc._vY, mSrc._vZ, mSrc._vW ); }


		void Init( const JpVector4& vX, const JpVector4& vY, const JpVector4& vZ, const JpVector4& vW )
		{
			_vX = vX;
			_vY = vY;
			_vZ = vZ;
			_vW = vW;
		}

		void Init
		(
			float f00, float f01, float f02, float f03,
			float f10, float f11, float f12, float f13,
			float f20, float f21, float f22, float f23,
			float f30, float f31, float f32, float f33
		);

		void Identity()
		{
			_vX.Init( 1.0f, 0.0f, 0.0f, 0.0f );
			_vY.Init( 0.0f, 1.0f, 0.0f, 0.0f );
			_vZ.Init( 0.0f, 0.0f, 1.0f, 0.0f );
			_vW.Init( 0.0f, 0.0f, 0.0f, 1.0f );
		}


		void SetTranslation( float fX, float fY, float fZ ) { _vW.Init( fX, fY, fZ ); }

		void SetTranslation( const JpVector4& v )
		{
			_vW.X = v.X;
			_vW.Y = v.Y;
			_vW.Z = v.Z;
		}

		void Translate( const JpVector4& v )
		{
			_vW.X += v.X;
			_vW.Y += v.Y;
			_vW.Z += v.Z;
		}

		void Translate( float fX, float fY, float fZ )
		{
			_vW.X += fX;
			_vW.Y += fY;
			_vW.Z += fZ;
		}


		void SetupScale( float fScaleX, float fScaleY, float fScaleZ );
		void ScaleX( float fS ) { M00 *= fS; }
		void ScaleY( float fS ) { M11 *= fS; }
		void ScaleZ( float fS ) { M22 *= fS; }
		void Scale( float fS ) 
		{ 
			M00 *= fS; 
			M11 *= fS; 
			M22 *= fS; 
		}

		void SetBasisVectors( const JpVector4& vXAxis, const JpVector4& vYAxis, const JpVector4& vZAxis);

		void Transpose4x4();
		void Transpose3x3();

		JpVector4 ApplyTransform( const JpVector4& vVector );
		JpVector4 ApplyRotation( const JpVector4& vVector );
		JpVector4 ApplyInverseRotation( const JpVector4& vVector );

		// the axes are in the columns
		JpVector4 GetXAxis()       { return _vX; }
		JpVector4 GetYAxis()       { return _vY; }
		JpVector4 GetZAxis()       { return _vZ; }
		JpVector4 GetTranslation() { return _vW; }

		void SetupRotationX( float fRot );
		void SetupRotationY( float fRot );
		void SetupRotationZ( float fRot );
		void SetupRotationArbitraryAxis( JpVector4 vAxis, float fRot );

		void RotateX( float fRot );
		void RotateY( float fRot );
		void RotateZ( float fRot );
		void RotateArbitraryAxis( JpVector4 vAxis, float fRot );


		bool AlignMatrixToVectorWithZAxis( const JpVector4& vDir );

	private:
};



// GLOBAL MATRIX FUNCTIONS
void MultiplyMatrices4x4( JpMatrix* Dest, const JpMatrix& L, const JpMatrix& R );

// GLOBAL MATRIX OVERLOADED OPERATORS
JpMatrix operator * ( const JpMatrix& mLeftHandArg, const JpMatrix& mRightHandArg );



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
inline void JpMatrix::Init
(
	float f00, float f01, float f02, float f03,
	float f10, float f11, float f12, float f13,
	float f20, float f21, float f22, float f23,
	float f30, float f31, float f32, float f33
)
{
	_vX.Init( f00, f10, f20, f30 );
	_vY.Init( f01, f11, f21, f31 );
	_vZ.Init( f02, f12, f22, f32 );
	_vW.Init( f03, f13, f23, f33 );
}




/******************************************************************************
 FUNCTION:  Multiply4x4
******************************************************************************/
inline void MultiplyMatrices4x4( JpMatrix* Dest, const JpMatrix& L, const JpMatrix& R )
{
	Dest->M00 = ( L.M00 * R.M00 ) + ( L.M01 * R.M10 ) + ( L.M02 * R.M20 ) + ( L.M03 * R.M30 );
	Dest->M10 = ( L.M10 * R.M00 ) + ( L.M11 * R.M10 ) + ( L.M12 * R.M20 ) + ( L.M13 * R.M30 );
	Dest->M20 = ( L.M20 * R.M00 ) + ( L.M21 * R.M10 ) + ( L.M22 * R.M20 ) + ( L.M23 * R.M30 );
	Dest->M30 = ( L.M30 * R.M00 ) + ( L.M31 * R.M10 ) + ( L.M32 * R.M20 ) + ( L.M33 * R.M30 );

	Dest->M01 = ( L.M00 * R.M01 ) + ( L.M01 * R.M11 ) + ( L.M02 * R.M21 ) + ( L.M03 * R.M31 );
	Dest->M11 = ( L.M10 * R.M01 ) + ( L.M11 * R.M11 ) + ( L.M12 * R.M21 ) + ( L.M13 * R.M31 );
	Dest->M21 = ( L.M20 * R.M01 ) + ( L.M21 * R.M11 ) + ( L.M22 * R.M21 ) + ( L.M23 * R.M31 );
	Dest->M31 = ( L.M30 * R.M01 ) + ( L.M31 * R.M11 ) + ( L.M32 * R.M21 ) + ( L.M33 * R.M31 );

	Dest->M02 = ( L.M00 * R.M02 ) + ( L.M01 * R.M12 ) + ( L.M02 * R.M22 ) + ( L.M03 * R.M32 );
	Dest->M12 = ( L.M10 * R.M02 ) + ( L.M11 * R.M12 ) + ( L.M12 * R.M22 ) + ( L.M13 * R.M32 );
	Dest->M22 = ( L.M20 * R.M02 ) + ( L.M21 * R.M12 ) + ( L.M22 * R.M22 ) + ( L.M23 * R.M32 );
	Dest->M32 = ( L.M30 * R.M02 ) + ( L.M31 * R.M12 ) + ( L.M32 * R.M22 ) + ( L.M33 * R.M32 );

	Dest->M03 = ( L.M00 * R.M03 ) + ( L.M01 * R.M13 ) + ( L.M02 * R.M23 ) + ( L.M03 * R.M33 );
	Dest->M13 = ( L.M10 * R.M03 ) + ( L.M11 * R.M13 ) + ( L.M12 * R.M23 ) + ( L.M13 * R.M33 );
	Dest->M23 = ( L.M20 * R.M03 ) + ( L.M21 * R.M13 ) + ( L.M22 * R.M23 ) + ( L.M23 * R.M33 );
	Dest->M33 = ( L.M30 * R.M03 ) + ( L.M31 * R.M13 ) + ( L.M32 * R.M23 ) + ( L.M33 * R.M33 );
}



/******************************************************************************
 FUNCTION:  SetupRotationX
******************************************************************************/
inline void JpMatrix::SetupRotationX( float fRot )
{
	Identity();
	M11 = cosf( fRot );
	M12 = -sinf( fRot );
	M21 = sinf( fRot );
	M22 = cosf( fRot );
}



/******************************************************************************
 FUNCTION:  SetupRotationY
******************************************************************************/
inline void JpMatrix::SetupRotationY( float fRot )
{
	Identity();
	M00 = cosf( fRot );
	M02 = sinf( fRot );
	M20 = -sinf( fRot );
	M22 = cosf( fRot );
}




/******************************************************************************
 FUNCTION:  SetupRotationZ
******************************************************************************/
inline void JpMatrix::SetupRotationZ( float fRot )
{
	Identity();
	M00 = cosf( fRot );
	M01 = -sinf( fRot );
	M10 = sinf( fRot );
	M11 = cosf( fRot );
}




/******************************************************************************
 FUNCTION:  SetupArbitraryAxisRotation
 PURPOSE:   Setup a rotation around an arbitrary axis.
 ARGS:      vAxis is the axis around with you want to rotate.  
            fRot is the amount of rotation to apply
 RETURN:    NA
 NOTES:     Taken from Graphic Gems I
******************************************************************************/
inline void JpMatrix::SetupRotationArbitraryAxis( JpVector4 vAxis, float fRot )
{
	float s, c, t;
	float tx, ty, tz;
	float sx, sy, sz;
		
	s = sinf( fRot );
	c = cosf( fRot );
	t = 1.0f - c;

	tx = t * vAxis.X;	ty = t * vAxis.Y;	tz = t * vAxis.Z;
	sx = s * vAxis.X;	sy = s * vAxis.Y;	sz = s * vAxis.Z;

	M00 = tx*vAxis.X + c;
	M01 = tx*vAxis.Y - sz;
	M02 = tx*vAxis.Z + sy;

	M10 = tx*vAxis.Y + sz;
	M11 = ty*vAxis.Y + c;
	M12 = ty*vAxis.Z - sx;

	M20 = tx*vAxis.Z - sy;
	M21 = ty*vAxis.Z + sx;
	M22 = tz*vAxis.Z + c;


	//Make the unfilled parts identity.
	M30 = 0.0f;   M31 = 0.0f;   M32 = 0.0f;
	M03 = 0.0f;   M13 = 0.0f;   M23 = 0.0f;
	M33 = 1.0f;	
}



/******************************************************************************
 FUNCTION:  RotateX
******************************************************************************/
inline void JpMatrix::RotateX( float fRot )
{
	JpMatrix mTemp;
	mTemp.SetupRotationX( fRot );
	MultiplyMatrices4x4( this, mTemp, *this );
}



/******************************************************************************
 FUNCTION:  RotateY
******************************************************************************/
inline void JpMatrix::RotateY( float fRot )
{
	JpMatrix mTemp;
	mTemp.SetupRotationY( fRot );
	MultiplyMatrices4x4( this, mTemp, *this );
}




/******************************************************************************
 FUNCTION:  RotateZ
******************************************************************************/
inline void JpMatrix::RotateZ( float fRot )
{
	JpMatrix mTemp;
	mTemp.SetupRotationZ( fRot );
	MultiplyMatrices4x4( this, mTemp, *this );
}




/******************************************************************************
 FUNCTION:  ArbitraryAxisRotate
******************************************************************************/
inline void JpMatrix::RotateArbitraryAxis( JpVector4 vAxis, float fRot )
{
	JpMatrix mTemp;
	mTemp.SetupRotationArbitraryAxis( vAxis, fRot );
	MultiplyMatrices4x4( this, mTemp, *this );
}




/******************************************************************************
 FUNCTION:  SetupScale
******************************************************************************/
inline void JpMatrix::SetupScale( float fScaleX, float fScaleY, float fScaleZ )
{
	Init
	(
		fScaleX, 0.0f, 0.0f, 0.0f,
		0.0f, fScaleY, 0.0f, 0.0f,
		0.0f, 0.0f, fScaleZ, 0.0f,
		0.0f, 0.0f, 0.0f, 1.0f
	);
}



/******************************************************************************
 FUNCTION:  Transpose4x4
 PURPOSE:   transpose the elements in the matrix
 NOTES:     The transpose of a matrix is when the columns become the 
            rows and vice versa. Here is an example:

            the transpose of [a][b]  is  [a][c]
                             [c][d]      [b][d]
******************************************************************************/
inline void JpMatrix::Transpose4x4()
{
	JpMatrix mTemp;

	mTemp.M00 = M00;
	mTemp.M01 = M10;
	mTemp.M02 = M20;
	mTemp.M03 = M30;

	mTemp.M10 = M01;
	mTemp.M11 = M11;
	mTemp.M12 = M21;
	mTemp.M13 = M31;

	mTemp.M20 = M02;
	mTemp.M21 = M12;
	mTemp.M22 = M22;
	mTemp.M23 = M32;

	mTemp.M30 = M03;
	mTemp.M31 = M13;
	mTemp.M32 = M23;
	mTemp.M33 = M33;

	*this = mTemp;
}



/******************************************************************************
 FUNCTION:  Transpose3x3
 PURPOSE:   Same as Transpose4x4, but slightly faster since it only transposes
            the 3x3 rotation matrix.
******************************************************************************/
inline void JpMatrix::Transpose3x3()
{
	JpMatrix mTemp;

	mTemp.M00 = M00;
	mTemp.M01 = M10;
	mTemp.M02 = M20;

	mTemp.M10 = M01;
	mTemp.M11 = M11;
	mTemp.M12 = M21;

	mTemp.M20 = M02;
	mTemp.M21 = M12;
	mTemp.M22 = M22;

	*this = mTemp;
}




/******************************************************************************
 FUNCTION:  SetBasisVectors
 PURPOSE:   Place Basis Vector information into the Matrix
 ARGS:      the three vectors are the 3 basis vectors
 RETURN:    NA
 NOTES:     
******************************************************************************/
inline void JpMatrix::SetBasisVectors
( 
	const JpVector4& vXAxis,
	const JpVector4& vYAxis, 
	const JpVector4& vZAxis
)
{
	_vX.X = vXAxis.X;
	_vX.Y = vXAxis.Y;
	_vX.Z = vXAxis.Z;

	_vY.X = vYAxis.X;
	_vY.Y = vYAxis.Y;
	_vY.Z = vYAxis.Z;

	_vZ.X = vZAxis.X;
	_vZ.Y = vZAxis.Y;
	_vZ.Z = vZAxis.Z;
}




/******************************************************************************
 FUNCTION:  *
 PURPOSE:   for multiplying two matrices
 ARGS:      self-explanatory
 RETURN:    returns the result of mLeftHandArg * mRightHandArg
 NOTES:     
******************************************************************************/
inline JpMatrix operator * ( const JpMatrix& mLeftHandArg, const JpMatrix& mRightHandArg )
{
	JpMatrix mResult;
	MultiplyMatrices4x4( &mResult, mLeftHandArg, mRightHandArg );
	return mResult;
}



/******************************************************************************
 FUNCTION:  ApplyTransform
 PURPOSE:   Apply the rotation and translation to the given vector
 NOTES:     
******************************************************************************/
inline JpVector4 JpMatrix::ApplyTransform( const JpVector4& vVector )
{
	// multiply each ROW of the matrix by the vector
	float x = 
		M00 * vVector.X + 
		M01 * vVector.Y + 
		M02 * vVector.Z + 
		M03;// * vVector.W;

	float y = 
		M10 * vVector.X + 
		M11 * vVector.Y + 
		M12 * vVector.Z + 
		M13;// * vVector.W;

	float z = 
		M20 * vVector.X +
		M21 * vVector.Y +
		M22 * vVector.Z + 
		M23;// * vVector.W;

	float w = 
		M30 * vVector.X +
		M31 * vVector.Y +
		M32 * vVector.Z +
		M33;// * vVector.W;

	JpVector4 vTemp( x, y, z, w );

	return vTemp;
}



/******************************************************************************
 FUNCTION:  ApplyRotation
 PURPOSE:   Apply just the rotation to the given vector
******************************************************************************/
inline JpVector4 JpMatrix::ApplyRotation( const JpVector4& vVector )
{
	// multiply each ROW of the matrix by the vector
	float x = 
		M00 * vVector.X + 
		M01 * vVector.Y + 
		M02 * vVector.Z;

	float y = 
		M10 * vVector.X + 
		M11 * vVector.Y + 
		M12 * vVector.Z;

	float z = 
		M20 * vVector.X +
		M21 * vVector.Y +
		M22 * vVector.Z;

	JpVector4 vTemp( x, y, z, vVector.W );

	return vTemp;
}



/******************************************************************************
 FUNCTION:  ApplyInverseRotation
 PURPOSE:   A common need is to rotate a vector by the inverse of a given
            matrix.  This is often done by Transposing the matrix, and then
            using the transpose to rotate the vector.  However, using this 
            function is equivalent, faster, and cleaner.
******************************************************************************/
inline JpVector4 JpMatrix::ApplyInverseRotation( const JpVector4& vVector )
{
	// multiply each COLUMN of the matrix by the vector
	float x = 
		M00 * vVector.X + 
		M10 * vVector.Y + 
		M20 * vVector.Z;

	float y = 
		M01 * vVector.X + 
		M11 * vVector.Y + 
		M21 * vVector.Z;

	float z = 
		M02 * vVector.X +
		M12 * vVector.Y +
		M22 * vVector.Z;

	JpVector4 vTemp( x, y, z, vVector.W );

	return vTemp;
}




/******************************************************************************
 FUNCTION:  AlignMatrixToVectorWithZAxis
 PURPOSE:   Orient the matrix so that the z axis points in the same direction
			as the given vector.
 NOTES:     this has not yet been thoroughly tested
			20050507 mdomowicz
******************************************************************************/
inline bool JpMatrix::AlignMatrixToVectorWithZAxis( const JpVector4& vDir )
{
	JpVector4 vFront = Normalize( vDir );
	JpVector4 vReferenceUp( 0.0, 1.0, 0.0 );

	if ( vDir.X == 0.0 && vDir.Z == 0.0 )
		vReferenceUp.Init( 1.0, 0.0, 0.0 );

	JpVector4 vSide = CrossProduct( vReferenceUp, vFront );
	JpVector4 vUp = CrossProduct( vFront, vSide );
	vSide.Normalize();
	vUp.Normalize();

	SetBasisVectors( vSide, vUp, vFront ); 
	return true;
}



#endif // __JpMatrix_H
