/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#ifndef __JpRandom_h
#define __JpRandom_h

#include "stdafx.h"

#include <stdio.h>
#include <stdlib.h>


#include "JpDebug.h"
#include "JpVector.h"

/*
	This is an implementation of "GameRand", which is a very decent 
	and fast PRNG.  It's not as strong as the Mersenne Twister, but
	it's pretty much good enough for most game needs.  I got the 
	info on GameRand from this article:

		http://mjolnirstudios.com/IanBullard/files/79ffbca75a75720f066d491e9ea935a0-10.php

	This implementation is based off a sample implementation I found here:

		http://pastebin.com/f7612a8d6

	And here is the sample implementation:

		class RandomGen
		{
				unsigned int high, low;
 
		public:
 
				RandomGen(unsigned int seed)
				{
						high = 0xDEADBEEF ^ seed;
						low  = 0x49616E42 ^ high;
				}
 
				unsigned int Next()
				{
						high = (high << 16) + (high >> 16);
						high += low;
						low += high;
 
						return high;
				}
		};


	mdomowicz 2009_04_09
*/

typedef unsigned int uint32;


class JpRandom
{
	public:
		//-----------------------
		// public functions
		JpRandom() { Seed( 0x5045bcac ); }

		JpRandom( unsigned int uiSeed ) { Seed( uiSeed ); }

		~JpRandom() {}


		void Seed(unsigned int seed)
		{
			high = 0xDEADBEEF ^ seed;
			low  = 0x49616E42 ^ high;
		}

		unsigned int GetInt()
		{
			high = (high << 16) + (high >> 16);
			high += low;
			low += high;

			return high;
		}

		// Return a random float in the range 0.0f <= r < 1.0f
		float GetFloat()
		{
			union
			{
				uint32	u32;
				float	f32;
			} u;

			// 0xff800000 = 1111.1111.1000.0000.0000.0000.0000.0000   =  -INF
			// 0x3f800000 = 0011.1111.1000.0000.0000.0000.0000.0000   =  1.0f

			// This essentially "starts with" the hex for 1.0, and fills in the
			// 23 bit mantissa with random numbers.  Thus we get a value
			// in the range: [1.0f, 2.0f)
			u.u32 = ( ( uint32 )( GetInt() >> 8 ) & ~0xff800000 ) | 0x3f800000;

			return u.f32 - 1.0f;
		}

		float GetFloatOneToMinusOne()
		{
			return ( 0.5f - GetFloat() ) * 2.0f;
		}

		int GetSign()
		{
			if ( GetFloat() < 0.5f )
				return -1;

			return 1;
		}

		bool GetBool()
		{
			if ( GetFloat() < 0.5f )
				return true;

			return false;
		}

		JpVector4f GetVector4f() 
		{ 
			return JpVector4f
			( 
				GetFloatOneToMinusOne(), 
				GetFloatOneToMinusOne(), 
				GetFloatOneToMinusOne(),
				GetFloatOneToMinusOne()
			); 
		}

		JpVector2f GetVector2f() 
		{ 
			return JpVector2f
			( 
				GetFloatOneToMinusOne(), 
				GetFloatOneToMinusOne()
			); 
		}

		JpVector2f GetNormal2f() 
		{ 
			JpVector2f vDir = GetVector2f();
			if ( vDir.IsZero() )
				vDir.X = 1.0f;

			vDir.Normalize();

			return vDir;
		}

		int GetIntRanged( int iMinInclusive, int iMaxInclusive )
		{
			ASSERTF( iMaxInclusive >= iMinInclusive, ( "\nBad range to GetInRanged." ) );
			int iDivisor = ( iMaxInclusive - iMinInclusive ) + 1;
			return iMinInclusive + ( GetInt() % iDivisor );
		}

		// These accessors are useful if/when the exact state
		// of the rng needs to be saved and restored.
		unsigned int GetHigh() { return high; }
		void SetHigh( unsigned int u ) { high = u; }

		unsigned int GetLow() { return low; }
		void SetLow( unsigned int u ) { low = u; }


		static JpRandom G;

	private:

		//-----------------------
		// private variables

		unsigned int high, low;
};


#endif //__JpRandom_h



