/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "JpSpline.h"
#include "JpUtils.h"

#include "mmgrChooser.h"



/******************************************************************************
 FUNCTION:  GetPointOnArray
 NOTES:     This function makes it easier to interpolate over a spline defined
            by a long array of points - an otherwise complex thing to do 
            because the spline functions only work with four points and a t 
            value between [0.0f, 1.0f).

            With this function, for example, you can provide a path with, say
            100 vertices, and simply "walk along" that path with a continuously
            growing value for t.
******************************************************************************/
void JpSpline::GetPointOnArray
(
	JpVector4f* pvResult,
	float t,
	JpVector4f* pavPoints,
	int iNumPoints,
	bool bWrapLower,
	bool bWrapUpper,
	SPLINE_TYPE eSplineType
)
{
	int i1;
	int i2;
	int i3;
	int i4;
	float fNormalizedTime;

	DetermineSplineParametersForArray
	( 
		&i1, &i2, &i3, &i4, 
		&fNormalizedTime,	
		t, 
		iNumPoints, 
		bWrapLower, 
		bWrapUpper 
	);
	

	JpVector4f* p = pavPoints;
	switch ( eSplineType )
	{
		case SPLINE_TYPE_BEZIER:
			GetPointOnBezierSpline( pvResult, fNormalizedTime, &p[ i1 ], &p[ i2 ], &p[ i3 ], &p[ i4 ] );
			break;

		case SPLINE_TYPE_BSPLINE:
			GetPointOnBSpline( pvResult, fNormalizedTime, &p[ i1 ], &p[ i2 ], &p[ i3 ], &p[ i4 ] );
			break;

		case SPLINE_TYPE_CATMULLROM:
			GetPointOnCatmullRomSpline( pvResult, fNormalizedTime, &p[ i1 ], &p[ i2 ], &p[ i3 ], &p[ i4 ] );
			break;

		default:
			ASSERTF( 0, ( "\nUnknown spline type." ) );
	}
}



/******************************************************************************
 FUNCTION:  DetermineSplineParametersForArray
 PURPOSE:   The spline interpolating functions only work with four control
            points and a time t that is between [0.0f, 1.0f).  So if we have
            an array of more than 4 points, we can just use the first 4
            points as long as t = [0.0f, 1.0f).  But what if t is >= 1.0f or
            < 0?  

            This function determines the correct points to use given 
            a number of points >= 4, and any value for t.  This function
            also returns a normalized value for t, ie in range [0.0f, 1.0f).

            Example 1: if we have 5 points, and t = 2.3, we return:

                iResult1 = 2;
                iResult2 = 3;
                iResult3 = 4;
                iResult4 = 0;  // or 4 if wrapping is off
                fNormalizedTimeResult = 0.3f;


            Example 2: if we have 5 points, and t = -8.3, we return:

                iResult1 = 1;
                iResult2 = 2;
                iResult3 = 3;
                iResult4 = 4;
                fNormalizedTimeResult = 0.7f;

 NOTES:     
            This chart helps when reasoning about the logic	for
            determining the index from t, and for how to deal with
            index wrapping.

            The chart shows that, "if t is between 2 and 3, then
            P1 = 2, P2 = 3, ... "

                t    =  -2  -1   0   1   2   3   4
                t+1  =  -1   0   1   2   3   4   5
                P1   = |-2 |-1 | 0 | 1 | 2 | 3 | 4 |
                P2   = |-1 | 0 | 1 | 2 | 3 | 4 | 5 |
                P3   = | 0 | 1 | 2 | 3 | 4 | 5 | 6 |
                P4   = | 1 | 2 | 3 | 4 | 5 | 6 | 7 |

******************************************************************************/
void JpSpline::DetermineSplineParametersForArray
(
	int* piResult1,
	int* piResult2,
	int* piResult3,
	int* piResult4,
	float* pfNormalizedTimeResult,
	float t,
	int iNumPoints,
	bool bWrapLower,
	bool bWrapUpper
)
{
	//-----------------------
	// calculate a value for t between [0.0f, 1.0f)
	// if t =  8.3, then we want 0.3
	// if t = -8.3, then we want 0.7
	float fIntegerResultUnused;
	float fFractionResult;
	if ( t >= 0 )
		fFractionResult = modff( t, &fIntegerResultUnused );
	else
		fFractionResult = 1.0f - modff( t, &fIntegerResultUnused );


	//-----------------------
	// find the first index.  floorf turns 8.3 into 8.0,
	// and -8.3 into -9.0.
	int i = (int)floorf( t );


	//-----------------------
	// if the first index is less than 0, 
	// wrap it or clamp to 0.
	if ( i < 0 )
	{
		if ( bWrapLower )
		{
			// This gives us the correct positive starting index.
			// ex 1. if iNumPoints = 5, and i = -1, then i is changed to 4.  
			// ex 2. if iNumPoints = 5, and i = -6, then i is changed to 4
			// ex 3. if iNumPoints = 5, and i = -9, then i is changed to 1.
			i = iNumPoints + ( i % iNumPoints );
		}
		else
		{
			i = 0;
		}
	}


	//-----------------------
	// from the first index, find the rest.
	int iP1 = i + 0;
	int iP2 = i + 1;
	int iP3 = i + 2;
	int iP4 = i + 3;


	//-----------------------
	// handle indices that are "past the end" of the
	// array - either wrap or clamp them.
	if ( bWrapUpper )
	{
		iP1 = ( iP1 % iNumPoints );
		iP2 = ( iP2 % iNumPoints );
		iP3 = ( iP3 % iNumPoints );
		iP4 = ( iP4 % iNumPoints );
	}
	else
	{
		iP1 = ClampMax( iP1, iNumPoints );
		iP2 = ClampMax( iP2, iNumPoints );
		iP3 = ClampMax( iP3, iNumPoints );
		iP4 = ClampMax( iP4, iNumPoints );
	}


	//-----------------------
	// return the results.
	*piResult1 = iP1;
	*piResult2 = iP2;
	*piResult3 = iP3;
	*piResult4 = iP4;
	*pfNormalizedTimeResult = fFractionResult;
}

