/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/



#pragma once

#include "stdafx.h"

#include "JpDebug.h"
#include "JpVector.h"

/* 
	A collection of spline interpolation functions.  

	This site has useful info and examples on the major spline
	types and their properties.

	http://www.blackpawn.com/texts/splines/
*/

class JpSpline
{
	public:

		enum SPLINE_TYPE
		{
			SPLINE_TYPE_BEZIER,
			SPLINE_TYPE_BSPLINE,
			SPLINE_TYPE_CATMULLROM,
			SPLINE_TYPE_LAST
		};


		static void GetPointOnArray
		(
			JpVector4f* pvResult,
			float t,
			JpVector4f* pavPoints,
			int iNumPoints,
			bool bWrapLower,
			bool bWrapUpper,
			SPLINE_TYPE eSplineType
		);


		// Four point Bezier curves hit the first and last points,
		// but only approach the middle points.  Thus:
		// when t = 0, result = P1
		// when t = 1, result = P4
		// when 0 < t < 1, result = something between, but NOT P2 or P3
		static inline void GetPointOnBezierSpline
		( 
			JpVector4f* pvResult,
			float t,
			JpVector4f* pvP1,
			JpVector4f* pvP2,
			JpVector4f* pvP3,
			JpVector4f* pvP4
		)
		{
			float t2 = t*t;
			float t3 = t*t*t;

			float fA = ( -1*t3 + 3*t2 - 3*t + 1);
			float fB = (  3*t3 - 6*t2 + 3*t + 0);
			float fC = ( -3*t3 + 3*t2 + 0*t + 0);
			float fD = (  1*t3 + 0*t2 + 0*t + 0);


			pvResult->X = 
			(
				pvP1->X * fA + 
				pvP2->X * fB +
				pvP3->X * fC +
				pvP4->X * fD
			);

			pvResult->Y = 
			(
				pvP1->Y * fA + 
				pvP2->Y * fB +
				pvP3->Y * fC +
				pvP4->Y * fD
			);

			pvResult->Z = 
			(
				pvP1->Z * fA + 
				pvP2->Z * fB +
				pvP3->Z * fC +
				pvP4->Z * fD
			);
		}

		static inline void GetPointOnBezierSpline2d
		( 
			JpVector2f* pvResult,
			float t,
			JpVector2f* pvP1,
			JpVector2f* pvP2,
			JpVector2f* pvP3,
			JpVector2f* pvP4
		)
		{
			float t2 = t*t;
			float t3 = t*t*t;

			float fA = ( -1*t3 + 3*t2 - 3*t + 1);
			float fB = (  3*t3 - 6*t2 + 3*t + 0);
			float fC = ( -3*t3 + 3*t2 + 0*t + 0);
			float fD = (  1*t3 + 0*t2 + 0*t + 0);


			pvResult->X = 
			(
				pvP1->X * fA + 
				pvP2->X * fB +
				pvP3->X * fC +
				pvP4->X * fD
			);

			pvResult->Y = 
			(
				pvP1->Y * fA + 
				pvP2->Y * fB +
				pvP3->Y * fC +
				pvP4->Y * fD
			);
		}


		// Four point BSpline curves start near (but not necessarily on) the 2nd point 
		// and end near (but not necessarily on) the 3rd point.
		// The First and last points only serve to control the shape of the curve. Thus:
		// when t = 0, result ~= P2
		// when t = 1, result ~= P3
		static inline void GetPointOnBSpline
		( 
			JpVector4f* pvResult,
			float t,
			JpVector4f* pvP1,
			JpVector4f* pvP2,
			JpVector4f* pvP3,
			JpVector4f* pvP4
		)
		{
			float t2 = t*t;
			float t3 = t*t*t;

			float fA = ( -1*t3 + 3*t2 - 3*t + 1) / 6;
			float fB = (  3*t3 - 6*t2 - 0*t + 4) / 6;
			float fC = ( -3*t3 + 3*t2 + 3*t + 1) / 6;
			float fD = (  1*t3 + 0*t2 + 0*t + 0) / 6;


			pvResult->X = 
			(
				pvP1->X * fA + 
				pvP2->X * fB +
				pvP3->X * fC +
				pvP4->X * fD
			);

			pvResult->Y = 
			(
				pvP1->Y * fA + 
				pvP2->Y * fB +
				pvP3->Y * fC +
				pvP4->Y * fD
			);

			pvResult->Z = 
			(
				pvP1->Z * fA + 
				pvP2->Z * fB +
				pvP3->Z * fC +
				pvP4->Z * fD
			);
		}

		static inline void GetPointOnBSpline2d
		( 
			JpVector2f* pvResult,
			float t,
			JpVector2f* pvP1,
			JpVector2f* pvP2,
			JpVector2f* pvP3,
			JpVector2f* pvP4
		)
		{
			float t2 = t*t;
			float t3 = t*t*t;

			float fA = ( -1*t3 + 3*t2 - 3*t + 1) / 6;
			float fB = (  3*t3 - 6*t2 - 0*t + 4) / 6;
			float fC = ( -3*t3 + 3*t2 + 3*t + 1) / 6;
			float fD = (  1*t3 + 0*t2 + 0*t + 0) / 6;

			pvResult->X = 
			(
				pvP1->X * fA + 
				pvP2->X * fB +
				pvP3->X * fC +
				pvP4->X * fD
			);

			pvResult->Y = 
			(
				pvP1->Y * fA + 
				pvP2->Y * fB +
				pvP3->Y * fC +
				pvP4->Y * fD
			);
		}


		// Four point Catmull-Rom splines start exactly on P2 and end exactly on P3.
		// The First and last points only serve to control the shape of the curve. Thus:
		// when t = 0, result = P2
		// when t = 1, result = P3
		//
		// This is a good choice if an interpolation path requires that each point
		// in the path is reached exactly.  Camera tracking, and animation curves are
		// example applications.
		static inline void GetPointOnCatmullRomSpline
		( 
			JpVector4f* pvResult,
			float t,
			JpVector4f* pvP1,
			JpVector4f* pvP2,
			JpVector4f* pvP3,
			JpVector4f* pvP4
		)
		{
			float t2 = t*t;
			float t3 = t*t*t;

			float fA = ( -1*t3 + 2*t2 - 1*t + 0) / 2;
			float fB = (  3*t3 - 5*t2 + 0*t + 2) / 2;
			float fC = ( -3*t3 + 4*t2 + 1*t + 0) / 2;
			float fD = (  1*t3 - 1*t2 + 0*t + 0) / 2;


			pvResult->X = 
			(
				pvP1->X * fA + 
				pvP2->X * fB +
				pvP3->X * fC +
				pvP4->X * fD
			);

			pvResult->Y = 
			(
				pvP1->Y * fA + 
				pvP2->Y * fB +
				pvP3->Y * fC +
				pvP4->Y * fD
			);

			pvResult->Z = 
			(
				pvP1->Z * fA + 
				pvP2->Z * fB +
				pvP3->Z * fC +
				pvP4->Z * fD
			);
		}

		static inline void GetPointOnCatmullRomSpline2d
		( 
			JpVector2f* pvResult,
			float t,
			JpVector2f* pvP1,
			JpVector2f* pvP2,
			JpVector2f* pvP3,
			JpVector2f* pvP4
		)
		{
			float t2 = t*t;
			float t3 = t*t*t;

			float fA = ( -1*t3 + 2*t2 - 1*t + 0) / 2;
			float fB = (  3*t3 - 5*t2 + 0*t + 2) / 2;
			float fC = ( -3*t3 + 4*t2 + 1*t + 0) / 2;
			float fD = (  1*t3 - 1*t2 + 0*t + 0) / 2;


			pvResult->X = 
			(
				pvP1->X * fA + 
				pvP2->X * fB +
				pvP3->X * fC +
				pvP4->X * fD
			);

			pvResult->Y = 
			(
				pvP1->Y * fA + 
				pvP2->Y * fB +
				pvP3->Y * fC +
				pvP4->Y * fD
			);
		}


	private:

		static void DetermineSplineParametersForArray
		(
			int* piResult1,
			int* piResult2,
			int* piResult3,
			int* piResult4,
			float* pfNormalizedTimeResult,
			float t,
			int iNumPoints,
			bool bWrapLower,
			bool bWrapUpper
		);
};



