/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "JpHashTable.h"
#include "JpString.h"
#include "mmgrChooser.h"




/******************************************************************************
 FUNCTION:  ClearString
******************************************************************************/
void JpString::ClearString()
{
	if ( _pCString )
	{
		delete[] _pCString;
		_pCString = NULL;
	}
}




/******************************************************************************
 FUNCTION:  SetCString
******************************************************************************/
void JpString::SetCString( const char* pacString )
{
	ASSERTF( _pCString == NULL, ( "\nString already set.  Clear before reuse." ) );

	if ( !pacString )
	{
		ClearString();
	}
	else if ( !_pCString )
	{
		_iLength = strlen( pacString ) + 1; //strlen does not include the null terminator.
		_pCString = new char[ _iLength ];
		strncpy( _pCString, pacString, _iLength );
		_pCString[ _iLength - 1 ] = 0; // null terminator, just in case.
	}
}




/******************************************************************************
 FUNCTION:  SetByPrintf
 NOTES:     1. _vsnprintf is a Microsoft specific function.

            2. Most of the complexity below is due to the fact that there is
            no good way to determine the final size of the string

            3. There are alternative functions to _vsnprintf available that 
            would remove most of the complexity below, but they are only 
            available in VC7 and above.

            4. If this were ported to gcc, it will of course need to be 
            rewritten.  However, gcc also has better functions for determining
            the final string size.  Look up vsprintf, vsnprintf, and so on.
******************************************************************************/
void JpString::SetByPrintf( char* pacFormat, ... )
{
	#if DCG_IS_TARGET_WIN32
		va_list args;

		va_start( args, pacFormat );

		//-----------------------
		// _vsnprintf returns the number of characters in the final
		// string, or -1 if the buffer is not big enough.
		// So here we first try to see if the string will fit in
		// a reasonable start length.
		int iCurrentLength = 1024;
		_pCString = new char[ iCurrentLength ];
		int iResult = _vsnprintf( _pCString, iCurrentLength, pacFormat, args );
			
		//-----------------------
		// if the string doesn't fit.  Try again, but with a buffer
		// that's double in size from the previous try.  Repeat until
		// the results fit in the buffer.
		while ( iResult == -1 )
		{
			delete[] _pCString;

			iCurrentLength *= 2;
			_pCString = new char[ iCurrentLength ];
			iResult = _vsnprintf( _pCString, iCurrentLength, pacFormat, args );
		}

		//-----------------------
		// Since the results fit, iResult holds the length of the string.
		// so we can now reallocate the string a final time, but this
		// time to the exact required size.
		delete[] _pCString;
		_pCString = new char[ iResult + 1 ];
		_iLength = _vsnprintf( _pCString, iResult, pacFormat, args );


		//-----------------------
		// Lots of opportunity for errors here, so...
		// adding a null, just in case.
		_pCString[ iResult ] = 0; 
	#else
		ASSERTF( 0, ( "\nSetByPrintf not yet implemented for this platform.  See comments." ) );
	#endif
}





/******************************************************************************
 FUNCTION:  MatchesString
******************************************************************************/
bool JpString::MatchesString( JpString* psString )
{
	return ( strcmp( GetCString(), psString->GetCString() ) == 0 );
}



/******************************************************************************
 FUNCTION:  MatchesCString
******************************************************************************/
bool JpString::MatchesCString( char* pacString )
{
	return ( strcmp( GetCString(), pacString ) == 0 );
}



/******************************************************************************
 FUNCTION:  EndsWith
******************************************************************************/
bool JpString::EndsWith( char* pacEnd )
{
	int iEndLength = strlen( pacEnd ) + 1;

	for ( int i = 0; i < iEndLength; i ++ )
	{
		int i2 = ( _iLength - 1 ) - i;
		int i3 = ( iEndLength - 1 ) - i;

		if ( _pCString[ i2 ] != pacEnd[ i3 ] )
			return false;
	}

	return true;
}



/******************************************************************************
 FUNCTION:  ParseInt
******************************************************************************/
int JpString::ParseInt( const char* pacString ) 
{
	int i;
	int iErrorCode = sscanf( pacString, "%i", &i );
	JP_UNUSED_VAR( iErrorCode );
	ASSERTF( iErrorCode != EOF, ( "\nerror converting string" ) );
	return i;
}



/******************************************************************************
 FUNCTION:  ParseHash
******************************************************************************/
hash JpString::ParseHash( const char* pacString ) 
{
	hash h;
	int iErrorCode = sscanf( pacString, "%x", &h );
	JP_UNUSED_VAR( iErrorCode );
	ASSERTF( iErrorCode != EOF, ( "\nerror converting string" ) );
	return h;
}



/******************************************************************************
 FUNCTION:  ParseFloat
******************************************************************************/
float JpString::ParseFloat( const char* pacString ) 
{
	float f;
	int iErrorCode = sscanf( pacString, "%f", &f );
	JP_UNUSED_VAR( iErrorCode );
	ASSERTF( iErrorCode != EOF, ( "\nerror converting string" ) );
	return f;
}



/******************************************************************************
 FUNCTION:  ParseBool
******************************************************************************/
bool JpString::ParseBool( const char* pacString ) 
{
	if ( JpHash( pacString ) == 0x0b069958 /* "false" */ )
		return false;
	else if ( JpHash( pacString ) == 0x4db211e5 /* "true" */ )
		return true;

	int i = ParseInt( pacString );
	return ( i != 0 );
}



/******************************************************************************
 FUNCTION:  BoolToString
******************************************************************************/
std::string JpString::BoolToString( bool b )
{
	if ( b )
		return "true";

	return "false";
}



/******************************************************************************
 FUNCTION:  UnsignedCharToString
******************************************************************************/
std::string JpString::UnsignedCharToString( unsigned char uc )
{
	char acTemp[ 32 ];
	sprintf( acTemp, "%u", uc );
	return acTemp;
}



/******************************************************************************
 FUNCTION:  CharToString
******************************************************************************/
std::string JpString::CharToString( char c )
{
	char acTemp[ 32 ];
	sprintf( acTemp, "%c", c );
	return acTemp;
}



/******************************************************************************
 FUNCTION:  UnsignedIntToString
******************************************************************************/
std::string JpString::UnsignedIntToString( unsigned int ui )
{
	char acTemp[ 32 ];
	sprintf( acTemp, "%u", ui );
	return acTemp;
}



/******************************************************************************
 FUNCTION:  IntToString
******************************************************************************/
std::string JpString::IntToString( int i )
{
	char acTemp[ 32 ];
	sprintf( acTemp, "%i", i );
	return acTemp;
}



/******************************************************************************
 FUNCTION:  FloatToString
******************************************************************************/
std::string JpString::FloatToString( float f )
{
	char acTemp[ 32 ];
	sprintf( acTemp, "%f", f );
	return acTemp;
}



/******************************************************************************
 FUNCTION:  HashToString
******************************************************************************/
std::string JpString::HashToString( hash h )
{
	char acTemp[ 32 ];
	sprintf( acTemp, "%x", h );
	return acTemp;
}
