/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#ifndef __JpUtils_h
#define __JpUtils_h

#include "stdafx.h"

#include <math.h>
#include "string.h"

#include "cocos2d.h"

#include "JpDebug.h"
#include "JpRandom.h"
#include "JpVector.h"


USING_NS_CC;



// From google's c++ style guide:
// A macro to disallow the copy constructor and operator= functions
// This should be used in the private: declarations for a class
#define DISALLOW_COPY_AND_ASSIGN(TypeName) \
  TypeName(const TypeName&);               \
  void operator=(const TypeName&)

// A similar macro for disallowing default construction.
#define DISALLOW_DEFAULT_CONSTRUCTOR(TypeName) \
  TypeName() {}                                \


// Wrap unused variables in this macro to prevent compiler warnings.
#define JP_UNUSED_VAR(unusedparam) (void)unusedparam


#define IS_POWER_OF_TWO( iValue )	( ( ( iValue ) & ( ( iValue ) - 1 ) ) == 0 )


#define DEG2RAD		( JpUtils::_fPi / 180.0f )
#define RAD2DEG		( 180.0f / JpUtils::_fPi )




/******************************************************************************
 Misc functions
******************************************************************************/
inline bool IsOdd( int i ) { return ( i & 1 ); }

inline float Square( float f ) { return ( f * f ); }

inline void Swap( float &f1, float &f2 )
{
	float fTemp = f1;
	f1 = f2;
	f2 = fTemp;
}

inline float Round( float f )
{
	return ( f > 0.0f ) ? floorf( f + 0.5f ) : ceilf( f - 0.5f );
}

// This function returns a value between 0 and 1, but at quantized increments.
// For example, if iNumQuantities = 3, then the return value will either be
// 0.0f, 0.3333f, or 0.6666f.  This function may also return 1.0f, but only if
// fValueToQuantize >= fRangeMax.
inline float Quantize( float fValueToQuantize, float fRangeMin, float fRangeMax, int iNumQuantities )
{
	float fRange = ( fRangeMax - fRangeMin );

	if ( fRange > 0.0f )
	{
		float fNumQuantities = (float)( iNumQuantities );
		float fValue = ( fValueToQuantize - fRangeMin );
		float fPercent = ( fValue / fRange );

		if ( fPercent < 0.0f )
			fPercent = 0.0f;

		if ( fPercent > 1.0f )
			fPercent = 1.0f;
		
		int iQuantum = (int)( fPercent * fNumQuantities );

		return (float)( iQuantum ) / fNumQuantities;
	}

	return 0.0f;
}




#if defined USING_OGRE
	inline bool OgreSceneNodeExists
	( 
		Ogre::SceneManager* pSceneManager, 
		const Ogre::String& sNodeName
	)
	{
		// As far as I can tell, using an exception handler is the
		// only way to test if a SceneNode already exits (in ogre 1.7.3).
		// imo, exceptions should only fire when there is unintended behavior,
		// not as a path for actual results.  But what's worse is that Visual
		// Studio outputs an error message whenever an exception fails, 
		// cluttering the output with false positives when this function is
		// used.  Sooo, basically use this only if you *really* need to, or for 
		// strictly temp purposes.  Otherwise find a work around.  
		// mdomowicz 2011_12_21
		/*
		try
		{
			pSceneManager->getSceneNode( sNodeName );

			// At this point, the scenenode exists
			return true;
		}
		catch (Ogre::Exception ex)
		{
			// The scenenode doesn't exist
			return false;
		}

		return false;
		*/

		// er...heh.  mdomowicz 2012_04_11
		// (probably don't need a function just for this, but keeping it here
		// for reference, because clearly i need it.)
		return pSceneManager->hasSceneNode( sNodeName );
	}
#endif




/******************************************************************************
 Ogre::Vector helpers
******************************************************************************/
#if defined USING_OGRE
	inline Ogre::Vector2 GetVector2DroppingZ( const Ogre::Vector3& v3 )
	{
		return Ogre::Vector2( v3.x, v3.y );
	}

	inline Ogre::Vector2 GetVector2UsingZasY( const Ogre::Vector3& v3 )
	{
		return Ogre::Vector2( v3.x, v3.z );
	}

	inline Ogre::Vector2 GetVector2Perpendicular( const Ogre::Vector2& v2 )
	{
		return Ogre::Vector2( -v2.y, v2.x );
	}

	inline Ogre::Vector2 Rotate( const Ogre::Vector2& v, float fRadians )
	{
		float fCos = cosf( fRadians );
		float fSin = sinf( fRadians );

		float X = ( v.x * fCos ) - ( v.y * fSin );
		float Y = ( v.y * fCos ) + ( v.x * fSin );

		return Ogre::Vector2( X, Y );
	}

	inline Ogre::Vector2 Normalize( const Ogre::Vector2& v )
	{
		float fLength = v.length();
		float fOneOverLength = 0.0f;

		if ( fLength > 0.0f )
			fOneOverLength = ( 1.0f/fLength );

		return Ogre::Vector2( v.x * fOneOverLength, v.y * fOneOverLength );
	}

	inline Ogre::Vector3 Normalize( const Ogre::Vector3& v )
	{
		float fLength = v.length();
		float fOneOverLength = 0.0f;

		if ( fLength > 0.0f )
			fOneOverLength = ( 1.0f/fLength );

		return Ogre::Vector3
		( 
			v.x * fOneOverLength, 
			v.y * fOneOverLength,
			v.z * fOneOverLength 
		);
	}

	inline float Length( const Ogre::Vector2& v )
	{
		return sqrtf( ( v.x * v.x ) + ( v.y * v.y ) );
	}

	inline float Length( const Ogre::Vector3& v )
	{
		return sqrtf( ( v.x * v.x ) + ( v.y * v.y ) + ( v.z * v.z ) );
	}

	inline float LengthSquared( const Ogre::Vector2& v )
	{
		return ( ( v.x * v.x ) + ( v.y * v.y ) );
	}

	inline float LengthSquared( const Ogre::Vector3& v )
	{
		return ( ( v.x * v.x ) + ( v.y * v.y ) + ( v.z * v.z ) );
	}

	inline float DotProduct( const Ogre::Vector2& v1, const Ogre::Vector2& v2 )
	{
		return ( ( v1.x * v2.x ) + ( v1.y * v2.y ) );
	}

	inline float DotProduct( const Ogre::Vector3& v1, const Ogre::Vector3& v2 )
	{
		return ( ( v1.x * v2.x ) + ( v1.y * v2.y ) + ( v1.z * v2.z ) );
	}
#else

	inline JpVector2f GetVector2Perpendicular( const JpVector2f& v2 )
	{
		return JpVector2f( -v2.Y, v2.X );
	}

#endif



/******************************************************************************
 FUNCTION:  GetSignedAngleBetween2d
 PURPOSE:   Returns the signed angle to v2B from v2A
 NOTES:     All input vectors must be normalized.
******************************************************************************/
inline float GetSignedAngleBetween2d( const JpVector2f& v2A, const JpVector2f& v2B )
{
	#if 0
		float fAngle = ( atan2f( v2B.Y, v2B.X ) - atan2f( v2A.Y, v2A.X ) );

		// Sometimes the calculated angle will be more than 180 deg, or 
		// less than -180 deg.  For example if the angle of vector A is 
		// 170 deg, and the angle of vector B is -45 deg, then then above
		// calculation would return an angle of -215.
		//
		//             +y
		//   A         |                AngleA = 170 degrees
		//     *--..__ |                AngleB = -45 degrees
		//     --------+--------+x
		//             |\               AngleB - AngleA = -215 degrees
		//             | \
		//             |  \ B
		//
		//
		// This can be interpreted as, "to turn towards B from A, turn 
		// -215 degrees".  But that's the long way to do it!  What we
		// prefer is "turn +145 degrees".  The following code makes this
		// adjustment when it's needed.

		if ( fAngle > JpUtils::_fPi )
		{
			fAngle -= JpUtils::_fTwoPi;
		}
		else if ( fAngle < -JpUtils::_fPi )
		{
			fAngle += JpUtils::_fTwoPi;
		}

		return fAngle;
	#else

		// This version uses the "perp dot product", which is the dot
		// product of one vector against the perpendicular of the other.
		// This version is better than the previous version because
		// it uses only one call to atan2, and it is branchless.
		// Previous version is kept for reference.
		// mdomowicz 2012_05_07


		// Dot( A, B ) can be thought of as the projection of B on A.
		float fDot =  ( v2A.X * v2B.X + v2A.Y * v2B.Y );

		// PerpDot( A, B ) can be thought of as the projection of B 
		// on the perpendicular of A.
		float fPerpDot = ( v2A.X * v2B.Y - v2A.Y * v2B.X );

		// Thus, together the Dot and PerpDot give projections along
		// an "x axis", and a "y axis", respectively.
		// Recall the atan2 parameters: atan2( y, x )
		return atan2( fPerpDot, fDot );

	#endif
}



/******************************************************************************
 FUNCTION:  GetAngleBetween3d
 NOTES:     All input vectors must be normalized.
            This returned angle is NOT signed.

            This was ported over from the Lovecraft prototype.  As of this
            comment, it's not been tested properly.  Use with caution, 
            and remove this comment once you've verified that it works.
            mdomowicz 2011_08_09
******************************************************************************/
inline float GetAngleBetween3d( const JpVector4f& v1, const JpVector4f& v2 )
{
	float fDot = ( v1.X * v2.X ) + ( v1.Y * v2.Y ) + ( v1.Z * v2.Z );

	// I was getting values just slightly outside of range, which makes acosf
	// freak out.  So clamping here.
	if ( fDot < -1.0f )
		fDot = -1.0f;
	else if ( fDot > 1.0f )
		fDot = 1.0f;

	if ( fDot != 1.0f ) 
		return acosf( fDot ); 

	return 0.0f;
}



/******************************************************************************
 choice selection helpers
******************************************************************************/
inline float Pick1From3f
( 
	int iChoice, 
	float fReturnOnError,
	float f0, 
	float f1, 
	float f2
)
{
	switch ( iChoice )
	{
		case 0: return f0;
		case 1: return f1;
		case 2: return f2;
	}

	return fReturnOnError;
}


inline float Pick1From5f
( 
	int iChoice, 
	float fReturnOnError,
	float f0, 
	float f1, 
	float f2,
	float f3,
	float f4
)
{
	switch ( iChoice )
	{
		case 0: return f0;
		case 1: return f1;
		case 2: return f2;
		case 3: return f3;
		case 4: return f4;
	}

	return fReturnOnError;
}




inline int Pick1From3i
( 
	int iChoice, 
	int iReturnOnError, 
	int i0, 
	int i1, 
	int i2
)
{
	switch ( iChoice )
	{
		case 0: return i0;
		case 1: return i1;
		case 2: return i2;
	}

	return iReturnOnError;
}

inline int Pick1From5i
( 
	int iChoice, 
	int iReturnOnError, 
	int i0, 
	int i1, 
	int i2,
	int i3,
	int i4
)
{
	switch ( iChoice )
	{
		case 0: return i0;
		case 1: return i1;
		case 2: return i2;
		case 3: return i3;
		case 4: return i4;
	}

	return iReturnOnError;
}






/******************************************************************************
 Macros for creating binary literals.
 NOTE:
      Got this from: 

          http://groups.google.com/group/comp.arch.embedded/msg/9d430b6d3da12c8f

      Sample Usage:

          char Sample1 = B8(01010101);  // = 85

          short int Sample2 = B16(10101010,01010101); // = 43605

          int Sample3 = B32(10000000,11111111,10101010,01010101); // = 2164238933 

******************************************************************************/
/* Binary constant generator macro
   By Tom Torfs - donated to the public domain
*/

/* All macro's evaluate to compile-time constants */

/* turn a numeric literal into a hex constant
   (avoids problems with leading zeroes)
   8-bit constants max value 0x11111111, always fits in unsigned long
*/
#define HEX__(n) 0x##n##LU

/* 8-bit conversion function */
#define B8__(x) ((x&0x0000000FLU)?1:0)      \
               +((x&0x000000F0LU)?2:0)      \
               +((x&0x00000F00LU)?4:0)      \
               +((x&0x0000F000LU)?8:0)      \
               +((x&0x000F0000LU)?16:0)     \
               +((x&0x00F00000LU)?32:0)     \
               +((x&0x0F000000LU)?64:0)     \
               +((x&0xF0000000LU)?128:0)

/* for upto 8-bit binary constants */
#define B8(d) ((unsigned char)B8__(HEX__(d)))

/* for upto 16-bit binary constants, MSB first */
#define B16(dmsb,dlsb) (((unsigned short)B8(dmsb)<<8)     \
                        + B8(dlsb))

/* for upto 32-bit binary constants, MSB first */
#define B32(dmsb,db2,db3,dlsb) (((unsigned long)B8(dmsb)<<24)      \
                                  + ((unsigned long)B8(db2)<<16) \
                                  + ((unsigned long)B8(db3)<<8)    \
                                  + B8(dlsb))




/******************************************************************************
 Powers of Two utilities
 NOTE: Got these from wikipedia (http://en.wikipedia.org/wiki/Power_of_two)

       These have not yet been thoroughly tested as of 2009_02_05.
       mdomowicz 2009_02_05
******************************************************************************/
inline bool IsPowerOfTwo( int iValue )
{
	return ( ( ( iValue ) & ( ( iValue ) - 1 ) ) == 0 );
}

inline int GetModuloPowerOfTwo( unsigned int uiValue, unsigned int uiPowerOfTwo )
{
	//-----------------------
	// this returns ( uiValue % uiPowerOfTwo )
	return ( uiValue & ( uiPowerOfTwo - 1 ) );
}

inline int GetNextHigherPowerOfTwo( int k )
{
	//-----------------------
	// returns the next highest power of two.
	// 
	// examples;
	//  120093   returns  131072   -- ie 2^17
	//  1025     returns  2048     -- notice, NOT 1024 here.
	//  -1000    returns  0        -- negative numbers always return 0

	k--;

	for ( int i = 1; i < sizeof(int) * 8; i <<= 1 )
		k = k | k >> i;

	return k + 1;
}





/******************************************************************************
 Min Max functions
******************************************************************************/
inline const int Min( int a, int b )
{
	return( (a < b) ? a : b) ;
}

inline const int Max( int a, int b )
{
	return( (a > b) ? a : b) ;
}

inline const float Min( float a, float b )
{
	return( (a < b) ? a : b) ;
}

inline const float Max( float a, float b )
{
	return( (a > b) ? a : b) ;
}




/******************************************************************************
 Clamp functions
******************************************************************************/
inline float Clamp( float fValue, float fMin, float fMax ) 
{ 
	if ( fValue < fMin )
		return fMin;
	else if ( fValue > fMax )
		return fMax;

	return fValue;
};

inline int Clamp( int iValue, int iMin, int iMax ) 
{ 
	if ( iValue < iMin )
		return iMin;
	else if ( iValue > iMax )
		return iMax;

	return iValue;
};

inline char Clamp( char cValue, char cMin, char cMax ) 
{ 
	if ( cValue < cMin )
		return cMin;
	else if ( cValue > cMax )
		return cMax;

	return cValue;
};



inline float ClampMax( float fValue, float fMax ) 
{ 
	if ( fValue > fMax )
		return fMax;

	return fValue;
};

inline int ClampMax( int iValue, int iMax ) 
{ 
	if ( iValue > iMax )
		return iMax;

	return iValue;
};

inline char ClampMax( char cValue, char cMax ) 
{ 
	if ( cValue > cMax )
		return cMax;

	return cValue;
};



inline float ClampMin( float fValue, float fMin ) 
{ 
	if ( fValue < fMin )
		return fMin;

	return fValue;
};

inline int ClampMin( int iValue, int iMin ) 
{ 
	if ( iValue < iMin )
		return iMin;

	return iValue;
};

inline char ClampMin( char cValue, char cMin ) 
{ 
	if ( cValue < cMin )
		return cMin;

	return cValue;
};




/******************************************************************************
 Distance functions
******************************************************************************/
inline float Dist2dSq
(
	float fX1, float fY1, 
	float fX2, float fY2 
)
{
	float fDistX = fX2 - fX1;
	float fDistY = fY2 - fY1;

	return ( fDistX * fDistX ) + ( fDistY * fDistY );
};


inline float Dist2d
( 
	float fX1, float fY1, 
	float fX2, float fY2 
)
{
	return sqrtf( Dist2dSq( fX1, fY1, fX2, fY2 ) );
};


inline float Dist3dSq
( 
	float fX1, float fY1, float fZ1,
	float fX2, float fY2, float fZ2
)
{
	float fDistX = fX2 - fX1;
	float fDistY = fY2 - fY1;
	float fDistZ = fZ2 - fZ1;

	return ( fDistX * fDistX ) + ( fDistY * fDistY ) + ( fDistZ * fDistZ );
};

inline float Dist3d
( 
	float fX1, float fY1, float fZ1,
	float fX2, float fY2, float fZ2
)
{
	return sqrtf( Dist3dSq( fX1, fY1, fZ1, fX2, fY2, fZ2 ) );
};






/******************************************************************************
 Interpolation functions
******************************************************************************/
inline const float Interpolate( float f1, float f2, float fTween )
{
	return f1 + ( f2 - f1 ) * fTween;
}

inline const float InterpolateClamped( float f1, float f2, float fTween )
{
	fTween = Clamp( fTween, 0.0f, 1.0f );
	return f1 + ( f2 - f1 ) * fTween;
}

inline const float ReverseInterpolate( float f1, float f2, float fInterpolate )
{
	return ( fInterpolate - f1 ) / ( f2 - f1 );
}


inline const JpVector2f Interpolate( const JpVector2f & v1, const JpVector2f & v2, float fTween )
{
	float fX = Interpolate( v1.X, v2.X, fTween ); 
	float fY = Interpolate( v1.Y, v2.Y, fTween );

	JpVector2f Final( fX, fY );
	return Final;
}


inline const JpVector2f InterpolateClamped( const JpVector2f & v1, const JpVector2f & v2, float fTween )
{
	fTween = Clamp( fTween, 0.0f, 1.0f );
	float fX = Interpolate( v1.X, v2.X, fTween ); 
	float fY = Interpolate( v1.Y, v2.Y, fTween );

	JpVector2f Final( fX, fY );
	return Final;
}


inline const ccColor3B Interpolate( const ccColor3B& Color1, const ccColor3B& Color2, float fTween )
{
	float fR = Interpolate( Color1.r, Color2.r, fTween ); 
	float fG = Interpolate( Color1.g, Color2.g, fTween );
	float fB = Interpolate( Color1.b, Color2.b, fTween );

	return ccc3( fR, fG, fB );

}

#if defined USING_OGRE
	inline const Ogre::Vector3 Interpolate( const Ogre::Vector3& v1, const Ogre::Vector3& v2, float fTween )
	{
		float fX = Interpolate( v1.x, v2.x, fTween ); 
		float fY = Interpolate( v1.y, v2.y, fTween );
		float fZ = Interpolate( v1.z, v2.z, fTween );

		Ogre::Vector3 Final( fX, fY, fZ );
		return Final;
	}

	inline const Ogre::Radian Interpolate( Ogre::Radian r1, Ogre::Radian r2, float fTween )
	{
		Ogre::Quaternion q1( r1, Ogre::Vector3::UNIT_Z );
		Ogre::Quaternion q2( r2, Ogre::Vector3::UNIT_Z );
		Ogre::Quaternion qResult = Ogre::Quaternion::Slerp( fTween, q1, q2, true /*shortestPath*/ );
		return qResult.getRoll();
	}


	inline const Ogre::ColourValue Interpolate( Ogre::ColourValue color1, Ogre::ColourValue color2, float fTween )
	{
		float fR = Interpolate( color1.r, color2.r, fTween ); 
		float fG = Interpolate( color1.g, color2.g, fTween );
		float fB = Interpolate( color1.b, color2.b, fTween );
		float fA = Interpolate( color1.a, color2.a, fTween );

		Ogre::ColourValue Final( fR, fG, fB, fA );
		return Final;
	}
#endif



/******************************************************************************
 S-Curve Interpolation.  (easy in, easy out.)
 NOTE:  The base function is this:

            y = x / (x + e^(c1 - c2 * x))

        Where the values for c1 and c2 can be tweaked to change the form 
        of the curve.

        c1 = 1.0f, and c2 = 4.0f gives a nice "standard" s-curve.

        c1 = 4.0f, and c2 = 8.0f gives a nice "standard" s-curve, but with
        more curvature in the shape

        c1 = 1.0f, and c2 = 8.0f gives an s-curve with a standard s between
        0 and 0.5, and a long asymptotic approach to 1.0 between 0.5 and 1.0

        c1 = 3.0f, and c2 = 4.0f gives a curve with a flat beginning.  Note
        this doesn't reach 1 when x = 1.  (see note below)

        Use www.wolframalpha.com (or your fav graphic calculator) to view
        and toy around with other curve settings.

        Note that y doesn't always reach 1 when x = 1, above.  For example,
		this is true when c1 = 2.7f, and c2 = 4.0f.  Therefore, f2
		may never be reached if t is limited to [0, 1].  Another reason to
		check your curves with a graphing calculator.
******************************************************************************/
inline const float InterpolateSCurve
( 
	float f1, 
	float f2, 
	float c1, 
	float c2, 
	float t 
)
{
	float e = 2.71828f;
	float y = t / ( t + powf( e, (c1 - c2 * t) ) );
	return f1 + ( f2 - f1 ) * y;
}



/******************************************************************************
 FUNCTION:  InterpolateSCurve
******************************************************************************/
inline JpVector2f InterpolateSCurve
( 
	const JpVector2f& v1, 
	const JpVector2f& v2, 
	float c1, 
	float c2, 
	float fTween 
)
{
	float fX = InterpolateSCurve( v1.X, v2.X, c1, c2, fTween );
	float fY = InterpolateSCurve( v1.Y, v2.Y, c1, c2, fTween );
	return JpVector2f( fX, fY );
}



/******************************************************************************
 Fast In Slow Out Interpolation.
 NOTE:  The larger fMaxX, the faster we approach f2.  Compare fMaxX = 1.0f,
        vs fMaxX = 1000.0f;
******************************************************************************/
inline const float InterpolateFastInSlowOut( float f1, float f2, float fMaxX, float fInterpolate )
{
	fInterpolate = Clamp( fInterpolate, 0.0f, 1.0f );
	float fX = Interpolate( 0.0001f, fMaxX, fInterpolate );

	float fY = log10f( fX );

	// log of 0.0001f is -4, so we offset back to 0.0f
	fY += 4.0f;

	// normalize
	float fMaxY = log10f( fMaxX );
	float fRange = ( 4.0f + fMaxY );
	fY /= fRange;  

	return Interpolate( f1, f2, fY );
}





/******************************************************************************
 A routine for finding the intersection point between two line segments.

 Note: This is probably not the best place for this code, but I couldn't think 
 of a more appropriate spot.   Here is good for now, I suppose.

 Note: I pretty much just stole this code from the following links:
	http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d/
	http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d/example.cpp

	Theory by: Paul Bourke (April 1989)
	Example Code by: Damian Coventry (Tuesday, 9 January 2007)

******************************************************************************/
#if defined USING_OGRE
	#define NOT_INTERESECTING 0
	#define PARALLEL 1
	#define COINCIDENT 2
	#define INTERESECTING 3

	inline int GetLineSegmentIntersection2d
	(
		Ogre::Vector2* pvResult, 
		const Ogre::Vector2 vStartA,
		const Ogre::Vector2 vEndA,
		const Ogre::Vector2 vStartB,
		const Ogre::Vector2 vEndB
	)
	{
		float denom = ((vEndB.y - vStartB.y)*(vEndA.x - vStartA.x)) -
					  ((vEndB.x - vStartB.x)*(vEndA.y - vStartA.y));

		float nume_a = ((vEndB.x - vStartB.x)*(vStartA.y - vStartB.y)) -
					   ((vEndB.y - vStartB.y)*(vStartA.x - vStartB.x));

		float nume_b = ((vEndA.x - vStartA.x)*(vStartA.y - vStartB.y)) -
					   ((vEndA.y - vStartA.y)*(vStartA.x - vStartB.x));

		if(denom == 0.0f)
		{
			if(nume_a == 0.0f && nume_b == 0.0f)
			{
				return COINCIDENT;
			}
			return PARALLEL;
		}

		float ua = nume_a / denom;
		float ub = nume_b / denom;

		if(ua >= 0.0f && ua <= 1.0f && ub >= 0.0f && ub <= 1.0f)
		{
			pvResult->x = vStartA.x + ua*(vEndA.x - vStartA.x);
			pvResult->y = vStartA.y + ua*(vEndA.y - vStartA.y);

			return INTERESECTING;
		}

		return NOT_INTERESECTING;
	}
#endif





/******************************************************************************
 String functions, etc
******************************************************************************/
class JpUtils
{
	public:

		static float _fPi;
		static float _fHalfPi;
		static float _fTwoPi;

		static ccColor3B _Black;
		static ccColor3B _White;
		static ccColor3B _Red;
		static ccColor3B _Green;
		static ccColor3B _Blue;
		static ccColor3B _Yellow;
		static ccColor3B _Orange;
		static ccColor3B _Magenta;
		static ccColor3B _Cyan;
		static ccColor3B _Purple;
		static ccColor3B _Grey16;
		static ccColor3B _Grey32;
		static ccColor3B _Grey64;
		static ccColor3B _Grey96;
		static ccColor3B _Grey128;
		static ccColor3B _Grey160;
		static ccColor3B _Grey192;
		static ccColor3B _Grey224;
		static ccColor3B _Grey240;


		static void StripPath( char acFileName[], const char acPathAndFileName[] )
		{
			int iFileNameStart = strlen( acPathAndFileName );

			while ( acPathAndFileName[ iFileNameStart ] != '\\' && acPathAndFileName[ iFileNameStart ] != '/' )
			{
				iFileNameStart --;
			}

			iFileNameStart ++;
			int iIndex = 0;
			while ( acPathAndFileName[ iFileNameStart ] != 0 )
			{
				acFileName[ iIndex ] = acPathAndFileName[ iFileNameStart ];
				iFileNameStart ++;
				iIndex ++;
			}

			acFileName[ iIndex ] = 0;
		}

		static void StripExtension( char acStringSansExt[], const char acStringWithExt[] )
		{
			int iExtStart = strlen( acStringWithExt );

			while ( acStringWithExt[ iExtStart ] != '.' )
			{
				iExtStart --;
			}

			int iIndex = 0;
			while ( iIndex < iExtStart )
			{
				acStringSansExt[ iIndex ] = acStringWithExt[ iIndex ];
				iIndex ++;
			}

			acStringSansExt[ iIndex ] = 0;
		}

		static void GetExtension( char acStringWithExt[], char acStringSansExt[] )
		{
			int iStrLen = strlen( acStringWithExt );

			int iExtStart = iStrLen;

			while ( acStringWithExt[ iExtStart ] != '.' )
			{
				iExtStart --;
			}

			// Skip the dot
			iExtStart ++;

			int iIndex    = iExtStart;
			while ( iIndex < iStrLen )
			{
				acStringSansExt[ ( iIndex - iExtStart) ] = acStringWithExt[ iIndex ];
				iIndex ++;
			}

			acStringSansExt[ ( iIndex - iExtStart) ] = 0;
		}

		static void CopyString( char* pacDest, const char* pacSrc, int iMaxCharsIncludingTerminator )
		{
            #ifdef _DEBUG // this check prevents an unused variable warning
                int iLength = strlen( pacSrc ) + 1; //strlen does not include the null terminator.
                ASSERTF( iLength <= iMaxCharsIncludingTerminator, ( "\nstring is too long" ) );
            #endif
            
			strncpy( pacDest, pacSrc, iMaxCharsIncludingTerminator );
			pacDest[ iMaxCharsIncludingTerminator - 1 ] = 0; // null terminator, just in case.
		}

		static void ConcatString( char* pacDest, const char* pacSrc, int iMaxDestSizeIncludingTerminator )
		{
			int iLengthDest = strlen( pacDest );
			int iLengthSrc = strlen( pacSrc );

			int iLengthCurr = iLengthDest + iLengthSrc + 1; //strlen does not include the null terminator.
			int iLengthRemaining = iMaxDestSizeIncludingTerminator - iLengthCurr;

			ASSERTF( iLengthRemaining > 0, ( "\nConcatenated string is longer than limit." ) );
			if ( iLengthRemaining > 0 )
			{
				strncat( pacDest, pacSrc, iLengthRemaining );
			}

			pacDest[ iMaxDestSizeIncludingTerminator - 1 ] = 0; // null terminator, just in case.
		}

	
		static char ColorFloatToChar( float fValue );

		// int version
		static void QuickSortArray( int* ipPointers, int Left, int Right );
		static int  PartitionArray(  int* ipPointers, int Left, int Right );
		
		// float version
		static void QuickSortArray( float* ipPointers, int Left, int Right );
		static int  PartitionArray(  float* ipPointers, int Left, int Right );

		static JpVector2i GetMaxWindowedResolution();
		static JpVector2i GetMaxFullScreenResolution();


	private:

};




/******************************************************************************
 FUNCTION: QuickSortArray   INTEGER
 PURPOSE:  A fast, recursive, in place, sorting routine
 ARGS:     iLeft is the left most index of the array that needs to be sorted.
           iRight is the right most index
           Values is the array of values to be sorted
 RETURN:   NA
 USAGE:    note that if Values is an array of 100 elements, 0 should be passed
           in for iLeft, and 99 should be passed in for iRight           
******************************************************************************/
inline void JpUtils::QuickSortArray( int* paiPointers, int Left, int Right )
{
	ASSERTF( 0, ( "\nYou are the first to use this function.  Please verify that your results are correct." ) );

	int Middle; 

	if (Left < Right)
	{
		Middle = PartitionArray( paiPointers, Left, Right );
		QuickSortArray( paiPointers, Left, Middle - 1 );
		QuickSortArray( paiPointers, Middle + 1, Right );
	}
}



/******************************************************************************
 FUNCTION: PartitionArray   INTEGER
 PURPOSE:  Used by QuickSortArray to do the dirty work of the sort
 ARGS:     iLeft and iRight are parameters sent by (and only by) QuickSort 
 RETURN:   NA
 USAGE:    This function should only be called by QuickSortArray
******************************************************************************/
inline int JpUtils::PartitionArray(  int* paiPointers, int Left, int Right )
{
	double Pivot = paiPointers[ Left ];
	int L = Left + 1;
	int R = Right;
 
	while (L <= R)
	{
		while (L <= Right && paiPointers[ L ] <= Pivot)
			L ++;

		while (R >= Left && paiPointers[ R ] > Pivot)
			R --;

		if (L < R)
		{
			int Temp = paiPointers[L];
			paiPointers[L] = paiPointers[R];
			paiPointers[R] = Temp;
			L ++;
			R --;
		}
	}


	int Temp2 = paiPointers[Left];
	paiPointers[Left] = paiPointers[R];
	paiPointers[R] = Temp2;

	return R;
}



/******************************************************************************
 FUNCTION: QuickSortArray   FLOAT
 PURPOSE:  A fast, recursive, in place, sorting routine
 ARGS:     iLeft is the left most index of the array that needs to be sorted.
           iRight is the right most index
           Values is the array of values to be sorted
 RETURN:   NA
 USAGE:    note that if Values is an array of 100 elements, 0 should be passed
           in for iLeft, and 99 should be passed in for iRight           
******************************************************************************/
inline void JpUtils::QuickSortArray( float* pafPointers, int Left, int Right )
{
	ASSERTF( 0, ( "\nYou are the first to use this function.  Please verify that your results are correct." ) );
	int Middle; 

	if (Left < Right)
	{
		Middle = PartitionArray( pafPointers, Left, Right );
		QuickSortArray( pafPointers, Left, Middle - 1 );
		QuickSortArray( pafPointers, Middle + 1, Right );
	}
}



/******************************************************************************
 FUNCTION: PartitionArray   FLOAT
 PURPOSE:  Used by QuickSortArray to do the dirty work of the sort
 ARGS:     iLeft and iRight are parameters sent by (and only by) QuickSort 
 RETURN:   NA
 USAGE:    This function should only be called by QuickSortArray
******************************************************************************/
inline int JpUtils::PartitionArray(  float* pafPointers, int Left, int Right )
{
	double Pivot = pafPointers[ Left ];
	int L = Left + 1;
	int R = Right;
 
	while (L <= R)
	{
		while (L <= Right && pafPointers[ L ] <= Pivot)
			L ++;

		while (R >= Left && pafPointers[ R ] > Pivot)
			R --;

		if (L < R)
		{
			float Temp = pafPointers[L];
			pafPointers[L] = pafPointers[R];
			pafPointers[R] = Temp;
			L ++;
			R --;
		}
	}


	float Temp2 = pafPointers[Left];
	pafPointers[Left] = pafPointers[R];
	pafPointers[R] = Temp2;

	return R;
}



inline char JpUtils::ColorFloatToChar( float fValue ) 
{	
	return (char)Clamp( ( fValue * 255.0f ), 0.0f, 255.0f ); 
}



/******************************************************************************
 FUNCTION: GetMaxWindowedResolution
 PURPOSE:  Returns the maximum size a full window (with all borders) can
           be.  Note a truly maximized window has no side/bottom borders,
		   which is not what this function returns.
******************************************************************************/
inline JpVector2i JpUtils::GetMaxWindowedResolution()
{
	#if DCG_IS_TARGET_WIN32
		RECT WinRect;
		SystemParametersInfo( SPI_GETWORKAREA, 0, &WinRect, 0 );

		int iMaxX = WinRect.right - WinRect.left;
		int iMaxY = WinRect.bottom - WinRect.top;

		int iBorderX = GetSystemMetrics( SM_CXSIZEFRAME ) * 2;
		int iBorderY = GetSystemMetrics( SM_CYSIZEFRAME ) * 2;
		int iBorderTitle = GetSystemMetrics( SM_CYCAPTION );

		iMaxX -= iBorderX;
		iMaxY -= ( iBorderY + iBorderTitle );

		return JpVector2i( iMaxX, iMaxY );
	#endif

	return JpVector2i( 0, 0 );
}



/******************************************************************************
 FUNCTION: GetMaxWindowedResolution
 PURPOSE:  Returns the maximum full screen size.
******************************************************************************/
inline JpVector2i JpUtils::GetMaxFullScreenResolution()
{
	#if DCG_IS_TARGET_WIN32
		int iScreenX = GetSystemMetrics( SM_CXSCREEN );
		int iScreenY = GetSystemMetrics( SM_CYSCREEN );

		return JpVector2i( iScreenX, iScreenY );
	#endif

	return JpVector2i( 0, 0 );
}



/******************************************************************************
 CLASS:    JpRasterLineIterator
 PURPOSE:  Makes it easy to walk along a bresenham line, point by point,
           without needing/using an array of generated points. 
******************************************************************************/
class JpRasterLineIterator
{
	public:
		JpRasterLineIterator( int iX0, int iY0, int iX1, int iY1 ) 
		{
			Init( iX0, iY0, iX1, iY1 );
		}

		void Init( int iX0, int iY0, int iX1, int iY1 ) 
		{
			_iX0 = iX0;
			_iY0 = iY0;
			_iX1 = iX1;
			_iY1 = iY1;
			_bHasMore = true;
			_bRunOnce = false;
			Reset();
		}

		void Reset()
		{
			// Set the current node to the head 
			// to support WHILE construction like:
			// 
			//    RasterLineIterator Iter( 1, 2, 3, 4 );
			//    while ( Iter.Next() ) {
			//        int iX = Iter.GetX();
			//        int iY = Iter.GetY(); }
			//
			_iCurX = -1;
			_iCurY = -1;
			_bRunOnce = false;

			_bFirstErrorRecoreded = false;
			_iAccumulatedError = 0;
		}


		bool First() 
		{ 
			// Set the current node to the first item
			// to support FOR construction like:
			// 
			//    RasterLineIterator Iter( 1, 2, 3, 4 );
			//    for ( Iter.First(); Iter.HasMore(); Iter.Next() ) {
			//        int iX = Iter.GetX(); 
			//        int iY = Iter.GetY(); }
			//
			_iCurX = _iX0;
			_iCurY = _iY0;
			_bRunOnce = true;

			_bFirstErrorRecoreded = false;
			_iAccumulatedError = 0;
			return true;
		}

		bool Next();

		bool HasMore() { return _bHasMore; }

		int GetX() { return _iCurX; }
		int GetY() { return _iCurY; }

	private:

		JpRasterLineIterator() {} // default constructor is private so it can't be used.

		int _iX0;
		int _iY0;
		int _iX1;
		int _iY1;
		int _iCurX;
		int _iCurY;
		bool _bHasMore;
		bool _bRunOnce;

		int _iAccumulatedError;
		bool _bFirstErrorRecoreded;
};



#endif //__JpUtils_h


