/*-----------------------------------------------------------------------------
 |  NesConfig.cpp
 |  Copyright (c) 2005 Mark F. Domowicz    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "NeConfig.h"
#include "NeFile.h"

#include "mmgrChooser.h"



/******************************************************************************
 FUNCTION:  ~NesConfig
******************************************************************************/
NeConfig::~NeConfig()
{
	Clear();
}



/******************************************************************************
 FUNCTION:  Clear
 PURPOSE:   delete every line.
******************************************************************************/
void NeConfig::Clear()
{
	JpList< Line >::Node* pCurrent = _LineList.GetFirst();
	for ( ; pCurrent != _LineList.GetHead(); pCurrent = pCurrent->GetNext() )
	{
		Line* pLine = pCurrent->GetData();
		delete pLine;
	}
	_LineList.RemoveAll();
}



/******************************************************************************
 FUNCTION:  FindLine
 PURPOSE:   find the line who's first word is the same as the given word.
******************************************************************************/
NeConfig::Line* NeConfig::FindLine( const char* pacFirstItemOnLine )
{
	JpList< Line >::Node* pCurrent = _LineList.GetFirst();
	for ( ; pCurrent != _LineList.GetHead(); pCurrent = pCurrent->GetNext() )
	{
		Line* pLine = pCurrent->GetData();
		Item* pItem = pLine->FindItemByNumber( 0 );		

		bool bSameText = ( strcmp( pItem->GetString(), pacFirstItemOnLine ) == 0 );
		if ( bSameText )
			return pLine;
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  CreateLine
******************************************************************************/
NeConfig::Line* NeConfig::CreateLine( const char* pacFirstItemOnLine )
{
	if ( pacFirstItemOnLine && pacFirstItemOnLine[ 0 ] != 0 )
	{
		Line* pNewLine = new Line;
		_LineList.AddLast( pNewLine );

		Item* pNewItem = new Item;
		pNewItem->SetString( pacFirstItemOnLine );
		pNewLine->AddItem( pNewItem );

		return pNewLine;
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  FindOrCreateLine
******************************************************************************/
NeConfig::Line* NeConfig::FindOrCreateLine( const char* pacFirstItemOnLine )
{
	Line* pLine = FindLine( pacFirstItemOnLine );

	if ( !pLine )
		pLine = CreateLine( pacFirstItemOnLine );

	return pLine;
}



/******************************************************************************
 FUNCTION:  ClearLineKeepingFirstItem
******************************************************************************/
void NeConfig::ClearLineKeepingFirstItem( NeConfig::Line* pLine )
{
	if ( pLine )
	{
		Item* pItem = pLine->FindItemByNumber( 0 );		

		std::string sText = pItem->GetString();

		pLine->RemoveAndDeleteAllItems();

		AppendStringToLine( pLine, sText.c_str() );
	}
}



/******************************************************************************
 FUNCTION:  DeleteLine
******************************************************************************/
void NeConfig::DeleteLine( NeConfig::Line* pLine )
{
	_LineList.Remove( pLine );
	delete pLine;
}



/******************************************************************************
 FUNCTION:  AppendIntToLine
******************************************************************************/
void NeConfig::AppendIntToLine( NeConfig::Line* pLine, int i )
{
	char acText[ 32 ];
	sprintf( acText, "%i", i );

	Item* pNewItem = new Item;
	pNewItem->SetString( acText );
	pLine->AddItem( pNewItem );
}



/******************************************************************************
 FUNCTION:  AppendFloatToLine
******************************************************************************/
void NeConfig::AppendFloatToLine( NeConfig::Line* pLine, float f )
{
	char acText[ 32 ];
	sprintf( acText, "%f", f );

	Item* pNewItem = new Item;
	pNewItem->SetString( acText );
	pLine->AddItem( pNewItem );
}



/******************************************************************************
 FUNCTION:  AppendBoolToLine
******************************************************************************/
void NeConfig::AppendBoolToLine( NeConfig::Line* pLine, bool b )
{
	Item* pNewItem = new Item;

	if ( b )
		pNewItem->SetString( "true" );
	else
		pNewItem->SetString( "false" );

	pLine->AddItem( pNewItem );
}



/******************************************************************************
 FUNCTION:  AppendStringToLine
******************************************************************************/
void NeConfig::AppendStringToLine( Line* pLine, const char* pacText )
{
	Item* pNewItem = new Item;
	pNewItem->SetString( pacText );
	pLine->AddItem( pNewItem );
}



/******************************************************************************
 FUNCTION:  ReadFile
 PURPOSE:   Parse the given file, separating each line into separate "line" 
			objects which are themselves made up of separate "item" objects.
 RETURN:    returns false on error.
******************************************************************************/
bool NeConfig::ReadFile( const char* pacPathAndFileName )
{
	NeFile File;
	bool bSuccess = File.OpenFile( pacPathAndFileName, "r" );

	if ( bSuccess )
	{
		ReadAllLinesFsm( &File );

		File.CloseFile();
		return true;
	}

	return false;
}



/******************************************************************************
 FUNCTION:  WriteFile
******************************************************************************/
bool NeConfig::WriteFile( const char* pacPathAndFileName )
{
	FILE* pFile;
	pFile = fopen( pacPathAndFileName, "w" );
	ASSERTF( pFile, ( "\nUnable to open file to write." ) );

	if ( pFile )
	{
		JpList< Line >::Iterator LineIter( GetLineList() );

		while ( LineIter.Next() )
		{
			Line* pLine = LineIter.GetNodeData();

			JpList< Item >::Iterator ItemIter( pLine->GetItemList() );
			while ( ItemIter.Next() )
			{
				Item* pItem = ItemIter.GetNodeData();

				fputs( pItem->GetString(), pFile );

				// If it's not the last item in the row, add a space
				if ( ItemIter.GetNode() != pLine->GetItemList()->GetLast() )
					fputs( " ", pFile );
			}

			// If it's not the last line, add a newline
			if ( LineIter.GetNode() != GetLineList()->GetLast() )
				fputs( "\n", pFile );
		}

		fclose( pFile );

		return true;
	}

	return false;
}



/******************************************************************************
 FUNCTION:  ReadAllLinesFsm
******************************************************************************/
void NeConfig::ReadAllLinesFsm( NeFile* pFile )
{
	struct Local
	{
		static bool IsWhitespace( char c )
		{
			bool bIsWhiteSpace = 
			(
				c == ' ' ||
				c == '\n' ||
				c == '\t' ||
				c == '\r' 
			);

			return bIsWhiteSpace;
		}
	};


	char acText[ MAX_ITEM_LENGTH ];
	memset( acText, 0, MAX_ITEM_LENGTH );
	pFile->Seek( 0, SEEK_SET );
	char c = 0;


	const int STATE_EAT_WHITE_SPACE_BEFORE_NEXT_LINE = 10;
	const int STATE_ADD_WORD_TO_LINE = 20;
	const int STATE_EAT_WHITE_SPACE_BEFORE_NEXT_WORD = 30;
	const int STATE_HANDLE_WORD_LENGTH_EXCEEDED = 40;
	const int STATE_EAT_COMMENT = 50;
	int iCurrentState = STATE_EAT_WHITE_SPACE_BEFORE_NEXT_LINE;


	while ( c != (char)EOF )
	{
		switch( iCurrentState )
		{
			case STATE_EAT_WHITE_SPACE_BEFORE_NEXT_LINE:
			{
				c = pFile->GetChar();

				if ( !Local::IsWhitespace( c ) && c != (char)EOF )
				{
					Line* pNewLine = new Line;
					_LineList.AddLast( pNewLine );
					iCurrentState = STATE_ADD_WORD_TO_LINE;
				}

				break;
			}

			case STATE_ADD_WORD_TO_LINE:
			{
				memset( acText, 0, MAX_ITEM_LENGTH );

				int iNumChars = 0;

				//-----------------------
				// read the word
				while 
				( 
					iNumChars < MAX_ITEM_LENGTH && 
					!Local::IsWhitespace( c ) &&
					c != (char)EOF
				)
				{
					acText[ iNumChars ] = c;
					c = pFile->GetChar();
					iNumChars ++;
				}

				//-----------------------
				// if it's a comment handle it.
				bool bIsComment = ( iNumChars >= 2 && acText[ 0 ] == '/' && acText[ 1 ] == '/' );
				if ( bIsComment )
				{
					Line* pCurrentLine = _LineList.GetLast()->GetData();

					// If no words are yet added to this line, then this entire line is a comment
					// and we must remove the line from the line list.
					// If the line list already has words in it, then only the end of the line
					// is a comment, and we should not remove the line from the line list.
					if ( pCurrentLine->GetItemList()->GetNumEntries() == 0 )
					{
						_LineList.Remove( pCurrentLine );
						delete pCurrentLine;
					}

					iCurrentState = STATE_EAT_COMMENT;
				}


				//-----------------------
				// ELSE add the word as an item to 
				// the current line.
				else
				{
					//-----------------------
					// if we successfully read a string, create a new item
					if ( acText[ 0 ] != 0 )
					{
						Line* pNewLine = _LineList.GetLast()->GetData();

						Item* pNewItem = new Item;
						pNewItem->SetString( acText );
						pNewLine->AddItem( pNewItem );
					}

					//-----------------------
					// choose a next state depending on how the cell reading
					// was terminated.
					if ( iNumChars >= MAX_ITEM_LENGTH )
					{
						acText[ MAX_ITEM_LENGTH - 1 ] = 0;
						iCurrentState = STATE_HANDLE_WORD_LENGTH_EXCEEDED;
					}
					else if ( c == ' ' || c == '\t' || c == '\r' )
						iCurrentState = STATE_EAT_WHITE_SPACE_BEFORE_NEXT_WORD;
					else if ( c == '\n' )
						iCurrentState = STATE_EAT_WHITE_SPACE_BEFORE_NEXT_LINE;
				}

				break;
			}

			case STATE_EAT_WHITE_SPACE_BEFORE_NEXT_WORD:
			{
				c = pFile->GetChar();

				if ( c == '\n' )
					iCurrentState = STATE_EAT_WHITE_SPACE_BEFORE_NEXT_LINE;
				else if ( c != ' ' && c != '\t' && c != '\r' && c != (char)EOF )
					iCurrentState = STATE_ADD_WORD_TO_LINE;

				break;
			}

			case STATE_HANDLE_WORD_LENGTH_EXCEEDED:
			{
				//-----------------------
				// just eat characters until we get to a new
				// word or new line.		
				if ( c == ' ' || c == '\t' || c == '\r' )
					iCurrentState = STATE_EAT_WHITE_SPACE_BEFORE_NEXT_WORD;
				else if ( c == '\n' )
					iCurrentState = STATE_EAT_WHITE_SPACE_BEFORE_NEXT_LINE;
				else
				{
					c = pFile->GetChar();
				}
				
				break;
			}

			case STATE_EAT_COMMENT:
			{
				//-----------------------
				// just eat characters until we get to a new line
				if ( c == '\n' )
					iCurrentState = STATE_EAT_WHITE_SPACE_BEFORE_NEXT_LINE;
				else
					c = pFile->GetChar();

				break;
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  FindLineByNumber
******************************************************************************/
NeConfig::Line* NeConfig::FindLineByNumber( int iLineNumber )
{
	int iCurrent = 0;

	JpList< Line >::Node* pCurrent = _LineList.GetFirst();
	for ( ; pCurrent != _LineList.GetHead(); pCurrent = pCurrent->GetNext() )
	{
		Line* pLine = pCurrent->GetData();
		if ( iCurrent == iLineNumber )
			return pLine;

		iCurrent ++;
	}

	return NULL;
}
