/*-----------------------------------------------------------------------------
 |  NesConfig.h
 |  Copyright (c) 2000 Mark F. Domowicz    All Rights Reserved
 +----------------------------------------------------------------------------*/


#pragma once

#include "stdafx.h"
#include "cocos2d.h"

#include "stdio.h"
#include "JpList.h"
#include "JpHashTable.h"

USING_NS_CC;


/*
	How to use this class.

	Assume we have a file called "test.ini" that looks like this:

		Test hello 1 32.4
		PlayMusic false


	Use this code to read in the values

		NesConfig Config;
		Config.ReadFile( "test.ini" );
		NesConfig::Line* pLine = Config.FindLine( "Test" );
		if ( pLine )
		{
			char* pTest = pLine->GetString( 1 ); // hello
			int i = pLine->GetInt( 2 ); // 1
			float f = pLine->GetFloat( 3 ); // 32.4
		}


		pLine = Config.FindLine( "PlayMusic" );
		if ( pLine )
		{
			bool bPlayMusic = pLine->GetBool( 1 );
		}


	You can also use this class to write files.

		// Get an NeConfig object.  Alternatively, of course, you can just instance an NeConfig object here.
		NeConfig* pConfig = SoGlobals::GetNeConfig();

		// If the desired line is already in the config, return that line.  Otherwise, return the existing line.
		NeConfig::Line* pLine = pConfig->FindOrCreateLine( "iResolution" );   

		// If the line is, say, "iResolution 1024 768", this function will drop everying after the first item,
		// thus making it just, "iResolution"
		pConfig->ClearLineKeepingFirstItem( pLine );      

		// After these calls, the line will read, "iResolution 1280 1024"
		pConfig->AppendFloatToLine( pLine, 1280 );
		pConfig->AppendFloatToLine( pLine, 1024 );

		// Write the file to disk
		pConfig->WriteFile( "MyConfig.ini" );
*/





#define MAX_ITEM_LENGTH 64


class NeFile;

class NeConfig
{
	public:

		//-----------------------
		// public classes and enums
		class Item
		{
			public:
				Item() 
				{
					_acString[ 0 ] = 0;
				}

				void SetString( const char* pacText )
				{
					strncpy( _acString, pacText, MAX_ITEM_LENGTH );
					_acString[ MAX_ITEM_LENGTH - 1 ] = 0;
				}

				char* GetString() { return _acString; }

				int GetInt() 
				{
					int i;
					int iErrorCode = sscanf( GetString(), "%i", &i );
					ASSERTF( iErrorCode != EOF, ( "\nerror converting string" ) );
					return i;					
				}

				float GetFloat() 
				{
					float f;
					int iErrorCode = sscanf( GetString(), "%f", &f );
					ASSERTF( iErrorCode != EOF, ( "\nerror converting string" ) );
					return f;					
				}

				bool GetBool() 
				{
					if ( strcmp( GetString(), "false" ) == 0 )
						return false;
					else if ( strcmp( GetString(), "true" ) == 0 )
						return true;

					int i = GetInt();
					return ( i != 0 );
				}

			private:
				char _acString[ MAX_ITEM_LENGTH ];
		};


		class Line
		{
			public:
				Line() {}

				virtual ~Line() { RemoveAndDeleteAllItems(); }

				void AddItem( Item* pItem )
				{
					_ItemList.Add( pItem, 0 );
				}

				void RemoveAndDeleteAllItems()
				{
					JpList< Item >::Node* pCurrent = _ItemList.GetFirst();
					for ( ; pCurrent != _ItemList.GetHead(); pCurrent = pCurrent->GetNext() )
					{
						Item* pItem = pCurrent->GetData();
						delete pItem;
					}
					_ItemList.RemoveAll();
				}

				Item* FindItemByNumber( int iItemNumber )
				{
					int iCurrent = 0;

					JpList< Item >::Node* pCurrent = _ItemList.GetFirst();
					for ( ; pCurrent != _ItemList.GetHead(); pCurrent = pCurrent->GetNext() )
					{
						Item* pItem = pCurrent->GetData();
						if ( iCurrent == iItemNumber )
							return pItem;

						iCurrent ++;
					}

					return NULL;
				}

				char* GetString( int iItemNumber ) 
				{
					Item* pItem = FindItemByNumber( iItemNumber );
					if ( pItem != NULL )
						return pItem->GetString();
					
					return NULL;
				}

				int GetInt( int iItemNumber ) 
				{
					Item* pItem = FindItemByNumber( iItemNumber );
					if ( pItem != NULL )
						return pItem->GetInt();
					
					return 0;
				}

				float GetFloat( int iItemNumber ) 
				{
					Item* pItem = FindItemByNumber( iItemNumber );
					if ( pItem != NULL )
						return pItem->GetFloat();

					return 0.0f;
				}

				bool GetBool( int iItemNumber ) 
				{
					Item* pItem = FindItemByNumber( iItemNumber );
					if ( pItem != NULL )
						return pItem->GetBool();

					return false;
				}

				ccColor3B GetColor( int iItemNumberRed, int iItemNumberGreen, int iItemNumberBlue )
				{
					ccColor3B Color;
					Color.r = GetInt( iItemNumberRed );
					Color.g = GetInt( iItemNumberGreen );
					Color.b = GetInt( iItemNumberBlue );
					return Color;
				}

				JpList< Item >* GetItemList() { return &_ItemList; }

			private:
				JpList< Item > _ItemList;
		};


		//-----------------------
		// public functions
		NeConfig() {}
		virtual ~NeConfig();

		bool ReadFile( const char* pacPathAndFileName );
		bool WriteFile( const char* pacPathAndFileName );
		Line* FindLine( const char* pacFirstItemOnLine );
		
		Line* CreateLine( const char* pacFirstItemOnLine );
		Line* FindOrCreateLine( const char* pacFirstItemOnLine );
		void ClearLineKeepingFirstItem( Line* pLine );
		void DeleteLine( Line* pLine );
		void AppendIntToLine( Line* pLine, int i );
		void AppendFloatToLine( Line* pLine, float f );
		void AppendBoolToLine( Line* pLine, bool b );
		void AppendStringToLine( Line* pLine, const char* pacText );

		void Clear();

		Line* FindLineByNumber( int iLineNumber );

		JpList< Line >* GetLineList() { return &_LineList; }


		//-----------------------
		// public variables


	private:

		//-----------------------
		// private variables
		JpList< Line > _LineList;


		//-----------------------
		// private functions
		void ReadAllLinesFsm( NeFile* pFile );
};

