/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

// Some of this was based off of: http://www.ogre3d.org/tikiwiki/ConsoleCode


#pragma once

#include "stdafx.h"

#include <iostream>

#if DCG_IS_TARGET_WIN32
	#include "OIS.h"
#endif

#include "JpHashTable.h"
#include "JpList.h"


class NeConsole;



enum NeConsoleItemType
{
	NECONSOLE_BOOL,
	NECONSOLE_UCHAR,
	NECONSOLE_CHAR,
	NECONSOLE_UINT,
	NECONSOLE_INT,
	NECONSOLE_FLOAT,
	NECONSOLE_STRING,
	NECONSOLE_FUNCTION
};


typedef void (*NeConsoleFunction)( NeConsole* pInstance, const std::vector< std::string > & );

#define NeConsoleItemTable JpHashTable< NeConsoleItem, 256 >


#ifdef _FINAL
	#define CDEC_FLOAT( f, v ) static float f = v;
	#define CSET_FLOAT( f )

	#define CDEC_INT( i, v ) static int i = v;
	#define CSET_INT( i )

	#define CDEC_UINT( ui, v ) static unsigned int ui = v;
	#define CSET_UINT( ui )

	#define CDEC_CHAR( c, v ) static char c = v;
	#define CSET_CHAR( c )

	#define CDEC_UCHAR( uc, v ) static unsigned char uc = v;
	#define CSET_UCHAR( uc )

	#define CDEC_STRING( s, v ) static std::string s = v;
	#define CSET_STRING( s )
#else
	#define CDEC_BOOL( b, v ) static bool b = v; NeConsole::GetInstance()->AddItem( #b, &b, NECONSOLE_BOOL );
	#define CSET_BOOL( b ) NeConsole::GetInstance()->AddItem( #b, &b, NECONSOLE_BOOL );

	#define CDEC_FLOAT( f, v ) static float f = v; NeConsole::GetInstance()->AddItem( #f, &f, NECONSOLE_FLOAT );
	#define CSET_FLOAT( f ) NeConsole::GetInstance()->AddItem( #f, &f, NECONSOLE_FLOAT );

	#define CDEC_INT( i, v ) static int i = v; NeConsole::GetInstance()->AddItem( #i, &i, NECONSOLE_INT );
	#define CSET_INT( i ) NeConsole::GetInstance()->AddItem( #i, &i, NECONSOLE_INT );

	#define CDEC_UINT( ui, v ) static unsigned int ui = v; NeConsole::GetInstance()->AddItem( #ui, &ui, NECONSOLE_UINT );
	#define CSET_UINT( ui ) NeConsole::GetInstance()->AddItem( #ui, &ui, NECONSOLE_UINT );

	#define CDEC_CHAR( c, v ) static char c = v; NeConsole::GetInstance()->AddItem( #c, &c, NECONSOLE_CHAR );
	#define CSET_CHAR( c ) NeConsole::GetInstance()->AddItem( #c, &c, NECONSOLE_CHAR );

	#define CDEC_UCHAR( uc, v ) static unsigned char uc = v; NeConsole::GetInstance()->AddItem( #uc, &uc, NECONSOLE_UCHAR );
	#define CSET_UCHAR( uc ) NeConsole::GetInstance()->AddItem( #uc, &uc, NECONSOLE_UCHAR );

	#define CDEC_STRING( s, v ) static std::string s = v; NeConsole::GetInstance()->AddItem( #s, &s, NECONSOLE_STRING );
	#define CSET_STRING( s ) NeConsole::GetInstance()->AddItem( #s, &s, NECONSOLE_STRING );
#endif


/******************************************************************************
**************************************************  CLASS:  ConsoleItem
******************************************************************************/
typedef struct 
{
	std::string _sName;
	NeConsoleItemType _Type;

	union
	{
		NeConsoleFunction pFunction;
		void *pVariable;
	};
} NeConsoleItem;



/******************************************************************************
**************************************************  CLASS:  NeConsole
******************************************************************************/
class NeConsole
{
	public:
		//-----------------------
		// public functions
		NeConsole() {}
		virtual ~NeConsole() {}
		void Init();
		void Deinit();
		virtual void Update( float dt );

		void AddItem( const char* pacName, void *pPointer, NeConsoleItemType Type );
		void DeleteItem( const char* pacName );

		virtual void AddOutputString( const std::string& s ) = 0;
		virtual void SetCurrentCommandLine( const std::string& s ) = 0;
		virtual void SetCurrentSuggestion( const std::string& s ) = 0;
		virtual void ClearConsole() = 0;


        #if DCG_IS_TARGET_WIN32
			void UpdateInputPressed( OIS::KeyCode eKeyCode );
			void UpdateInputReleased( OIS::KeyCode eKeyCode );

			void UpdateInputPressedForCommand( OIS::KeyCode eKeyCode );
			void UpdateInputPressedForAnswerString( OIS::KeyCode eKeyCode );
        #endif


		void UpdateVirtualKeyboardInputPressed( char cKey );
		void UpdateVirtualKeyboardDeleteBackward();


		void UpdateRepeatBackspace( float dt );

		NeConsoleItemTable* GetItemTable() { return &_ItemTable; }
		void SetDefaultCommand( NeConsoleFunction Def ) { DefaultCommand = Def; }
		void SendMessage( std::string sCommand );

		void Print( std::string sString );
		bool ParseCommandLine( std::string sCommandLine );
		void ParseCommandQueue();
		std::string PadString( std::string sString, int iWidth );
		void SetSuggestion();


		void WaitForUserAnswerString();
		std::string GetUserAnswerString() { return _sUserAnswerString; }
		void ClearUserAnswerString();


		static NeConsole* GetInstance() { return _pInstance; }
		static void SetInstance( NeConsole* p );

		//-----------------------
		// built-in console functions
		static void ErrorFunction( NeConsole* pInstance, const std::vector<std::string> & args );
		static void RemoveCommand( NeConsole* pInstance, const std::vector<std::string> & args );
		static void PrintList( NeConsole* pInstance, const std::vector<std::string> & args );
		static void Clear( NeConsole* pInstance, const std::vector<std::string> & args );
		

	private:

		//-----------------------
		// private functions
		void TokenizeString( std::vector<std::string>* psResults, std::string sToBeTokenized );
		inline void StringToUpper( char* cpString );
		int CustomStrspn( std::string sString1, std::string sPattern );


		//-----------------------
		// private variables
		static NeConsole* _pInstance;

		NeConsoleFunction DefaultCommand;

		NeConsoleItemTable _ItemTable;
		JpList< std::string > _CommandList;
		JpList< std::string > _CommandQueue;
		JpList< std::string > _SuggestionList;


		int   _iCommandListSize;
		int   _iCommandListOffset;
		int   _iSuggestionListOffset;

		bool  _bBackspaceIsDown;
		bool  _bSkipAutoSuggest;

		float _fBackspaceCooldown;

		#if DCG_IS_TARGET_WIN32
			OIS::KeyCode _ePrevInputPressed;
			OIS::KeyCode _eLastInputPressed;
			OIS::KeyCode _eLastKeyReleased;
		#endif


		std::string _sCommand;
		std::string _sOutput;
		std::string _sSuggestionPartial;
		std::string _sSuggestionFull;


		bool _bNextStringIsAUserAnswerNotACommand;
		std::string _sUserAnswerString;
};


