/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

#include "stdafx.h"

#include <stdio.h>

#include "JpUtils.h"
#include "NeCsvIn.h"
#include "NeFile.h"

#include "mmgrChooser.h"




/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void NeCsvIn::Init()
{
	_papCells = NULL;
	_iNumRows = 0;
	_iNumCols = 0;
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void NeCsvIn::Deinit()
{
	_RowIndicesTable.RemoveAll();
	_ColIndicesTable.RemoveAll();

	DeleteAllCells();
}



/******************************************************************************
 FUNCTION:  Clear
******************************************************************************/
void NeCsvIn::Clear()
{
	Deinit();
	Init();
}



/******************************************************************************
 FUNCTION:  DeleteAllCells
******************************************************************************/
void NeCsvIn::DeleteAllCells()
{
	int iNumCells = _iNumRows * _iNumCols;
	for ( int i = 0; i < iNumCells; i ++ )
	{
		if ( _papCells[ i ] )
		{
			delete _papCells[ i ];
			_papCells[ i ] = NULL;
		}
	}


	if ( _papCells )
	{
		delete[] _papCells;
		_papCells = NULL;
	}


	_iNumRows = 0;
	_iNumCols = 0;
}



/******************************************************************************
 FUNCTION:  ReadCsvFile
******************************************************************************/
bool NeCsvIn::ReadCsvFile( const char acPathAndFileName[] )
{
	NeFile File;
	const char* pacFullPathAndFilename = CCFileUtils::sharedFileUtils()->fullPathFromRelativePath( acPathAndFileName );
	bool bSuccess = File.OpenFile( pacFullPathAndFilename, "r" );

	// HACK HACK HACK HACK HACK
	// On android (and possibly on iOS and other platforms), the location of data that is INSTALLED
	// with the app is different than data that is SAVED by the app at runtime.
	// This function is used to read both types of files, and is used differently in each case.
	// In the INSTALLED case, callers expect the incoming path to be altered as needed to find the file.
	// In the SAVED case, callers assume they must provide the exact path, and that the given path is
	// not altered.
	// A proper fix would be to make it so the usage assumptions are the same in both cases.  However,
	// at the current stage of things, it's probably easiest and safest to simply accommodate both uses cases.
	// Thus the below code.  Note that this could lead to a bug if a SAVED file has the same name as an
	// INSTALLED file, but for manhattan, this should not be a problem.
	// mdomowicz 2015_11_30
	if ( !bSuccess )
	{
		bSuccess = File.OpenFile( acPathAndFileName, "r" );
	}

	if ( bSuccess )
	{
		DetermineDimensions( &_iNumRows, &_iNumCols, &File );

		_papCells = CreateCells( _iNumRows, _iNumCols );

		ReadCells( &File );

		File.CloseFile();

		return true;
	}

	return false;
}



/******************************************************************************
 FUNCTION:  MakeHashIndicesFromColValuesAtRow
******************************************************************************/
void NeCsvIn::MakeHashIndicesFromColValuesAtRow( int iRow )
{
	for ( int iCol = 0; iCol < _iNumCols; iCol ++ )
	{
		Cell* pCell = GetCell( iRow, iCol );

		if ( pCell )
		{
			// note, hash tables only store pointers, but since all we want to do
			// is store a single integer, we just cast to pretend it's a pointer.
			_ColIndicesTable.Store( JpHash( pCell->GetString() ), (int*) iCol );
		}
	}
}



/******************************************************************************
 FUNCTION:  MakeHashIndicesFromRowValuesAtCol
******************************************************************************/
void NeCsvIn::MakeHashIndicesFromRowValuesAtCol( int iCol )
{
	for ( int iRow = 0; iRow < _iNumRows; iRow ++ )
	{
		Cell* pCell = GetCell( iRow, iCol );

		if ( pCell )
		{
			// note, hash tables only store pointers, but since all we want to do
			// is store a single integer, we just cast to pretend it's a pointer.
			_RowIndicesTable.Store( JpHash( pCell->GetString() ), (int*) iRow );
		}
	}
}



/******************************************************************************
 FUNCTION:  DetermineDimensions
******************************************************************************/
void NeCsvIn::DetermineDimensions( int* piNumRows, int* piNumCols, NeFile* pFile )
{
	pFile->Seek( 0, SEEK_SET );

	int iNumRows = 0;
	int iNumCols = 0;

	bool bIsAtEof = false;

	while ( bIsAtEof == false )
	{
		//-----------------------
		// read off the stream until
		// we hit the first non
		// white space character.
		char c = pFile->GetChar();
		while ( IsWhitespace( c ) )
			c = pFile->GetChar();


		//-----------------------
		// note if we are at a new line
		// or at end of file.
		bIsAtEof = ( c == (char)EOF );

		bool bRowTallied = false;
		int iNumColsInThisLine = 0;
		while ( c != '\n' && c != (char)EOF )
		{
			// only tally a row if the line has something on it.
			if ( !bRowTallied )
			{
				bRowTallied = true;
				iNumRows ++;
			}


			if ( c == ',' )
				iNumColsInThisLine ++;

			c = pFile->GetChar();

			if ( c == '\n' || c == (char)EOF )
				iNumColsInThisLine ++;
		}


		if ( iNumColsInThisLine > iNumCols )
			iNumCols = iNumColsInThisLine;
	}


	*piNumRows = iNumRows;
	*piNumCols = iNumCols;	
}



/******************************************************************************
 FUNCTION:  ReadCells
******************************************************************************/
void NeCsvIn::ReadCells( NeFile* pFile )
{

	pFile->Seek( 0, SEEK_SET );

	char acString[ MAX_CELL_STRING_LENGTH ];

	int iX = 0;
	int iY = 0;

	hash uiState = 0xa1357080 /* "EatWhiteSpace" */;
	char c = 0;


	//-----------------------
	// A state machine for reading the data
	// into our table of cells

	while ( c != (char)EOF )
	{
		switch ( uiState )
		{
			case 0xa1357080 /* "EatWhiteSpace" */:
			{
				c = pFile->GetChar();

				// handles the case where the first cell is blank
				if ( c == ',' ) 
				{
					iX ++;
					uiState = 0x65a9acc2 /* "HandleComma" */;
				}
				else if ( !IsWhitespace( c ) )
					uiState = 0x994a02b5 /* "ReadCell" */;

				break;
			}


			case 0x994a02b5 /* "ReadCell" */:
			{
				memset( acString, 0, MAX_CELL_STRING_LENGTH );

				int iNumChars = 0;

				//-----------------------
				// read as much of the cell as possible
				while 
				( 
					iNumChars < MAX_CELL_STRING_LENGTH && 
					c != ',' &&
					c != '\n' &&

					#if !DCG_IS_TARGET_WIN32
						c != '\r' &&
					#endif

					c != (char)EOF
				)
				{
					acString[ iNumChars ] = c;
					c = pFile->GetChar();
					iNumChars ++;
				}
				

				//-----------------------
				// if we successfully read a string, create a new cell
				if ( acString[ 0 ] != 0 )
				{
					Cell* pCell = new Cell;
					pCell->SetString( acString );

					ASSERTF( iY < _iNumRows && iX < _iNumCols, ( "\nCell indices are out of range!" ) );

					if ( iY < _iNumRows && iX < _iNumCols )
						_papCells[ ( iY * _iNumCols ) + iX ] = pCell;


					iX ++;
				}


				//-----------------------
				// choose a next state depending on how the cell reading
				// was terminated.
				if ( iNumChars >= MAX_CELL_STRING_LENGTH )
				{
					acString[ MAX_CELL_STRING_LENGTH - 1 ] = 0;
					uiState = 0x94a4ab4c /* "HandleCellOverage" */;
				}
				else if ( c == ',' )
					uiState = 0x65a9acc2 /* "HandleComma" */;
				else if ( c == '\n' )
					uiState = 0x0a5c2929 /* "HandleNewLine" */;

				#if !DCG_IS_TARGET_WIN32
					else if ( c == '\r' )
						uiState = 0x0a5c2929 /* "HandleNewLine" */;
				#endif

				break;
			}


			case 0x94a4ab4c /* "HandleCellOverage" */:
			{
				if ( c == ',' )
					uiState = 0x65a9acc2 /* "HandleComma" */;
				else if ( c == '\n' )
					uiState = 0x0a5c2929 /* "HandleNewLine" */;
				else
					c = pFile->GetChar();
				
				break;
			}

			case 0x65a9acc2 /* "HandleComma" */:
			{
				c = pFile->GetChar();

				if ( c == ',' )
					iX ++;

				else if ( c == '\n' )
					uiState = 0x0a5c2929 /* "HandleNewLine" */;

				else if ( !IsWhitespace( c ) )
					uiState = 0x994a02b5 /* "ReadCell" */;

				break;
			}

			case 0x0a5c2929 /* "HandleNewLine" */:
			{
				iX = 0;
				iY ++;
				uiState = 0xa1357080 /* "EatWhiteSpace" */;
				break;
			}

			default:	
				ASSERTF( 0, ( "\nUnknown State" ) );
		}

	}
}



/******************************************************************************
 FUNCTION:  IsWhitespace
 RETURN:    true, if the given character is whitespace.
******************************************************************************/
bool NeCsvIn::IsWhitespace( char c )
{
	bool bIsWhiteSpace = 
	(
		#if !DCG_IS_TARGET_WIN32
			c == '\r' ||
		#endif

		c == ' ' ||
		c == '\n' ||
		c == '\t' 
	);

	return bIsWhiteSpace;
}



/******************************************************************************
 FUNCTION:  CreateCells
******************************************************************************/
NeCsvIn::Cell** NeCsvIn::CreateCells( int iNumRows, int iNumCols )
{
	NeCsvIn::Cell** papCells = NULL;
	
	if ( iNumRows > 0 && iNumCols > 0 )
	{
		int iNumCells = iNumRows * iNumCols;
		papCells = new Cell* [ iNumCells ];

		for ( int i = 0; i < iNumCells; i ++ )
			papCells[ i ] = NULL;
	}

	return papCells;
}



/******************************************************************************
 FUNCTION:  GetCell
******************************************************************************/
NeCsvIn::Cell* NeCsvIn::GetCell( int iRow, int iCol )
{
	if ( _papCells )
		return _papCells[ ( iRow * _iNumCols ) + iCol ];

	return NULL;
}	



/******************************************************************************
 FUNCTION:  GetCellByHashIndices
******************************************************************************/
NeCsvIn::Cell* NeCsvIn::GetCellByHashIndices
( 
	hash hRowHash, 
	hash hColHash 
)
{
	if ( _RowIndicesTable.Exists( hRowHash ) && _ColIndicesTable.Exists( hColHash ) )
	{
		// note: we hackily stored row/col integers in the hash table as int*
		// since hash tables only take pointers.  Thus we have to cast back to
		// int here.
		int iRow = (int)_RowIndicesTable.Find( hRowHash );
		int iCol = (int)_ColIndicesTable.Find( hColHash );
		
		return GetCell( iRow, iCol );	
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  GetCellByIntAndHash
******************************************************************************/
NeCsvIn::Cell* NeCsvIn::GetCellByIntAndHash( int iRow, hash hColHash )
{
	if ( _ColIndicesTable.Exists( hColHash ) )
	{
		// see GetCellByHashIndices for a note on the cast here.
		int iCol = (int)_ColIndicesTable.Find( hColHash );
		return GetCell( iRow, iCol );	
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  GetCellByHashAndInt
******************************************************************************/
NeCsvIn::Cell* NeCsvIn::GetCellByHashAndInt( hash hRowHash, int iCol )
{
	if ( _RowIndicesTable.Exists( hRowHash ) )
	{
		// see GetCellByHashIndices for a note on the cast here.
		int iRow = (int)_RowIndicesTable.Find( hRowHash );
		return GetCell( iRow, iCol );	
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  GetCellByTextIndices
******************************************************************************/
NeCsvIn::Cell* NeCsvIn::GetCellByTextIndices( char* pacRowText, char* pacColText ) 
{ 
	return GetCellByHashIndices( JpHash( pacRowText ), JpHash( pacColText ) ); 
}



/******************************************************************************
 FUNCTION:  GetString
******************************************************************************/
char* NeCsvIn::GetString( int iRow, int iCol )
{
	Cell* pCell = GetCell( iRow, iCol );
	if ( pCell )
		return pCell->GetString();

	return NULL;
}	



/******************************************************************************
 FUNCTION:  GetStringByHashIndices
******************************************************************************/
char* NeCsvIn::GetStringByHashIndices
( 
	hash hRowHash, 
	hash hColHash
)
{
	Cell* pCell = GetCellByHashIndices( hRowHash, hColHash );
	if ( pCell )
		return pCell->GetString();

	return NULL;
}



/******************************************************************************
 FUNCTION:  GetStringByIntAndHash
******************************************************************************/
char* NeCsvIn::GetStringByIntAndHash( int iRow, hash hColHash )
{
	Cell* pCell = GetCellByIntAndHash( iRow, hColHash );
	if ( pCell )
		return pCell->GetString();

	return NULL;
}



/******************************************************************************
 FUNCTION:  GetStringByHashAndInt
******************************************************************************/
char* NeCsvIn::GetStringByHashAndInt( hash hRowHash, int iCol )
{
	Cell* pCell = GetCellByHashAndInt( hRowHash, iCol );
	if ( pCell )
		return pCell->GetString();

	return NULL;
}



/******************************************************************************
 FUNCTION:  GetStringByTextIndices
******************************************************************************/
char* NeCsvIn::GetStringByTextIndices
( 
	char* pacRowText, 
	char* pacColText
)
{
	Cell* pCell = GetCellByTextIndices( pacRowText, pacColText );
	if ( pCell )
		return pCell->GetString();

	return NULL;
}



/******************************************************************************
 FUNCTION:  GetInt
******************************************************************************/
int NeCsvIn::GetInt( int iRow, int iCol, int iReturnOnError )
{
	Cell* pCell = GetCell( iRow, iCol );
	if ( pCell )
		return pCell->GetInt();

	return iReturnOnError;
}	



/******************************************************************************
 FUNCTION:  GetIntByHashIndices
******************************************************************************/
int NeCsvIn::GetIntByHashIndices
( 
	hash hRowHash, 
	hash hColHash, 
	int iReturnOnError 
)
{
	Cell* pCell = GetCellByHashIndices( hRowHash, hColHash );
	if ( pCell )
		return pCell->GetInt();

	return iReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetIntByIntAndHash
******************************************************************************/
int NeCsvIn::GetIntByIntAndHash( int iRow, hash hColHash, int iReturnOnError )
{
	Cell* pCell = GetCellByIntAndHash( iRow, hColHash );
	if ( pCell )
		return pCell->GetInt();

	return iReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetIntByHashAndInt
******************************************************************************/
int NeCsvIn::GetIntByHashAndInt( hash hRowHash, int iCol, int iReturnOnError )
{
	Cell* pCell = GetCellByHashAndInt( hRowHash, iCol );
	if ( pCell )
		return pCell->GetInt();

	return iReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetIntByTextIndices
******************************************************************************/
int NeCsvIn::GetIntByTextIndices
( 
	char* pacRowText, 
	char* pacColText, 
	int iReturnOnError 
)
{
	Cell* pCell = GetCellByTextIndices( pacRowText, pacColText );
	if ( pCell )
		return pCell->GetInt();

	return iReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetHash
******************************************************************************/
hash NeCsvIn::GetHash( int iRow, int iCol, hash hReturnOnError )
{
	Cell* pCell = GetCell( iRow, iCol );
	if ( pCell )
		return pCell->GetHash();

	return hReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetHashByHashIndices
******************************************************************************/
hash NeCsvIn::GetHashByHashIndices( hash hRowHash, hash hColHash, hash hReturnOnError )
{
	Cell* pCell = GetCellByHashIndices( hRowHash, hColHash );
	if ( pCell )
		return pCell->GetHash();

	return hReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetHashByIntAndHash
******************************************************************************/
hash NeCsvIn::GetHashByIntAndHash( int iRow, hash hColHash, hash hReturnOnError )
{
	Cell* pCell = GetCellByIntAndHash( iRow, hColHash );
	if ( pCell )
		return pCell->GetHash();

	return hReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetHashByHashAndInt
******************************************************************************/
hash NeCsvIn::GetHashByHashAndInt( hash hRowHash, int iCol, hash hReturnOnError )
{
	Cell* pCell = GetCellByHashAndInt( hRowHash, iCol );
	if ( pCell )
		return pCell->GetHash();

	return hReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetHashByTextIndices
******************************************************************************/
hash NeCsvIn::GetHashByTextIndices( char* pacRowText, char* pacColText, hash hReturnOnError )
{
	Cell* pCell = GetCellByTextIndices( pacRowText, pacColText );
	if ( pCell )
		return pCell->GetHash();

	return hReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetFloat
******************************************************************************/
float NeCsvIn::GetFloat( int iRow, int iCol, float fReturnOnError )
{
	Cell* pCell = GetCell( iRow, iCol );
	if ( pCell )
		return pCell->GetFloat();

	return fReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetFloatByHashIndices
******************************************************************************/
float NeCsvIn::GetFloatByHashIndices
( 
	hash hRowHash, 
	hash hColHash, 
	float fReturnOnError 
)
{
	Cell* pCell = GetCellByHashIndices( hRowHash, hColHash );
	if ( pCell )
		return pCell->GetFloat();

	return fReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetFloatByIntAndHash
******************************************************************************/
float NeCsvIn::GetFloatByIntAndHash( int iRow, hash hColHash, float fReturnOnError )
{
	Cell* pCell = GetCellByIntAndHash( iRow, hColHash );
	if ( pCell )
		return pCell->GetFloat();

	return fReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetFloatByHashAndInt
******************************************************************************/
float NeCsvIn::GetFloatByHashAndInt( hash hRowHash, int iCol, float fReturnOnError )
{
	Cell* pCell = GetCellByHashAndInt( hRowHash, iCol );
	if ( pCell )
		return pCell->GetFloat();

	return fReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetFloatByTextIndices
******************************************************************************/
float NeCsvIn::GetFloatByTextIndices
( 
	char* pacRowText, 
	char* pacColText, 
	float fReturnOnError 
)
{
	Cell* pCell = GetCellByTextIndices( pacRowText, pacColText );
	if ( pCell )
		return pCell->GetFloat();

	return fReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetBool
******************************************************************************/
bool NeCsvIn::GetBool( int iRow, int iCol, bool bReturnOnError )
{
	Cell* pCell = GetCell( iRow, iCol );
	if ( pCell )
		return pCell->GetBool();

	return bReturnOnError;
}	



/******************************************************************************
 FUNCTION:  GetBoolByHashIndices
******************************************************************************/
bool NeCsvIn::GetBoolByHashIndices
( 
	hash hRowHash, 
	hash hColHash, 
	bool bReturnOnError 
)
{
	Cell* pCell = GetCellByHashIndices( hRowHash, hColHash );
	if ( pCell )
		return pCell->GetBool();

	return bReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetBoolByIntAndHash
******************************************************************************/
bool NeCsvIn::GetBoolByIntAndHash( int iRow, hash hColHash, bool bReturnOnError )
{
	Cell* pCell = GetCellByIntAndHash( iRow, hColHash );
	if ( pCell )
		return pCell->GetBool();

	return bReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetBoolByHashAndInt
******************************************************************************/
bool NeCsvIn::GetBoolByHashAndInt( hash hRowHash, int iCol, bool bReturnOnError )
{
	Cell* pCell = GetCellByHashAndInt( hRowHash, iCol );
	if ( pCell )
		return pCell->GetBool();

	return bReturnOnError;
}



/******************************************************************************
 FUNCTION:  GetBoolByTextIndices
******************************************************************************/
bool NeCsvIn::GetBoolByTextIndices
( 
	char* pacRowText, 
	char* pacColText, 
	bool bReturnOnError 
)
{
	Cell* pCell = GetCellByTextIndices( pacRowText, pacColText );
	if ( pCell )
		return pCell->GetBool();

	return bReturnOnError;
}
