/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#pragma once

#include "stdafx.h"

#include "string.h"
#include "JpHashTable.h"


#define MAX_CELL_STRING_LENGTH  90

#define MAX_HASH_INDICES_ROW 8192
#define MAX_HASH_INDICES_COL 64


/*
	This class is for reading in CSV (Comma Separated Values) files.
	This is often useful when tables of data must be read in since most spreadsheet
	programs support import/edit/export of CSV

	Notes:

		1. This class creates a 2d map of Cell pointers.  

		2. This class is two-pass.  Pass 1 gets the Cell map dimensions. Pass 2 gets the cell values.

		3. Sparse tables are memory inefficient.  In example, we have a 3x5 table.  15 pointers
		   will be created, even though only 7 cells have data.

			   1, 2, 3, 4, 5
			   a
			   b

		4. Tables are kept in memory as long as the NeCsv object exists.

		5. The first row is 0, the first col is 0.

		6. Using MakeHashIndicesFromColValuesAtRow and its partner function, a single row and a
		   single col can be used to create indices from the values in them.  So with the table

			   0,     cat, dog, ferret, gerbil
			   red,   1,     2,      3,      4
			   blue,  5,     6,      7,      8
			   black, 9,    10,     11,     12

		   a user could look up the value for ("blue", "ferret"), and get 7.
		   Alternatively, the user could still use the absolute indexes, (2, 3) to get 7.



	Usage Example:

		Assume Test.csv looks like this:

			a, 2, 3, 4, , sdf, , 5, now is the time for all good men to come to the aid of their country the big brown fox jumped over the lazy dog
			b, 2, 3, 4, 5, 6, 7, 8


			c, 2, 3


			d 

			e


		Here is code that uses Test.csv:

			NeCsvIn MyCsv;
			MyCsv.ReadCsvFile( "Test.csv" );

			MyCsv.MakeHashIndicesFromColValuesAtRow( 0 );
			MyCsv.MakeHashIndicesFromRowValuesAtCol( 0 );

			NeCsv::Cell* pCell = NULL;
			pCell = MyCsv.GetCellByHashIndices( JpHash("b"), JpHash("sdf") );
			float fVal = pCell->GetFloat(); 

			// fVal == 6.0f


			pCell = MyCsv.GetCell( 1, 5 );
			int iVal = pCell->GetInt(); 

			// iVal == 6

*/

class NeFile;

class NeCsvIn
{
	public:

		//-----------------------
		// public classes and enums
		class Cell
		{
			public:
				Cell() 
				{
					_acString[ 0 ] = 0;
				}

				void SetString( char* pacText )
				{
					strncpy( _acString, pacText, MAX_CELL_STRING_LENGTH );
					_acString[ MAX_CELL_STRING_LENGTH - 1 ] = 0;
				}

				char* GetString() { return _acString; }

				int GetInt() 
				{
					int i;
					int iErrorCode = sscanf( GetString(), "%i", &i );
					JP_UNUSED_VAR( iErrorCode );
					ASSERTF( iErrorCode != EOF, ( "\nerror converting string" ) );
					return i;					
				}

				hash GetHash() 
				{
					hash h;
					int iErrorCode = sscanf( GetString(), "%u", &h );
					JP_UNUSED_VAR( iErrorCode );
					ASSERTF( iErrorCode != EOF, ( "\nerror converting string" ) );
					return h;					
				}

				float GetFloat() 
				{
					float f;
					int iErrorCode = sscanf( GetString(), "%f", &f );
					JP_UNUSED_VAR( iErrorCode );
					ASSERTF( iErrorCode != EOF, ( "\nerror converting string" ) );
					return f;					
				}

				bool GetBool() 
				{
					if ( JpHash( GetString() ) == 0x0b069958 /* "false" */ )
						return false;
					else if ( JpHash( GetString() ) == 0x4db211e5 /* "true" */ )
						return true;

					int i = GetInt();
					return ( i != 0 );
				}

			private:
				char _acString[ MAX_CELL_STRING_LENGTH ];
		};




		//-----------------------
		// public functions
		NeCsvIn() { Init(); }

		~NeCsvIn() { Deinit(); }

		void Init();

		void Deinit();

		bool IsClear() { return _papCells == NULL; }
		void Clear();

		bool ReadCsvFile( const char acPathAndFileName[] );

		void DetermineDimensions( int* piNumRows, int* piNumCols, NeFile* pFile );

		int GetNumRows() { return _iNumRows; }
		int GetNumCols() { return _iNumCols; }

		void MakeHashIndicesFromColValuesAtRow( int iRow );
		void MakeHashIndicesFromRowValuesAtCol( int iCol );

		bool RowHashIndexExists( hash hRowHash ) { return _RowIndicesTable.Exists( hRowHash ); }
		bool ColHashIndexExists( hash hColHash ) { return _ColIndicesTable.Exists( hColHash ); }

		Cell* GetCell( int iRow, int iCol );
		Cell* GetCellByHashIndices( hash hRowHash, hash hColHash );
		Cell* GetCellByIntAndHash( int iRow, hash hColHash );
		Cell* GetCellByHashAndInt( hash hRowHash, int iCol );
		Cell* GetCellByTextIndices( char* pacRowText, char* pacColText );


		char* GetString( int iRow, int iCol );
		char* GetStringByHashIndices( hash hRowHash, hash hColHash );
		char* GetStringByIntAndHash( int iRow, hash hColHash );
		char* GetStringByHashAndInt( hash hRowHash, int iCol );
		char* GetStringByTextIndices( char* pacRowText, char* pacColText );

		int GetInt( int iRow, int iCol, int iReturnOnError );
		int GetIntByHashIndices( hash hRowHash, hash hColHash, int iReturnOnError );
		int GetIntByIntAndHash( int iRow, hash hColHash, int iReturnOnError );
		int GetIntByHashAndInt( hash hRowHash, int iCol, int iReturnOnError );
		int GetIntByTextIndices( char* pacRowText, char* pacColText, int iReturnOnError );

		hash GetHash( int iRow, int iCol, hash hReturnOnError );
		hash GetHashByHashIndices( hash hRowHash, hash hColHash, hash hReturnOnError );
		hash GetHashByIntAndHash( int iRow, hash hColHash, hash hReturnOnError );
		hash GetHashByHashAndInt( hash hRowHash, int iCol, hash hReturnOnError );
		hash GetHashByTextIndices( char* pacRowText, char* pacColText, hash hReturnOnError );

		float GetFloat( int iRow, int iCol, float fReturnOnError );
		float GetFloatByHashIndices( hash hRowHash, hash hColHash, float fReturnOnError );
		float GetFloatByIntAndHash( int iRow, hash hColHash, float fReturnOnError );
		float GetFloatByHashAndInt( hash hRowHash, int iCol, float fReturnOnError );
		float GetFloatByTextIndices( char* pacRowText, char* pacColText, float fReturnOnError );

		bool GetBool( int iRow, int iCol, bool bReturnOnError );
		bool GetBoolByHashIndices( hash hRowHash, hash hColHash, bool bReturnOnError );
		bool GetBoolByIntAndHash( int iRow, hash hColHash, bool bReturnOnError );
		bool GetBoolByHashAndInt( hash hRowHash, int iCol, bool bReturnOnError );
		bool GetBoolByTextIndices( char* pacRowText, char* pacColText, bool bReturnOnError );


		void DeleteAllCells();

	private:


		//-----------------------
		// private variables
		int _iNumRows;
		int _iNumCols;

		Cell** _papCells;



		//-----------------------
		// private functions

		void ReadCells( NeFile* pFile );
	
		bool IsWhitespace( char c );

		Cell** CreateCells( int iNumRows, int iNumCols );


		JpHashTable< int, MAX_HASH_INDICES_ROW * 2 > _RowIndicesTable;
		JpHashTable< int, MAX_HASH_INDICES_COL * 2 > _ColIndicesTable;
};



