/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/



#pragma once

#include "stdafx.h"
#include "cocos2d.h"

#include "JpList.h"
#include "JpRandom.h"

USING_NS_CC;


/*
	This class is for creating a pool of candidate values and then randomly picking one value
	from that pool.  Example usage:

		SoRandomPoolSelector PoolSelector;
		PoolSelector.AddValueInt( (int)SUB_TYPE_A, 22 );
		PoolSelector.AddValueInt( (int)SUB_TYPE_B, 10 );

		SUB_TYPE eSubType = (SUB_TYPE)PoolSelector.SelectValueFromPool();


	In the calls to AddValueInt(), the first parameter is the value being added to the pool.
	Later calls to PoolSelector.SelectValueFromPool() will return one of those values. 

	The second parameter to AddValueInt is a pool "count".  That is, it's the number of times
	the value is "added" to the pool.  Adjusting the the count equates to adjusting the relative
	probability that the corresponding value will be selected.  In the above example, SUB_TYPE_A
	is slightly more than two times as likely to be selected than SUB_TYPE_B, because there are 
	slightly more that 2x the number of SUB_TYPE_A values in the pool.

	Note 1: As you can see in the implementation, this class doesn't actually add "count" number
	of values to the pool.  So in the example above, there are technically only two entries
	in the class.  There are NOT 22 separate entries for SUB_TYPE_A, and 10 for SUB_TYPE_B.
	However, the probability results will act AS IF there was indeed a pool of 32 values to
	select from.

	Note 2: When a value is selected by a call to SelectValueFromPool() the value is NOT
	removed.  The pool is not altered in any way.  In the above example, the probably of 
	selecting SUB_TYPE_A would always be 22/32, not matter how many times either value was
	return to the user.

	mdomowicz 2011_09_01
*/

class NeRandomPoolPicker
{
	public:

		//-----------------------
		// public classes

		class Entry
		{
			public:
				Entry() 
				{
					_iValue = 0;
					_iRangeMin = 0;
					_iRangeMax = 0;
				}

				~Entry() {}

				void SetRange( int iRangeMin, int iRangeMax )
				{
					_iRangeMin = iRangeMin;
					_iRangeMax = iRangeMax;
				}

				bool IsInRange( int i ) { return ( i >= _iRangeMin && i < _iRangeMax ); }

				int GetIntValue() { return _iValue; }
				void SetIntValue( int i ) { _iValue = i; }

			private:
				int _iValue;
				int _iRangeMin;
				int _iRangeMax;
		};


		//-----------------------
		// public functions
		NeRandomPoolPicker() { _iCurrentMax = 0; }

		virtual ~NeRandomPoolPicker() { _EntryList.DeleteAll(); }

		void AddValueInt( int iValue, int iCount );

		int SelectValueFromPool( JpRandom* pRng );

		int GetNumEntries() { return _EntryList.GetNumEntries(); }

	private:

		//-----------------------
		// private variables
		int _iCurrentMax;

		JpList< Entry > _EntryList;
};