/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#if DCG_IS_TARGET_WIN32
	#include "windows.h"
#elif DCG_IS_TARGET_IOS
	#include "unistd.h"
#endif


#include "fmod.hpp"
#include "fmod_errors.h"

#include "JpUtils.h"

#include "NeSoundEffect.h"
#include "NeSoundManager.h"

#include "mmgrChooser.h"




/******************************************************************************
 FUNCTION:  NeSoundManager
******************************************************************************/
NeSoundManager::NeSoundManager() :	
_pFmodSystem( NULL ),
_pMusicTrack( NULL ),
_pMusicTrackChannel( NULL ),
_pucMusicTrackDataForAndroid( NULL )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void NeSoundManager::Init()
{
	#if DCG_IS_TARGET_IOS || DCG_IS_TARGET_ANDROID

		FMOD_RESULT      result; 
		unsigned int     version;
	

		result = FMOD::System_Create(&_pFmodSystem); 
		ERRCHECK(result);
	
		result = _pFmodSystem->getVersion(&version);
		ERRCHECK(result);
	
		if (version < FMOD_VERSION)
		{
			fprintf(stderr, "You are using an old version of FMOD %08x.  This program requires %08x\n", version, FMOD_VERSION);
			exit(-1);
		}

		#if DCG_IS_TARGET_ANDROID					
			result = _pFmodSystem->init(32, FMOD_INIT_NORMAL, NULL);
		#else
			result = _pFmodSystem->init(32, FMOD_INIT_NORMAL | FMOD_INIT_ENABLE_PROFILE, NULL);
		#endif
	
		NeSoundEffect::SetFmodSoundSystem( _pFmodSystem );

	#else
	
		//-----------------------
		// Most of this code comes from from the API document, 
		// "Getting Started with FMOD for Windows.pdf".  The doc
		// insisted that the windows code this way.  
		//
		// In their words:
		// "Note: The following code must be used for shipping games! 
		// Do not ship a game without a startup sequence based on 
		// this code!"
		//
		// mdomowicz 2011_05_25

		FMOD_RESULT      result; 
		unsigned int     version; 
		int              numdrivers; 
		FMOD_SPEAKERMODE speakermode; 
		FMOD_CAPS        caps; 
		char             name[256]; 
		 
		/* 
			Create a System object and initialize. 
		*/ 
		result = FMOD::System_Create(&_pFmodSystem); 
		ERRCHECK(result); 
		 
		result = _pFmodSystem->getVersion(&version); 
		ERRCHECK(result); 
		 
		if (version < FMOD_VERSION) 
		{ 
			printf("Error!  You are using an old version of FMOD %08x.  This program requires %08x\n", version, FMOD_VERSION); 
			return; 
		} 
		 
		result = _pFmodSystem->getNumDrivers(&numdrivers); 
		ERRCHECK(result); 
		 
		if (numdrivers == 0) 
		{ 
			result = _pFmodSystem->setOutput(FMOD_OUTPUTTYPE_NOSOUND); 
			ERRCHECK(result); 
		} 
		else 
		{ 
			result = _pFmodSystem->getDriverCaps(0, &caps, 0, 0, &speakermode); 
			ERRCHECK(result); 
		 
			/* 
				Set the user selected speaker mode. 
			*/ 
			result = _pFmodSystem->setSpeakerMode(speakermode); 
			ERRCHECK(result); 
		 
			if (caps & FMOD_CAPS_HARDWARE_EMULATED) 
			{ 
				/* 
					The user has the 'Acceleration' slider set to off!  This is really bad  
					for latency! You might want to warn the user about this. 
				*/ 
				result = _pFmodSystem->setDSPBufferSize(1024, 10); 
				ERRCHECK(result); 
			} 
		 
			result = _pFmodSystem->getDriverInfo(0, name, 256, 0); 
			ERRCHECK(result); 
		 
			if (strstr(name, "SigmaTel")) 
			{ 
				/* 
					Sigmatel sound devices crackle for some reason if the format is PCM 16bit. 
					PCM floating point output seems to solve it. 
				*/ 
				result = _pFmodSystem->setSoftwareFormat(48000, FMOD_SOUND_FORMAT_PCMFLOAT, 0,0, FMOD_DSP_RESAMPLER_LINEAR); 
				ERRCHECK(result); 
			} 
		} 

		result = _pFmodSystem->init(100, FMOD_INIT_NORMAL, 0); 
		if (result == FMOD_ERR_OUTPUT_CREATEBUFFER) 
		{ 
			/* 
				Ok, the speaker mode selected isn't supported by this soundcard.  Switch it  
				back to stereo... 
			*/ 
			result = _pFmodSystem->setSpeakerMode(FMOD_SPEAKERMODE_STEREO); 
			ERRCHECK(result); 
		 
			/* 
				... and re-init. 
			*/ 
			result = _pFmodSystem->init(100, FMOD_INIT_NORMAL, 0); 
			ERRCHECK(result); 
		}


		NeSoundEffect::SetFmodSoundSystem( _pFmodSystem );

	#endif
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void NeSoundManager::Deinit()
{
	UnloadMusicTrack();
	UnloadAllSoundEffects();

	FMOD_RESULT result;

	result = _pFmodSystem->close();
	ERRCHECK(result);
	result = _pFmodSystem->release();
	ERRCHECK(result);

}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void NeSoundManager::Update( float dt )
{
	// API (fmod.hpp) says this should be called every frame.
	_pFmodSystem->update();
}



/******************************************************************************
 FUNCTION:  ERRCHECK
 NOTES:     This comes from the Fmod docs
******************************************************************************/
void NeSoundManager::ERRCHECK( FMOD_RESULT result )
{
	if (result != FMOD_OK)
	{
		printf("FMOD error! (%d) %s\n", result, FMOD_ErrorString(result));
		assert(false);
	}
}



/******************************************************************************
 FUNCTION:  LoadMusicTrack
******************************************************************************/
void NeSoundManager::LoadMusicTrack( const char* pacPathAndFileName )
{
	// FMOD doesn't seem to like .mp3 files on iPad1.  I noticed these problems
	// in ManhattanC.  It will load .mp3 music very very slowly (so slow, it looks
	// like the app crashed), and often it will crash when loading mp3 sound
	// effects.  Thus this check and warning.
	// mdomowicz 2013_08_18
	char acResult[ 128 ] = "";
	JpUtils::GetExtension( (char*)pacPathAndFileName, acResult );

	bool bIsMp3 = ( JpHash( acResult ) == JpHash( "mp3" ) );
	ASSERTF( !bIsMp3, ( "\nAvoid using mp3s.  FMOD is slow and crashy when using mp3's on iPad 1." ) );


	#if DCG_IS_TARGET_ANDROID

		/*
			From researching the fmod forums, and from a conversation with Matthew B at Firelight,
			I've learned that FMOD requires special handling to load audio data on Android.  Basically,
			if you're not loading data off the SDCard, then you have to use the "load from memory"
			technique, as see in the loadfrommemory example included with the Android API.
			mdomowicz 2013_10_31
		*/

		unsigned long DataSizeResult = 0;

		const char* pFullPathAndFilename = CCFileUtils::sharedFileUtils()->fullPathFromRelativePath( pacPathAndFileName );
		_pucMusicTrackDataForAndroid = CCFileUtils::sharedFileUtils()->getFileData( pFullPathAndFilename, "rb", &DataSizeResult );

		// Copied from fmod example, "loadfrommemory"
		FMOD_CREATESOUNDEXINFO exinfo;
		memset( &exinfo, 0, sizeof( FMOD_CREATESOUNDEXINFO ) );
		exinfo.cbsize = sizeof(FMOD_CREATESOUNDEXINFO);
		exinfo.length = DataSizeResult;

		FMOD_RESULT result;

		result = _pFmodSystem->createStream
		( 
			(const char*)_pucMusicTrackDataForAndroid, 
			FMOD_HARDWARE | FMOD_LOOP_NORMAL | FMOD_OPENMEMORY, 
			&exinfo,
			&_pMusicTrack
		);

		_pMusicTrack->setLoopCount( -1 );

	#else
		FMOD_RESULT result;

		result = _pFmodSystem->createStream
		( 
			CCFileUtils::sharedFileUtils()->fullPathFromRelativePath( pacPathAndFileName ), 
			FMOD_HARDWARE | FMOD_LOOP_NORMAL, 
			NULL, // FMOD_CREATESOUNDEXINFO *exinfo
			&_pMusicTrack
		);

		_pMusicTrack->setLoopCount( -1 );
	#endif

	ERRCHECK(result);
}



/******************************************************************************
 FUNCTION:  UnloadMusicTrack
******************************************************************************/
void NeSoundManager::UnloadMusicTrack()
{
	if ( _pMusicTrack )
	{
		FMOD_RESULT result;
		result = _pMusicTrack->release();
		ERRCHECK(result);
		
		_pMusicTrack = NULL;
		_pMusicTrackChannel = NULL;

		if ( _pucMusicTrackDataForAndroid )
		{
			delete _pucMusicTrackDataForAndroid;
			_pucMusicTrackDataForAndroid = NULL;
		}
	}
}



/******************************************************************************
 FUNCTION:  PlayMusic
******************************************************************************/
void NeSoundManager::PlayMusic()
{
	FMOD_RESULT result;
	
	if ( _pFmodSystem && _pMusicTrack )
	{
		result = _pFmodSystem->playSound
		(
			FMOD_CHANNEL_FREE, 
			_pMusicTrack, 
			false, // bPaused
			&_pMusicTrackChannel
		);

		ERRCHECK(result);
	}
}



/******************************************************************************
 FUNCTION:  StopMusic
******************************************************************************/
void NeSoundManager::StopMusic()
{
	if ( _pMusicTrackChannel )
	{
		FMOD_RESULT result;
		result = _pMusicTrackChannel->stop();
		ERRCHECK(result);
	}
}



/******************************************************************************
 FUNCTION:  PauseMusic
******************************************************************************/
void NeSoundManager::PauseMusic( bool bPause )
{
	if ( _pMusicTrackChannel )
	{
		FMOD_RESULT result;
		result = _pMusicTrackChannel->setPaused( bPause );
		ERRCHECK(result);
	}
}



/******************************************************************************
 FUNCTION:  LoadSoundEffect
******************************************************************************/
NeSoundEffect* NeSoundManager::LoadSoundEffect( const char* pacPathAndFileName )
{
	// FMOD doesn't seem to like .mp3 files on iPad1.  I noticed these problems
	// in ManhattanC.  It will load .mp3 music very very slowly (so slow, it looks
	// like the app crashed), and often it will crash when loading mp3 sound
	// effects.  Thus this check and warning.
	// mdomowicz 2013_08_18
	char acResult[ 128 ] = "";
	JpUtils::GetExtension( (char*)pacPathAndFileName, acResult );

	bool bIsMp3 = ( JpHash( acResult ) == JpHash( "mp3" ) );
	ASSERTF( !bIsMp3, ( "\nAvoid using mp3s.  FMOD is slow and crashy when using mp3's on iPad 1." ) );


	FMOD::Sound* pFmodSound = NULL;
	FMOD_RESULT result;

	#if DCG_IS_TARGET_ANDROID

		unsigned long DataSizeResult = 0;

		const char* pFullPathAndFilename = CCFileUtils::sharedFileUtils()->fullPathFromRelativePath( pacPathAndFileName );
		unsigned char* pData = CCFileUtils::sharedFileUtils()->getFileData( pFullPathAndFilename, "rb", &DataSizeResult );

		// Copied from fmod example, "loadfrommemory"
		FMOD_CREATESOUNDEXINFO exinfo;
		memset( &exinfo, 0, sizeof( FMOD_CREATESOUNDEXINFO ) );
		exinfo.cbsize = sizeof(FMOD_CREATESOUNDEXINFO);
		exinfo.length = DataSizeResult;

		result = _pFmodSystem->createSound
		( 
			(const char*)pData, 
			FMOD_SOFTWARE | FMOD_LOOP_NORMAL | FMOD_OPENMEMORY, 
			&exinfo,
			&pFmodSound
		);

		delete[] pData;

	#else

		result = _pFmodSystem->createSound
		( 
			CCFileUtils::sharedFileUtils()->fullPathFromRelativePath( pacPathAndFileName ), 
			FMOD_SOFTWARE | FMOD_LOOP_NORMAL, 
			0, // FMOD_CREATESOUNDEXINFO *exinfo
			&pFmodSound 
		);		

	#endif


	ERRCHECK(result);

	if (result == FMOD_OK)
	{
		NeSoundEffect* pSound = new NeSoundEffect();
		pSound->SetSound( pFmodSound );

		char acFilename[ 128 ] = "";
		JpUtils::StripPath( acFilename, pacPathAndFileName );
		_SoundEffectTable.Store( JpHash( acFilename ), pSound );

		return pSound;
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  UnloadSoundEffect
******************************************************************************/
void NeSoundManager::UnloadSoundEffect( hash hSound )
{
	NeSoundEffect* pSound = _SoundEffectTable.Find( hSound );

	if ( pSound )
	{
		FMOD_RESULT result;
		FMOD::Sound* pFmodSound = pSound->GetSound();
		result = pFmodSound->release();
		ERRCHECK(result);
		
		_SoundEffectTable.Remove( hSound );
		delete pSound;
	}
}



/******************************************************************************
 FUNCTION:  UnloadAllSoundEffects
******************************************************************************/
void NeSoundManager::UnloadAllSoundEffects()
{
	NeSoundEffectTable::Iterator Iter( &_SoundEffectTable );

	int iNumFmodSoundReleased = 0;
	while ( Iter.Next() )
	{
		NeSoundEffect* pSound = Iter.GetData();
		FMOD::Sound* pFmodSound = pSound->GetSound();

		FMOD_RESULT result;
		result = pFmodSound->release();
		iNumFmodSoundReleased ++;
		ERRCHECK(result);
	}

	// This is here just to make sure the hash table iterator works.  It's not 
	// battle tested yet.
	ASSERTF( iNumFmodSoundReleased == _SoundEffectTable.GetNumEntries(), ( "\nSomething isn't right with hash table iterator" ) );

	_SoundEffectTable.DeleteAll();
}



/******************************************************************************
 FUNCTION:  GetSoftwareFormat
******************************************************************************/
void NeSoundManager::GetSoftwareFormat( SoftwareFormatParams* pResults )
{
	FMOD_RESULT result;

	result = _pFmodSystem->getSoftwareFormat
	( 
		&pResults->_iSampleRate, 
		&pResults->_eSoundFormat, 
		&pResults->_iNumOutputChannels, 
		&pResults->_iMaxChannels, 
		&pResults->_eResampleMethod, 
		&pResults->_iBits 
	);

	ERRCHECK(result);
}



