/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include <stdio.h>


#include "JpUtils.h"
#include "NeConfig.h"
#include "NeCsvIn.h"
#include "NeStamp.h"
#include "NeTgaIo.h"

#include "mmgrChooser.h"


NeStampManager* NeStampManager::_pInstance = NULL;





/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void NeStamp::Init()
{
	_iWidth = 0;
	_iHeight = 0;
	_iNumChannels = 0;
	_pacData = NULL;
}





/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void NeStamp::Deinit()
{
	if ( _pacData )
		delete[] _pacData;

}




/******************************************************************************
 FUNCTION:  LoadFromImageFile
******************************************************************************/
void NeStamp::LoadFromImageFile( char acPathAndFileName[] )
{
	bool bSuccess = false;

	unsigned char* pacData;
	int iBitsPerPixel = 0;

	NeTgaIo TgaIo;

	bSuccess = TgaIo.Read
	( 
		&pacData, 
		&_iWidth, 
		&_iHeight, 
		&iBitsPerPixel, 
		acPathAndFileName
	);


	ASSERTF( bSuccess, ( "\nFailed to load stamp image." ) );
	if ( bSuccess )
	{
		_iNumChannels = iBitsPerPixel / 8;
		_pacData = pacData;
	}
}




/******************************************************************************
 FUNCTION:  IsWhiteAtCoordinates
******************************************************************************/
bool NeStamp::IsWhiteAtCoordinates( int iX, int iY )
{
	ccColor4F Color = GetColorAtCoordinates( iX, iY );

	return 
	( 
		Color.r == 1.0f &&
		Color.g == 1.0f &&
		Color.b == 1.0f
	);
}




/******************************************************************************
 FUNCTION:  SameColorAtCoordinates
******************************************************************************/
bool NeStamp::SameColorAtCoordinates( int iX, int iY, const ccColor4F& CompareColor )
{
	ccColor4F Color = GetColorAtCoordinates( iX, iY );
	return 
	( 
		Color.r == CompareColor.r &&
		Color.g == CompareColor.g &&
		Color.b == CompareColor.b 
	);
}




/******************************************************************************
 FUNCTION:  GetColorAtCoordinates
******************************************************************************/
ccColor4F NeStamp::GetColorAtCoordinates( int iX, int iY )
{
	unsigned char ucR = 0;
	unsigned char ucG = 0;
	unsigned char ucB = 0;

	if ( _pacData )
	{
		int iIndex = _iNumChannels * ( ( iY * _iWidth ) + iX );
		ucR = _pacData[ iIndex + 0 ];
		ucG = _pacData[ iIndex + 1 ];
		ucB = _pacData[ iIndex + 2 ];
	}

	float fR = ( ucR / 255.0f );
	float fG = ( ucG / 255.0f );
	float fB = ( ucB / 255.0f );

	ccColor4F Color = { fR, fG, fB, 1.0f };
	return Color;
}











/******************************************************************************
 FUNCTION:  LoadStampsFromIniFile

 PURPOSE:   This reads in bunch a Stamps from a list specified in the
            provided file.  

 NOTES:     The list must adhere to the following format:

            One Stamp is specified per line in the list.
            Each line in the list has three columns.

               1st column: always the word "stamp"
               2nd column: path and file name of the Stamp
               3rd column: The look up key name


            Example:

               stamp   Data\Stamps\Stamp01.tga      Stamp01


            Note, whitespace is ignored between columns.
******************************************************************************/
void NeStampManager::LoadStampsFromIniFile( char acPathAndFileName[] )
{
	NeConfig Config;
	Config.ReadFile( acPathAndFileName );

	//-----------------------
	// reach each line in the config
	// file.  If the keyworld "Stamp"
	// is the first line of the
	// current line, try to read the
	// Stamp.
	JpList< NeConfig::Line >* pLineList = Config.GetLineList();
	JpList< NeConfig::Line >::Iterator Iter( pLineList );
	while ( Iter.Next() )
	{
		NeConfig::Line* pLine = Iter.GetNodeData();

		if ( strcmp( pLine->GetString( 0 ), "stamp" ) == 0 )
		{
			LoadStampFromFile
			(
				pLine->GetString( 1 ),
				JpHash( pLine->GetString( 2 ) )
			);
		} 
	}
}




/******************************************************************************
 FUNCTION:  LoadStampsFromCsv
******************************************************************************/
void NeStampManager::LoadStampsFromCsv( char acPathAndFileName[] )
{
	NeCsvIn Csv;
	Csv.ReadCsvFile( acPathAndFileName );
	Csv.MakeHashIndicesFromColValuesAtRow( 0 );
	Csv.MakeHashIndicesFromRowValuesAtCol( 0 );

	int iNumRows = Csv.GetNumRows();

	for ( int i = 1; i < iNumRows; i ++ )
	{
		char* pacPathAndFilename = Csv.GetStringByIntAndHash( i, 0xd588e307 /* "sPathAndFilename" */ );
		LoadStampFromFile( pacPathAndFilename, 0 );
	}
}



/******************************************************************************
 FUNCTION:  LoadStampFromFile
 NOTES:		This has to be in the CPP because it uses new/delete, and "mmgr",
			which I am using to track memory allocations, requires that all
			new/delete stuff is done in the cpp.
			mdomowicz 2005_08_22
******************************************************************************/
NeStamp* NeStampManager::LoadStampFromFile
( 
	char acPathAndFileName[], 
	unsigned int iLookupKey 
)
{
	NeStamp* pNewStamp = new NeStamp;
	pNewStamp->LoadFromImageFile( acPathAndFileName );

	//-----------------------
	// if no lookup key is 
	// provided, then create one
	// using the filename.
	if ( iLookupKey == 0 )
	{
		char acFileName[256];
		JpUtils::StripPath( acFileName, acPathAndFileName );
		iLookupKey = JpHash( acFileName );
	}

	bool bStored = StoreStamp( iLookupKey, pNewStamp );	
	if ( bStored == false )
	{
		delete pNewStamp;
		pNewStamp = NULL;
	}

	return pNewStamp;
}



/******************************************************************************
 FUNCTION:  RemoveAndDeleteStamp
 NOTES:		This has to be in the CPP because it uses new/delete, and "mmgr",
			which I am using to track memory allocations, requires that all
			new/delete stuff is done in the cpp.
			mdomowicz 2005_08_22
******************************************************************************/
void NeStampManager::RemoveAndDeleteStamp( unsigned int uiKey )
{
	NeStamp* pStamp = FindStamp( uiKey );
	if ( pStamp )
	{
		delete pStamp;
		_StampTable.Remove( uiKey );
	}
}



/******************************************************************************
 FUNCTION:  RemoveAndDeleteAll
 NOTES:		This has to be in the CPP because it uses new/delete, and "mmgr",
			which I am using to track memory allocations, requires that all
			new/delete stuff is done in the cpp.
			mdomowicz 2005_08_22
******************************************************************************/
void NeStampManager::RemoveAndDeleteAll()
{
	_StampTable.DeleteAll();
}



/******************************************************************************
 FUNCTION:  CreateInstance
******************************************************************************/
void NeStampManager::CreateInstance()
{
	ASSERTF( _pInstance == NULL, ( "\nCan't create more than one instance.  Not allowed." ) );

	if ( !_pInstance )
		_pInstance = new NeStampManager;
}



/******************************************************************************
 FUNCTION:  DeleteInstance
******************************************************************************/
void NeStampManager::DeleteInstance()
{
	if ( _pInstance )
	{
		delete _pInstance;
		_pInstance = NULL;
	}
}








