/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

#include "stdafx.h"

#include <string.h>

#include "JpDebug.h"

#include "NeConfig.h"

#include "SoActor.h"
#include "SoActorManager.h"
#include "SoConfigManager.h"
#include "SoLabelBmFont.h"
#include "SoParticleManager.h"
#include "SoSaveSystem.h"
#include "SoSoundManager.h"
#include "SoSprite.h"
#include "SoTextureUtils.h"
#include "SoUtils.h"

#include "mmgrChooser.h"



/******************************************************************************
 FUNCTION:  SoActor
******************************************************************************/
SoActor::SoActor() :
_pParentNode( NULL ),
_gGuid( 0 )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoActor::Init( CCNode* pParentNode )
{
	_gGuid = 0;

	SetParentNode( pParentNode );

	GenerateAndSetGuid();
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoActor::Deinit()
{	
	if ( _pParentNode )
		_pParentNode = NULL;

	ASSERTF( _gGuid, ( "\nAttempting to deinit actor more than once!" ) );

	_gGuid = 0;
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoActor::Update( float dt )
{
}



/******************************************************************************
 FUNCTION:  OnDeath
******************************************************************************/
void SoActor::OnDeath()
{
	SoParticleManager::GetInstance()->LoadCsvEventParticle( SoParticleEvent::TYPE_ONDEATH, this, JpVector2f( 0.0f, 0.0f ) );	
	SoSoundManager::GetInstance()->PlaySoundEffectOnDeath( this );	

	SoActorManager* pActorManager = SoActorManager::GetInstance();
	pActorManager->RemoveAndDeleteActor( this );
}



/******************************************************************************
 FUNCTION:  AddToActorManagerByPriority
 NOTE:      !!!! POTENTIAL GOTCHA !!!!
			If this is ever reached via a constructor you may not get the 
			expected results due to the call to GetType().  This is because 
			GetType is a virtual function, and virtual functions don't work 
			properly when called at construction time.
			mdomowicz 2009_01_21
******************************************************************************/
void SoActor::AddToActorManagerByPriority()
{
	int iPriority = 0;
	SoActorTypes::TYPE eActorType = GetType();
	
	switch ( eActorType )
	{
		case SoActorTypes::TYPE_ACTOR:
		default:
			break;
	}

	GenerateAndSetGuid();

	SoActorManager* pActorManager = SoActorManager::GetInstance();
	pActorManager->AddActor( this, iPriority );
}



/******************************************************************************
 FUNCTION:  GenerateAndSetGuid
******************************************************************************/
void SoActor::GenerateAndSetGuid()
{
	// sometimes things get removed and readded to the
	// actor manager.  We should make sure not to change
	// the actors guid in these situations.
	// mdomowicz 2009_04_29
	if ( _gGuid == 0 )
	{
		_gGuid = GenerateGuid();
	}
}



/******************************************************************************
 FUNCTION:  GenerateGuid
******************************************************************************/
guid SoActor::GenerateGuid()
{
	SoActorManager* pActorManager = SoActorManager::GetInstance();

	unsigned int uiGuid = JpRandom::G.GetInt();

	while ( pActorManager->GetActorByGuid( uiGuid ) != NULL )
	{
		uiGuid = JpRandom::G.GetInt();
		uiGuid = SoUtils::AlterGuidForCollisionPrevention( uiGuid );
	}

	return uiGuid;
}



/******************************************************************************
 FUNCTION:  draw
******************************************************************************/
void SoActor::draw(void)
{
}



/******************************************************************************
 FUNCTION:  CreateSprite
******************************************************************************/
SoSprite* SoActor::CreateSprite( CCNode* pParent, const char* pacPathAndFilename )
{
	SoSprite* pSprite = NULL;

	if ( UseResampledTexture() && SoTextureUtils::GetTextureScaler() != 1.0f )
		pSprite = SoSprite::CreateSpriteResampledFromFilename( pacPathAndFilename, SoTextureUtils::GetTextureScaler() );
	else
		pSprite = SoSprite::CreateScaledToDevice( pacPathAndFilename );


	pSprite->retain();
	pSprite->setPosition( ccp( 0.0f, 0.0f ) );
	pParent->addChild( pSprite );
	pSprite->setVisible( false );
	return pSprite;
}



/******************************************************************************
 FUNCTION:  CreateBmFontLabel
******************************************************************************/
SoLabelBmFont* SoActor::CreateBmFontLabel( CCNode* pParent, const std::string& sText, const std::string& sFontName, cocos2d::CCTextAlignment Alignment )
{
	CCSize size = CCDirector::sharedDirector()->getWinSize();

	SoLabelBmFont* pLabel = SoLabelBmFont::create
	(
		sText.c_str(), 
		SoUtils::AddRelativePathToFont( sFontName.c_str() ), 
		size.width, 
		Alignment
	);

	CCPoint AnchorPoint = CCPoint( 0, 0 );
	if ( Alignment == kCCTextAlignmentCenter )
		AnchorPoint = CCPoint( 0.5f, 0.5f );

	pLabel->setColor( JpUtils::_White );
	pLabel->setOpacity( 255.0f );
	pLabel->retain();
	pLabel->setAnchorPoint( AnchorPoint );
	pLabel->setScaleX( SoUtils::GetiPadRatioX() );
	pLabel->setScaleY( SoUtils::GetiPadRatioY() );

	pParent->addChild( pLabel );

	return pLabel;
}



/******************************************************************************
 FUNCTION:  RemoveAndDestroyNode
******************************************************************************/
void SoActor::RemoveAndDestroyNode( CCNode* pParent, CCNode* pNodeToDestroy )
{
	ASSERTF( pParent, ( "\npParent is NULL!" ) );
	if ( pParent && pNodeToDestroy )
	{
		pParent->removeChild( pNodeToDestroy, DCG_CLEANUP_ALL_RUNNING_ACTIONS );
		pNodeToDestroy->release();
	}
}



/******************************************************************************
 FUNCTION:  SaveToFile
******************************************************************************/
void SoActor::SaveToFile( SoSaveSystem* pSaveSystem )
{
	NeCsvOut* pCsvWriter = pSaveSystem->GetSharedCsvWriter();
	pCsvWriter->AddNewRow();
	pCsvWriter->AppendUnsignedIntToLastRow( _gGuid );
}



/******************************************************************************
 FUNCTION:  LoadFromFile
******************************************************************************/
void SoActor::LoadFromFile( SoSaveSystem* pSaveSystem )
{
	NeCsvIn* pCsvReader = pSaveSystem->GetSharedCsvReader();
	int iReaderIndex = pSaveSystem->GetSharedCsvReaderIndex();
	_gGuid = pCsvReader->GetHashByIntAndHash( iReaderIndex, 0xab3a05b2 /* "hGuid" */, 0 );
}



/******************************************************************************
 FUNCTION:  DefineCsvColumns
******************************************************************************/
void SoActor::DefineCsvColumns( SoSaveSystem* pSaveSystem )
{
	NeCsvOut* pCsvWriter = pSaveSystem->GetSharedCsvWriter();
	ASSERTF( pCsvWriter->IsClear(), ( "\nCsv is not empty!  This should be the very first thing to write to the csv." ) );
	pCsvWriter->Clear();
	pCsvWriter->AddNewRow();
	pCsvWriter->AppendStringToLastRow( "hGuid" );
}

























