/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

#include "stdafx.h"


#include <string.h>

#include "JpDebug.h"
#include "JpFileUtils.h"

#include "SoActorManager.h"
#include "SoSaveSystem.h"

#include "mmgrChooser.h"


SoActorManager* SoActorManager::_pInstance = NULL;



/******************************************************************************
 FUNCTION:  SoActorManager
******************************************************************************/
SoActorManager::SoActorManager() :
_pCachedPlayerActor( NULL ),
_uiCachedPlayerActorNameHash( 0 ),
_pParentNode( NULL )
{
}



/******************************************************************************
 FUNCTION:  RemoveAndDeleteActor
******************************************************************************/
void SoActorManager::RemoveAndDeleteActor( SoActor* pActor )
{
	//-----------------------
	// NOTE: we do not call _ActorList.Remove anymore because
	// RemoveAndDeleteActor is often called from an actor being 
	// updated via SoActorManager::Update, which walks the _ActorList, 
	// and if an actor is removed from the _ActorList while the 
	// ActorList is being walked, the list could become broken.
	// 2008_07_20 mdomowicz

	//_ActorList.Remove( pActor );

	if ( pActor != NULL )
	{
		_PendingDeletionList.AddUniquely( pActor, 0 );
		_PreservedActorList.Remove( pActor );

		if ( pActor == _pCachedPlayerActor )
		{
			_uiCachedPlayerActorNameHash = 0;
			_pCachedPlayerActor = NULL;
		}
	}
}




/******************************************************************************
 FUNCTION:  RemoveAndDeleteAllActors
******************************************************************************/
void SoActorManager::RemoveAndDeleteAllActors()
{
	JpList< SoActor >::Node* pCurrent = _ActorList.GetFirst();
	for ( ; pCurrent != _ActorList.GetHead(); pCurrent = pCurrent->GetNext() )
	{	
		SoActor* pActor = pCurrent->GetData();
		_PendingDeletionList.AddUniquely( pActor, 0 );
	}

	_ActorList.RemoveAll();


	_uiCachedPlayerActorNameHash = 0;
	_pCachedPlayerActor = NULL;


	_ActorGuidTable.RemoveAll();
	_PreservedActorList.RemoveAll();
}




/******************************************************************************
 FUNCTION:  RemoveAndDeleteAllActorsNow
******************************************************************************/
void SoActorManager::RemoveAndDeleteAllActorsNow()
{
	RemoveAndDeleteAllActors();
	DeleteAllActorsInPendingDeletionList();
}




/******************************************************************************
 FUNCTION:  DeleteAllActorsInPendingDeletionList
******************************************************************************/
void SoActorManager::DeleteAllActorsInPendingDeletionList()
{
	//-----------------------	
	// for all actors pending deletion
	// remove from the actor list
	// and delete the actor.
	JpList< SoActor >::Node* pCurrent = _PendingDeletionList.GetFirst();
	for ( ; pCurrent != _PendingDeletionList.GetHead(); pCurrent = pCurrent->GetNext() )
	{
		SoActor* pActor = pCurrent->GetData();

		//-----------------------	
		// This is the only truly safe place 
		// an actor can be removed from the
		// _ActorList
		// 2008_07_20 mdomowicz
		_ActorList.Remove( pActor );

		_ActorGuidTable.Remove( pActor->GetGuid() );

		pCurrent->GetData()->Deinit();
		delete pCurrent->GetData();
	}

	_PendingDeletionList.RemoveAll();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoActorManager::Update( float dt )
{
	JpList< SoActor >::Node* pCurrent = _ActorList.GetFirst();
	for ( ; pCurrent != _ActorList.GetHead(); )
	{
		//-----------------------	
		// Saving pNext deals with what happens if
		// this current node is deleted in
		// this update, but it does not correctly
		// handle what happens if the next node 
		// is deleted (in this update)!  That's why 
		// RemoveAndDeleteActor is implemented as it 
		// is above.  See notes in 
		// RemoveAndDeleteActor for more info.
		// 2008_07_20 mdomowicz
		JpList< SoActor >::Node* pNextNode = pCurrent->GetNext();

		SoActor* pActor = pCurrent->GetData();

		pActor->Update( dt );
	
		pCurrent = pNextNode;
	}


	DeleteAllActorsInPendingDeletionList();
}




/******************************************************************************
 FUNCTION:  AppendFilterdActors
 PURPOSE:	Finds all the actors of the given type, and adds them to the given
			list.
******************************************************************************/
int SoActorManager::AppendFilterdActors
(
	JpList< SoActor >* pActorList, 
	SoActorTypes::TYPE eActorType 
)
{
	int iNumAppended = 0;

	JpList< SoActor >::Node* pCurrent = _ActorList.GetFirst();
	for ( ; pCurrent != _ActorList.GetHead(); pCurrent = pCurrent->GetNext() )
	{
		SoActor* pActor = pCurrent->GetData();
		if ( pActor->GetType() == eActorType )
		{
			iNumAppended ++;
			pActorList->Add( pActor, 0 );
		}
	}

	return iNumAppended;
}



/******************************************************************************
 FUNCTION:  GetCountByType
 PURPOSE:	returns the number of actors of the given type
******************************************************************************/
int SoActorManager::GetCountByType
(
	SoActorTypes::TYPE eActorType 
)
{
	int iCount = 0;

	JpList< SoActor >::Node* pCurrent = _ActorList.GetFirst();
	for ( ; pCurrent != _ActorList.GetHead(); pCurrent = pCurrent->GetNext() )
	{
		SoActor* pActor = pCurrent->GetData();
		if ( pActor->GetType() == eActorType )
			iCount ++;
	}

	return iCount;
}



/******************************************************************************
 FUNCTION:  GetActorByGuid
******************************************************************************/
SoActor* SoActorManager::GetActorByGuid( unsigned int uiGuid )
{
	return _ActorGuidTable.Find( uiGuid );
}



/******************************************************************************
 FUNCTION:  GenerateGuid
******************************************************************************/
guid SoActorManager::GenerateGuid()
{
	unsigned int uiGuid = JpRandom::G.GetInt();

	while ( GetActorByGuid( uiGuid ) != NULL )
		uiGuid = JpRandom::G.GetInt();

	return uiGuid;
}




/******************************************************************************
 FUNCTION:  RemoveAndDeleteUnpreservedActorsNow
 PURPOSE:	Remove all actors from the actor manager, EXCEPT those that are
            are in the _PreservedActorList
******************************************************************************/
void SoActorManager::RemoveAndDeleteUnpreservedActorsNow()
{
	//-----------------------
	// manually pull all "preserved actors" from
	// the main actor list
	JpList< SoActor >::Iterator PreservedIter( &_PreservedActorList );
	while ( PreservedIter.Next() )
	{
		SoActor* pActor = PreservedIter.GetNodeData();
		_ActorList.Remove( pActor );
	}


	//-----------------------
	// Empty out the actor list.
	// Note this is ALMOST like a call to 
	// RemoveAndDeleteAllActorsNow,
	// but unlike that function, this code
	// does not clear _PreservedActorList
	JpList< SoActor >::Iterator Iter( &_ActorList );
	while ( Iter.Next() )
	{
		SoActor* pActor = Iter.GetNodeData();
		_PendingDeletionList.AddUniquely( pActor, 0 );
	}

	_ActorList.RemoveAll();

	_uiCachedPlayerActorNameHash = 0;
	_pCachedPlayerActor = NULL;

	_ActorGuidTable.RemoveAll();

	DeleteAllActorsInPendingDeletionList();


	//-----------------------
	// readd preserved actors back into the manager.
	PreservedIter.Reset();
	while ( PreservedIter.Next() )
	{
		SoActor* pActor = PreservedIter.GetNodeData();
		pActor->AddToActorManagerByPriority();
	}
}




/******************************************************************************
 FUNCTION:  SetActorPreservedStatus
******************************************************************************/
void SoActorManager::SetActorPreservedStatus
( 
	SoActor* pActor, 
	bool bPreserve 
)
{
	if ( bPreserve )
	{
		_PreservedActorList.AddUniquely( pActor, 0 );
	}
	else
	{
		_PreservedActorList.Remove( pActor );
	}
}



/******************************************************************************
 FUNCTION:  CreateInstance
******************************************************************************/
void SoActorManager::CreateInstance()
{
	ASSERTF( _pInstance == NULL, ( "\nCan't create more than one instance.  Not allowed." ) );

	if ( !_pInstance )
		_pInstance = new SoActorManager;
}



/******************************************************************************
 FUNCTION:  DeleteInstance
******************************************************************************/
void SoActorManager::DeleteInstance()
{
	if ( _pInstance )
	{
		_pInstance->RemoveAndDeleteAllActorsNow();
		delete _pInstance;
		_pInstance = NULL;
	}
}


/******************************************************************************
 FUNCTION:  SaveToFile
******************************************************************************/
void SoActorManager::SaveToFile( SoSaveSystem* pSaveSystem )
{
	struct Local
	{
		static void Save( SoActorList* pActorList, SoSaveSystem* pSaveSystem, char* pacFilename )
		{
			NeCsvOut* pCsvWriter = pSaveSystem->GetSharedCsvWriter();

			pCsvWriter->Clear();
			SoActorList::Iterator Iter( pActorList );
			while ( Iter.Next() )
				Iter.GetNodeData()->SaveToFile( pSaveSystem );
			
			if ( pActorList->GetNumEntries() )
				pCsvWriter->WriteToFile( pSaveSystem->ConstructPathAndFileName( pacFilename, -1 ) );
			else
				JpFileUtils::DeleteFile( pSaveSystem->ConstructPathAndFileName( pacFilename, -1 ) );
		}
	};

	SoActorList HeroList;

	
	//-----------------------
	// collect the actors into separate lists
	SoActorList* pActorList = SoActorManager::GetActorList();
	SoActorList::Iterator Iter( pActorList );
	while ( Iter.Next() )
	{
		SoActor* pActor = Iter.GetNodeData();

		//switch ( pActor->GetType() )
		//{
			//case SoActorTypes::TYPE_ACTOR_HERO: HeroList.AddLast( pActor ); break;
		//}
	}

	//-----------------------
}



/******************************************************************************
 FUNCTION:  LoadFromFile
******************************************************************************/
void SoActorManager::LoadFromFile( SoSaveSystem* pSaveSystem )
{
	struct Local
	{
		static SoActor* NewActorFromType( SoActorTypes::TYPE eActorType )
		{
			//switch ( eActorType )
			//{
			//}

			ASSERTF( 0, ( "\nUnsupported actor type." ) );
			return NULL;
		}

		static void Load( SoActorManager* pActorManager, SoActorTypes::TYPE eActorType, SoSaveSystem* pSaveSystem, char* pacFilename )
		{
			NeCsvIn* pCsvReader = pSaveSystem->GetSharedCsvReader();

			pSaveSystem->SetSharedCsvReaderIndex( 0 );
			pCsvReader->Clear();
			pCsvReader->ReadCsvFile( pSaveSystem->ConstructPathAndFileName( pacFilename, -1 ) );
			pCsvReader->MakeHashIndicesFromColValuesAtRow( 0 );

			for ( int i = 1; i < pCsvReader->GetNumRows(); i ++ )
			{
				// shared reader index must be incremented for every row.
				pSaveSystem->IncrementSharedCsvReaderIndex();

				guid gGuid = pCsvReader->GetHashByIntAndHash( i, 0xab3a05b2 /* "hGuid" */, 0 );
				
				if ( !pActorManager->Exists( gGuid ) )
				{
					SoActor* pNewActor = NewActorFromType( eActorType );
					pNewActor->Init( pActorManager->GetParentNode() );
					pNewActor->LoadFromFile( pSaveSystem );
				}
			}
		}
	};


	//-----------------------
	// load the rest of the actors.
	//Local::Load( this, SoActorTypes::TYPE_ACTOR_TERRAIN, pSaveSystem, "SoActorTerrain" );
}



/******************************************************************************
 FUNCTION:  LoadHeroFromFile
 NOTES:     The hero is a special case, and must be loaded earlier than
			the rest of the actors.  hence a dedicated function.
******************************************************************************/
void SoActorManager::LoadHeroFromFile( SoSaveSystem* pSaveSystem )
{
}
