/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

#include "stdafx.h"

#include "JpString.h"
#include "SoActorManager.h"
#include "SoActorVisibleBoardSpot.h"
#include "SoActorVisibleTile.h"
#include "SoCamera.h"
#include "SoFontDefs.h"
#include "SoGameBitPlayer.h"
#include "SoGameBitPlayerHuman.h"
#include "SoGameClient.h"
#include "SoGameGlobals.h"
#include "SoLabelBmFont.h"
#include "SoTextureUtils.h"

#include "mmgrChooser.h"




/******************************************************************************
 FUNCTION:  SoActorVisibleBoardSpot
******************************************************************************/
SoActorVisibleBoardSpot::SoActorVisibleBoardSpot() :
_pGameBitTile( NULL ),
_pLabel( NULL ),
_iRenderZOrder( -1 ),
_fTimeHighlighted( 0.0f )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoActorVisibleBoardSpot::Init( CCNode* pParentNode )
{
	SoActorVisible::Init( pParentNode );
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoActorVisibleBoardSpot::Deinit()
{	
	RemoveAndDestroyNode( _pParentNode, _pSprite );

	if ( _pLabel )	
		RemoveAndDestroyNode( _pParentNode, _pLabel );

	SoActorVisible::Deinit();
}



/******************************************************************************
 FUNCTION:  SynchronizeAfterLoad
******************************************************************************/
void SoActorVisibleBoardSpot::SynchronizeAfterLoad()
{	
	if ( !_pSprite )
		CreateBoardSpotSprite();

	_vWorldPosition = GetTargetPosition();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoActorVisibleBoardSpot::Update( float dt )
{
	SoActorVisible::Update( dt );

	if ( _pGameBitTile )
	{
		int iTileY = _pGameBitTile->GetTileY();
		int iZOrder = BASE_RENDER_ORDER_BOARD_SPOT + ( iTileY * 10 );
		SetRenderZOrder( iZOrder );
		

		if ( !_pSprite )
		{
			CreateBoardSpotSprite();

			// Set the sprite to reasonable start position.
			_vWorldPosition = GetTargetPosition();		
		}

		if ( !_pLabel && VISIBLE_BOARD_SPOT_SHOW_COORDINATE_LABELS )
			CreateLabel();
	}


	//-----------------------

	JpVector2f vTargetPos = GetTargetPosition();

	float fDt = ClampMax( dt, 0.1f );
	_vWorldPosition = InterpolateClamped( _vWorldPosition, vTargetPos, 5.0f * fDt );


	JpVector2f vScreenSpacePos = SoCameraManager::GetInstance()->GetCurrentCamera()->TransformWorldToScreenSpace( _vWorldPosition );
	
	if ( _pSprite )
	{
		_pSprite->setVisible( true );
		if ( SoGameGlobals::IsLoadingGame() )
		{
			_pSprite->setVisible( false );
		}

		_pSprite->SetPosition( vScreenSpacePos );

		if ( ShouldHighlightForHumanPlayer() || ShouldHighlightForDebugPlayer() )
		{
			_fTimeHighlighted += dt;

			// This is some cheesy code to make it so when a new tile is drawn
			// the corresponding board spot flashes for a bit.  Could use some polish
			// but I'll wait until closer to final art is in place.
			// mdomowicz 2014_02_27
			if ( _fTimeHighlighted < 2.0f )
			{
				float fTween = 1.0f + sinf( _fTimeHighlighted * 12.0f );
				fTween /= 2.0f;
				ccColor3B FinalColor = Interpolate( JpUtils::_Grey224, JpUtils::_White, fTween );
				_pSprite->setColor( FinalColor );
			}
			else if ( _fTimeHighlighted < 4.0f )
			{
				float fDeltaTime = ClampMin( 4.0f - _fTimeHighlighted, 0.0f );
				float fDeltaTimeNormalized = ( fDeltaTime / 2.0f );

				float fTween = ( 1.0f - fDeltaTimeNormalized );
				fTween *= fTween;

				ccColor3B FinalColor = Interpolate( JpUtils::_White, JpUtils::_Grey224, fTween );
				_pSprite->setColor( FinalColor );
			}
			else
			{
				_pSprite->setColor( JpUtils::_Grey224 );
			}
		}
		else	
			_pSprite->setColor( JpUtils::_Grey160 );
	}

	if ( _pLabel )
	{
		_pLabel->SetPosition( vScreenSpacePos );
	}
}



/******************************************************************************
 FUNCTION:  SetScale
******************************************************************************/
void SoActorVisibleBoardSpot::SetScale( const JpVector2f& vScale )
{
	if ( _pSprite )
	{
		_pSprite->setScaleX( vScale.X * SoUtils::GetiPadRatioX() );
		_pSprite->setScaleY( vScale.Y * SoUtils::GetiPadRatioY() );
	}

	if ( _pLabel )
	{
		_pLabel->setScaleX( vScale.X * SoUtils::GetiPadRatioX() );
		_pLabel->setScaleY( vScale.Y * SoUtils::GetiPadRatioY() );
	}
}



/******************************************************************************
 FUNCTION:  SetGameBitTile
******************************************************************************/
void SoActorVisibleBoardSpot::SetGameBitTile( SoGameBitTile* p )
{ 
	_pGameBitTile = p; 

	// When loading a save game, it could happen that parking spots were created
	// before their corresponding VisibleBoardSpots got their RenderZOrder properly
	// set.  Thus parking spots would render in the wrong z order.  By setting the 
	// Z order here, we ensure ActorVisibleBoardSpot instances know their z order 
	// as soon as possible.  mdomowicz 2016_01_31
	if ( _pGameBitTile )
	{
		int iTileY = _pGameBitTile->GetTileY();
		int iZOrder = BASE_RENDER_ORDER_BOARD_SPOT + ( iTileY * 10 );
		SetRenderZOrder( iZOrder );
	}
}



/******************************************************************************
 FUNCTION:  CreateBoardSpotSprite
******************************************************************************/
void SoActorVisibleBoardSpot::CreateBoardSpotSprite()
{
	if ( !_pSprite )
	{
		_pSprite = CreateSprite( _pParentNode, SoUtils::AddRelativePathToTexture( "BoardSpot.png" ) );
		_pSprite->setVisible( false );
		_pSprite->setColor( JpUtils::_Grey192 );

		_pParentNode->reorderChild( _pSprite, GetRenderZOrder() );

		SetScale( (1.0f / SoTextureUtils::GetTextureScaler()) * JpVector2f( SoUtils::GetSpriteZoom(), SoUtils::GetSpriteZoom() ) );
	}
}



/******************************************************************************
 FUNCTION:  CreateLabel
******************************************************************************/
void SoActorVisibleBoardSpot::CreateLabel()
{
	if ( !_pLabel && _pGameBitTile )
	{
		std::string sLabelText = SoActorVisibleTile::GenerateTileLabelTextFromXy( _pGameBitTile->GetTileX(), _pGameBitTile->GetTileY() );
		_pLabel = CreateBmFontLabel( _pParentNode, sLabelText, SoFontDefs::GetBoardSpot(), kCCTextAlignmentCenter );
		_pLabel->setVisible( true );
		_pLabel->setColor( JpUtils::_Black );
		SoUtils::SetRenderZOrder( _pParentNode, _pLabel, GetRenderZOrder() + 1 );
		SetScale( (1.0f / SoTextureUtils::GetTextureScaler()) * JpVector2f( SoUtils::GetSpriteZoom(), SoUtils::GetSpriteZoom() ) );
	}
}



/******************************************************************************
 FUNCTION:  GetTargetPosition
******************************************************************************/
JpVector2f SoActorVisibleBoardSpot::GetTargetPosition()
{
	JpVector2f vPosition( 0.0f, 0.0f );


	if ( _pSprite && _pGameBitTile )
	{
		int iSpriteSize = _pSprite->getContentSize().width;

		int iTileX = _pGameBitTile->GetTileX();
		int iTileY = _pGameBitTile->GetTileY();

		int iPosX = iTileX * iSpriteSize;
		int iPosY = iTileY * iSpriteSize;

		vPosition.Init( iPosX, -iPosY );
		vPosition += SoGameClient::GetGlobalCenterOffset( _pGame );
	}

	return vPosition;
}



/******************************************************************************
 FUNCTION:  ShouldHighlightForHumanPlayer
******************************************************************************/
bool SoActorVisibleBoardSpot::ShouldHighlightForHumanPlayer()
{
	if ( _pGameBitTile )
	{
		//-----------------------
		// Get the player that owns the tile
		SoGameBitPlayer* pOwnerPlayer = _pGame->GetPlayerByPlayerNumber( _pGameBitTile->GetOwnerPlayerNumber() );
		SoGameBitPlayerHuman* pOwnerPlayerHuman = NULL;
		if ( pOwnerPlayer )
			pOwnerPlayerHuman = pOwnerPlayer->GetAsPlayerHuman();
		
		//-----------------------
		bool bThereIsOnlyOneHumanPlayer = ( _pGame->GetOwningClient()->GetNumHumanPlayersOnThisClient() == 1 );

		bool bThisPlayerIsCurrentPlayer = 
		( 
			pOwnerPlayer &&
			_pGame->GetCurrentPlayerNumber() == pOwnerPlayer->GetPlayerNumber() 
		);

		bool bViewTileIsToggledOn = ( pOwnerPlayerHuman && pOwnerPlayerHuman->GetIsViewTilesToggleOn() );

		if ( bThereIsOnlyOneHumanPlayer )
		{
			bool bTheOwnerPlayerIsHuman = ( pOwnerPlayerHuman != NULL );
			return bTheOwnerPlayerIsHuman;
		}
		else
			return ( bThisPlayerIsCurrentPlayer && bViewTileIsToggledOn );
	}

	return false;
}



/******************************************************************************
 FUNCTION:  ShouldHighlightForDebugPlayer
******************************************************************************/
bool SoActorVisibleBoardSpot::ShouldHighlightForDebugPlayer()
{
	if ( SoDebug::GetPlayerWithDebugHighlightEnabled() != PLAYER_UNDEFINED_PLAYER_NUMBER )
		return ( _pGameBitTile->GetOwnerPlayerNumber() == SoDebug::GetPlayerWithDebugHighlightEnabled() );

	return false;
}
