/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

#include "stdafx.h"

#include "JpString.h"
#include "SoActorManager.h"
#include "SoActorVisibleBoardSpot.h"
#include "SoActorVisiblePlayer.h"
#include "SoActorVisibleHotel.h"
#include "SoActorVisibleTile.h"
#include "SoLabelBmFont.h"
#include "SoCamera.h"
#include "SoFontDefs.h"
#include "SoGameBitPlayer.h"
#include "SoTextureUtils.h"

#include "mmgrChooser.h"



float SoActorVisibleHotel::_afSpriteHeightToWidthRatios[ GAMERULES_NUM_HOTELS ];



/******************************************************************************
 FUNCTION:  SoActorVisibleHotel
******************************************************************************/
SoActorVisibleHotel::SoActorVisibleHotel() :
_pGameBitHotel( NULL ),
_iRenderOrderState( 0 )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoActorVisibleHotel::Init( CCNode* pParentNode )
{
	SoActorVisible::Init( pParentNode );

	_bIsUserMovable = true;
	_fUserMoveMaxDist = 5000.0f;


	// Yes, this static array gets re-inited every time,
	// but that's ok.  This is a convenient place to init the array.
	// mdomowicz 2014_03_06
	for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
		_afSpriteHeightToWidthRatios[ i ] = 1.0f;
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoActorVisibleHotel::Deinit()
{	
	RemoveAndDestroyNode( _pParentNode, _pSprite );
	SoActorVisible::Deinit();
}



/******************************************************************************
 FUNCTION:  SynchronizeAfterLoad
******************************************************************************/
void SoActorVisibleHotel::SynchronizeAfterLoad()
{	
	_vWorldPosition = GetTargetPosition();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoActorVisibleHotel::Update( float dt )
{
	SoActorVisible::Update( dt );

	if ( _pGameBitHotel )
	{
		if ( !_pSprite )
		{
			CreateHotelSprite();

			// Set the sprite to reasonable start position.
			_vWorldPosition = GetTargetPosition();		
		}
	}


	//-----------------------

	JpVector2f vTargetPos = GetTargetPosition();

	float fDt = ClampMax( dt, 0.1f );
	_vWorldPosition = InterpolateClamped( _vWorldPosition, vTargetPos, 3.0f * fDt );



	if ( _pGameBitHotel && _pGameBitHotel->IsPlaced() )
	{
		_bIsUserMovable = false;

		// If a tile is placed, we want to zero out the user move offset, but
		// we don't want to make it zero immediately, as that will cause a noticeable
		// jump in the tiles position before it starts moving.  So we interpolate here.
		_vUserMoveOffset = InterpolateClamped( _vUserMoveOffset, SoUtils::_v2fZero, 5.0f * fDt );
	}


	JpVector2f vScreenSpacePos = SoCameraManager::GetInstance()->GetCurrentCamera()->TransformWorldToScreenSpace( _vWorldPosition + _vUserMoveOffset );


	if ( _pSprite )
	{
		_pSprite->SetPosition( vScreenSpacePos );
	}


	UpdateRenderOrder( dt );
}



/******************************************************************************
 FUNCTION:  UpdateRenderOrder
******************************************************************************/
void SoActorVisibleHotel::UpdateRenderOrder( float dt )
{
	if ( _pGameBitHotel )
	{
		const int STATE_NOT_IN_PLAY = 0;
		const int STATE_PLACING_ONTO_BOARD = 20;
		const int STATE_ON_BOARD = 30;

		switch ( _iRenderOrderState )
		{
			case STATE_NOT_IN_PLAY:
			{
				if ( _pGameBitHotel->GetPlacementTileId() != TILE_UNDEFINED_ID )
				{
					SetRenderZOrder( BASE_RENDER_ORDER_HOTEL + _iVisibleActorNumber );
					_iRenderOrderState = STATE_PLACING_ONTO_BOARD;
				}

				break;
			}

			case STATE_PLACING_ONTO_BOARD:
			{
				if ( _pGameBitHotel->GetPlacementTileId() != TILE_UNDEFINED_ID )
				{
					SoGameBitTile* pTile = _pGame->GetTileSet()->GetTileById( _pGameBitHotel->GetPlacementTileId() );
					SoActorVisibleBoardSpot* pVisibleBoardSpot = pTile->GetActorVisibleTile()->GetVisibleBoardSpot();
					JpVector2f vBoardSpotPos = pVisibleBoardSpot->GetWorldPosition();

					if ( Length( vBoardSpotPos - _vWorldPosition ) < 1.0f )
					{
						SetRenderZOrder( pVisibleBoardSpot->GetRenderZOrder() + OFFSET_RENDER_ORDER_HOTEL );
						_iRenderOrderState = STATE_ON_BOARD;
					}
				}
				else
				{
					_iRenderOrderState = STATE_NOT_IN_PLAY;
				}

				break;
			}

			case STATE_ON_BOARD:
			{
				if ( _pGameBitHotel->GetPlacementTileId() == TILE_UNDEFINED_ID )
				{
					SetRenderZOrder( BASE_RENDER_ORDER_HOTEL + _iVisibleActorNumber );
					_iRenderOrderState = STATE_NOT_IN_PLAY;
				}

				break;
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  SetRenderZOrder
******************************************************************************/
void SoActorVisibleHotel::SetRenderZOrder( int iRenderZOrder )
{
	_iRenderOrder = ( iRenderZOrder );
	SoUtils::SetRenderZOrder( _pParentNode, _pSprite, _iRenderOrder );
}



/******************************************************************************
 FUNCTION:  SetScale
******************************************************************************/
void SoActorVisibleHotel::SetScale( const JpVector2f& vScale )
{
	if ( _pSprite )
	{
		_pSprite->setScaleX( vScale.X * SoUtils::GetiPadRatioX() );
		_pSprite->setScaleY( vScale.Y * SoUtils::GetiPadRatioY() );
	}
}



/******************************************************************************
 FUNCTION:  CreateHotelSprite
******************************************************************************/
void SoActorVisibleHotel::CreateHotelSprite()
{
	if ( !_pSprite && _pGameBitHotel )
	{
		char acTemp[ 128 ] = "";
		sprintf( acTemp, "HotelTile%i.png", _pGameBitHotel->GetHotelId() );

		_pSprite = CreateSprite( _pParentNode, SoUtils::AddRelativePathToTexture( acTemp ) );
		_pSprite->setVisible( true );
		//_pSprite->setColor( GetHotelColorById( _pGameBitHotel->GetHotelId() ) );

		SetSpriteHeightToWidthRatio( _pGameBitHotel->GetHotelId(), _pSprite );

		_iRenderOrder = ( BASE_RENDER_ORDER_HOTEL + _iVisibleActorNumber );
		_pParentNode->reorderChild( _pSprite, _iRenderOrder );

		SetScale( (1.0f / SoTextureUtils::GetTextureScaler()) * JpVector2f( SoUtils::GetSpriteZoom(), SoUtils::GetSpriteZoom() ) );
	}
}



/******************************************************************************
 FUNCTION:  GetTargetPosition
******************************************************************************/
JpVector2f SoActorVisibleHotel::GetTargetPosition()
{
	JpVector2f vPosition( 0.0f, 0.0f );
	float fRandomDistRange = 0.0f;

	if ( _pGameBitHotel )
	{
		if ( _pGameBitHotel->IsPlaced() )
		{
			SoGameBitTile* pTile = _pGame->GetTileSet()->GetTileById( _pGameBitHotel->GetPlacementTileId() );
			SoActorVisibleBoardSpot* pVisibleBoardSpot = pTile->GetActorVisibleTile()->GetVisibleBoardSpot();
			vPosition = pVisibleBoardSpot->GetWorldPosition();

			// Adjust the height of the hotel tile so that it appears to
			// sit on top of the tile in an isometric way.
			switch ( _pGameBitHotel->GetHotelId() )
			{
				case 0:
				case 1:
					vPosition.Y += 15.0f;
					break;
				case 2:
				case 3:
					vPosition.Y += 20.0f;
					break;
				case 4:
					vPosition.Y += 25.0f;
					break;
				case 5:
				case 6:
					vPosition.Y += 25.0f;
					break;
			}
		}
		else
		{
			vPosition.Init( 0.0f, 1700.0f );
			fRandomDistRange = 250.0f;
		}
	}


	if ( fRandomDistRange > 0.0f )
	{
		JpRandom LocalRng( GetGuid() );
		JpVector2f vRandomDir = LocalRng.GetNormal2f();
		float fRandomDist = Interpolate( 0.0f, fRandomDistRange, LocalRng.GetFloat() );
		JpVector2f vRandomOffset = vRandomDir * fRandomDist;

		vPosition += vRandomOffset;
	}

	return vPosition;
}



/******************************************************************************
 FUNCTION:  GetHotelColorById
******************************************************************************/
ccColor3B SoActorVisibleHotel::GetHotelColorById( int iHotelId )
{
	switch ( iHotelId )
	{
		case HOTEL_UNDEFINED_ID: return JpUtils::_Grey64;

		case 0: return JpUtils::_Red;
		case 1: return JpUtils::_Green;
		case 2: return JpUtils::_Blue;
		case 3: return JpUtils::_Yellow;
		case 4: return JpUtils::_Orange;
		case 5: return JpUtils::_Magenta;
		case 6: return JpUtils::_Cyan;

		default: 
			ASSERTF( 0, ( "\nUnhandled hotel id" ) );
	}

	return JpUtils::_Grey64;
}



/******************************************************************************
 FUNCTION:  GetHoteNameById
******************************************************************************/
std::string SoActorVisibleHotel::GetHoteNameById( int iHotelId )
{
	switch ( iHotelId )
	{
		case 0: return "Remmiw";
		case 1: return "Glue";
		case 2: return "BrenTec";
		case 3: return "Yang";
		case 4: return "Omit";
		case 5: return "Marie";
		case 6: return "Conover";

		default: 
			ASSERTF( 0, ( "\nUnhandled hotel id" ) );
	}

	return "ERROR";
}



/******************************************************************************
 FUNCTION:  GetSpriteHeightToWidthRatio
******************************************************************************/
float SoActorVisibleHotel::GetSpriteHeightToWidthRatio( int iHotelId )
{
	if ( iHotelId >= 0 && iHotelId < GAMERULES_NUM_HOTELS )
		return _afSpriteHeightToWidthRatios[ iHotelId ];

	ASSERTF( 0, ( "\nBad Hotel ID" ) );
	return 0.0f;
}



/******************************************************************************
 FUNCTION:  SetSpriteHeightToWidthRatio
******************************************************************************/
void SoActorVisibleHotel::SetSpriteHeightToWidthRatio( int iHotelId, SoSprite* pSprite )
{
	float fRatio = ( (float)pSprite->getContentSize().height / (float)pSprite->getContentSize().width );
	
	ASSERTF( iHotelId >= 0 && iHotelId < GAMERULES_NUM_HOTELS, ( "\nBad Hotel ID" ) );

	if ( iHotelId >= 0 && iHotelId < GAMERULES_NUM_HOTELS )
		_afSpriteHeightToWidthRatios[ iHotelId ] = fRatio;
}



