/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

#include "stdafx.h"

#include "JpString.h"
#include "SoActorManager.h"
#include "SoActorVisibleBoardSpot.h"
#include "SoActorVisibleHotel.h"
#include "SoActorVisiblePlayer.h"
#include "SoActorVisibleTile.h"
#include "SoLabelBmFont.h"
#include "SoCamera.h"
#include "SoFontDefs.h"
#include "SoGameBitPlayer.h"
#include "SoGameBitPlayerHuman.h"
#include "SoGameClient.h"
#include "SoGameGlobals.h"
#include "SoTextureUtils.h"

#include "mmgrChooser.h"



SoActorVisibleTile::OrganizePos SoActorVisibleTile::_OrganizePosArrays[ GAMERULES_MAX_PLAYERS ][ GAMERULES_NUM_TILES_PER_HAND ];




/******************************************************************************
 FUNCTION:  SoActorVisibleTile
******************************************************************************/
SoActorVisibleTile::SoActorVisibleTile() :
_pGameBitTile( NULL ),
_pVisibleBoardSpot( NULL ),
_pLabel( NULL ),
_vLastCameraPos( 0.0f, 0.0f ),
_pTileBackSprite( NULL ),
_iOrganizeState( 0 ),
_iOrganizeIndex( -1 ),
_iRenderOrderState( 0 ),
_iPreviousHotelId( HOTEL_UNDEFINED_ID )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoActorVisibleTile::Init( CCNode* pParentNode )
{
	SoActorVisible::Init( pParentNode );

	_bIsUserMovable = true;
	_fUserMoveMaxDist = 5000.0f;
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoActorVisibleTile::Deinit()
{	
	RemoveAndDestroyNode( _pParentNode, _pSprite );
	RemoveAndDestroyNode( _pParentNode, _pTileBackSprite );

	if ( _pLabel )
		RemoveAndDestroyNode( _pParentNode, _pLabel );

	SoActorVisible::Deinit();
}



/******************************************************************************
 FUNCTION:  SynchronizeAfterLoad
******************************************************************************/
void SoActorVisibleTile::SynchronizeAfterLoad()
{	
	_vWorldPosition = GetTargetPosition();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoActorVisibleTile::Update( float dt )
{
	SoActorVisible::Update( dt );

	if ( _pGameBitTile && _pVisibleBoardSpot )
	{
		if ( !_pSprite && !_pTileBackSprite )
		{
			CreateTileSprites();

			// Set the sprite to reasonable start position.
			_vWorldPosition = GetTargetPosition();		
		}

		if ( !_pLabel && VISIBLE_TILE_SHOW_COORDINATE_LABELS )
			CreateLabel();
	}


	//-----------------------
	// When a tile is being held, the gameplay camera goes in a special mode
	// in which, if the player is touching near the edge of the window/screen,
	// the camera will move towards that edge.  This allows for tiles to
	// be dragged far distances which is useful at times.
	// When the camera is in this special mode, this code is required to make the
	// tile "keep up" with the moving camera.  It's slightly hard to explain,
	// so the best way to see what this code does, is simply comment it out, and
	// try dragging a tile off screen.  mdomowicz 2013_10_19
	SoCameraManager* pCameraManager = SoCameraManager::GetInstance();

	if ( pCameraManager )
	{
		SoCamera* pCamera = pCameraManager->FindCamera( "Gameplay" );

		if ( _bUserIsTouching )
			_vUserMoveOffset += ( pCamera->GetPosition() - _vLastCameraPos );

		_vLastCameraPos = pCamera->GetPosition();
	}

	UpdateTileSprite( dt );

	// Appears to be a relic of a much earlier version of Achoir.  
	// Removing for now because it's causing a load-game bug where
	// tiles will not SynchronizeAfterLoad() correctly.
	// This should be delete eventually, if it is truly never used.
	// mdomowicz 2016_01_31
	//UpdateOrganizePos( dt );

	UpdateTilePosition( dt );

	UpdateTileVisibility( dt );

	UpdateRenderOrder( dt );
}



/******************************************************************************
 FUNCTION:  UpdateTilePosition
******************************************************************************/
void SoActorVisibleTile::UpdateTilePosition( float dt )
{
	//-----------------------
	// Update the tile position
	JpVector2f vTargetPos = GetTargetPosition();

	float fDt = ClampMax( dt, 0.1f );
	_vWorldPosition = InterpolateClamped( _vWorldPosition, vTargetPos, 5.0f * fDt );

	if ( _pGameBitTile && _pGameBitTile->IsPlaced() )
	{
		_bIsUserMovable = false;

		// If a tile is placed, we want to zero out the user move offset, but
		// we don't want to make it zero immediately, as that will cause a noticeable
		// jump in the tiles position before it starts moving.  So we interpolate here.
		_vUserMoveOffset = InterpolateClamped( _vUserMoveOffset, SoUtils::_v2fZero, 5.0f * fDt );
	}

	JpVector2f vScreenSpacePos = SoCameraManager::GetInstance()->GetCurrentCamera()->TransformWorldToScreenSpace
	( 
		_vWorldPosition + _vUserMoveOffset + _vOrganizeOffset
	);


	if ( _pSprite )
	{
		_pSprite->SetPosition( vScreenSpacePos );
	}

	if ( _pTileBackSprite )
	{
		_pTileBackSprite->SetPosition( vScreenSpacePos );
	}

	if ( _pLabel )
	{
		_pLabel->SetPosition( vScreenSpacePos );
	}
}



/******************************************************************************
 FUNCTION:  UpdateTileVisibility
******************************************************************************/
void SoActorVisibleTile::UpdateTileVisibility( float dt )
{
	if ( _pGameBitTile && _pSprite && _pTileBackSprite )
	{
		if ( SoGameGlobals::IsLoadingGame() )
		{
			_pSprite->setVisible( false );
			_pTileBackSprite->setVisible( false );	
		}
		else if ( _pGameBitTile->IsPlaced() )
		{
			_pSprite->setVisible( true );
			_pTileBackSprite->setVisible( false );

			if ( _pLabel )
				_pLabel->setVisible( true );
		}
		else
		{
			_pSprite->setVisible( false );
			_pTileBackSprite->setVisible( false );

			if ( _pLabel )
				_pLabel->setVisible( false );
		}
	}
}



/******************************************************************************
 FUNCTION:  UpdateOrganizePos
******************************************************************************/
void SoActorVisibleTile::UpdateOrganizePos( float dt )
{
	if ( _pGameBitTile )
	{
		const int STATE_NOT_IN_PLAY = 0;
		const int STATE_IN_PLAYERS_HAND = 10;
		const int STATE_ON_BOARD = 20;

		switch ( _iOrganizeState )
		{
			case STATE_NOT_IN_PLAY:
			{
				_vOrganizeOffset = InterpolateClamped( _vOrganizeOffset, SoUtils::_v2fZero, 5.0f * dt );

				if ( _pGameBitTile->GetOwnerPlayerNumber() != PLAYER_UNDEFINED_PLAYER_NUMBER )
				{
					_iOrganizeIndex = AssignOrganizePos( _pGameBitTile );
					_iOrganizeState = STATE_IN_PLAYERS_HAND;
				}

				break;
			}
			case STATE_IN_PLAYERS_HAND:
			{
				JpVector2f vAssignedOffset = GetOrganizePosFromIndex( _pGameBitTile, _iOrganizeIndex );

				_vOrganizeOffset = InterpolateClamped( _vOrganizeOffset, vAssignedOffset, 5.0f * dt );

				if ( _pGameBitTile->GetOwnerPlayerNumber() == PLAYER_UNDEFINED_PLAYER_NUMBER )
				{
					UnassignOrganizePos( _pGameBitTile );
					_iOrganizeState = STATE_NOT_IN_PLAY;
				}
				else if ( _pGameBitTile->IsPlaced() )
				{
					UnassignOrganizePos( _pGameBitTile );
					_iOrganizeState = STATE_ON_BOARD;
				}

				break;
			}
			case STATE_ON_BOARD:
			{
				_vOrganizeOffset = InterpolateClamped( _vOrganizeOffset, SoUtils::_v2fZero, 5.0f * dt );

				if ( !_pGameBitTile->IsPlaced() )
				{
					_iOrganizeIndex = AssignOrganizePos( _pGameBitTile );
					_iOrganizeState = STATE_IN_PLAYERS_HAND;
				}

				break;
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  UpdateRenderOrder
******************************************************************************/
void SoActorVisibleTile::UpdateRenderOrder( float dt )
{
	if ( _pGameBitTile )
	{
		const int STATE_NOT_IN_PLAY = 0;
		const int STATE_IN_PLAYERS_HAND = 10;
		const int STATE_PLACING_ONTO_BOARD = 20;
		const int STATE_ON_BOARD = 30;

		switch ( _iRenderOrderState )
		{
			case STATE_NOT_IN_PLAY:
			{
				if ( _pGameBitTile->GetOwnerPlayerNumber() != PLAYER_UNDEFINED_PLAYER_NUMBER )
				{
					SetRenderZOrder( BASE_RENDER_ORDER_TILE );
					_iRenderOrderState = STATE_IN_PLAYERS_HAND;
				}
				else if ( _pGameBitTile->IsPlaced() )
				{
					_iRenderOrderState = STATE_PLACING_ONTO_BOARD;
				}

				break;
			}

			case STATE_IN_PLAYERS_HAND:
			{
				if ( _pGameBitTile->GetOwnerPlayerNumber() == PLAYER_UNDEFINED_PLAYER_NUMBER )
				{
					_iRenderOrderState = STATE_NOT_IN_PLAY;
				}
				else if ( _pGameBitTile->IsPlaced() )
				{
					_iRenderOrderState = STATE_PLACING_ONTO_BOARD;
				}

				break;
			}

			case STATE_PLACING_ONTO_BOARD:
			{
				JpVector2f vBoardSpotPos = _pVisibleBoardSpot->GetWorldPosition();
				if ( Length( vBoardSpotPos - _vWorldPosition ) < 10.0f )
				{
					SetRenderZOrder( _pVisibleBoardSpot->GetRenderZOrder() + OFFSET_RENDER_ORDER_TILE );
					_iRenderOrderState = STATE_ON_BOARD;
				}

				break;
			}

			case STATE_ON_BOARD:
			{
				if ( !_pGameBitTile->IsPlaced() )
				{
					SetRenderZOrder( BASE_RENDER_ORDER_TILE );
					_iRenderOrderState = STATE_IN_PLAYERS_HAND;
				}

				break;
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  UpdateTileSprite
******************************************************************************/
void SoActorVisibleTile::UpdateTileSprite( float dt )
{
	if ( _iPreviousHotelId != _pGameBitTile->GetHotelId() )
	{
		RecreateMainSprite();
		_iPreviousHotelId = _pGameBitTile->GetHotelId();
	}
}



/******************************************************************************
 FUNCTION:  SetRenderZOrder
******************************************************************************/
void SoActorVisibleTile::SetRenderZOrder( int iRenderZOrder )
{
	_iRenderOrder = ( iRenderZOrder );
	SoUtils::SetRenderZOrder( _pParentNode, _pSprite, _iRenderOrder );
	SoUtils::SetRenderZOrder( _pParentNode, _pTileBackSprite, _iRenderOrder );

	if ( _pLabel )
		SoUtils::SetRenderZOrder( _pParentNode, _pLabel, _iRenderOrder + 1 );
}



/******************************************************************************
 FUNCTION:  SetScale
******************************************************************************/
void SoActorVisibleTile::SetScale( const JpVector2f& vScale )
{
	if ( _pSprite )
	{
		_pSprite->setScaleX( vScale.X * SoUtils::GetiPadRatioX() );
		_pSprite->setScaleY( vScale.Y * SoUtils::GetiPadRatioY() );
	}

	if ( _pTileBackSprite )
	{
		_pTileBackSprite->setScaleX( vScale.X * SoUtils::GetiPadRatioX() );
		_pTileBackSprite->setScaleY( vScale.Y * SoUtils::GetiPadRatioY() );
	}

	if ( _pLabel )
	{
		_pLabel->setScaleX( vScale.X * SoUtils::GetiPadRatioX() );
		_pLabel->setScaleY( vScale.Y * SoUtils::GetiPadRatioY() );
	}
}



/******************************************************************************
 FUNCTION:  RecreateMainSprite
******************************************************************************/
void SoActorVisibleTile::RecreateMainSprite()
{
	RemoveAndDestroyNode( _pParentNode, _pSprite );
	_pSprite = NULL;

	if ( _pGameBitTile )
	{
		std::string sFilename;
		ccColor3B Color;

		int iHotelId = _pGameBitTile->GetHotelId();
		if ( iHotelId != HOTEL_UNDEFINED_ID )
		{
			sFilename = "Tile";
			sFilename += JpString::IntToString( iHotelId );

			// Each hotel tile has 3 variations...pick one.
			JpRandom LocalRng( GetGuid() );
			int iChoice = LocalRng.GetInt() % 3;
			if ( iChoice == 1 )
				sFilename += "b";
			else if ( iChoice == 2 )
				sFilename += "c";

			sFilename += ".png";

			Color = JpUtils::_White;
		}
		else
		{
			sFilename = "Tile.png";
			Color = JpUtils::_White;
		}

		_pSprite = CreateSprite( _pParentNode, SoUtils::AddRelativePathToTexture( sFilename.c_str() ) );
		_pSprite->setVisible( true );
		_pSprite->setColor( Color );


		SoUtils::SetRenderZOrder( _pParentNode, _pSprite, _iRenderOrder );
		SetScale( (1.0f / SoTextureUtils::GetTextureScaler()) * JpVector2f( SoUtils::GetSpriteZoom(), SoUtils::GetSpriteZoom() ) );
	}
}



/******************************************************************************
 FUNCTION:  CreateTileSprites
******************************************************************************/
void SoActorVisibleTile::CreateTileSprites()
{
	if ( !_pSprite )
	{
		_pSprite = CreateSprite( _pParentNode, SoUtils::AddRelativePathToTexture( "Tile.png" ) );
		_pSprite->setVisible( false );
		_pSprite->setColor( JpUtils::_White );

		_pTileBackSprite = CreateSprite( _pParentNode, SoUtils::AddRelativePathToTexture( "TileBack.png" ) );
		_pTileBackSprite->setVisible( true );
		_pTileBackSprite->setColor( JpUtils::_Grey64 );

		_iRenderOrder = ( BASE_RENDER_ORDER_TILE + _iVisibleActorNumber );
		_pParentNode->reorderChild( _pSprite, _iRenderOrder );
		_pParentNode->reorderChild( _pTileBackSprite, _iRenderOrder );

		SetScale( (1.0f / SoTextureUtils::GetTextureScaler()) * JpVector2f( SoUtils::GetSpriteZoom(), SoUtils::GetSpriteZoom() ) );
	}
}



/******************************************************************************
 FUNCTION:  CreateLabel
******************************************************************************/
void SoActorVisibleTile::CreateLabel()
{
	if ( !_pLabel && _pGameBitTile )
	{
		std::string sLabelText = GenerateTileLabelTextFromXy( _pGameBitTile->GetTileX(), _pGameBitTile->GetTileY() );
		_pLabel = CreateBmFontLabel( _pParentNode, sLabelText, SoFontDefs::GetBoardSpot(), kCCTextAlignmentCenter );
		_pLabel->setVisible( true );
		_pLabel->setColor( JpUtils::_White );

		_iRenderOrder = ( BASE_RENDER_ORDER_TILE + _iVisibleActorNumber );
		SoUtils::SetRenderZOrder( _pParentNode, _pLabel, _iRenderOrder + 1 );

		SetScale( (1.0f / SoTextureUtils::GetTextureScaler()) * JpVector2f( SoUtils::GetSpriteZoom(), SoUtils::GetSpriteZoom() ) );
	}
}



/******************************************************************************
 FUNCTION:  GenerateTileLabelTextFromXy
******************************************************************************/
std::string SoActorVisibleTile::GenerateTileLabelTextFromXy( int iX, int iY )
{
	std::string sResult = JpString::IntToString( iX + 1 );

	ASSERTF( iY >= 0 && iY < 26, ( "\nCurrent tile height is unsupported." ) );

	if ( iY >= 0 && iY < 26 )
	{
		char acLetters[ 26 ] = { 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z' };
		sResult += acLetters[ iY ];
	}

	return sResult;
}



/******************************************************************************
 FUNCTION:  GetTargetPosition
******************************************************************************/
JpVector2f SoActorVisibleTile::GetTargetPosition()
{
	JpVector2f vPosition( 0.0f, 0.0f );
	float fRandomDistRange = 0.0f;

	if ( _pGameBitTile )
	{
		if ( _pGameBitTile->IsPlaced() )
		{
			vPosition =  _pVisibleBoardSpot->GetWorldPosition();

			if ( _pLabel )
			{
				_pLabel->setVisible( true );
			}
		}
		else if ( _pGameBitTile->GetOwnerPlayerNumber() != PLAYER_UNDEFINED_PLAYER_NUMBER )
		{
			vPosition = SoActorVisiblePlayer::GetBoardWorldPositionForPlayer( _pGameBitTile->GetOwnerPlayerNumber() );
			vPosition *= 3.0f;
			fRandomDistRange = 10.0f;

			if ( _pLabel )
			{
				_pLabel->setVisible( true );
			}
		}
		else
		{
			vPosition.Init( 0.0f, 1700.0f );
			fRandomDistRange = 250.0f;

			if ( _pLabel )
			{
				_pLabel->setVisible( false );
			}
		}
	}


	if ( fRandomDistRange > 0.0f )
	{
		JpRandom LocalRng( GetGuid() );
		JpVector2f vRandomDir = LocalRng.GetNormal2f();
		float fRandomDist = Interpolate( 0.0f, fRandomDistRange, LocalRng.GetFloat() );
		JpVector2f vRandomOffset = vRandomDir * fRandomDist;

		vPosition += vRandomOffset;
	}

	return vPosition;
}



/******************************************************************************
 FUNCTION:  IsFrontVisible
******************************************************************************/
bool SoActorVisibleTile::IsFrontVisible()
{
	if ( _pGameBitTile )
	{
		//-----------------------
		// for starters, if the tile is dealt and showing above 
		// the main board, it is visible.
		bool bTileIsVisible = false;

		//-----------------------
		// Get the player that is taking its turn right now.
		SoGameBitPlayer* pCurrentPlayer = _pGame->GetCurrentPlayer();
		SoGameBitPlayerHuman* pCurrentPlayerHuman = NULL;
		if ( pCurrentPlayer )
			pCurrentPlayerHuman = pCurrentPlayer->GetAsPlayerHuman();

		//-----------------------
		// Get the player that owns this tile
		SoGameBitPlayer* pOwnerPlayer = _pGame->GetPlayerByPlayerNumber( _pGameBitTile->GetOwnerPlayerNumber() );
		SoGameBitPlayerHuman* pOwnerPlayerHuman = NULL;
		if ( pOwnerPlayer )
			pOwnerPlayerHuman = pOwnerPlayer->GetAsPlayerHuman();
		
		//-----------------------
		// named bools - so the IF block is readable
		bool bTheCurrentPlayerIsHuman = ( pCurrentPlayerHuman != NULL );
		bool bTheTilesOwnerIsHuman = ( pOwnerPlayerHuman != NULL );
		bool bTheCurrentPlayerOwnsThisTile = ( _pGameBitTile->GetOwnerPlayerNumber() == _pGame->GetCurrentPlayerNumber() );
		bool bThereIsOnlyOneHumanPlayer = ( _pGame->GetOwningClient()->GetNumHumanPlayersOnThisClient() == 1 );


		//-----------------------
		// The core logic for visibility rules for tiles that
		// players already own.
		if ( bThereIsOnlyOneHumanPlayer )
		{
			if ( bTheTilesOwnerIsHuman )
				bTileIsVisible = true;
		}
		else
		{
			if ( bTheCurrentPlayerOwnsThisTile )
			{
				if ( bTheCurrentPlayerIsHuman )
					bTileIsVisible = pCurrentPlayerHuman->GetIsViewTilesToggleOn();
			}
		}

		return bTileIsVisible;
	}

	return false;
}



/******************************************************************************
 FUNCTION:  InitializeOrganizePosArrays
******************************************************************************/
void SoActorVisibleTile::InitializeOrganizePosArrays()
{
	for ( int i = 0; i < GAMERULES_MAX_PLAYERS; i ++ )
	{
		JpVector2f vPlayerPos = SoActorVisiblePlayer::GetBoardWorldPositionForPlayer( i );

		for ( int j = 0; j < GAMERULES_NUM_TILES_PER_HAND; j ++ )
		{
			// Set the position
			float fXBase = -150.0f * ( GAMERULES_NUM_TILES_PER_HAND / 2 );

			float fX = ( j * 150.0f ) + fXBase;
			float fY = 0.0f;

			JpVector2f vOffset( fX, fY );
			_OrganizePosArrays[ i ][ j ].SetOffset( vOffset );


			// Clear the owning tile
			_OrganizePosArrays[ i ][ j ].SetOwningTile( TILE_UNDEFINED_ID );
		}
	}
}



/******************************************************************************
 FUNCTION:  AssignOrganizePos
******************************************************************************/
int SoActorVisibleTile::AssignOrganizePos( SoGameBitTile* pGameBitTile )
{
	int iPlayer = pGameBitTile->GetOwnerPlayerNumber();
	int iTileId = pGameBitTile->GetId();

	if ( iPlayer >= 0 && iPlayer < GAMERULES_MAX_PLAYERS )
	{
		for ( int i = 0; i < GAMERULES_NUM_TILES_PER_HAND; i ++ )
		{
			if ( _OrganizePosArrays[ iPlayer ][ i ].GetOwningTile() == TILE_UNDEFINED_ID )
			{
				_OrganizePosArrays[ iPlayer ][ i ].SetOwningTile( iTileId );
				return i;
			}
		}
	}

	// There is an actual bug that causes the assert to fire.
	// However, since tiles are offscreen now, it sorta doesn't matter.
	// So for now, I'm just commenting this out.  Will need to fix
	// later if tiles are made visible to players again.
	// mdomowicz 2014_02_20 
	//ASSERTF( 0, ( "\nCould not assign index." ) );
	return -1;
}



/******************************************************************************
 FUNCTION:  UnassignOrganizePos
******************************************************************************/
void SoActorVisibleTile::UnassignOrganizePos( SoGameBitTile* pGameBitTile )
{
	int iPlayer = pGameBitTile->GetOwnerPlayerNumber();
	int iTileId = pGameBitTile->GetId();

	if ( iPlayer >= 0 && iPlayer < GAMERULES_MAX_PLAYERS )
	{
		for ( int i = 0; i < GAMERULES_NUM_TILES_PER_HAND; i ++ )
		{
			if ( _OrganizePosArrays[ iPlayer ][ i ].GetOwningTile() == iTileId )
				_OrganizePosArrays[ iPlayer ][ i ].SetOwningTile( TILE_UNDEFINED_ID );
		}
	}
}



/******************************************************************************
 FUNCTION:  GetOrganizePosFromIndex
******************************************************************************/
JpVector2f SoActorVisibleTile::GetOrganizePosFromIndex( SoGameBitTile* pGameBitTile, int iIndex )
{
	if ( iIndex >= 0 && iIndex < GAMERULES_NUM_TILES_PER_HAND )
		return _OrganizePosArrays[ pGameBitTile->GetOwnerPlayerNumber() ][ iIndex ].GetOffset();

	return SoUtils::_v2fZero;	
}



