/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "JpRandom.h"

#include "SoCamera.h"
#include "SoCameraGameplay.h"
#include "SoBillboardMist.h"
#include "SoSprite.h"
#include "SoUtils.h"

#include "mmgrChooser.h"


/******************************************************************************
 FUNCTION:  SoMistCloud
******************************************************************************/
SoMistCloud::SoMistCloud() :
_pSprite( NULL ),
_pParentNode( NULL )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoMistCloud::Init( CCNode* pParentNode, const char* pacMistTexturePathAndFilename )
{
	_pParentNode = pParentNode;
	_pSprite = CreateSprite( _pParentNode, pacMistTexturePathAndFilename );
	SoUtils::SetRenderZOrder( _pParentNode, _pSprite, BASE_RENDER_ORDER_MIST_ABOVE );

	_eCurrentSubState = eSubState_FadeIn;

	_fRotation = JpRandom::G.GetFloat() * 360.0f * DEG2RAD;
	_fRotationRate = 8.0f * JpRandom::G.GetFloatOneToMinusOne() * DEG2RAD;

	_fTimer = 0.0f;
	_fLifetime = Interpolate( 5.0f, 30.0f, JpRandom::G.GetFloat() );

	_fAlpha = 0.0f;
	_fScale = Interpolate( 15.0f, 21.0f, JpRandom::G.GetFloat() );

	_fScaleRate = Interpolate( 0.1f, 0.5f, JpRandom::G.GetFloat() );

	UpdateBillboard();
}




/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoMistCloud::Deinit()
{
	_pParentNode->removeChild( _pSprite, DCG_CLEANUP_ALL_RUNNING_ACTIONS );
	_pSprite->release();
}




/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoMistCloud::Update( float dt )
{
	SoCameraManager* pCameraManager = SoCameraManager::GetInstance();
	SoCamera* pCamera = pCameraManager->GetCurrentCamera();
	JpVector2f vCameraPos = pCamera->GetPosition();

	_fTimer += dt;

	float fDistToCameraSq = LengthSquared( _vPosition - vCameraPos );

	switch ( _eCurrentSubState )
	{
		case eSubState_FadeIn:
		{
			_fAlpha += dt * 0.0125f;
			_fAlpha = Min( _fAlpha, 0.25f );
			if ( _fAlpha == 0.25f )
			{
				_fTimer = 0.0f;
				_eCurrentSubState = eSubState_Show;
			}

			break;
		}

		case eSubState_Show:
		{
			if ( _fTimer > _fLifetime )
			{
				_fTimer = 0.0f;
				_eCurrentSubState = eSubState_FadeOut;
			}

			break;
		}

		case eSubState_FadeOut:
		{
			_fAlpha -= dt * 0.0125f;
			_fAlpha = Max( _fAlpha, 0.0f );
			if ( _fAlpha == 0.0f )
			{
				_fTimer = 0.0f;
				_eCurrentSubState = eSubState_Done;				
			}

			break;
		}

		case eSubState_Done:
		{
			break;
		}

		default:
			ASSERTF( 0, ( "\nUnkonw substate" ) );
	}


	_fRotation += ( _fRotationRate * dt );
	_fScale += ( _fScaleRate * dt );
	_fScale = Clamp( _fScale, 5.0f, 20.0f );

	UpdateBillboard();
}



/******************************************************************************
 FUNCTION:  UpdateBillboard
******************************************************************************/
void SoMistCloud::UpdateBillboard()
{
	float fWidth = _pSprite->getContentSize().width;
	float fHeight = _pSprite->getContentSize().height;

 	_pSprite->setScaleX( _fScale * SoUtils::GetiPadRatioX() * SoUtils::GetSpriteZoom() );
 	_pSprite->setScaleY( _fScale * SoUtils::GetiPadRatioY() * SoUtils::GetSpriteZoom() );
 	_pSprite->setColor( ccc3( 255, 255, 255 ) );
	_pSprite->setOpacity( 128.0f * _fAlpha );
 	_pSprite->setRotation( _fRotation * RAD2DEG );
	SetPosition( _vPosition );
}



/******************************************************************************
 FUNCTION:  SetPosition
******************************************************************************/
void SoMistCloud::SetPosition( JpVector2f& v )
{ 
	_vPosition = v; 
	if ( _pSprite )
	{	
		JpVector2f vScreenPos = SoUtils::TransformWorldToScreenSpace( _vPosition );
		_pSprite->SetPosition( vScreenPos );
	}
}



/******************************************************************************
 FUNCTION:  CreateSprite
******************************************************************************/
SoSprite* SoMistCloud::CreateSprite( CCNode* pParent, const char* pacPathAndFilename )
{
	SoSprite* pSprite = SoSprite::CreateScaledToDevice( pacPathAndFilename );
	pSprite->retain();
	pParent->addChild( pSprite );
	pSprite->setVisible( true );
	pSprite->setAnchorPoint( ccp( 0.5f, 0.5f ) );
	return pSprite;
}










/******************************************************************************
 FUNCTION:  SoBillboardMist
******************************************************************************/
SoBillboardMist::SoBillboardMist() :
_pParentNode( NULL ),
_iMinCloudCount( 0 ),
_iMaxCloudCount( 0 )
{
}



/******************************************************************************
 FUNCTION:  ~SoBillboardMist
******************************************************************************/
SoBillboardMist::~SoBillboardMist()
{
}


/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoBillboardMist::Init
( 
	CCNode* pParentNode, 
	const char* pacMistTexturePathAndFilename, 
	int iMinBillboardCount 
)
{
	_pParentNode = pParentNode;
	_sMistTexturePathAndFilename = pacMistTexturePathAndFilename;

	//-----------------------
	// create the billboard set
	_iMinCloudCount = iMinBillboardCount;
	_iMaxCloudCount = _iMinCloudCount * 2;
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoBillboardMist::Deinit()
{
	SoMistCloudList::Iterator Iter( &_CloudList );
	while ( Iter.Next() )
	{
		SoMistCloud* pCloud = Iter.GetNodeData();
		pCloud->Deinit();
		delete pCloud;
	}

	_CloudList.RemoveAll();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoBillboardMist::Update( float dt )
{
	SoMistCloudList::Iterator Iter( &_CloudList );
	while ( Iter.Next() )
	{
		SoMistCloud* pCloud = Iter.GetNodeData();

		if ( pCloud->IsDone() )
		{
			_CloudList.Remove( pCloud );
			pCloud->Deinit();
			delete pCloud;
		}
		else
		{
			pCloud->Update( dt );
		}
	}

	if ( _CloudList.GetNumEntries() < _iMinCloudCount )
	{
		int iNumNewClouds = 1;//JpRandom::G.GetInt() % 10;
		for ( int i = 0; i < iNumNewClouds; i ++ )
			CreateAndAddCloud();
	}
}



/******************************************************************************
 FUNCTION:  CreateAndAddCloud
******************************************************************************/
void SoBillboardMist::CreateAndAddCloud()
{
	if ( _CloudList.GetNumEntries() < _iMaxCloudCount )
	{
		SoCameraManager* pCameraManager = SoCameraManager::GetInstance();
		SoCamera* pCamera = pCameraManager->GetCurrentCamera();

		JpVector2f vCameraPos = SoUtils::_v2fZero;

		JpVector2f vOffset = SoUtils::_v2fZero;
		float fMaxRangeX = CAMERA_GAMEPLAYER_MAX_WORLD_X * 2.0f;
		float fMaxRangeY = CAMERA_GAMEPLAYER_MAX_WORLD_Y * 2.0f;
		vOffset.X += Interpolate( -fMaxRangeX, fMaxRangeX, JpRandom::G.GetFloat() );
		vOffset.Y += Interpolate( -fMaxRangeY, fMaxRangeY, JpRandom::G.GetFloat() );
		
		JpVector2f vCloudPos = vCameraPos + vOffset;


		SoMistCloud* pNewCloud = new SoMistCloud;
		pNewCloud->Init( _pParentNode, _sMistTexturePathAndFilename.c_str() );
		pNewCloud->SetPosition( vCloudPos );

		_CloudList.AddFirst( pNewCloud );
	}
}


