/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "JpDebug.h"
#include "JpNoise.h"
#include "JpUtils.h"

#include "SoActorManager.h"
#include "SoActorVisible.h"
#include "SoCameraGameplay.h"
#include "SoTextureUtils.h"

#include "mmgrChooser.h"



/******************************************************************************
 FUNCTION:  SoCameraGameplay
******************************************************************************/
SoCameraGameplay::SoCameraGameplay() :
_fTimeElapsed( 0.0f ),
_gFocusActor( 0 ),
_fCurrentZoom( 1.0f ),
_fTargetZoom( 1.0f ),
_fShakeDuration( 0.0f ),
_fShakeMagnitude( 0.0f ),
_fShakeTimer( 0.0f ),
_vVelocity( 0.0f, 0.0f ),
_fSpeedMax( 0.0f ),
_fDrag( 0.0f ),
_vInputTouchVelocity( 0.0f, 0.0f ),
_vMoveToTargetPos( 0.0f, 0.0f ),
_vStartPos( 0.0f, 0.0f ),
_vEdgePanDir( 0.0f, 0.0f ),
_bUserIsTouching( false ),
_eState( STATE_STANDARD ),
_vStandardPos( 0.0f, 0.0f ),
_fStandardZoom( 1.0f )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoCameraGameplay::Init()
{
	SoCamera::Init();

	_fTimeElapsed = 0.0f;

	_fSpeedMax = 50.0f;
	_fDrag = 0.9f;

	_vStandardPos = CAMERA_GAMEPLAY_STANDARD_POS_1;
	_fStandardZoom = CAMERA_GAMEPLAY_STANDARD_ZOOM_1;
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoCameraGameplay::Deinit()
{
	SoCamera::Deinit();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoCameraGameplay::Update( float dt )
{
	_fTimeElapsed += dt;

	switch ( _eState )
	{
		case STATE_STANDARD:
		{
			if ( _fCurrentZoom != _fTargetZoom )
			{
				_fCurrentZoom = InterpolateClamped( _fCurrentZoom, _fTargetZoom, 10.0f * dt );

				if ( fabsf( _fCurrentZoom - _fTargetZoom ) < 0.005f )
					_fCurrentZoom = _fTargetZoom;

				SetZoom( _fCurrentZoom );
			}

			if ( Length( _vVelocity ) > _fSpeedMax )
			{
				_vVelocity.Normalize();
				_vVelocity *= _fSpeedMax;
			}

			_vVelocity -= _vVelocity * ( _fDrag * dt );

			JpVector2f vOldCameraPos = GetPosition();
			
			JpVector2f vNewCameraPos = vOldCameraPos + _vVelocity;


			vNewCameraPos.X = Clamp( vNewCameraPos.X, -CAMERA_GAMEPLAYER_MAX_WORLD_X, CAMERA_GAMEPLAYER_MAX_WORLD_X );
			vNewCameraPos.Y = Clamp( vNewCameraPos.Y, -CAMERA_GAMEPLAYER_MAX_WORLD_Y, CAMERA_GAMEPLAYER_MAX_WORLD_Y );


			// If the camera is near the a snap point, snap to it.
			// If snapping occurs, it overrides the vNewCameraPos calculations above..
			// Update: disabling snapping for Achoir, for now.  mdomowicz 2014_02_01
			// SnapToPosition( &vNewCameraPos, _vStartPos, dt );

			SetPosition( vNewCameraPos );

			break;
		}

		case STATE_EDGE_PAN:
		{
			JpVector2f vNewPos = GetPosition() + ( _vEdgePanDir * 1000.0f * dt );
			vNewPos.X = Clamp( vNewPos.X, -CAMERA_GAMEPLAYER_MAX_WORLD_X, CAMERA_GAMEPLAYER_MAX_WORLD_X );
			vNewPos.Y = Clamp( vNewPos.Y, -CAMERA_GAMEPLAYER_MAX_WORLD_Y, CAMERA_GAMEPLAYER_MAX_WORLD_Y );

			SetPosition( vNewPos );
			break;
		}

		case STATE_MOVE_TO_TARGET:
		{
			JpVector2f vDir = Normalize( _vMoveToTargetPos - _vPosition );

			JpVector2f vNewPos = GetPosition() + ( vDir * 10000.0f * dt );

			float fDistToTargetSq = Length( GetPosition() - _vMoveToTargetPos );
			float fDistToNewPosSq = Length( GetPosition() - vNewPos );

			if ( fDistToNewPosSq > fDistToTargetSq )
				vNewPos = _vMoveToTargetPos;

			SetPosition( vNewPos );

			if ( vNewPos == _vMoveToTargetPos )
				_eState = STATE_STANDARD;

			break;
		}

		case STATE_MOVE_TO_TARGET_AND_ZOOM:
		{
			float fInterpolateSpeed = 10.0f;

			if ( _fCurrentZoom != _fTargetZoom )
			{
				_fCurrentZoom = InterpolateSCurve( _fCurrentZoom, _fTargetZoom, 1.0f, 4.0f, fInterpolateSpeed * dt );

				if ( fabsf( _fCurrentZoom - _fTargetZoom ) < 0.00005f )
					_fCurrentZoom = _fTargetZoom;

				SetZoom( _fCurrentZoom );
			}


			JpVector2f vNewPos = InterpolateSCurve( _vPosition, _vMoveToTargetPos, 1.0f, 4.0f, fInterpolateSpeed * dt );

			SetPosition( vNewPos );

			if ( vNewPos == _vMoveToTargetPos )
				_eState = STATE_STANDARD;

			break;
		}

		default:
			ASSERTF( 0, ( "\nUnknown camera state." ) );
	}
}



/******************************************************************************
 FUNCTION:  MoveCameraTargetPosition
******************************************************************************/
void SoCameraGameplay::MoveCameraTargetPosition( JpVector2f vTargetPos )
{
	_eState = STATE_MOVE_TO_TARGET;
	_vMoveToTargetPos = vTargetPos;
}



/******************************************************************************
 FUNCTION:  MoveCameraTargetPositionAndZoom
******************************************************************************/
void SoCameraGameplay::MoveCameraTargetPositionAndZoom( JpVector2f vTargetPos, float fZoom )
{
	_eState = STATE_MOVE_TO_TARGET_AND_ZOOM;
	_vMoveToTargetPos = vTargetPos;
	_fTargetZoom = fZoom;
}



/******************************************************************************
 FUNCTION:  SetCameraTargetPositionAndZoom
******************************************************************************/
void SoCameraGameplay::SetCameraTargetPositionAndZoom( JpVector2f vPos, float fZoom )
{
	_eState = STATE_STANDARD;
	_vVelocity.Zero();
	_fTargetZoom = fZoom;
	_vMoveToTargetPos = vPos;
	SetZoom( fZoom );
	SetPosition( vPos );
}



/******************************************************************************
 FUNCTION:  MoveToOpenWindowPositionAndZoom
******************************************************************************/
void SoCameraGameplay::MoveToOpenWindowPositionAndZoom()
{
	JpVector2f vPos( 750.0f, -650.0f );
	MoveCameraTargetPositionAndZoom( vPos, 0.28f );
}



/******************************************************************************
 FUNCTION:  MoveToStandardPositionAndZoom
******************************************************************************/
void SoCameraGameplay::MoveToStandardPositionAndZoom()
{
	MoveCameraTargetPositionAndZoom( _vStandardPos, _fStandardZoom );
}



/******************************************************************************
 FUNCTION:  JumpToStandardPositionAndZoom
******************************************************************************/
void SoCameraGameplay::JumpToStandardPositionAndZoom()
{
	SetCameraTargetPositionAndZoom( _vStandardPos, _fStandardZoom );
}



/******************************************************************************
 FUNCTION:  SnapToPosition
******************************************************************************/
void SoCameraGameplay::SnapToPosition( JpVector2f* pvResult, JpVector2f vTarget, float dt  )
{
	if ( !_bUserIsTouching )
	{
		JpVector2f vOldCameraPos = GetPosition();

		float fDistToTarget = ( vTarget - vOldCameraPos ).Length();

		if ( fDistToTarget < 100.0f )
		{
			JpVector2f vDirToTarget = Normalize( vTarget - vOldCameraPos );

			// Make sure we never overshoot the Position.
			JpVector2f vFinalOffset = vDirToTarget * 500.0f * dt;
			if ( vFinalOffset.Length() > fDistToTarget )
				vFinalOffset = vDirToTarget * fDistToTarget;

			*pvResult = vOldCameraPos + vFinalOffset;
		}
	}
}



/******************************************************************************
 FUNCTION:  Activate
******************************************************************************/
void SoCameraGameplay::Activate()
{
	SoCamera::Activate();
}



/******************************************************************************
 FUNCTION:  Deactivate
******************************************************************************/
void SoCameraGameplay::Deactivate()
{
	SoCamera::Deactivate();
}



/******************************************************************************
 FUNCTION:  OnInput
******************************************************************************/
void SoCameraGameplay::OnInput( SoInputDispatcher::InputPacket* pInput )
{
	SoInputDispatcher::InputPacket* pI = pInput; // for conciseness

	if ( pI->_pSoCcTouchesDownList )
	{
		_vInputTouchVelocity.Zero();
		SetVelocity( _vInputTouchVelocity );
		_bUserIsTouching = true;
	}
	else if ( pI->_pSoCcTouchesUpList )
	{
		if ( _vInputTouchVelocity.Length() > 5.0f )
			SetVelocity( _vInputTouchVelocity );

		_bUserIsTouching = false;
	}
	else if ( pI->_pSoCcTouchesMoveList )
	{
		SoCcTouchList* pNonConstTouchList = (SoCcTouchList*)pI->_pSoCcTouchesMoveList;

		if ( _eState == STATE_STANDARD )
		{			
			if ( pNonConstTouchList->GetNumEntries() > 0 )
			{
				CCTouch* pTouch0 = pNonConstTouchList->GetAtIndex( 0 )->GetData();
				JpVector2f vP0 = SoUtils::ToJpVector2f( pTouch0->locationInView() );
				vP0.Y *= -1.0f;

				CCTouch* pTouch1 = pNonConstTouchList->GetAtIndex( 0 )->GetData();
				JpVector2f vP1 = SoUtils::ToJpVector2f( pTouch1->previousLocationInView() );
				vP1.Y *= -1.0f;

				JpVector2f vDelta = ( vP1 - vP0 ) * ( 1.0f / SoUtils::GetSpriteZoom() );
				vDelta.X *= ( 1.0f / SoUtils::GetiPadRatioX() );
				vDelta.Y *= ( 1.0f / SoUtils::GetiPadRatioY() );

				JpVector2f vOldCameraPos = GetPosition();
				JpVector2f vNewCameraPos = vOldCameraPos + vDelta;
				SetPosition( vNewCameraPos );


				_vInputTouchVelocity = vDelta;
				SetVelocity( JpVector2f( 0.0f, 0.0f ) );
			}
		}
		else if ( _eState == STATE_EDGE_PAN )
		{
			if ( pNonConstTouchList->GetNumEntries() > 0 )
			{
				//-----------------------
				// get the touch position
				CCTouch* pTouch0 = pNonConstTouchList->GetAtIndex( 0 )->GetData();
				JpVector2f vP0 = SoUtils::ToJpVector2f( pTouch0->locationInView() );

				//-----------------------
				// Determine which edges of the screen the user is touching, if any.
				CCSize SizeInPixels = CCDirector::sharedDirector()->getWinSize();
				float fEdgeSize = 100.0f * SoUtils::GetiPadRatio();

				bool bIsRight = ( vP0.X > ( SizeInPixels.width - fEdgeSize ) );
				bool bIsLeft = ( vP0.X < fEdgeSize );
				bool bIsUp = ( vP0.Y > ( SizeInPixels.height - fEdgeSize ) );
				bool bIsDown = ( vP0.Y < fEdgeSize );


				//-----------------------
				// determine a pan direction based on which edges are being touched.
				_vEdgePanDir.Zero();

				if ( bIsLeft && bIsUp )
					_vEdgePanDir.Init( -1.0f, -1.0f );
				else if ( bIsRight && bIsUp )
					_vEdgePanDir.Init( 1.0f, -1.0f );
				else if ( bIsRight && bIsDown )
					_vEdgePanDir.Init( 1.0f, 1.0f );
				else if ( bIsLeft && bIsDown )
					_vEdgePanDir.Init( -1.0f, 1.0f );
				else if ( bIsLeft )
					_vEdgePanDir.Init( -1.0f, 0.0f );
				else if ( bIsRight )
					_vEdgePanDir.Init( 1.0f, 0.0f );
				else if ( bIsUp )
					_vEdgePanDir.Init( 0.0f, -1.0f );
				else if ( bIsDown )
					_vEdgePanDir.Init( 0.0f, 1.0f );

				_vEdgePanDir.Normalize();
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  GetFocusActor
******************************************************************************/
SoActor* SoCameraGameplay::GetFocusActor() 
{ 
	SoActor* pFocusActor = SoActorManager::GetInstance()->GetActorByGuid( _gFocusActor );
	return pFocusActor; 
}



/******************************************************************************
 FUNCTION:  SetFocusActor
******************************************************************************/
void SoCameraGameplay::SetFocusActor( SoActor* pActor ) 
{ 
	_gFocusActor = 0;
	if ( pActor )
		_gFocusActor = pActor->GetGuid(); 
}



/******************************************************************************
 FUNCTION:  ResetFocus
******************************************************************************/
void SoCameraGameplay::ResetFocus()
{
	SoActor* pFocusActor = SoActorManager::GetInstance()->GetActorByGuid( _gFocusActor );

	if ( pFocusActor )
	{
		SoActorVisible* pActorVisible = pFocusActor->GetAsVisible();
		if ( pActorVisible )
		{
			JpVector2f vActorPos = pActorVisible->GetWorldPosition();
			SetPosition( vActorPos );
		}
	}
}



/******************************************************************************
 FUNCTION:  SetZoom
******************************************************************************/
void SoCameraGameplay::SetZoom( float f )
{
	_fCurrentZoom = f;

	SoUtils::SetSpriteZoom( f );

	JpVector2f vScale( f, f );
	SoActorList* pActorList = SoActorManager::GetInstance()->GetActorList();
	SoActorList::Iterator Iter( pActorList );
	while ( Iter.Next() )
	{
		SoActor* pActor = Iter.GetNodeData();
		
		SoActorVisible* pActorVisible = pActor->GetAsVisible();
		if ( pActorVisible )
		{
			if ( pActorVisible->UseResampledTexture() )
				pActorVisible->SetScale( vScale * ( 1.0f / SoTextureUtils::GetTextureScaler() ) );
			else
				pActorVisible->SetScale( vScale );
		}
	}	
}



/******************************************************************************
 FUNCTION:  InitShake
******************************************************************************/
void SoCameraGameplay::InitCameraShake( float fMagnitude, float fDuration )
{
	_fShakeMagnitude = fMagnitude;
	_fShakeDuration = fDuration;
	_fShakeTimer = _fShakeDuration;
}
