/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "SoCamera.h"
#include "SoConsole.h"
#include "SoCurl.h"
#include "SoFsmStateMain.h"
#include "SoSprite.h"
#include "SoUtils.h"

#include "mmgrChooser.h"


USING_NS_CC;


SoCurl* SoCurl::_pInstance = NULL;



/******************************************************************************
 FUNCTION:  SoCurl
******************************************************************************/
SoCurl::SoCurl() :
_pCurlHandle( NULL ),
_pMultiHandle( NULL ),
_iNumRunningOperations( 0 ),
_bListenerListLocked( false )
{
	Init();
}



/******************************************************************************
 FUNCTION:  ~SoCurl
******************************************************************************/
SoCurl::~SoCurl()
{
	Deinit();
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoCurl::Init()
{
	curl_global_init(CURL_GLOBAL_ALL);

	_pMultiHandle = curl_multi_init();

	SetUrlToPhpFile( CURL_URL_MAIN_PHP );

	_sPostData = "";
	_sGetUrl = "";
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoCurl::Deinit()
{
	curl_multi_cleanup( _pMultiHandle );

	curl_easy_cleanup( _pCurlHandle );

	curl_global_cleanup();

	RemoveAllListeners();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoCurl::Update( float dt )
{
	//-----------------------
	// Here we detect curl operations that are completed,
	// and clean up after them.
	CURLMsg* pCurlMsg;
	int iNumMessageInQueue = 0;

	while ( (pCurlMsg = curl_multi_info_read(_pMultiHandle, &iNumMessageInQueue) ) ) 
	{
		if (pCurlMsg->msg == CURLMSG_DONE) 
		{
			char *url;
			CURL *e = pCurlMsg->easy_handle;

			curl_easy_getinfo(pCurlMsg->easy_handle, CURLINFO_PRIVATE, &url);

			if ( pCurlMsg->data.result != CURLE_OK )
			{
				ASSERTF( 0, ( "\nlibcurl operation didn't work correctly." ) );
				CCLOG( "R: %d - %s <%s>\n", pCurlMsg->data.result, curl_easy_strerror(pCurlMsg->data.result), url );
			}

			curl_multi_remove_handle(_pMultiHandle, e);
			curl_easy_cleanup(e);

			if ( _pCurlHandle == e )
				_pCurlHandle = NULL;
		}
	}


	//-----------------------
	// Update libcurl
	CURLMcode MCode = curl_multi_perform( _pMultiHandle, &_iNumRunningOperations );
	ASSERTF( MCode == CURLM_OK, ( "\nCurl muilti perform failed." ) );



	//-----------------------
	// This is was my old method for detecting when the Curl handle was
	// ready for clean up.  I think the version above is more robust though.
	// Keeping this here for reference.
	// mdomowicz 2013_09_29
	/*
	if ( _iNumRunningOperations == 0 )
	{
		if ( _pCurlHandle )
		{
			curl_multi_remove_handle( _pMultiHandle, _pCurlHandle );
			curl_easy_cleanup( _pCurlHandle );
			_pCurlHandle = NULL;
		}
	}
	*/
}



/******************************************************************************
 FUNCTION:  DoPost
 NOTES:     pacPostData is a string with data, in a format matching how it 
            would be sent to a website in the body of a POST command.
            eg: "param1=5&param2=9" - key value pairs, separated by an &
******************************************************************************/
void SoCurl::DoPost( const char* pacPostData )
{
	if ( _pCurlHandle == NULL )
	{
		// Save the POST data string!  libcurl only stores a pointer to this data, and 
		// DOES NOT COPY IT, so we must guarantee it persistence, else libcurl may 
		// send junk data.  This fixes a pretty bad bug I discovered in the ManhattanC
		// project.  mdomowicz 2013_09_29
		_sPostData = pacPostData;

		// Create a curl handle
		_pCurlHandle = curl_easy_init() ;

		// Setup the URL
		curl_easy_setopt( _pCurlHandle, CURLOPT_URL, _sUrlToPhpFile.c_str() );

		// Setup the POST data
		curl_easy_setopt( _pCurlHandle , CURLOPT_POSTFIELDS , _sPostData.c_str() ) ;

		// Tell libcurl to follow redirection, if the site is redirected
		curl_easy_setopt( _pCurlHandle, CURLOPT_FOLLOWLOCATION, 1L );

		// Set the callback for the returned data
		curl_easy_setopt( _pCurlHandle, CURLOPT_WRITEFUNCTION, &SoCurl::CallbackOnWrite );

		// Set the user data for the callback, in case we need to talk to this SoCurl object
		curl_easy_setopt( _pCurlHandle, CURLOPT_WRITEDATA, this ); 

		// Add the _pCurlHandle to the multi stack
		CURLMcode ResultCurlM = curl_multi_add_handle( _pMultiHandle, _pCurlHandle );
		CheckResultCurlM( ResultCurlM );

		// Start processing
		ResultCurlM = curl_multi_perform( _pMultiHandle, &_iNumRunningOperations );
		CheckResultCurlM( ResultCurlM );
	}
}



/******************************************************************************
 FUNCTION:  DoGet
 NOTES:     pacGetData is a string with data, in a format matching how it 
            would be sent to a website in the URL as part of a GET command.
            For example, to construct this GET request:

               http://www.example.com/index.php?TestId=1&TestId=2&TestId=3

            pacData should be the string:

                TestId=1&TestId=2&TestId=3

            Notice that this doesn't include a '?' at the beginning.  The
            '?' is automatically added.
******************************************************************************/
void SoCurl::DoGet( const char* pacGetData )
{
	if ( _pCurlHandle == NULL )
	{
		// Create a curl handle
		_pCurlHandle = curl_easy_init() ;

		// Build the final URL, which includes the GET data.
		// Note that this URL is saved in a persistent string.  Although the online docs didn't
		// say this was required, it IS required for POST operations (see my notes in DoPost()),
		// and so I'm doing it this way just to be on the safe side.
		// mdomowicz 2013_09_29
		_sGetUrl = _sUrlToPhpFile;
		_sGetUrl += "?";
		_sGetUrl += pacGetData;

		// Setup the URL
		curl_easy_setopt( _pCurlHandle, CURLOPT_URL, _sGetUrl.c_str() );

		// Tell libcurl to follow redirection, if the site is redirected
		curl_easy_setopt( _pCurlHandle, CURLOPT_FOLLOWLOCATION, 1L );

		// Set the callback for the returned data
		curl_easy_setopt( _pCurlHandle, CURLOPT_WRITEFUNCTION, &SoCurl::CallbackOnWrite );

		// Set the user data for the callback, in case we need to talk to this SoCurl object
		curl_easy_setopt( _pCurlHandle, CURLOPT_WRITEDATA, this ); 

		// Add the _pCurlHandle to the multi stack
		CURLMcode ResultCurlM = curl_multi_add_handle( _pMultiHandle, _pCurlHandle );
		CheckResultCurlM( ResultCurlM );

		// Start processing
		ResultCurlM = curl_multi_perform( _pMultiHandle, &_iNumRunningOperations );
		CheckResultCurlM( ResultCurlM );
	}
}



/******************************************************************************
 FUNCTION:  CurlTest
******************************************************************************/
void SoCurl::CurlTest()
{
	if ( _pCurlHandle == NULL )
	{
		// Create a curl handle
		_pCurlHandle = curl_easy_init() ;

		// Set the URL
		curl_easy_setopt( _pCurlHandle, CURLOPT_URL, "http://www.strangejacket.com/manhattan_old.php" );

		// Tell libcurl to follow redirection, if the site is redirected
		curl_easy_setopt( _pCurlHandle, CURLOPT_FOLLOWLOCATION, 1L );

		// Set the callback for the returned data
		curl_easy_setopt( _pCurlHandle, CURLOPT_WRITEFUNCTION, &SoCurl::CallbackOnWrite );

		// Set the user data for the callback, in case we need to talk to this SoCurl object
		curl_easy_setopt( _pCurlHandle, CURLOPT_WRITEDATA, this );

		// Add the _pCurlHandle to the multi stack
		CURLMcode ResultCurlM = curl_multi_add_handle( _pMultiHandle, _pCurlHandle );
		CheckResultCurlM( ResultCurlM );

		// Start processing
		ResultCurlM = curl_multi_perform( _pMultiHandle, &_iNumRunningOperations );
		CheckResultCurlM( ResultCurlM );
	}
}



/******************************************************************************
 FUNCTION:  CurlTestGet
******************************************************************************/
void SoCurl::CurlTestGet()
{
	DoGet( "ComicId=1" );
}



/******************************************************************************
 FUNCTION:  CurlTestPost
******************************************************************************/
void SoCurl::CurlTestPost()
{
	DoPost( "param1=5&param2=9" );
}



/******************************************************************************
 FUNCTION:  CallbackOnWrite
 NOTE:      pUserData should be either NULL, or an SoCurl* 

            !!! Any other value will cause a crash or memory corruption !!!!

******************************************************************************/
size_t SoCurl::CallbackOnWrite( char* pacBuffer, size_t iSizeOfType, size_t iSizeOfBuffer, void* pUserData )
{
	std::string sData;

	for ( size_t c = 0; c < ( iSizeOfType * iSizeOfBuffer ); c++ )
	{
		sData.push_back(pacBuffer[c]);
	}

	if ( pUserData )
	{
		SoCurl* pCurl = (SoCurl*)pUserData;
		pCurl->BroadcastDataToListeners( sData );
	}

	// For debugging
	//CCLOG( sData.c_str() );
	//SoConsole* pConsole = (SoConsole*)NeConsole::GetInstance();
	//pConsole->AddOutputString( sData );
	//pConsole->SetCurrentTextColor( JpUtils::_White );


	return ( iSizeOfType * iSizeOfBuffer ); //tell curl how many bytes we handled
}



/******************************************************************************
 FUNCTION:  CheckResultCurl
******************************************************************************/
void SoCurl::CheckResultCurl( CURLcode Result )
{
	ASSERTF( Result == 0, ( "\nBad result in SoCurl." ) );
}



/******************************************************************************
 FUNCTION:  CheckResultCurlM
******************************************************************************/
void SoCurl::CheckResultCurlM( CURLMcode Result )
{
	ASSERTF( Result == 0, ( "\nBad result in SoCurl." ) );
}



/******************************************************************************
 FUNCTION:  AddListener
******************************************************************************/
void SoCurl::AddListener( SoCurl::Listener* pListener )
{
	ASSERTF( _bListenerListLocked == false, ("Can't add while list is locked.") )
	if ( _bListenerListLocked == false )
		_ListenerList.AddUniquely( pListener, 0 );
}



/******************************************************************************
 FUNCTION:  RemoveListener
******************************************************************************/
void SoCurl::RemoveListener( SoCurl::Listener* pListener )
{
	ASSERTF( _bListenerListLocked == false, ("Can't remove while list is locked.") )
	if ( _bListenerListLocked == false )
		_ListenerList.Remove( pListener );
}



/******************************************************************************
 FUNCTION:  RemoveAllListeners
******************************************************************************/
void SoCurl::RemoveAllListeners()
{
	ASSERTF( _bListenerListLocked == false, ("Can't remove while list is locked.") )
	if ( _bListenerListLocked == false )
		_ListenerList.RemoveAll();
}



/******************************************************************************
 FUNCTION:  BroadcastDataToListeners
******************************************************************************/
void SoCurl::BroadcastDataToListeners( const std::string& sData )
{
	_bListenerListLocked = true;

	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnCurlDataReceived( sData );
	}

	_bListenerListLocked = false;
}



/******************************************************************************
 FUNCTION:  CreateInstance
******************************************************************************/
void SoCurl::CreateInstance()
{
	ASSERTF( !_pInstance, ( "\nCan't create a second instance." ) )
	if ( !_pInstance )
	{
		_pInstance = new SoCurl;
	}
}



/******************************************************************************
 FUNCTION:  DeleteInstance
******************************************************************************/
void SoCurl::DeleteInstance()
{
	if ( _pInstance )
	{
		delete _pInstance;
		_pInstance = NULL;
	}
}



