/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#pragma once

#include "stdafx.h"
#include "cocos2d.h"

#include "SoGameBit.h"


#define PLAYER_UNDEFINED_PLAYER_NUMBER  -1

#define SoGameBitPlayerList JpList< SoGameBitPlayer >
#define SoGameBitPlayerArray JpArray< SoGameBitPlayer >

/*
	Because I keep forgetting.  Here are the different player classes and their purposes:

	SoGameBitPlayer              - The base class for all players

	SoGameBitPlayerAi            - The class used for all Ai Players.  Currently, they use MCTS, but
	                               this would be a good starting point for changing to NeuralNets or
	                               or another method, should that be desired.

	SoGameBitPlayerDumb          - Since there can only be a single instance of either a PlayerHuman or a 
	                               PlayerAi per seat in the game, there is the problem of what to do in 
	                               network multiplayer games, where each seat on each client must be 
	                               represented by a Player instance.  PlayerDumb instances address this.  
	                               When a client must have a Player instance locally to represent a 
	                               PlayerAi/Human on a different client, it creates a PlayerDumb.  For 
	                               example, in a two player game where each player is human, and each 
	                               human is playing on a different client, four Players instances would 
	                               be created, and it would look like this:

	                                           Client 1                     Client 2
	                                   Seat 1  PlayerHuman                  PlayerDumb
	                                   Seat 2  PlayerDumb                   PlayerHuman


	SoGameBitPlayerHuman         - This represents human players.  Contains logic for acquiring choices
	                               from the user.

	SoGameBitPlayerMcts          - This is used whenever the MCTS Ai algorithm is used.  Essentially, the
	                               algorithm creates a PlayerMcts for every Player in the game, and uses
	                               the PlayerMcts to generate moves as part of the MCTS simulation.
	                               Currently PlayerAi is setup to use MCTS, and so PlayerMcts is used
	                               whenever PlayerAi is one of the players.
	                               Note: Technically, what is currently used is PlayerMctsRandom, which
	                               derives from PlayerMcts.  See SoGameStateMcts::MakeCopy()

	SoGameBitPlayerMctsRandom    - This is a flavor of PlayerMcts that makes moves almost completely
	                               randomly, without any heuristic input.  Currently, the game is setup
	                               to use this flavor, but new flavors could be made and used.
	                               To use new flavors, see SoGameStateMcts::MakeCopy()

	Note: PlayerAi and PlayerHuman are, in a sense, the only actual players.  That is, you can
	can think of those classes as being the ones that "take a seat at the table" when a game starts,
	and are the classes that ultimately decide what moves to make.  The other classes are all in 
	service of these two.

	mdomowicz 2015_01_11
*/


class SoGameBitPlayerDumb;
class SoGameState;


class SoGameBitPlayer : public SoGameBit
{
	public:

		//-----------------------
		// public enums
		enum STATE
		{
			STATE_INIT,
			STATE_WAIT,

			STATE_SELECT_ACTION_INIT,
			STATE_SELECT_ACTION,

			STATE_SELECT_NEW_HOTEL_INIT,
			STATE_SELECT_NEW_HOTEL,

			STATE_SELECT_MERGE_SURVIVOR_HOTEL_INIT,
			STATE_SELECT_MERGE_SURVIVOR_HOTEL,

			STATE_SELECT_MERGE_VICTIM_HOTEL_INIT,
			STATE_SELECT_MERGE_VICTIM_HOTEL,

			STATE_SELECT_STOCK_PURCHASE_INIT,
			STATE_SELECT_STOCK_PURCHASE,

			STATE_SELECT_MANAGE_OLD_STOCK_INIT,
			STATE_SELECT_MANAGE_OLD_STOCK,

			STATE_DECLARE_LAST_TURN_INIT,
			STATE_DECLARE_LAST_TURN,

			STATE_CANCEL_ACTION_TIMEOUT,

			STATE_MESSAGE_BOX_UPDATE,

			STATE_UNDO,
			STATE_RESTART_TURN,
			STATE_END_TURN,
			STATE_LAST
		};

		enum PLAYER_COLOR
		{
			PLAYER_COLOR_BLUE,
			PLAYER_COLOR_GREEN,
			PLAYER_COLOR_PURPLE,
			PLAYER_COLOR_RED,
			PLAYER_COLOR_WHITE,
			PLAYER_COLOR_YELLOW,
			PLAYER_COLOR_LAST,
		};

		enum AI_PLAYER_STRENGTH
		{
			AI_PLAYER_STRENGTH_MEDIUM,
			AI_PLAYER_STRENGTH_MEDIUM2,
			AI_PLAYER_STRENGTH_LAST
		};

		//-----------------------
		// public functions
		SoGameBitPlayer();
		virtual ~SoGameBitPlayer() {}

		void MakeCopy( SoGameBitPlayer* pSource );

		virtual void Init( CCNode* pParentNode ); 
		virtual void Deinit();

		virtual void Update( float dt );

		virtual TYPE GetType() { return TYPE_GAMEBIT_PLAYER; }
		virtual SoGameBitPlayer* GetAsPlayer() { return this; }
		virtual SoGameBitPlayerDumb* GetAsPlayerDumb() { return NULL; }


		// !!! VERY IMPORTANT !!!
		// Any functions like these, that change the players state directly, must be
		// overloaded in SoGameBitPlayerDumb so that the Dumb versions are just empty functions
		// that do nothing at all.  See SoGameBitPlayerDumb for examples of this in action.
		// If you don't do this, then Dumb players that are created to represent real (ai or human)
		// players on other clients will have their internal states changed, and when MCTS simulation
		// copies the Dumb players to create Mcts Players, and then simulates forward, the Mcts players
		// will all be in incorrect/unexpected states - leading to lots of problems.
		// mdomowicz 2014_01_22
		virtual void CommandSelectAction() { _eRequestedState = STATE_SELECT_ACTION_INIT; }
		virtual void CommandSelectStockPurchases() { _eRequestedState = STATE_SELECT_STOCK_PURCHASE_INIT; }
		virtual void CommandSelectNewHotel() { _eRequestedState = STATE_SELECT_NEW_HOTEL_INIT; }
		virtual void CommandSelectMergeHotels() { _eRequestedState = STATE_SELECT_MERGE_SURVIVOR_HOTEL_INIT; }
		virtual void CommandSelectManageOldStock() { _eRequestedState = STATE_SELECT_MANAGE_OLD_STOCK_INIT; }

		STATE GetState() { return _eState; }
		void SetState( STATE e ) { _eState = e; }
		STATE GetRequestedState() { return _eRequestedState; }
		void SetRequestedState( STATE e ) { _eRequestedState = e; }

		int GetPlayerNumber() { return _iPlayerNumber; }
		void SetPlayerNumber( int i ) { _iPlayerNumber = i; }

		int GetMoney() { return _iMoney; }
		void SetMoney( int i ) { _iMoney = i; }
		void AdjustMoney( int i );

		int GetStockCount( int iHotelId );
		void SetStockCount( int iHotelId, int iAmount );
		void AdjustStockCount( int iHotelId, int iAdjustment );

		bool IsFirstMoveOfPlayersTurn() { return _bIsFirstMoveOfPlayersTurn; }
		void SetIsFirstMoveOfPlayersTurn( bool b ) { _bIsFirstMoveOfPlayersTurn = b; }

		int GetNumStocksPurchasedThisTurn() { return _iNumStocksPurchasedThisTurn; }
		void SetNumStocksPurchasedThisTurn( int i ) { _iNumStocksPurchasedThisTurn = i; }
		void AdjustNumStocksPurchasedThisTurn( int i );

		bool IsAllowedToPurchaseStock() { return _bIsAllowedToPurchaseStock; }
		void SetIsAllowedToPurchaseStock( bool b ) { _bIsAllowedToPurchaseStock = b; }

		bool MustPrintNonStockPurchaseToLog() { return _bMustPrintNonStockPurchaseToLog; }
		void SetMustPrintNonStockPurchaseToLog( bool b ) { _bMustPrintNonStockPurchaseToLog  = b; }

		bool IsWinner() { return _bIsWinner; }
		void SetIsWinner( bool b ) { _bIsWinner = b; }

		SoGameState* GetOwningGame() { return _pOwningGame; }
		void SetOwningGame( SoGameState* p ) { _pOwningGame = p; }

		AI_PLAYER_STRENGTH GetAiPlayerStrength() { return _eAiPlayerStrength; }
		void SetAiPlayerStrength( AI_PLAYER_STRENGTH e ) { _eAiPlayerStrength = e; }

		PLAYER_COLOR GetPlayerColor() { return GetPlayerColorByNumber( _iPlayerNumber ); }
		static PLAYER_COLOR GetPlayerColorByNumber( int iPlayerNumber );
		static void SetPlayerColorByNumber( int iPlayerNumber, PLAYER_COLOR ePlayerColor );


	protected:

		//-----------------------
		// protected variables
		STATE _eState;
		STATE _eRequestedState;

		int _iPlayerNumber;
		int _iMoney;

		int _iNumStocksPurchasedThisTurn;
		bool _bIsAllowedToPurchaseStock;
		bool _bIsFirstMoveOfPlayersTurn;

		// This is a bit bad, since it's really just for the visible
		// rendering, and not for game rules at all.  This basically
		// prevents the "buys no stock" log text from appearing twice.
		// A previous implementation used a magic value (-1) for
		// the _iNumStocksPurchasedThisTurn to signal this state
		// but the magic number was causing subtle bugs elsewhere,
		// so I've decided to use this less evil solution instead.
		// mdomowicz 2014_02_11
		bool _bMustPrintNonStockPurchaseToLog;

		bool _bIsWinner;

		SoGameState* _pOwningGame;

		int _aiStockCounts[ GAMERULES_NUM_HOTELS ];

		AI_PLAYER_STRENGTH _eAiPlayerStrength;

		static PLAYER_COLOR _aePlayerColors[ GAMERULES_MAX_PLAYERS ];


		//-----------------------
		// protected functions

		void PrintToConsole( const std::string& s );
		void CPrintF( const char* pFormat, ... );
};


