/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

#include "stdafx.h"

#include "NeRandomPoolPicker.h"

#include "SoActorManager.h"
#include "SoFsmStateMain.h"
#include "SoGameBitPlayerMctsHeuristic.h"
#include "SoGameGlobals.h"
#include "SoGameState.h"
#include "SoGuiManager.h"
#include "SoGuiMessage.h"

#include "mmgrChooser.h"


#define STOCK_OWNERSHIP_STATUS_NONE          0
#define STOCK_OWNERSHIP_STATUS_MAJORITY      1
#define STOCK_OWNERSHIP_STATUS_MAJORITY_TIE  2
#define STOCK_OWNERSHIP_STATUS_MINORITY      3
#define STOCK_OWNERSHIP_STATUS_MINORITY_TIE  4

#define HEURISTIC_THOUGHT_CHANCE  0.5f


/******************************************************************************
 FUNCTION:  SoActorPlayerMctsHeuristic
******************************************************************************/
SoGameBitPlayerMctsHeuristic::SoGameBitPlayerMctsHeuristic() :
_PlaceableTilesInHandArray( GAMERULES_NUM_TILES_PER_HAND )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::Init( CCNode* pParentNode )
{
	SoGameBitPlayerMcts::Init( pParentNode );
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::Deinit()
{	
	SoGameBitPlayerMcts::Deinit();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::Update( float dt )
{
	SoGameBitPlayerMcts::Update( dt );
}


/******************************************************************************
 FUNCTION:  SelectAction
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::SelectAction()
{
	if 
	( 
		_pOwningGame->GetLastPlacedTileId() == TILE_UNDEFINED_ID && 
		!_pOwningGame->IsLastTilePlacementSkipped() 
	)
	{
		DoTilePlacement();
	}
	else
	{
		if ( _pOwningGame->CanDeclareLastTurn() )
		{
			// -Heuristic- only declare last turn, if it wins the game for us.
			if ( CanWinIfTurnEndsNow() )
			{
				DoDeclareLastTurn();
			}
			else
			{
				DoEndTurn();
			}
		}
		else
		{
			DoEndTurn();
		}
	}
}



/******************************************************************************
 FUNCTION:  DoTilePlacement
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::DoTilePlacement()
{
	//-----------------------
	// Do this first!  Before any further Ai processing.
	// It updates _PlaceableTilesInHandArray, just below!
	UpdatePlaceableTilesInHandArray();


	//-----------------------
	// Here we select a tile.  Either randomly, or based on a tile score.
	// The potential random selection is needed for the sake of the Mcts
	// algorithm.
	JpRandom* pRng = _pOwningGame->GetMasterRng();
	SoGameBitTile* pSelectedTile = _PlaceableTilesInHandArray.GetRandom( pRng );

	if ( pRng->GetFloat() < HEURISTIC_THOUGHT_CHANCE )
	{
		//-----------------------
		// For use by the Ai from this point on, we calculate
		// and internally cache this player majority and minority stock holder 
		// status each hotel chain, and also note any ties.
		UpdateStockCountStats();
		

		//-----------------------
		// For each tile we can place, calculate a score for that tile.
		// And add it to a list, sorted by tile score.
		SoGameBitTileList SortedTileList;
		SoGameBitTileArray::Iterator PlaceableTilesIter( &_PlaceableTilesInHandArray );
		while ( PlaceableTilesIter.Next() )
		{
			JpRandom* pRng = _pOwningGame->GetMasterRng();
			SoGameBitTile* pTile = PlaceableTilesIter.GetData();

			int iScore = ScoreTilePlacement( pTile );

			// make sure that tiles with equal scores don't order the same every time.
			iScore += pRng->GetIntRanged( 0, 20 );  

			SortedTileList.Add( pTile, iScore );
		}

		//-----------------------
		// select the tile with the highest score.
		SoGameBitTileList::Node* pLastNode = SortedTileList.GetLast();
		pSelectedTile = pLastNode->GetData();
	}


	if ( pSelectedTile )
	{
		_Message.Clear();
		SoCodecPlyrMovePlaceTile TilePlacementEncoder;
		TilePlacementEncoder._iPlayerNumber = GetPlayerNumber();
		TilePlacementEncoder._iPlacedTileId = pSelectedTile->GetId();
		TilePlacementEncoder._bSkippingTilePlacement = false;
		TilePlacementEncoder.Encode( &_Message, 0, 0 );

		SendMessageToMessageTarget( _Message );
	}
	else
	{
		//-----------------------
		// If we are here, then the player is unable to make a tile placement.
		// So we automatically send a message to the server, to signal 
		// that this player is intentionally skipping tile placement.
		// Note: even with the 2008 special rule that says a player can replace an
		// unplayable hand with new tiles at the start of his turn, it is still
		// possible that there are no new tiles left!  So no matter what, we still
		// must handle this corner case.  mdomowicz 2014_01_15
		_Message.Clear();
		SoCodecPlyrMovePlaceTile TilePlacementEncoder;
		TilePlacementEncoder._iPlayerNumber = GetPlayerNumber();
		TilePlacementEncoder._iPlacedTileId = TILE_UNDEFINED_ID;
		TilePlacementEncoder._bSkippingTilePlacement = true;
		TilePlacementEncoder.Encode( &_Message, 0, 0 );

		SendMessageToMessageTarget( _Message );
	}



	_eState = STATE_WAIT;
}



/******************************************************************************
 FUNCTION:  DoSelectStockPurchase
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::DoSelectStockPurchase()
{
	//-----------------------
	// Define local functions.
	struct Local
	{
		static bool IsStockPurchasable2( SoGameState* pGame, int iHotelId, int iCashAvailable, int iNumToPurchase, int* paiAlreadyPlannedPurchases )
		{
			SoGameBitHotel* pHotel = pGame->GetHotelByHotelId( iHotelId );

			int iCostPerStock = SoGameBitHotel::GetPricePerStockFromTable( iHotelId, pHotel->GetSize() );
			int iNumHotelStocksAvailable = pHotel->GetAvailableStock();
			int iNumAlreadyPlannedPurchases = paiAlreadyPlannedPurchases[ iHotelId ];

			bool bIsPlaced = pHotel->IsPlaced();
			bool bPlayerHasFunds = ( ( iCostPerStock * iNumToPurchase ) <= iCashAvailable );
			bool bHotelStocksAreAvailable = ( ( iNumAlreadyPlannedPurchases + iNumToPurchase ) <= iNumHotelStocksAvailable );

			return ( bIsPlaced && bPlayerHasFunds && bHotelStocksAreAvailable );
		}
	};


	//-----------------------
	// This must come before any further Ai processing!
	UpdateStockCountStats();


	//-----------------------
	int aiStocksToAvoid[ GAMERULES_NUM_HOTELS ];
	memset( aiStocksToAvoid, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	int aiPurchases[ GAMERULES_NUM_HOTELS ];
	memset( aiPurchases, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	int iNumStocksPurchased = 0;
	int iPlayerMoneyAfterPurchase = GetMoney();

	JpRandom* pRng = _pOwningGame->GetMasterRng();


	//-----------------------
	// potentially do some heuristic thinking.
	if ( pRng->GetFloat() < HEURISTIC_THOUGHT_CHANCE )
	{
		//-----------------------
		// collect info on stocks we want to avoid if possible.
		for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
		{
			SoGameBitHotel* pHotel = _pOwningGame->GetHotelByHotelId( i );

			// If we have more than half the available stocks, then nobody will ever beat us.  No need to buy more.
			if ( GetStockCount( i ) >= ( GAMERULES_NUM_STOCKS_PER_HOTEL / 2 ) + 1 )
			{
				aiStocksToAvoid[ i ] ++;
			}

			// If we are already in the lead, and have 4 or more stocks than second place, nobody can catch up in
			// a single turn, so no need to buy more.
			int iDeltaBetweenMajorAndMinor = ( _aiMostStockOwnedPerHotel[ i ] - _aiSecondMostStockOwnedPerHotel [ i ] );
			if ( _aiStockOwnershipStatus[ i ] == STOCK_OWNERSHIP_STATUS_MAJORITY && iDeltaBetweenMajorAndMinor >= 4 )
			{
				aiStocksToAvoid[ i ] ++;
			}

			// Avoid stocks that are about to become safe, until we are farther into the game.
			if ( _pOwningGame->GetNumMergersPerformed() <= 3 )
			{
				if ( pHotel->GetSize() > 8 )
					aiStocksToAvoid[ i ] ++;
			}
		}


		//-----------------------
		// In the special case where we have 1 stock and we are the leader (ie, nobody else has
		// any), then buy 3 more, to gain an instant commanding lead.  This situation comes up a LOT
		// when new corporations are created.  Note: only do this if we can buy 3.
		SoGameBitHotelList RankedHotelStocksList;

		for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
		{
			SoGameBitHotel* pHotel = _pOwningGame->GetHotelByHotelId( i );

			if ( GetStockCount( i ) == 1 && _aiStockOwnershipStatus[ i ] == STOCK_OWNERSHIP_STATUS_MAJORITY )
			{
				if ( Local::IsStockPurchasable2( _pOwningGame, i, iPlayerMoneyAfterPurchase, 3, &aiPurchases[0] ) )
				{
					int iPrestige = SoGameBitHotel::GetHotelPrestigeRating( i );
					int iRandom = pRng->GetIntRanged( 0, 20 );
					RankedHotelStocksList.Add( _pOwningGame->GetHotelByHotelId( i ), (100 * iPrestige) + iRandom );
				}
			}
		}

		if ( RankedHotelStocksList.GetNumEntries() > 0 )
		{
			SoGameBitHotel* pTempHotel = RankedHotelStocksList.RemoveLast();
			int iTempHotelId = pTempHotel->GetHotelId();
			aiPurchases[ iTempHotelId ] = 3;
			iNumStocksPurchased = 3;

			int iCostPerStock = SoGameBitHotel::GetPricePerStockFromTable( iTempHotelId, pTempHotel->GetSize() );
			iPlayerMoneyAfterPurchase -= iCostPerStock * 3;
		}
	}


	//-----------------------
	// The preceding logic may have purchased some stocks already.
	// But if less than three were purchased, we can purchase more.
	// We do so here, randomly.
	if ( iNumStocksPurchased < 3 )
	{
		int iMaxValidHotels = GAMERULES_NUM_HOTELS;
		int aiValidHotels[ GAMERULES_NUM_HOTELS ];

		for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
			aiValidHotels[ i ] = i;


		int iMaxStocksToBuy = pRng->GetInt() % (4 - iNumStocksPurchased);

		while ( iNumStocksPurchased < iMaxStocksToBuy && iMaxValidHotels > 0 )
		{
			//-----------------------
			// pick a random hotel
			int iChoice = pRng->GetInt() % iMaxValidHotels;
			int iHotelId = aiValidHotels[ iChoice ];
			SoGameBitHotel* pHotel = _pOwningGame->GetHotelByHotelId( iHotelId );


			//-----------------------
			// verify that it's purchasable
			bool bCanBuyThisStock = Local::IsStockPurchasable2( _pOwningGame, iHotelId, iPlayerMoneyAfterPurchase, 1, &aiPurchases[0] );
			bool bShouldAvoid = ( aiStocksToAvoid[ iHotelId ] > 0 );
			bool bShouldBuyThisStock = ( bCanBuyThisStock && !bShouldAvoid );


			//-----------------------
			// If it is purchasable, buy it.  If not, remove
			// it from the array of possible choices.
			if ( bShouldBuyThisStock )
			{
				aiPurchases[ iHotelId ] ++;
				iNumStocksPurchased ++;

				int iCostPerStock = SoGameBitHotel::GetPricePerStockFromTable( iHotelId, pHotel->GetSize() );
				iPlayerMoneyAfterPurchase -= iCostPerStock;
			}
			else
			{
				// Swap the chosen element with the last element.
				int iTemp = aiPurchases[ iChoice ];
				aiValidHotels[ iChoice ] = aiValidHotels[ iMaxValidHotels - 1 ];
				aiValidHotels[ iMaxValidHotels - 1 ] = iTemp;

				// Update the array limit.
				iMaxValidHotels --;
			}
		}
	}


	_Message.Clear();
	SoCodecPlyrMoveStockPurchases StockPurchasesEncoder;
	StockPurchasesEncoder._iPlayerNumber = GetPlayerNumber();
	StockPurchasesEncoder._iAmountOfHotelWithId0 = aiPurchases[ 0 ];
	StockPurchasesEncoder._iAmountOfHotelWithId1 = aiPurchases[ 1 ];
	StockPurchasesEncoder._iAmountOfHotelWithId2 = aiPurchases[ 2 ];
	StockPurchasesEncoder._iAmountOfHotelWithId3 = aiPurchases[ 3 ];
	StockPurchasesEncoder._iAmountOfHotelWithId4 = aiPurchases[ 4 ];
	StockPurchasesEncoder._iAmountOfHotelWithId5 = aiPurchases[ 5 ];
	StockPurchasesEncoder._iAmountOfHotelWithId6 = aiPurchases[ 6 ];
	StockPurchasesEncoder.Encode( &_Message, 0, 0 );
	SendMessageToMessageTarget( _Message );
	_eState = STATE_WAIT;
}



/******************************************************************************
 FUNCTION:  DoSelectNewHotel
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::DoSelectNewHotel()
{
	//-----------------------
	// This must come before any further Ai processing!
	UpdateStockCountStats();


	//-----------------------
	JpRandom* pRng = _pOwningGame->GetMasterRng();
	SoGameBitHotelList HotelChoices;


	//-----------------------
	// Collect a list of hotels that are placeable.
	for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
	{
		SoGameBitHotel* pHotel = _pOwningGame->GetHotelByHotelId( i );

		if ( !pHotel->IsPlaced() )
		{
			int iHotelPrestige = SoGameBitHotel::GetHotelPrestigeRating( i );

			// This adjusts the value so that relative hotel prestige is constant
			// but hotels with the same prestige are sorted randomly relative to each other.
			iHotelPrestige *= 1000;
			iHotelPrestige += pRng->GetIntRanged( 0, 20 );

			// and add to the list in prestige order.
			HotelChoices.Add( pHotel, iHotelPrestige );
		}
	}

	//-----------------------
	// select a hotel at random from the eligible hotels list.
	SoGameBitHotelList::Node* pRandomNode = HotelChoices.GetRandom( pRng );
	SoGameBitHotel* pSelectedHotel = pRandomNode->GetData();


	//-----------------------
	// however, tend to attempt a more intelligent selection
	if ( pRng->GetFloat() < HEURISTIC_THOUGHT_CHANCE || _pOwningGame->GetGameProgress() == SoGameState::GAME_PROGRESS_EARLY_GAME )
	{
		// Select the most "prestigious" hotel, given certain conditions are met.
		// Note this relies on the special prestige sorting done above.
		SoGameBitHotelList::Iterator Iter( &HotelChoices );
		while ( Iter.Prev() )
		{
			SoGameBitHotel* pHotel = Iter.GetNodeData();

			// If we have a majority (or majority tie), select this hotel.
			int iOwnershipStatus = _aiStockOwnershipStatus[ pHotel->GetHotelId() ];
			if ( iOwnershipStatus == STOCK_OWNERSHIP_STATUS_MAJORITY || iOwnershipStatus == STOCK_OWNERSHIP_STATUS_MAJORITY_TIE )
			{
				pSelectedHotel = pHotel;
				break;
			}

			// ELSE if we have a minority, sometimes select this hotel.
			else if ( iOwnershipStatus == STOCK_OWNERSHIP_STATUS_MINORITY && pRng->GetFloat() < 0.2f )
			{
				pSelectedHotel = pHotel;
				break;
			}
		}
	}


	//-----------------------
	// If no placeable hotel is found, something went wrong.  It is illegal to place 
	// a tile that makes a new hotel when none are available.  Check the tile selection code.
	ASSERTF( pSelectedHotel, ( "\nThere should always be a valid hotel at this point, else how did we get here?" ) );


	//-----------------------
	// send our new hotel selection.
	if ( pSelectedHotel )
	{
		_Message.Clear();

		SoCodecPlyrMoveNewHotelPick NewHotelPickEncoder;
		NewHotelPickEncoder._iPlayerNumber = GetPlayerNumber();
		NewHotelPickEncoder._iHotelId = pSelectedHotel->GetHotelId();
		NewHotelPickEncoder.Encode( &_Message, 0, 0 );

		SendMessageToMessageTarget( _Message );
	}

	_eState = STATE_WAIT;
}



/******************************************************************************
 FUNCTION:  DoSelectMergeSurviorHotel
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::DoSelectMergeSurviorHotel()
{
	JpRandom* pRng = _pOwningGame->GetMasterRng();
	SoGameBitTile* pLastPlacedTile = _pOwningGame->GetLastPlacedTile();


	//-----------------------
	// Pick the surviving hotel
	SoGameBitHotelList SurvivorHotels;
	_pOwningGame->BuildListOfPotentialMergeSurvivorHotels( &SurvivorHotels, pLastPlacedTile->GetTileX(), pLastPlacedTile->GetTileY() );

	SoGameBitHotelList::Node* pRandomSurvivorNode = SurvivorHotels.GetRandom( pRng );
	SoGameBitHotel* pSurvivorHotel = pRandomSurvivorNode->GetData();


	//-----------------------
	// Then pick the victim hotel
	SoGameBitHotelList VictimHotels;
	_pOwningGame->BuildListOfPotentialMergeVictimHotels( &VictimHotels, pLastPlacedTile->GetTileX(), pLastPlacedTile->GetTileY(), pSurvivorHotel );

	SoGameBitHotelList::Node* pRandomVictimNode = VictimHotels.GetRandom( pRng );
	SoGameBitHotel* pVictimHotel = pRandomVictimNode->GetData();


	//-----------------------
	// Send the result.
	_Message.Clear();

	SoCodecPlyrMoveMergeHotels MergeHotelsEncoder;
	MergeHotelsEncoder._iPlayerNumber = GetPlayerNumber();
	MergeHotelsEncoder._iSurvivorHotelId = pSurvivorHotel->GetHotelId(); 
	MergeHotelsEncoder._iVictimHotelId = pVictimHotel->GetHotelId();
	MergeHotelsEncoder.Encode( &_Message, 0, 0 );
	SendMessageToMessageTarget( _Message );

	_eState = STATE_WAIT;
}



/******************************************************************************
 FUNCTION:  DoManageOldStock
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::DoManageOldStock()
{
	int iHotelId = _pOwningGame->GetLastMergerVictimHotelId();
	int iAmountAvailable = GetStockCount( iHotelId );
	int iAmountToKeep = 0;
	int iAmountToTrade = 0;
	int iAmountToSell = 0;
	bool bStockHasBeenManaged = false;


	//-----------------------
	// This must come before any further Ai processing!
	UpdateStockCountStats();


	//-----------------------
	// Special handling if it is the early game.
	JpRandom* pRng = _pOwningGame->GetMasterRng();
	if ( _pOwningGame->GetGameProgress() == SoGameState::GAME_PROGRESS_EARLY_GAME && pRng->GetFloat() < HEURISTIC_THOUGHT_CHANCE )
	{
		if ( _aiStockOwnershipStatus[ iHotelId ] == STOCK_OWNERSHIP_STATUS_MAJORITY )
		{
			// we are in the lead, so figure out how much more stock we have than the runner-up
			int iAmountAhead = ( iAmountAvailable - _aiSecondMostStockOwnedPerHotel[ iHotelId ] );

			// If we are ahead by 4 or less, keep everything.
			// (since you can only buy 3 stocks per turn, a lead of 4 makes this player unbeatable
			// for at least 1 turn).
			int iSurplus = ClampMin( iAmountAhead - 4, 0 );
			iAmountToKeep = ( iAmountAhead - iSurplus );

			// anything left over, manage randomly
			int iAdditionalToKeep = 0;
			ManageOldStockRandomly( &iAdditionalToKeep, &iAmountToTrade, &iAmountToSell, iSurplus );
			iAmountToKeep += iAdditionalToKeep;
			bStockHasBeenManaged = true;
		}
	}


	//-----------------------
	// If no processing was done earlier, do it now.
	if ( !bStockHasBeenManaged )
	{
		ManageOldStockRandomly( &iAmountToKeep, &iAmountToTrade, &iAmountToSell, iAmountAvailable );
	}


	//-----------------------
	// send the message.
	_Message.Clear();
	SoCodecPlyrManageOldStock ManageOldStockEncoder;
	ManageOldStockEncoder._iPlayerNumber = GetPlayerNumber();
	ManageOldStockEncoder._iNumStocksToKeep = iAmountToKeep;
	ManageOldStockEncoder._iNumStocksToSell = iAmountToSell;
	ManageOldStockEncoder._iNumStocksToTrade = iAmountToTrade;
	ManageOldStockEncoder.Encode( &_Message, 0, 0 );
	SendMessageToMessageTarget( _Message );

	_eState = STATE_WAIT;
}



/******************************************************************************
 FUNCTION:  ManageOldStockRandomly
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::ManageOldStockRandomly
( 
	int* piAmountToKeepResult, 
	int* piAmountToTradeResult, 
	int* piAmountToSellResult,
	int iAmountAvailable 
)
{
	JpRandom* pRng = _pOwningGame->GetMasterRng();

	int iAmountToKeep = 0;
	int iAmountToTrade = 0;
	int iAmountToSell = 0;

	SoGameBitHotel* pSurvivingHotel = _pOwningGame->GetLastMergerSurvivorHotel();
	int iAmountSurviorHotelStockAvailable = pSurvivingHotel->GetAvailableStock();


	while ( iAmountAvailable > 0 )
	{
		if ( iAmountAvailable >= 2 && iAmountSurviorHotelStockAvailable > 0 )
		{
			int iChoice = ( pRng->GetInt() % 3 );
			switch ( iChoice )
			{
				case 0:
					iAmountToKeep ++;
					iAmountAvailable --;
					break;
				case 1:
					iAmountToSell ++;
					iAmountAvailable --;
					break;
				case 2:
					iAmountToTrade += 2;
					iAmountAvailable -= 2;
					iAmountSurviorHotelStockAvailable --;
					break;
				default:
					ASSERTF( 0, ( "\nUnhandled case" ) );
			}
		}	
		else
		{
			int iChoice = ( pRng->GetInt() % 2 );
			switch ( iChoice )
			{
				case 0:
					iAmountToKeep ++;
					iAmountAvailable --;
					break;
				case 1:
					iAmountToSell ++;
					iAmountAvailable --;
					break;
				default:
					ASSERTF( 0, ( "\nUnhandled case" ) );
			}
		}
	}

	*piAmountToKeepResult = iAmountToKeep;
	*piAmountToTradeResult = iAmountToTrade;
	*piAmountToSellResult = iAmountToSell;
}



/******************************************************************************
 FUNCTION:  DoDeclareLastTurn
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::DoDeclareLastTurn()
{
	_Message.Clear();

	SoCodecPlyrMoveDeclareLastTurn DeclareLastTurnEncoder;
	DeclareLastTurnEncoder._iPlayerNumber = GetPlayerNumber();
	DeclareLastTurnEncoder.Encode( &_Message, 0, 0 );
	SendMessageToMessageTarget( _Message );

	_eState = STATE_WAIT;
}



/******************************************************************************
 FUNCTION:  DoEndTurn
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::DoEndTurn()
{
	_Message.Clear();
	SoCodecPlyrMoveEndTurn Encoder;
	Encoder._iPlayerNumber = GetPlayerNumber();

	Encoder.Encode( &_Message, 0, 0 );
	SendMessageToMessageTarget( _Message );

	_eState = STATE_WAIT;
}


/******************************************************************************
 FUNCTION:  CanWinIfTurnEnds
******************************************************************************/
bool SoGameBitPlayerMctsHeuristic::CanWinIfTurnEndsNow()
{
	int aiPlayerScores[ GAMERULES_MAX_PLAYERS ];
	memset( aiPlayerScores, 0, GAMERULES_MAX_PLAYERS * sizeof( int ) );


	//-----------------------
	// Get each player's current score.
	SoGameBitPlayerArray::Iterator PlayerIter( _pOwningGame->GetPlayerArray() );
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();
		aiPlayerScores[ pPlayer->GetPlayerNumber() ] = pPlayer->GetMoney();
	}

	//-----------------------
	// Add the majority and minority stock holders bonuses
	// for every hotel that's currently active.
	for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
	{
		SoGameBitHotel* pHotel = _pOwningGame->GetHotelByHotelId( i );

		if ( pHotel->IsPlaced() )
		{
			int aiBonuses[ GAMERULES_MAX_PLAYERS ];
			memset( aiBonuses, 0, GAMERULES_MAX_PLAYERS * sizeof( int ) );

			_pOwningGame->CalculateMajorityMinorityStockOwnersBonusesBasic( i, aiBonuses );

			for ( int j = 0; j < GAMERULES_MAX_PLAYERS; j ++ )
				aiPlayerScores[ j ] += aiBonuses[ j ];
		}
	}

	//-----------------------
	// Add the money earned for selling all active hotel 
	// stock back to the bank
	PlayerIter.Reset();
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();

		for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
		{
			SoGameBitHotel* pHotel = _pOwningGame->GetHotelByHotelId( i );

			if ( pHotel->IsPlaced() )
			{
				int iNumStockOwned = pPlayer->GetStockCount( i );
				int iStockPrice = ( SoGameBitHotel::GetPricePerStockFromTable( i, pHotel->GetSize() ) );
				int iProfit = ( iStockPrice * iNumStockOwned );
				aiPlayerScores[ pPlayer->GetPlayerNumber() ] += iProfit;
			}
		}
	}

	//-----------------------
	// If this player's score is less than any other score,
	// this player is NOT a winner.
	int iMyScore = aiPlayerScores[ GetPlayerNumber() ];
	for ( int i = 0; i < GAMERULES_MAX_PLAYERS; i ++ )
	{
		if ( iMyScore < aiPlayerScores[ i ] )
			return false;
	}

	return true;
}



/******************************************************************************
 FUNCTION:  ScoreTilePlacement
******************************************************************************/
int SoGameBitPlayerMctsHeuristic::ScoreTilePlacement( SoGameBitTile* pTile )
{
	int iFinalScore = 0;
	int iTempResult = 0;
	SoTileSet* pTileSet = _pOwningGame->GetTileSet();
	int iTileX = pTile->GetTileX();
	int iTileY = pTile->GetTileY();


	if ( pTileSet->DoesTileMergeHotelsWhenPlaced( iTileX, iTileY ) )
	{
		iFinalScore = ScoreTilePlacementIfItMergesHotels( pTile );
	}
	else if ( pTileSet->DoesTileCreateNewHotelWhenPlaced( iTileX, iTileY ) )
	{
		iFinalScore = 100;
	}
	else if ( DoesTileExtendHotelChainIOwn( &iTempResult, pTile ) )
	{
		iFinalScore = iTempResult;
	}


	iFinalScore += GetScoreForPairingWithOtherTilesIOwn( pTile );

	return iFinalScore;
}


/******************************************************************************
 FUNCTION:  ScoreTilePlacementIfItMergesHotels
 NOTE:      This function assumes that the given tile is already known to cause
            a merger.  This function does not verify that a merger occurs.
******************************************************************************/
int SoGameBitPlayerMctsHeuristic::ScoreTilePlacementIfItMergesHotels( SoGameBitTile* pTile )
{
	int iFinalScore = 0;
	SoTileSet* pTileSet = _pOwningGame->GetTileSet();
	int iTileX = pTile->GetTileX();
	int iTileY = pTile->GetTileY();

	SoGameBitTile* pNeighborTile = NULL;
	int MergedHotels[ GAMERULES_NUM_HOTELS ];
	memset( MergedHotels, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	//-----------------------
	// Collect all the hotels that are going to be merged.
	pNeighborTile = pTileSet->GetTileByXy( iTileX + 1, iTileY + 0 );
	if ( pNeighborTile && pNeighborTile->GetHotelId() != HOTEL_UNDEFINED_ID )
		MergedHotels[ pNeighborTile->GetHotelId() ] = 1;

	pNeighborTile = pTileSet->GetTileByXy( iTileX - 1, iTileY + 0 );
	if ( pNeighborTile && pNeighborTile->GetHotelId() != HOTEL_UNDEFINED_ID )
		MergedHotels[ pNeighborTile->GetHotelId() ] = 1;

	pNeighborTile = pTileSet->GetTileByXy( iTileX + 0, iTileY + 1 );
	if ( pNeighborTile && pNeighborTile->GetHotelId() != HOTEL_UNDEFINED_ID )
		MergedHotels[ pNeighborTile->GetHotelId() ] = 1;

	pNeighborTile = pTileSet->GetTileByXy( iTileX + 0, iTileY - 1 );
	if ( pNeighborTile && pNeighborTile->GetHotelId() != HOTEL_UNDEFINED_ID )
		MergedHotels[ pNeighborTile->GetHotelId() ] = 1;


	//-----------------------
	// For each hotel that will be merged, see what our score will be as a result.
	// Return the highest score recorded.
	iFinalScore = 0;
	for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
	{
		if ( MergedHotels[ i ] != 0 )
		{
			iFinalScore += ScoreMergedHotel( i );
		}
	}

	return iFinalScore;
}



/******************************************************************************
 FUNCTION:  ScoreMergedHotel
******************************************************************************/
int SoGameBitPlayerMctsHeuristic::ScoreMergedHotel( int iHotelId )
{
	int iFinalScore = 0;

	int aiBonuses[ GAMERULES_MAX_PLAYERS ];
	memset( aiBonuses, 0, GAMERULES_MAX_PLAYERS * sizeof( int ) );

	_pOwningGame->CalculateMajorityMinorityStockOwnersBonusesBasic( iHotelId, aiBonuses );

	//-----------------------
	// Add points for the money we make, if any.
	if ( aiBonuses[ GetPlayerNumber() ] > 0 )
		iFinalScore = 100 + aiBonuses[ GetPlayerNumber() ];
		
	//-----------------------
	// But reduce the score if other players make money too.
	// Note: it's possible for scores to be negative! This means
	// other players do better than us.
	for ( int i = 0; i < GAMERULES_MAX_PLAYERS; i ++ )
	{
		if ( i != GetPlayerNumber() )
		{
			iFinalScore -= aiBonuses[ i ];
		}
	}

	return iFinalScore;
}



/******************************************************************************
 FUNCTION:  DoesTileExtendHotelChainIOwn
******************************************************************************/
bool SoGameBitPlayerMctsHeuristic::DoesTileExtendHotelChainIOwn( int* piScoreResult, SoGameBitTile* pTile )
{
	int iFinalScore = 0;
	SoTileSet* pTileSet = _pOwningGame->GetTileSet();
	int iTileX = pTile->GetTileX();
	int iTileY = pTile->GetTileY();

	SoGameBitTile* pNeighborTile = NULL;
	int NeighborHotels[ GAMERULES_NUM_HOTELS ];
	memset( NeighborHotels, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	//-----------------------
	// Collect all the hotels that neighbor the given tile
	pNeighborTile = pTileSet->GetTileByXy( iTileX + 1, iTileY + 0 );
	if ( pNeighborTile && pNeighborTile->GetHotelId() != HOTEL_UNDEFINED_ID )
		NeighborHotels[ pNeighborTile->GetHotelId() ] = 1;

	pNeighborTile = pTileSet->GetTileByXy( iTileX - 1, iTileY + 0 );
	if ( pNeighborTile && pNeighborTile->GetHotelId() != HOTEL_UNDEFINED_ID )
		NeighborHotels[ pNeighborTile->GetHotelId() ] = 1;

	pNeighborTile = pTileSet->GetTileByXy( iTileX + 0, iTileY + 1 );
	if ( pNeighborTile && pNeighborTile->GetHotelId() != HOTEL_UNDEFINED_ID )
		NeighborHotels[ pNeighborTile->GetHotelId() ] = 1;

	pNeighborTile = pTileSet->GetTileByXy( iTileX + 0, iTileY - 1 );
	if ( pNeighborTile && pNeighborTile->GetHotelId() != HOTEL_UNDEFINED_ID )
		NeighborHotels[ pNeighborTile->GetHotelId() ] = 1;


	//-----------------------
	// If this player owns any stock in any neighboring hotel chains,
	// give it a score.  Give bonus points for majoriy/minorty status.
	for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
	{
		if ( NeighborHotels[ i ] != 0 )
		{
			if ( GetStockCount( i ) > 0 )
			{
				iFinalScore += 25;
				switch ( _aiStockOwnershipStatus[ i ] )
				{
					case STOCK_OWNERSHIP_STATUS_MAJORITY:     iFinalScore += 60;  break;
					case STOCK_OWNERSHIP_STATUS_MAJORITY_TIE: iFinalScore += 50;  break;
					case STOCK_OWNERSHIP_STATUS_MINORITY:     iFinalScore += 40;  break;
					case STOCK_OWNERSHIP_STATUS_MINORITY_TIE: iFinalScore += 30;  break;
					default:
						ASSERTF( 0, ( "\nunknown stock ownership status." ) );
				}
			}
		}
	}

	*piScoreResult = iFinalScore;
	return ( iFinalScore > 0 );
}



/******************************************************************************
 FUNCTION:  GetScoreForPairingWithOtherTilesIOwn
******************************************************************************/
int SoGameBitPlayerMctsHeuristic::GetScoreForPairingWithOtherTilesIOwn( SoGameBitTile* pTile )
{
	int iFinalScore = 0;

	//-----------------------
	// get our four neighbor tiles
	SoTileSet* pTileSet = _pOwningGame->GetTileSet();
	int iX = pTile->GetTileX();
	int iY = pTile->GetTileY();
	SoGameBitTile* p0 = pTileSet->GetTileByXy( iX + 1, iY + 0 );
	SoGameBitTile* p1 = pTileSet->GetTileByXy( iX - 1, iY + 0 );
	SoGameBitTile* p2 = pTileSet->GetTileByXy( iX + 0, iY + 1 );
	SoGameBitTile* p3 = pTileSet->GetTileByXy( iX + 0, iY - 1 );

	//-----------------------
	// Count how many of the tiles in our hand are one of the
	// four neighbors.
	int iNumMatches = 0;
	SoGameBitTileArray::Iterator TileIter( &_PlaceableTilesInHandArray );
	while ( TileIter.Next() )
	{
		SoGameBitTile* p = TileIter.GetData();
		if ( p == p0 || p == p1 || p == p2 || p == p3 )
			iNumMatches ++;
	}

	//-----------------------
	// Score the result.
	switch ( iNumMatches )
	{
		case 0: iFinalScore = 0; break;
		case 1: iFinalScore = 50; break;
		case 2: iFinalScore = 70; break;
		case 3: iFinalScore = 90; break;
		case 4: iFinalScore = 110; break;
		default:
			ASSERTF( 0, ( "\nbad number of matches." ) );
	}

	return iFinalScore;
}



/******************************************************************************
 FUNCTION:  UpdateStockCountStats
 NOTE:      This function determines if this player is the majority or minority
            stock holder for each hotel chain, and notes any ties..  The results 
            are internally cached, so that all of the Ai routines can refer to 
            them.  As such, make sure this is called at the right time in the
            decision making process (ie, as early as possible).

            Much of this implementation is copied from SoGameStateMcts::Evaluate()
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::UpdateStockCountStats()
{
	//-----------------------
	// For every hotel, find out what is the most and second most stock
	// any player might own of that hotel.
	memset( _aiMostStockOwnedPerHotel, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );
	memset( _aiSecondMostStockOwnedPerHotel, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	SoGameBitPlayerArray::Iterator PlayerIter( _pOwningGame->GetPlayerArray() );
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();

		for ( int j = 0; j < GAMERULES_NUM_HOTELS; j ++ )
		{
			int iStockCount = pPlayer->GetStockCount( j );

			if ( iStockCount > _aiMostStockOwnedPerHotel[ j ] )
			{
				_aiSecondMostStockOwnedPerHotel[ j ] = _aiMostStockOwnedPerHotel[ j ];
				_aiMostStockOwnedPerHotel[ j ] = iStockCount;
			}
			else if 
			( 
				iStockCount != _aiMostStockOwnedPerHotel[ j ] &&
				iStockCount > _aiSecondMostStockOwnedPerHotel[ j ] 
			)
			{
				_aiSecondMostStockOwnedPerHotel[ j ] = iStockCount;
			}
		}		
	}

	//-----------------------
	// Count how many players have the most possible stock.
	// And count how many players have the second most possible stock.
	int aiNumOwnersOfMostStock[ GAMERULES_NUM_HOTELS ];	
	int aiNumOwnersOfSecondMostStock[ GAMERULES_NUM_HOTELS ];	
	memset( aiNumOwnersOfMostStock, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );
	memset( aiNumOwnersOfSecondMostStock, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	PlayerIter.Reset();
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();

		for ( int j = 0; j < GAMERULES_NUM_HOTELS; j ++ )
		{
			int iStockCount = pPlayer->GetStockCount( j );

			if ( iStockCount == _aiMostStockOwnedPerHotel[ j ] )
				aiNumOwnersOfMostStock[ j ] ++;
			else if ( iStockCount == _aiSecondMostStockOwnedPerHotel[ j ] )
				aiNumOwnersOfSecondMostStock[ j ] ++;
		}		
	}

	//-----------------------
	// With the info above, we can now determine if we are majority or minority
	// stock holders for each hotel, and if there are any ties.
	memset( _aiStockOwnershipStatus, STOCK_OWNERSHIP_STATUS_NONE, GAMERULES_NUM_HOTELS * sizeof( int ) );

	for ( int j = 0; j < GAMERULES_NUM_HOTELS; j ++ )
	{
		int iStockCount = GetStockCount( j );
		if ( iStockCount == _aiMostStockOwnedPerHotel[ j ] )
		{
			if ( aiNumOwnersOfMostStock[ j ] == 1 )
			{
				_aiStockOwnershipStatus[ j ] = STOCK_OWNERSHIP_STATUS_MAJORITY;
			}
			else
			{
				_aiStockOwnershipStatus[ j ] = STOCK_OWNERSHIP_STATUS_MAJORITY_TIE;
			}
		}
		else if ( iStockCount = _aiSecondMostStockOwnedPerHotel[ j ] )
		{
			if ( aiNumOwnersOfSecondMostStock[ j ] == 1 )
			{
				_aiStockOwnershipStatus[ j ] = STOCK_OWNERSHIP_STATUS_MINORITY;
			}
			else
			{
				_aiStockOwnershipStatus[ j ] = STOCK_OWNERSHIP_STATUS_MINORITY_TIE;
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  UpdatePlaceableTilesInHandArray
 NOTE:      This function cycles through all the tiles and builds a list of
            tiles that this player owns, and which are actually placeable.
            The results are cached, so that they can be used as needed.
            Thus, this function should be called before any Ai processing
            is done.
******************************************************************************/
void SoGameBitPlayerMctsHeuristic::UpdatePlaceableTilesInHandArray()
{
	_PlaceableTilesInHandArray.Clear();

	SoTileSet* pTileSet = _pOwningGame->GetTileSet();
	SoGameBitTileArray* pAllTiles = pTileSet->GetAllTiles();

	SoGameBitTileArray::Iterator TileIter( pAllTiles );
	while ( TileIter.Next() )
	{
		SoGameBitTile* pTile = TileIter.GetData();
		if ( pTile->GetOwnerPlayerNumber() == GetPlayerNumber() && !pTile->IsPlaced() )
		{
			bool bIsPlacable = 
			(
				!_pOwningGame->DoesTileCreateIllegalHotelWhenPlaced( pTile->GetTileX(), pTile->GetTileY() ) &&
				!_pOwningGame->DoesTileMergeSafeHotelsWhenPlaced( pTile->GetTileX(), pTile->GetTileY() )
			);

			if ( bIsPlacable )
			{
				_PlaceableTilesInHandArray.AddLast( pTile );
			}
		}
	}
}
