/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "JpString.h"

#include "SoActorManager.h"
#include "SoActorTypes.h"
#include "SoActorVisibleBoardSpot.h"
#include "SoActorVisibleCoordinate.h"
#include "SoActorVisibleHotel.h"
#include "SoActorVisiblePlayer.h"
#include "SoActorVisibleTile.h"
#include "SoCodec.h"
#include "SoConsole.h"
#include "SoFsmStateMain.h"
#include "SoGameClient.h"
#include "SoGameGlobals.h"
#include "SoGameState.h"
#include "SoGuiAnimWinMessageBox.h"
#include "SoGuiManager.h"
#include "SoGuiMessage.h"
#include "SoHud.h"
#include "SoNetworkComms.h"

#include "mmgrChooser.h"


USING_NS_CC;


/******************************************************************************
 FUNCTION:  SoGameClient
******************************************************************************/
SoGameClient::SoGameClient() :
_hServerId( 0 ),
_hClientId( 0 ),
_eCurrentState( STATE_LAST ),
_pGame( NULL ),
_iNumHumanPlayersOnThisClient( 0 ),
_gGuiMessageBox( 0 ),
_pGuiMessageBox( NULL ),
_iMessageBoxState( 0 ),
_fMessageBoxTimer( 0.0f ),
_iMessageBoxPlayerNumber( PLAYER_UNDEFINED_PLAYER_NUMBER ),
_iLastMessageNumReceivedFromServer( -1 ),
_iLastMessageNumSentToServer( -1 )
{
}



/******************************************************************************
 FUNCTION:  ~SoGameClient
******************************************************************************/
SoGameClient::~SoGameClient()
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoGameClient::Init()
{
	SoNetworkComms* pNetwork = SoGameGlobals::GetNetworkComms();
	_hClientId = pNetwork->RequestNewClientId();

	_eCurrentState = STATE_WAIT_FOR_SERVER_ADD;

	_pGame = new SoGameState;
	_pGame->Init();
	_pGame->SetOwningClientId( _hClientId );
	_pGame->SetOwningClient( this );
	_pGame->SetOwningServerId( 0 );

	SoHud::GetInstance()->GetHotelStats()->SetGameState( _pGame );

	_iNumHumanPlayersOnThisClient = 0;

	_iLastMessageNumReceivedFromServer = -1;
	_iLastMessageNumSentToServer = -1;

	SoGuiManager* pGuiManager = SoGuiManager::GetInstance();
	_pGuiMessageBox = new SoGuiAnimWinMessageBox;
	_pGuiMessageBox->SetGame( _pGame );
	_gGuiMessageBox = pGuiManager->AddGui( _pGuiMessageBox, NULL );
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoGameClient::Deinit()
{
	delete _pGame;
	_pGame = NULL;
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoGameClient::Update( float dt )
{
	SoHudLog* pLog = SoFsmMainLayer::GetSharedHudLog();
	SoConsole* pConsole = (SoConsole*)NeConsole::GetInstance();
	SoNetworkComms* pNetwork = SoGameGlobals::GetNetworkComms();
	SoGameMessage* pInMsg = pNetwork->GetNextMessageForClient( _hServerId, _hClientId );


	if ( !SoGameGlobals::IsLoadingGame() )
		UpdateAnotherPlayerIsActingMessageBox( dt );

	if ( _pGame )
		_pGame->Update( dt );

	//-----------------------
	// The Server sequentially numbers the messages it sends so that the Clients can be sure
	// to handle the messages in the order they were sent.  In local games, this is never actually
	// a problem, but in online games message order is not guaranteed.
	// Here we are checking to make sure that the next message coming in has the expected numbering.  
	// If it does not, then it is an out-of-sequence message, and we ignore it for now.  We must 
	// wait until the correct next message comes in.  mdomowicz 2013_11_16
	if ( pInMsg && pInMsg->GetMessageNumber() == _iLastMessageNumReceivedFromServer + 1 )
	{
		_iLastMessageNumReceivedFromServer ++;

		_RngForSaveLoadSupport.SetHigh( _pGame->GetMasterRng()->GetHigh() );
		_RngForSaveLoadSupport.SetLow( _pGame->GetMasterRng()->GetLow() );

		switch ( _eCurrentState )
		{
			case STATE_WAIT_FOR_SERVER_ADD:
			{
				if ( !pInMsg )
					break;

				if ( pInMsg->IsType( Msg::TYPE_SVR_COMMAND_ADD_CLI ) )
				{
					SoCodecSvrCommandAddCli Decoder;
					Decoder.Decode( pInMsg );
					CPrintF( "Client %x has connected to Server %x", _hClientId, _hServerId );

					_pGame->SeedMasterRng( Decoder._hRngSeed );
					CPrintF( "Master Rng Seed is: %x", Decoder._hRngSeed );

					pNetwork->RemoveAndDeleteMessageForClient( pInMsg );

					if ( pNetwork->GetAsNetworkCommsOnline() )
						pLog->AddMessageF( JpUtils::_Yellow, "Waiting for all clients to connect..." );

					_eCurrentState = STATE_LOBBY;
				}
				else
				{
					ASSERTF( 0, ( "\nClient Message handling problem.  Unexpected state for this message type" ) );
				}

				break;
			}
			case STATE_LOBBY:
			{
				if ( !pInMsg )
					break;

				if ( pInMsg->IsType( Msg::TYPE_SVR_COMMAND_SEAT_ASSIGNMNET ) )
				{
					SoCodecSvrCommandSeatAssignment Decoder;
					Decoder.Decode( pInMsg );

					if ( Decoder._hClientToReceiveSeat == _hClientId )
					{
						if ( Decoder._eGameBitTypeOfNewPlayer == SoGameBit::TYPE_GAMEBIT_PLAYER_AI )
						{
							SoGameBitPlayer* pNewPlayer = _pGame->AddPlayerAi();
							pNewPlayer->SetPlayerNumber( Decoder._iPlayerNumber );
							pNewPlayer->SetAiPlayerStrength( Decoder._eAiPlayerStrength );
							CPrintF( "Client %x has added an Ai Player", _hClientId );
						}
						else if ( Decoder._eGameBitTypeOfNewPlayer == SoGameBit::TYPE_GAMEBIT_PLAYER_MCTS_RANDOM )
						{
							SoGameBitPlayer* pNewPlayer = _pGame->AddPlayerMctsRandom();
							pNewPlayer->SetPlayerNumber( Decoder._iPlayerNumber );
							CPrintF( "Client %x has added an MctsPayout Player", _hClientId );
						}
						else if ( Decoder._eGameBitTypeOfNewPlayer == SoGameBit::TYPE_GAMEBIT_PLAYER_HUMAN )
						{
							SoGameBitPlayer* pNewPlayer = _pGame->AddPlayerHuman();
							pNewPlayer->SetPlayerNumber( Decoder._iPlayerNumber );
							CPrintF( "Client %x has added a Human Player", _hClientId );
							_iNumHumanPlayersOnThisClient ++;
						}
					}
					else
					{
						SoGameBitPlayer* pNewPlayer = _pGame->AddPlayerDumb();
						pNewPlayer->SetPlayerNumber( Decoder._iPlayerNumber );
						CPrintF( "Client %x has added a Dumb Player", _hClientId );
					}

					_pGame->SortPlayerArrayByPlayerNumber();


					pNetwork->RemoveAndDeleteMessageForClient( pInMsg );

					std::string sPlayerName = SoActorVisiblePlayer::GetPlayerNameByPlayerNumber( Decoder._iPlayerNumber );
					pLog->AddMessageF( JpUtils::_White, "Player '%s' joined the game", sPlayerName.c_str() );
				}
				else if ( pInMsg->IsType( Msg::TYPE_SVR_COMMAND_INIT_NEW_GAME ) )
				{
					pLog->AddMessageF( JpUtils::_White, "Preparing new game..." );

					_pGame->HandleMessage( pInMsg );
					pNetwork->RemoveAndDeleteMessageForClient( pInMsg );


					// This is a really cheesy way of making the above line render full opaque, before 
					// creating the visible actors.  mdomwicz 2013_11_21
					pLog->Update( 100.0f );

					CreateVisibleActors();

					_eCurrentState = STATE_START_NEW_GAME;
				}
				else
				{
					ASSERTF( 0, ( "\nClient Message handling problem.  Unexpected state for this message type" ) );
				}

				break;
			}
			case STATE_START_NEW_GAME:
			{
				if ( !pInMsg )
					break;

				if 
				( 
					pInMsg->IsType( Msg::TYPE_SVR_COMMAND_PLYR_TAKE_TURN ) ||
					pInMsg->IsType( Msg::TYPE_SVR_COMMAND_PLYR_PURCHASE_STOCK ) ||
					pInMsg->IsType( Msg::TYPE_SVR_COMMAND_PLYR_PICK_NEW_HOTEL ) ||
					pInMsg->IsType( Msg::TYPE_SVR_COMMAND_PLYR_PICK_MERGE_HOTELS ) ||
					pInMsg->IsType( Msg::TYPE_SVR_COMMAND_PLYR_MANAGE_OLD_STOCK ) ||

					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_UNDO ) ||
					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_RESTART_TURN ) ||
					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_END_TURN ) ||
					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_STOCK_PURCHASES ) ||
					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_NEW_HOTEL_PICK ) ||
					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_MERGE_HOTELS ) ||
					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_MANAGE_OLD_STOCK ) ||
					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_DECLARE_LAST_TURN ) ||
					pInMsg->IsType( Msg::TYPE_PLYR_MOVE_PLACE_TILE ) ||
					pInMsg->IsType( Msg::TYPE_DEBUG_COMMAND )
				)
				{
					_pGame->HandleMessage( pInMsg );
					pNetwork->RemoveAndDeleteMessageForClient( pInMsg );
				}
				else
				{
					ASSERTF( 0, ( "\nClient Message handling problem.  Unexpected state for this message type" ) );
				}

				break;
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  UpdateAnotherPlayerIsActingMessageBox
******************************************************************************/
void SoGameClient::UpdateAnotherPlayerIsActingMessageBox( float dt )
{
	_fMessageBoxTimer += dt;

	SoGuiManager* pGuiManager = SoGuiManager::GetInstance();
	char acText[ 128 ] = "";

	int iStockManagingPlayer = _pGame->GetPlayerNumberCommandedToManageOldStock();

	SoGameBitPlayer* pStockManagingPlayer = _pGame->GetPlayerByPlayerNumber( iStockManagingPlayer );

	bool bAPlayerIsManagingStock = ( iStockManagingPlayer != PLAYER_UNDEFINED_PLAYER_NUMBER );
	bool bStockManagingPlayerIsLocalHuman = ( pStockManagingPlayer && pStockManagingPlayer->GetAsPlayerHuman() != NULL );


	const int MESSAGEBOX_STATE_WAIT = 0;
	const int MESSAGEBOX_STATE_MANAGESTOCK_INIT = 10;
	const int MESSAGEBOX_STATE_MANAGESTOCK_UPDATE = 20;
	const int MESSAGEBOX_STATE_CLOSE_INIT = 50;
	const int MESSAGEBOX_STATE_CLOSE_UPDATE = 60;

	switch ( _iMessageBoxState )
	{
		case MESSAGEBOX_STATE_WAIT:
		{
			if ( bAPlayerIsManagingStock && !bStockManagingPlayerIsLocalHuman )
				_iMessageBoxState = MESSAGEBOX_STATE_MANAGESTOCK_INIT;

			break;
		}

		case MESSAGEBOX_STATE_MANAGESTOCK_INIT:
		{
			if ( iStockManagingPlayer == PLAYER_UNDEFINED_PLAYER_NUMBER )
			{
				// In fast Ai games, it can happen that, by the time this state is reached, the
				// player managing stock has already finished.  In that case, jump back to
				// the wait state.  mdomowicz 2014_03_31
				_iMessageBoxState = MESSAGEBOX_STATE_WAIT;
			}
			else
			{
				pGuiManager->SendMessageToGui( _gGuiMessageBox, &SoGuiMessage::_BasicTransitionIn );

				std::string sText = SoActorVisiblePlayer::GetPlayerNameByPlayerNumber( iStockManagingPlayer );
				sText += " is keeping/selling/trading defunct ";
				sText += SoActorVisibleHotel::GetHoteNameById( _pGame->GetLastMergerVictimHotelId() );
				sText += " stock.";
				_pGuiMessageBox->SetLabelText( sText );

				_iMessageBoxPlayerNumber = iStockManagingPlayer;
				_iMessageBoxState = MESSAGEBOX_STATE_MANAGESTOCK_UPDATE;
			}

			break;
		}

		case MESSAGEBOX_STATE_MANAGESTOCK_UPDATE:
		{
			bool bPlayerChanged = ( _iMessageBoxPlayerNumber != iStockManagingPlayer );

			if ( !bAPlayerIsManagingStock || bStockManagingPlayerIsLocalHuman || bPlayerChanged )
				_iMessageBoxState = MESSAGEBOX_STATE_CLOSE_INIT;		

			break;
		}

		case MESSAGEBOX_STATE_CLOSE_INIT:
		{
			pGuiManager->SendMessageToGui( _gGuiMessageBox, &SoGuiMessage::_BasicTransitionOut );

			_iMessageBoxPlayerNumber = PLAYER_UNDEFINED_PLAYER_NUMBER;
			_fMessageBoxTimer = 0.0f;

			_iMessageBoxState = MESSAGEBOX_STATE_CLOSE_UPDATE;
			break;
		}

		case MESSAGEBOX_STATE_CLOSE_UPDATE:
		{
			if ( _fMessageBoxTimer > 1.0f )
				_iMessageBoxState = MESSAGEBOX_STATE_WAIT;

			break;
		}

		default:
			ASSERTF( 0, ( "\nUnknown message box state." ) );
	}
}



/******************************************************************************
 FUNCTION:  ConnectToServer
******************************************************************************/
void SoGameClient::ConnectToServer( const std::string& sServerName )
{
	SoNetworkComms* pNetwork = SoGameGlobals::GetNetworkComms();

	if ( sServerName == "" )
		_hServerId = pNetwork->FindAvailableServer();
	else
		_hServerId = SoGameServer::GenerateServerIdFromString( sServerName );

	
	if ( _hServerId == 0 )
	{
		CPrintF( "No server exists.  Call \\CreateServer first." );
	}
	else if ( _hServerId )
	{
		SoGameMessage M( _hServerId, _hClientId, Msg::TYPE_CLI_REQUESTS_ADD_TO_SVR );
		SendMessageToServer( M );
	}	
}



/******************************************************************************
 FUNCTION:  SendMessageToServer
******************************************************************************/
void SoGameClient::SendMessageToServer( SoGameMessage& M )
{
	SoNetworkComms* pNetwork = SoGameGlobals::GetNetworkComms();
	M.SetServerId( _hServerId );
	M.SetClientId( _hClientId );

	_iLastMessageNumSentToServer ++;
	M.SetMessageNumber( _iLastMessageNumSentToServer );


	pNetwork->SendMessageToServer( M );
}



/******************************************************************************
 FUNCTION:  CreatePlayerHuman
******************************************************************************/
void SoGameClient::CreatePlayerHuman()
{
	SoCodecCliRequestsSeat Encoder;
	Encoder._eGameBitTypeOfNewPlayer = SoGameBit::TYPE_GAMEBIT_PLAYER_HUMAN;

	SoGameMessage M;
	Encoder.Encode( &M, _hServerId, _hClientId );

	SendMessageToServer( M );
}



/******************************************************************************
 FUNCTION:  CreatePlayerMctsRandom
******************************************************************************/
void SoGameClient::CreatePlayerMctsRandom()
{
	SoCodecCliRequestsSeat Encoder;
	Encoder._eGameBitTypeOfNewPlayer = SoGameBit::TYPE_GAMEBIT_PLAYER_MCTS_RANDOM;

	SoGameMessage M;
	Encoder.Encode( &M, _hServerId, _hClientId );

	SendMessageToServer( M );
}



/******************************************************************************
 FUNCTION:  CreatePlayerAi
******************************************************************************/
void SoGameClient::CreatePlayerAi()
{
	SoCodecCliRequestsSeat Encoder;
	Encoder._eGameBitTypeOfNewPlayer = SoGameBit::TYPE_GAMEBIT_PLAYER_AI;
	Encoder._eAiPlayerStrength = SoGameBitPlayer::AI_PLAYER_STRENGTH_MEDIUM;

	SoGameMessage M;
	Encoder.Encode( &M, _hServerId, _hClientId );

	SendMessageToServer( M );
}



/******************************************************************************
 FUNCTION:  CreatePlayerAiLevel2
******************************************************************************/
void SoGameClient::CreatePlayerAiLevel2()
{
	SoCodecCliRequestsSeat Encoder;
	Encoder._eGameBitTypeOfNewPlayer = SoGameBit::TYPE_GAMEBIT_PLAYER_AI;
	Encoder._eAiPlayerStrength = SoGameBitPlayer::AI_PLAYER_STRENGTH_MEDIUM2;

	SoGameMessage M;
	Encoder.Encode( &M, _hServerId, _hClientId );

	SendMessageToServer( M );
}



/******************************************************************************
 FUNCTION:  CreateVisibleActors
******************************************************************************/
void SoGameClient::CreateVisibleActors()
{
	SoActorVisible::ResetVisibleActorCounter();

	SoGameBitPlayer::SetPlayerColorByNumber( 0, SoGameBitPlayer::PLAYER_COLOR_BLUE );
	SoGameBitPlayer::SetPlayerColorByNumber( 1, SoGameBitPlayer::PLAYER_COLOR_GREEN );
	SoGameBitPlayer::SetPlayerColorByNumber( 2, SoGameBitPlayer::PLAYER_COLOR_PURPLE );
	SoGameBitPlayer::SetPlayerColorByNumber( 3, SoGameBitPlayer::PLAYER_COLOR_RED );
	SoGameBitPlayer::SetPlayerColorByNumber( 4, SoGameBitPlayer::PLAYER_COLOR_WHITE );
	SoGameBitPlayer::SetPlayerColorByNumber( 5, SoGameBitPlayer::PLAYER_COLOR_YELLOW );

	/*
	// For testing alternate color assignments.
	SoActorPlayer::SetPlayerColorByNumber( 0, SoActorPlayer::PLAYER_COLOR_RED );
	SoActorPlayer::SetPlayerColorByNumber( 1, SoActorPlayer::PLAYER_COLOR_YELLOW );
	SoActorPlayer::SetPlayerColorByNumber( 2, SoActorPlayer::PLAYER_COLOR_GREEN );
	SoActorPlayer::SetPlayerColorByNumber( 3, SoActorPlayer::PLAYER_COLOR_BLUE );
	SoActorPlayer::SetPlayerColorByNumber( 4, SoActorPlayer::PLAYER_COLOR_PURPLE );
	SoActorPlayer::SetPlayerColorByNumber( 5, SoActorPlayer::PLAYER_COLOR_WHITE );
	*/

	
	SoGameBitPlayer* pPlayer0 = _pGame->GetPlayerByPlayerNumber( 0 );
	SoGameBitPlayer* pPlayer1 = _pGame->GetPlayerByPlayerNumber( 1 );
	SoGameBitPlayer* pPlayer2 = _pGame->GetPlayerByPlayerNumber( 2 );
	SoGameBitPlayer* pPlayer3 = _pGame->GetPlayerByPlayerNumber( 3 );
	SoGameBitPlayer* pPlayer4 = _pGame->GetPlayerByPlayerNumber( 4 );
	SoGameBitPlayer* pPlayer5 = _pGame->GetPlayerByPlayerNumber( 5 );

	if ( pPlayer0 )
	{
		SoActorVisiblePlayer* pVisiblePlayer = new SoActorVisiblePlayer;
		pVisiblePlayer->Init( SoActorManager::GetInstance()->GetParentNode() );
		pVisiblePlayer->AddToActorManagerByPriority();
		pVisiblePlayer->SetGameBitPlayer( pPlayer0 );
		pVisiblePlayer->SetGameState( _pGame );

		SoHud::GetInstance()->AddPlayerGuide( pPlayer0 );
	}

	if ( pPlayer1 )
	{
		SoActorVisiblePlayer* pVisiblePlayer = new SoActorVisiblePlayer;
		pVisiblePlayer->Init( SoActorManager::GetInstance()->GetParentNode() );
		pVisiblePlayer->AddToActorManagerByPriority();
		pVisiblePlayer->SetGameBitPlayer( pPlayer1 );
		pVisiblePlayer->SetGameState( _pGame );

		SoHud::GetInstance()->AddPlayerGuide( pPlayer1 );
	}

	if ( pPlayer2 )
	{
		SoActorVisiblePlayer* pVisiblePlayer = new SoActorVisiblePlayer;
		pVisiblePlayer->Init( SoActorManager::GetInstance()->GetParentNode() );
		pVisiblePlayer->AddToActorManagerByPriority();
		pVisiblePlayer->SetGameBitPlayer( pPlayer2 );
		pVisiblePlayer->SetGameState( _pGame );

		SoHud::GetInstance()->AddPlayerGuide( pPlayer2 );
	}

	if ( pPlayer3 )
	{
		SoActorVisiblePlayer* pVisiblePlayer = new SoActorVisiblePlayer;
		pVisiblePlayer->Init( SoActorManager::GetInstance()->GetParentNode() );
		pVisiblePlayer->AddToActorManagerByPriority();
		pVisiblePlayer->SetGameBitPlayer( pPlayer3 );
		pVisiblePlayer->SetGameState( _pGame );

		SoHud::GetInstance()->AddPlayerGuide( pPlayer3 );
	}

	if ( pPlayer4 )
	{
		SoActorVisiblePlayer* pVisiblePlayer = new SoActorVisiblePlayer;
		pVisiblePlayer->Init( SoActorManager::GetInstance()->GetParentNode() );
		pVisiblePlayer->AddToActorManagerByPriority();
		pVisiblePlayer->SetGameBitPlayer( pPlayer4 );
		pVisiblePlayer->SetGameState( _pGame );

		SoHud::GetInstance()->AddPlayerGuide( pPlayer4 );
	}

	if ( pPlayer5 )
	{
		SoActorVisiblePlayer* pVisiblePlayer = new SoActorVisiblePlayer;
		pVisiblePlayer->Init( SoActorManager::GetInstance()->GetParentNode() );
		pVisiblePlayer->AddToActorManagerByPriority();
		pVisiblePlayer->SetGameBitPlayer( pPlayer5 );
		pVisiblePlayer->SetGameState( _pGame );

		SoHud::GetInstance()->AddPlayerGuide( pPlayer5 );
	}


	SoTileSet* pTileSet = _pGame->GetTileSet();
	SoGameBitTileArray* pAllTiles = pTileSet->GetAllTiles();
	SoGameBitTileArray::Iterator TileIter( pAllTiles );
	while ( TileIter.Next() )
	{
		SoGameBitTile* pTile = TileIter.GetData();
		
		SoActorVisibleBoardSpot* pNewSpot = new SoActorVisibleBoardSpot;
		pNewSpot->Init( SoActorManager::GetInstance()->GetParentNode() );
		pNewSpot->AddToActorManagerByPriority();
		pNewSpot->SetGameState( _pGame );
		pNewSpot->SetGameBitTile( pTile );


		SoActorVisibleTile* pNewVisibleTile = new SoActorVisibleTile;
		pNewVisibleTile->Init( SoActorManager::GetInstance()->GetParentNode() );
		pNewVisibleTile->AddToActorManagerByPriority();
		pNewVisibleTile->SetGameState( _pGame );

		pNewVisibleTile->SetGameBitTile( pTile );
		pTile->SetActorVisibleTile( pNewVisibleTile );

		pNewVisibleTile->SetVisibleBoardSpot( pNewSpot );
	}

	for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
	{
		SoGameBitHotel* pHotel = _pGame->GetHotelByHotelId( i );
		
		SoActorVisibleHotel* pNewVisibleHotel = new SoActorVisibleHotel;
		pNewVisibleHotel->Init( SoActorManager::GetInstance()->GetParentNode() );
		pNewVisibleHotel->AddToActorManagerByPriority();
		pNewVisibleHotel->SetGameState( _pGame );

		pNewVisibleHotel->SetGameBitHotel( pHotel );
	}

	SoActorVisibleTile::InitializeOrganizePosArrays();



	int iBoardWidth = _pGame->GetBoardWidth();
	int iBoardHeight = _pGame->GetBoardHeight();

	for ( int x = 0; x < iBoardWidth; x ++ )
	{
		SoActorVisibleCoordinate* pNewVisibleCoordinate = new SoActorVisibleCoordinate;
		pNewVisibleCoordinate->Init( SoActorManager::GetInstance()->GetParentNode() );
		pNewVisibleCoordinate->AddToActorManagerByPriority();

		pNewVisibleCoordinate->SetLabelText( JpString::IntToString( x + 1 ) ) ;
		JpVector2f vCoordinatePos = SoGameClient::GetGlobalCenterOffset( _pGame );
		vCoordinatePos.X += 0.0f + ( 128.0f * x );
		vCoordinatePos.Y += 96.0f;
		pNewVisibleCoordinate->SetWorldPosition( vCoordinatePos ); 
	}

	for ( int y = 0; y < iBoardHeight; y ++ )
	{
		SoActorVisibleCoordinate* pNewVisibleCoordinate = new SoActorVisibleCoordinate;
		pNewVisibleCoordinate->Init( SoActorManager::GetInstance()->GetParentNode() );
		pNewVisibleCoordinate->AddToActorManagerByPriority();

		pNewVisibleCoordinate->SetLabelText( JpString::CharToString( (char)( 65 + y ) ) ) ;
		JpVector2f vCoordinatePos = SoGameClient::GetGlobalCenterOffset( _pGame );
		vCoordinatePos.X -= 96.0f;
		vCoordinatePos.Y -= 128.0f * y;
		pNewVisibleCoordinate->SetWorldPosition( vCoordinatePos ); 
	}
}



/******************************************************************************
 FUNCTION:  GetGlobalCenterOffset
******************************************************************************/
JpVector2f SoGameClient::GetGlobalCenterOffset( SoGameState* pGameState )
{
	if ( pGameState )
	{
		int iWidthInTiles = pGameState->GetBoardWidth();
		int iHeightInTiles = pGameState->GetBoardHeight();

		int iTileWidth = 128;
		int iTileHeight = 128;

		float fTotalWidth = iWidthInTiles * iTileWidth;
		float fTotalHeight = iHeightInTiles * iTileHeight;

		float fX = -( fTotalWidth / 2 );
		float fY = ( fTotalHeight / 2 );

		// Honestly, I'm not sure why this needs to be here to center the board,
		// but it doesn't matter since this current placement code is probably temp.
		// mdomowicz 2013_12_13
		fX += 64.0f;
		fY -= 64.0f;

		return JpVector2f( fX, fY );

	}

	return JpVector2f( 0.0f, 0.0f );
}



/******************************************************************************
 FUNCTION:  WriteSaveGameCsvToDisk
******************************************************************************/
void SoGameClient::WriteSaveGameCsvToDisk()
{
	//-----------------------
	// save the data for this class
	NeCsvOut CsvOut;

	char acText[ 128 ] = "";
	sprintf( acText, "%i", _hClientId );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "_hClientId" );
	CsvOut.AppendStringToLastRow( acText );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "MasterRngSeed" );
	CsvOut.AppendIntToLastRow( _pGame->GetMasterRngSeed() );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "MasterRngSeedHigh" );
	CsvOut.AppendIntToLastRow( _RngForSaveLoadSupport.GetHigh() );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "MasterRngSeedLow" );
	CsvOut.AppendIntToLastRow( _RngForSaveLoadSupport.GetLow() );

	int iNext = _RngForSaveLoadSupport.GetInt();

	char acFullPathAndFilename[ 512 ] = "";
	SoUtils::CreateFullPathWithRelativePathAndFilename( acFullPathAndFilename, SoUtils::AddRelativePathToSaveData( "SavedData_Client.csv" ) );
	CsvOut.WriteToFile( acFullPathAndFilename );
}



/******************************************************************************
 FUNCTION:  InvalidateSaveData
******************************************************************************/
void SoGameClient::InvalidateSaveData()
{
	NeCsvOut CsvOut;

	char acText[ 128 ] = "";
	sprintf( acText, "%i", _hClientId );
	
	CsvOut.Clear();
	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "_hClientId" );
	CsvOut.AppendIntToLastRow( 0 );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "MasterRngSeed" );
	CsvOut.AppendIntToLastRow( 0 );

	char acFullPathAndFilename[ 512 ] = "";
	SoUtils::CreateFullPathWithRelativePathAndFilename( acFullPathAndFilename, SoUtils::AddRelativePathToSaveData( "SavedData_Client.csv" ) );
	CsvOut.WriteToFile( acFullPathAndFilename );
}



/******************************************************************************
 FUNCTION:  PrintToConsole
******************************************************************************/
void SoGameClient::PrintToConsole( const std::string& s )
{
	SoConsole* pConsole = (SoConsole*)NeConsole::GetInstance();
	pConsole->AddOutputString( s );
	pConsole->SetCurrentTextColor( JpUtils::_White );

	CCLOG( s.c_str() );
}



/******************************************************************************
 FUNCTION:  CPrintF
******************************************************************************/
void SoGameClient::CPrintF( const char* pFormat, ... )
{
	if ( strlen( pFormat ) < 2048 )
	{
		char buffer[2048];
		va_list args;
		va_start (args, pFormat);
		vsprintf (buffer,pFormat, args);
		va_end (args);

		PrintToConsole( buffer );
	}
}
