/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"


#include "JpRandomStrong.h"
#include "SoCodec.h"
#include "SoConfigManager.h"
#include "SoConsole.h"
#include "SoGameGlobals.h"
#include "SoGameMessage.h"
#include "SoGameServer.h"
#include "SoGameState.h"
#include "SoNetworkComms.h"

#include "mmgrChooser.h"


USING_NS_CC;


/******************************************************************************
 FUNCTION:  SoGameServer
******************************************************************************/
SoGameServer::SoGameServer() :
_eCurrentState( STATE_LAST ),
_hServerId( 0 ),
_pGame( NULL ),
_bIsLoadingGame( false )
{
}



/******************************************************************************
 FUNCTION:  ~SoGameServer
******************************************************************************/
SoGameServer::~SoGameServer()
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoGameServer::Init( const std::string& sServerName )
{
	if ( sServerName == "" )
		GenerateServerId();
	else
		_hServerId = GenerateServerIdFromString( sServerName );


	_eCurrentState = STATE_LOBBY;

	_pGame = new SoGameState;
	_pGame->Init();
	_pGame->SetOwningClientId( 0 );
	_pGame->SetOwningServerId( _hServerId );
	_pGame->SetOwningServer( this );

	CPrintF( "Game server created. Server Id: %x", _hServerId );


	//-----------------------
	// Check if the user specified a debug World Hash 
	NeConfig::Line* pLine = SoConfigManager::GetInstance()->FindLine( "DebugWorldName" );
	std::string sDebugWorldName = "";
	if ( pLine )
		sDebugWorldName = pLine->GetString( 1 );

	pLine = SoConfigManager::GetInstance()->FindLine( "DebugWorldHash" );
	std::string sDebugWorldHash = "";
	if ( pLine )
		sDebugWorldHash = pLine->GetString( 1 );


	//-----------------------
	// Seed the Rng - use one of the debug values if specified, 
	// else use the strong rng.
	if ( sDebugWorldName != "" )
	{
		_pGame->SeedMasterRng( JpHash( sDebugWorldName.c_str() ) );
		CPrintF( "Master Rng seeded with DebugWorldName read from .ini: %s", sDebugWorldName.c_str() );
	}
	else if ( sDebugWorldHash != "" )
	{
		hash hRngSeed;
		sscanf( sDebugWorldHash.c_str(), "%x", &hRngSeed );
		_pGame->SeedMasterRng( hRngSeed );

		CPrintF( "Master Rng seeded with DebugWorldHash read from .ini: %x", hRngSeed );
	}
	else
	{
		_pGame->SeedMasterRng( JpRandomStrong::G.GetInt() );
		CPrintF( "Master Rng seeded with JpRandomStrong: %x", _pGame->GetMasterRngSeed() );
	}


	CPrintF( "Server is in LOBBY state." );
}



/******************************************************************************
 FUNCTION:  InitForLoadGame
******************************************************************************/
void SoGameServer::InitForLoadGame( hash hServerId, hash hMasterRngSeed )
{
	_hServerId = hServerId;

	_pGame = new SoGameState;
	_pGame->Init();
	_pGame->SetOwningClientId( 0 );
	_pGame->SetOwningServerId( _hServerId );
	_pGame->SetOwningServer( this );
	_pGame->SeedMasterRng( hMasterRngSeed );

	_eCurrentState = STATE_LOBBY;

	CPrintF( "Game server created. Server Id: %x", _hServerId );
	CPrintF( "Master Rng seeded from load game: %x", _pGame->GetMasterRngSeed() );
	CPrintF( "Server is in LOBBY state." );
}


/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoGameServer::Deinit()
{
	SoClientInfoList::Iterator ClientIter( &_ClientInfoList );
	while ( ClientIter.Next() )
		delete ClientIter.GetNodeData();

	_ClientInfoList.RemoveAll();


	delete _pGame;
	_pGame = NULL;
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoGameServer::Update( float dt )
{
	SoNetworkComms* pNetwork = SoGameGlobals::GetNetworkComms();
	SoGameMessage* pInMsg = pNetwork->GetNextMessageForServer( _hServerId );

	if ( _pGame )
		_pGame->Update( dt );


	//-----------------------
	// Each Client sequentially numbers the messages they send, so that the Server can be sure
	// to handle the messages in the order they were sent.  In local games, this is never actually
	// a problem, but in online games message order is not guaranteed.
	// Here we are checking to make sure that the next message coming in has the expected numbering.  
	// If it does not, then it is an out-of-sequence message, and we ignore it for now.  We must 
	// wait until the correct next message comes in.  mdomowicz 2013_11_16
	bool bMessageIsExpected = false;
	if ( pInMsg )
	{
		hash hClientId = pInMsg->GetClientId();
		ClientInfo* pClientInfo = GetClientInfoFromClientId( hClientId );

		if ( !pClientInfo )
		{
			bMessageIsExpected = true;
		}
		else if ( pClientInfo && pInMsg->GetMessageNumber() == pClientInfo->_iLastMessageNumReceivedFromClient + 1 )
		{
			pClientInfo->_iLastMessageNumReceivedFromClient ++;
			bMessageIsExpected = true;
		}
	}

	//-----------------------
	// save every message, for save/load purposes.
	if ( bMessageIsExpected )
	{
		pInMsg->WriteToCsv( &_CsvOutMessageData );
	}

	//-----------------------
	switch ( _eCurrentState )
	{
		case STATE_LOBBY:
		{
			if ( !pInMsg || !bMessageIsExpected )
				break;
			
			if ( pInMsg->IsType( Msg::TYPE_CLI_REQUESTS_ADD_TO_SVR ) )
			{
				hash hClientId = pInMsg->GetClientId();
				ClientInfo* pNewClientInfo = new ClientInfo;

				pNewClientInfo->_hClientId = hClientId;
				pNewClientInfo->_iLastMessageNumReceivedFromClient ++;

				_ClientInfoList.AddLast( pNewClientInfo );


				SoGameMessage M;

				SoCodecSvrCommandAddCli Encoder;
				Encoder._hRngSeed = _pGame->GetMasterRngSeed();
				Encoder.Encode( &M, _hServerId, hClientId );

				SendMessageToClient( hClientId, M );

				pNetwork->RemoveAndDeleteMessageForServer( pInMsg );


				// Here we tell the new client about already existing/occupied seats.
				// An alternative way to do this is to just tell the new client how many
				// seats are occupied, and have the new client create that many dumb players.
				// However, in the near future, we will want to add more information about
				// existing players, such as color, portrait, etc, and so we actually
				// do "proper" player adds here.  mdomowicz 2013_09_24
				SoGameBitPlayerArray::Iterator Iter( _pGame->GetPlayerArray() );
				while ( Iter.Next() )
				{
					SoGameBitPlayer* pPlayer = Iter.GetData();
					SoCodecSvrCommandSeatAssignment Encoder2;
					Encoder2._hClientToReceiveSeat = -1;
					Encoder2._eGameBitTypeOfNewPlayer = SoGameBit::TYPE_GAMEBIT_PLAYER_DUMB;
					Encoder2._iPlayerNumber = pPlayer->GetPlayerNumber();
					Encoder2._eAiPlayerStrength = pPlayer->GetAiPlayerStrength();

					SoGameMessage AddPlayerMsg;
					Encoder2.Encode( &AddPlayerMsg, _hServerId, hClientId );
					SendMessageToClient( hClientId, AddPlayerMsg );
				}
			}

			if ( pInMsg->IsType( Msg::TYPE_CLI_REQUESTS_SEAT ) )
			{
				SoGameBitPlayer* pNewPlayer = _pGame->AddPlayerDumb();
				if ( pNewPlayer )
				{
					CPrintF( "Server %x has added a Dumb Player", _hServerId );

					SoCodecCliRequestsSeat Decoder;
					Decoder.Decode( pInMsg );

					SoCodecSvrCommandSeatAssignment Encoder;
					Encoder._hClientToReceiveSeat = pInMsg->GetClientId();
					Encoder._eGameBitTypeOfNewPlayer = Decoder._eGameBitTypeOfNewPlayer;
					Encoder._iPlayerNumber = pNewPlayer->GetPlayerNumber();
					Encoder._eAiPlayerStrength = Decoder._eAiPlayerStrength;

					SoGameMessage Msg;
					Encoder.Encode( &Msg, _hServerId, -1 );

					SendMessageToAllClients( Msg );
					pNetwork->RemoveAndDeleteMessageForServer( pInMsg );
				}
			}

			if ( pInMsg->IsType( Msg::TYPE_SVR_COMMAND_START_GAME ) )
			{
				_pGame->StartNewGameGame();
				pNetwork->RemoveAndDeleteMessageForServer( pInMsg );
				pNetwork->TellWebServerGameHasStarted();
				_eCurrentState = STATE_RUN_GAME;
			}
			break;
		}

		case STATE_RUN_GAME:
		{
			if ( bMessageIsExpected )
			{
				bool bMessageHandled = _pGame->UpdateFromServer( pInMsg );
				if ( bMessageHandled )
					pNetwork->RemoveAndDeleteMessageForServer( pInMsg );
			}
			else
			{
				_pGame->UpdateFromServer( NULL );
			}

			break;
		}

	
		case STATE_LAST:
		{
			break;
		}
	}
}



/******************************************************************************
 FUNCTION:  GenerateServerId
******************************************************************************/
hash SoGameServer::GenerateServerId()
{
	_hServerId = JpRandom::G.GetInt();
	return _hServerId;
}



/******************************************************************************
 FUNCTION:  GenerateServerIdFromString
******************************************************************************/
hash SoGameServer::GenerateServerIdFromString( const std::string& s )
{
	if ( s != "" )
		return JpHash( s.c_str() );

	return 0;
}



/******************************************************************************
 FUNCTION:  SendMessageToAllClients
******************************************************************************/
void SoGameServer::SendMessageToAllClients( SoGameMessage& Msg )
{
	SoNetworkComms* pNetwork = SoGameGlobals::GetNetworkComms();

	SoClientInfoList::Iterator Iter( &_ClientInfoList );
	while ( Iter.Next() )
	{
		ClientInfo* pClientInfo = Iter.GetNodeData();

		hash hClientId = pClientInfo->_hClientId;
		Msg.SetClientId( hClientId );


		pClientInfo->_iLastMessageNumSentToClient ++;
		Msg.SetMessageNumber( pClientInfo->_iLastMessageNumSentToClient );


		pNetwork->SendMessageToClient( Msg );
	}
}



/******************************************************************************
 FUNCTION:  SendMessageToClient
******************************************************************************/
void SoGameServer::SendMessageToClient( hash hClientId, SoGameMessage& Msg )
{
	SoNetworkComms* pNetwork = SoGameGlobals::GetNetworkComms();

	Msg.SetClientId( hClientId );

	ClientInfo* pClientInfo = GetClientInfoFromClientId( hClientId );
	pClientInfo->_iLastMessageNumSentToClient ++;	
	Msg.SetMessageNumber( pClientInfo->_iLastMessageNumSentToClient );

	pNetwork->SendMessageToClient( Msg );
}



/******************************************************************************
 FUNCTION:  GetClientInfoFromClientId
******************************************************************************/
SoGameServer::ClientInfo* SoGameServer::GetClientInfoFromClientId( hash hClientId )
{
	SoClientInfoList::Iterator ClientIter( &_ClientInfoList );
	while ( ClientIter.Next() )
	{
		ClientInfo* pClientInfo = ClientIter.GetNodeData();
		if ( pClientInfo->_hClientId == hClientId )
			return pClientInfo;
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  WriteSaveGameCsvToDisk
******************************************************************************/
void SoGameServer::WriteSaveGameCsvToDisk()
{
	//-----------------------
	// Save all the messages
	char acFullPathAndFilename1[ 512 ] = "";
	SoUtils::CreateFullPathWithRelativePathAndFilename( acFullPathAndFilename1, SoUtils::AddRelativePathToSaveData( "SavedData_Messages.csv" ) );
	_CsvOutMessageData.WriteToFile( acFullPathAndFilename1 );


	//-----------------------
	// save the data for this class
	NeCsvOut CsvOut;

	char acText[ 128 ] = "";
	sprintf( acText, "%i", _hServerId );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "_hServerId" );
	CsvOut.AppendStringToLastRow( acText );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "MasterRngSeed" );
	CsvOut.AppendIntToLastRow( _pGame->GetMasterRngSeed() );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "_bIsVisibleStock" );
	CsvOut.AppendBoolToLastRow( _pGame->IsVisibleStocks() );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "_bIsVisibleMoney" );
	CsvOut.AppendBoolToLastRow( _pGame->IsVisibleMoney() );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "MasterRngSeedHigh" );
	CsvOut.AppendIntToLastRow( _pGame->GetMasterRng()->GetHigh() );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "MasterRngSeedLow" );
	CsvOut.AppendIntToLastRow( _pGame->GetMasterRng()->GetLow() );

	char acFullPathAndFilename2[ 512 ] = "";
	SoUtils::CreateFullPathWithRelativePathAndFilename( acFullPathAndFilename2, SoUtils::AddRelativePathToSaveData( "SavedData_Server.csv" ) );
	CsvOut.WriteToFile( acFullPathAndFilename2 );
}



/******************************************************************************
 FUNCTION:  InvalidateSaveData
******************************************************************************/
void SoGameServer::InvalidateSaveData()
{
	NeCsvOut CsvOut;
	CsvOut.AddNewRow();
	CsvOut.AppendIntToLastRow( 0 );

	char acFullPathAndFilename1[ 512 ] = "";
	SoUtils::CreateFullPathWithRelativePathAndFilename( acFullPathAndFilename1, SoUtils::AddRelativePathToSaveData( "SavedData_Messages.csv" ) );
	CsvOut.WriteToFile( acFullPathAndFilename1 );

	char acText[ 128 ] = "";
	sprintf( acText, "%i", _hServerId );
	
	CsvOut.Clear();
	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "_hServerId" );
	CsvOut.AppendIntToLastRow( 0 );

	CsvOut.AddNewRow();
	CsvOut.AppendStringToLastRow( "MasterRngSeed" );
	CsvOut.AppendIntToLastRow( 0 );

	char acFullPathAndFilename2[ 512 ] = "";
	SoUtils::CreateFullPathWithRelativePathAndFilename( acFullPathAndFilename2, SoUtils::AddRelativePathToSaveData( "SavedData_Server.csv" ) );
	CsvOut.WriteToFile( acFullPathAndFilename2 );
}



/******************************************************************************
 FUNCTION:  GetLastMessageNumReceivedFromClient
******************************************************************************/
int SoGameServer::GetLastMessageNumReceivedFromClient( hash hClientId )
{
	ClientInfo* pClientInfo = GetClientInfoFromClientId( hClientId );
	if ( pClientInfo )
	{
		return pClientInfo->_iLastMessageNumReceivedFromClient;
	}

	return -1;
}


/******************************************************************************
 FUNCTION:  PrintToConsole
******************************************************************************/
void SoGameServer::PrintToConsole( const std::string& s )
{
	SoConsole* pConsole = (SoConsole*)NeConsole::GetInstance();
	pConsole->SetCurrentTextColor( JpUtils::_Magenta );
	pConsole->AddOutputString( s );
	pConsole->SetCurrentTextColor( JpUtils::_White );

	CCLOG( s.c_str() );
}



/******************************************************************************
 FUNCTION:  CPrintF
******************************************************************************/
void SoGameServer::CPrintF( const char* pFormat, ... )
{
	if ( strlen( pFormat ) < 2048 )
	{
		char buffer[2048];
		va_list args;
		va_start (args, pFormat);
		vsprintf (buffer,pFormat, args);
		va_end (args);

		PrintToConsole( buffer );
	}
}
