/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"


#include "NeCsvIn.h"
#include "SoActorManager.h"
#include "SoConfigManager.h"
#include "SoConsole.h"
#include "SoGameBitPlayerAi.h"
#include "SoGameBitPlayerDumb.h"
#include "SoGameBitPlayerHuman.h"
#include "SoGameBitPlayerMctsHeuristic.h"
#include "SoGameBitPlayerMctsRandom.h"
#include "SoGameClient.h"
#include "SoGameGlobals.h"
#include "SoGameMessage.h"
#include "SoGameServer.h"
#include "SoGameStateMcts.h"

#include "mmgrChooser.h"


USING_NS_CC;


/******************************************************************************
 FUNCTION:  SoGameMcts
******************************************************************************/
SoGameStateMcts::SoGameStateMcts()
{
	memset( _afPlayerEvaluations, 0.0f, GAMERULES_MAX_PLAYERS * sizeof( float ) );
}



/******************************************************************************
 FUNCTION:  ~SoGameMcts
******************************************************************************/
SoGameStateMcts::~SoGameStateMcts()
{
	Deinit();
}



/******************************************************************************
 FUNCTION:  MakeCopy
******************************************************************************/
void SoGameStateMcts::MakeCopy( SoGameState* pSource, SoGameBitPlayer::AI_PLAYER_STRENGTH eDesiredAiPlayerStrength )
{
	//-----------------------
	// copy the internal data
	MakeCopyHelperForMcts( pSource );


	//-----------------------
	// make an mcts copy of every player
	SoGameBitPlayerArray::Iterator PlayerIter( pSource->GetPlayerArray() );
	while ( PlayerIter.Next() )
	{
		// This is how we choose between different types of MCTS Ais
		SoGameBitPlayer* pNewPlayer = NULL;
		if ( eDesiredAiPlayerStrength == SoGameBitPlayer::AI_PLAYER_STRENGTH_MEDIUM )
			pNewPlayer = new SoGameBitPlayerMctsRandom;
		else if ( eDesiredAiPlayerStrength == SoGameBitPlayer::AI_PLAYER_STRENGTH_MEDIUM2 )
			pNewPlayer = new SoGameBitPlayerMctsHeuristic;

		// This catches an easily made mistake, with hard to debug consequences.
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();
		if ( pPlayer->GetAsPlayerDumb() )
		{
			// If this assert fires, then a Dumb player is getting its state altered somehow.
			// This will lead to all sorts of problems when doing mcts simulation, that can end
			// up being hard to diagnose.
			// The most likely cause of the state change is a command function being implemented
			// in SoGameBitPlayer.h, but not overridden (to do nothing) in SoGameBitPlayerDumb.h.
			// See those header files for examples and additional comments.
			// mdomowicz 2014_01_22
			ASSERTF( pPlayer->GetState() == SoGameBitPlayer::STATE_WAIT, ( "\nA Dumb player is in an unexpected state! See comments." ) );
		}

		pNewPlayer->Init( SoActorManager::GetInstance()->GetParentNode() );
		pNewPlayer->MakeCopy( pPlayer );
		pNewPlayer->SetOwningGame( this );

		_PlayerArray.AddLast( pNewPlayer );
	}
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoGameStateMcts::Init()
{
	SoGameState::Init();
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoGameStateMcts::Deinit()
{
	// Note: right now the parent class Deinit is called by
	// the parent class' destructor, so we don't do it here.
	// mdomowicz 2013_10_10
	//SoGame::Deinit()
	
	DeleteMessageLists();
}



/******************************************************************************
 FUNCTION:  UpdateFromServer
******************************************************************************/
bool SoGameStateMcts::UpdateFromServer( SoGameMessage* pInMsg )
{
	return SoGameState::UpdateFromServer( pInMsg );
}



/******************************************************************************
 FUNCTION:  Simulate
******************************************************************************/
void SoGameStateMcts::Simulate()
{
	//-----------------------
	// Get a message for the server (if there is one)
	SoGameMessageList::Node* pFirstNode = _MessagesForServers.GetFirst();
	SoGameMessage* pServerMsg = NULL;

	if ( pFirstNode )
		pServerMsg = pFirstNode->GetData();


	//-----------------------
	// update the game, and destroy the message
	bool bMessageHandled = UpdateFromServer( pServerMsg );

	if ( bMessageHandled )
	{
		if ( 0 )
		{
			int iPlayer = SoCodec::ExtractPlayerNumberFromMessage( pServerMsg );
			CCLOG( "Msg to server: P%i: %s", iPlayer, Msg::GetStringFromType( pServerMsg->GetType() ) );
		}

		_MessagesForServers.Remove( pServerMsg );
		delete pServerMsg;
		pServerMsg = NULL;
	}


	//-----------------------
	// Get a message for the client (if there is one)
	pFirstNode = _MessagesForClients.GetFirst();
	SoGameMessage* pClientMsg = NULL;

	if ( pFirstNode )
		pClientMsg = pFirstNode->GetData();


	//-----------------------
	// Handle the client message
	if ( pClientMsg )
	{
		if ( 0 )
		{
			int iPlayer = SoCodec::ExtractPlayerNumberFromMessage( pClientMsg );
			CCLOG( "Msg to client: P%i: %s", iPlayer, Msg::GetStringFromType( pClientMsg->GetType() ) );
		}

		HandleMessage( pClientMsg );

		_MessagesForClients.Remove( pClientMsg );
		delete pClientMsg;
		pClientMsg = NULL;
	}


	//-----------------------
	// Update all the players.
	SoGameBitPlayerArray::Iterator PlayerIter( GetPlayerArray() );
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();
		pPlayer->Update( 0.0f );
	}
}



/******************************************************************************
 FUNCTION:  SimulateUntilNextPlayerMoveCommand
******************************************************************************/
void SoGameStateMcts::SimulateUntilNextPlayerMoveCommand()
{
	bool bGameOver = false;
	bool bClientCommandReceived = false;

	while ( !bClientCommandReceived && !bGameOver )
	{
		//-----------------------
		// Get a message for the server (if there is one)
		SoGameMessageList::Node* pFirstNode = _MessagesForServers.GetFirst();
		SoGameMessage* pServerMsg = NULL;

		if ( pFirstNode )
			pServerMsg = pFirstNode->GetData();


		//-----------------------
		// update the game, and destroy the message
		bool bMessageHandled = UpdateFromServer( pServerMsg );

		if ( bMessageHandled )
		{
			_MessagesForServers.Remove( pServerMsg );
			delete pServerMsg;
			pServerMsg = NULL;
		}


		//-----------------------
		// Get a message for the client (if there is one)
		pFirstNode = _MessagesForClients.GetFirst();
		SoGameMessage* pClientMsg = NULL;

		if ( pFirstNode )
			pClientMsg = pFirstNode->GetData();


		//-----------------------
		// Handle the client message
		if ( pClientMsg )
		{
			HandleMessage( pClientMsg );

			_MessagesForClients.Remove( pClientMsg );
			delete pClientMsg;
			pClientMsg = NULL;

			bClientCommandReceived = true;
		}


		//-----------------------
		// We also exit the while loop if the game is won.
		bGameOver = IsGameOver();
	}
}



/******************************************************************************
 FUNCTION:  Evaluate
******************************************************************************/
void SoGameStateMcts::Evaluate()
{
	memset( _afPlayerEvaluations, 0.0f, GAMERULES_MAX_PLAYERS * sizeof( float ) );


	//-----------------------
	// Get the current value of each hotel (price per stock)
	int aiPricesPerStock[ GAMERULES_NUM_HOTELS ];	
	memset( aiPricesPerStock, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	for ( int i = 0; i < GAMERULES_NUM_HOTELS; i ++ )
	{
		SoGameBitHotel* pHotel = GetHotelByHotelId( i );
		int iSize = pHotel->GetSize();
		aiPricesPerStock[ i ] = SoGameBitHotel::GetPricePerStockFromTable( i, iSize );
	}


	//-----------------------
	// give points based on player resources
	SoGameBitPlayerArray::Iterator PlayerIter( GetPlayerArray() );
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();
		int i = pPlayer->GetPlayerNumber();


		// Give points based on money - and extra points if it's the last turn.
		float fMoneyValue = pPlayer->GetMoney();
		fMoneyValue *= 10.0f;

		if ( IsLastTurnDeclared() )
			fMoneyValue *= 2.0f;

		_afPlayerEvaluations[ i ] += fMoneyValue;


		// Give points for how much stock the player owns
		for ( int j = 0; j < GAMERULES_NUM_HOTELS; j ++ )
		{
			int iNumStockOwned = pPlayer->GetStockCount( j );
			float fStockValue = iNumStockOwned * aiPricesPerStock[ j ];
			fStockValue *= 10.0f;

			if ( IsLastTurnDeclared() )
				fStockValue *= 2.0f;

			_afPlayerEvaluations[ i ] += fStockValue;
		}
	}


	//-----------------------
	// see what is the most and second most any player 
	// has of every stock.
	int aiMostStockOwnedPerHotel[ GAMERULES_NUM_HOTELS ];	
	int aiSecondMostStockOwnedPerHotel[ GAMERULES_NUM_HOTELS ];
	memset( aiMostStockOwnedPerHotel, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );
	memset( aiSecondMostStockOwnedPerHotel, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	PlayerIter.Reset();
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();

		for ( int j = 0; j < GAMERULES_NUM_HOTELS; j ++ )
		{
			int iStockCount = pPlayer->GetStockCount( j );

			if ( iStockCount > aiMostStockOwnedPerHotel[ j ] )
			{
				aiSecondMostStockOwnedPerHotel[ j ] = aiMostStockOwnedPerHotel[ j ];
				aiMostStockOwnedPerHotel[ j ] = iStockCount;
			}
			else if 
			( 
				iStockCount != aiMostStockOwnedPerHotel[ j ] &&
				iStockCount > aiSecondMostStockOwnedPerHotel[ j ]
			)
			{
				aiSecondMostStockOwnedPerHotel[ j ] = iStockCount;
			}
		}		
	}


	//-----------------------
	// count how many players are tied for majority and minority
	// stock holders.
	int aiNumOwnersOfMostStock[ GAMERULES_NUM_HOTELS ];	
	int aiNumOwnersOfSecondMostStock[ GAMERULES_NUM_HOTELS ];	
	memset( aiNumOwnersOfMostStock, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );
	memset( aiNumOwnersOfSecondMostStock, 0, GAMERULES_NUM_HOTELS * sizeof( int ) );

	PlayerIter.Reset();
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();

		for ( int j = 0; j < GAMERULES_NUM_HOTELS; j ++ )
		{
			int iStockCount = pPlayer->GetStockCount( j );

			if ( iStockCount == aiMostStockOwnedPerHotel[ j ] )
				aiNumOwnersOfMostStock[ j ] ++;
			else if ( iStockCount == aiSecondMostStockOwnedPerHotel[ j ] )
				aiNumOwnersOfSecondMostStock[ j ] ++;
		}		
	}


	//-----------------------
	// finally we can award point bonuses based on if this player
	// is a majority or minority stock holder.
	PlayerIter.Reset();
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();

		for ( int j = 0; j < GAMERULES_NUM_HOTELS; j ++ )
		{
			int iStockCount = pPlayer->GetStockCount( j );

			if ( iStockCount > 0 )
			{
				SoGameBitHotel* pHotel = GetHotelByHotelId( j );
		
				float fBonus = 0.0f;

				if ( iStockCount == aiMostStockOwnedPerHotel[ j ] )
				{
					fBonus = SoGameBitHotel::GetMajorityBonusFromTable( j, pHotel->GetSize() );
					if ( aiNumOwnersOfMostStock[ j ] > 1 )
						fBonus /= aiNumOwnersOfMostStock[ j ];
				}
				else if ( iStockCount == aiSecondMostStockOwnedPerHotel[ j ] )
				{
					fBonus = SoGameBitHotel::GetMinorityBonusFromTable( j, pHotel->GetSize() );
					if ( aiNumOwnersOfSecondMostStock[ j ] > 1 )
						fBonus /= aiNumOwnersOfSecondMostStock[ j ];
				}

				_afPlayerEvaluations[ pPlayer->GetPlayerNumber() ] += ( fBonus * 20.0f );
			}
		}
	}


	//-----------------------
	// Give points for winning.  Less points if it's a tie.
	int iNumWinners = 0;
	PlayerIter.Reset();
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();

		if ( pPlayer->IsWinner() )
			iNumWinners ++;
	}	

	if ( iNumWinners > 0 )
	{
		PlayerIter.Reset();
		while ( PlayerIter.Next() )
		{
			SoGameBitPlayer* pPlayer = PlayerIter.GetData();

			if ( pPlayer->IsWinner() )
				_afPlayerEvaluations[ pPlayer->GetPlayerNumber() ] += ( 10000.0f / iNumWinners );		
		}	
	}
}



/******************************************************************************
 FUNCTION:  GetPlayerEvaluation
******************************************************************************/
float SoGameStateMcts::GetPlayerEvaluation( int iPlayerNumber )
{
	return _afPlayerEvaluations[ iPlayerNumber ];
}



/******************************************************************************
 FUNCTION:  SendMessageToAllClients
 NOTE:      Note that SoGameMcts simulates the client/server environment.

            So, here, instead of actually passing the message along to a client,
            this function saves the message to a list, which is later taken as
            "network" input for the "client", when we simulate.

            Compare this implementation to the parent (SoGame) version 
******************************************************************************/
void SoGameStateMcts::SendMessageToAllClients( SoGameMessage& M )
{
	if 
	( 
		/*
			In the normal client/server environment, all clients and
			the server have their own instance of an SoGame.  When players make a
			move a "PLYR_MOVE" message is sent to the server, which uses the message
			to update its SoGame.  Then the server forwards this same message
			to all the clients, so the client's can update their own SoGame's
			accordingly.

			SoGameMcts is a special version of SoGame, that uses a simulated
			client/server environment.  Whereas SoGame has to work with
			SoGameClient and SoGameServer, this is not true for SoGameMcts.

			Thus instead of this:

					SoGameClient    SoGameServer
					     |               |
					   SoGame          SoGame

			We have this:

					sim client     sim server
					         \    /
					       SoGameMcts

			The key difference is that, there is just one SoGameMcts instance
			at a time.  Thus the process of, "server updates its SoGame and then
			all clients update theirs" is invalid.  Only the server needs to
			update the SoGame, and the change will affect all clients automatically.

			And so, this is why here we only handle a subset of messages normally
			sent to clients.  We exclude all messages that the server will have
			already handled, and allow only "Player Move Command" messages that
			the server doesn't handle.

			See SoGameClient::Update() as reference, for the full list of messages
			normally handled by a client.

			mdomowicz 2013_09_03
		*/

		M.IsType( Msg::TYPE_SVR_COMMAND_PLYR_TAKE_TURN ) ||
		M.IsType( Msg::TYPE_SVR_COMMAND_PLYR_PURCHASE_STOCK ) ||
		M.IsType( Msg::TYPE_SVR_COMMAND_PLYR_PICK_NEW_HOTEL ) ||
		M.IsType( Msg::TYPE_SVR_COMMAND_PLYR_PICK_MERGE_HOTELS ) ||
		M.IsType( Msg::TYPE_SVR_COMMAND_PLYR_MANAGE_OLD_STOCK )
	)
	{
		SoGameMessage* pNewMessage = new SoGameMessage;
		pNewMessage->CopyMessage( M );

		_MessagesForClients.AddLast( pNewMessage );
	}
}



/******************************************************************************
 FUNCTION:  SendMessageToServer
 NOTE:      Note that SoGameMcts simulates the client/server environment.

            So, here, instead of actually passing the message along to a server,
            this function saves the message to a list, which is later taken as
            "network" input for the "server", when we simulate.

            Compare this implementation to the parent (SoGame) version 
******************************************************************************/
void SoGameStateMcts::SendMessageToServer( SoGameMessage& M )
{
	SoGameMessage* pNewMessage = new SoGameMessage;
	pNewMessage->CopyMessage( M );

	_MessagesForServers.AddLast( pNewMessage );
}



/******************************************************************************
 FUNCTION:  DeleteMessageLists
******************************************************************************/
void SoGameStateMcts::DeleteMessageLists()
{
	SoGameMessageList::Iterator Iter1( &_MessagesForServers );
	while ( Iter1.Next() )
		delete Iter1.GetNodeData();
	_MessagesForServers.RemoveAll();
	
	SoGameMessageList::Iterator Iter2( &_MessagesForClients );
	while ( Iter2.Next() )
		delete Iter2.GetNodeData();
	_MessagesForClients.RemoveAll();
}



/******************************************************************************
 FUNCTION:  CalculateChecksumForMcts
 NOTE:      This calculates a checksum from the current data-state of SoGame.
            This is for debug purposes only, mainly for testing the
            correctness of SoMctsTreeNode::EvaluateNode()
******************************************************************************/
ChecksumUint32 SoGameStateMcts::CalculateChecksumForMcts()
{
	JpChecksumFletcher32 Checksum;

	Checksum.AddUint32( (ChecksumUint32)_eCurrentState );
	Checksum.AddUint32( (ChecksumUint32)_iCurrentPlayerNum );

	Checksum.AddUint32( (ChecksumUint32)_iRound );


	// This changes between mcts simulations - don't checksum
	//Checksum.AddUint32( (ChecksumUint32)_hMasterRngSeed );
	//Checksum.AddUint32( (ChecksumUint32)_MasterRng.GetHigh() );
	//Checksum.AddUint32( (ChecksumUint32)_MasterRng.GetLow() );


	SoGameBitPlayerArray::Iterator PlayerIter( &_PlayerArray );
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();

		Checksum.AddUint32( (ChecksumUint32)pPlayer->GetState() );
		Checksum.AddUint32( (ChecksumUint32)pPlayer->GetRequestedState() );
		Checksum.AddUint32( (ChecksumUint32)pPlayer->GetPlayerNumber() );
		Checksum.AddUint32( (ChecksumUint32)pPlayer->GetMoney() );
		Checksum.AddUint32( (ChecksumUint32)pPlayer->IsFirstMoveOfPlayersTurn() );
	}

	return Checksum.GetChecksum();
}



/******************************************************************************
 FUNCTION:  WriteAllDataToDisk
 NOTES:     This is a tool for examining this SoGame's data.  It is intended
            for debug purposes only.
******************************************************************************/
void SoGameStateMcts::WriteAllDataToDisk( char* pacFilename )
{
	struct Local
	{
		static void WriteUint( NeCsvOut* pCsvOut, char acName[], unsigned int ui )
		{
			char acText[ 128 ] = "";
			sprintf( acText, "MCTS DATA: %s: ---- %i", acName, ui );
			pCsvOut->AddNewRow();
			pCsvOut->AppendStringToLastRow( acText );
		}
	};

	NeCsvOut CsvOut;


	Local::WriteUint( &CsvOut, "_eCurrentState", (unsigned int)_eCurrentState );
	Local::WriteUint( &CsvOut, "_iCurrentPlayerNum", (unsigned int)_iCurrentPlayerNum );

	Local::WriteUint( &CsvOut, "_iRound", (unsigned int)_iRound );


	// This changes between mcts simulations - don't checksum
	//Local::PrintUint( &CsvOut, "", (unsigned int)_hMasterRngSeed );
	//Local::PrintUint( &CsvOut, "", (unsigned int)_MasterRng.GetHigh() );
	//Local::PrintUint( &CsvOut, "", (unsigned int)_MasterRng.GetLow() );


	SoGameBitPlayerArray::Iterator PlayerIter( &_PlayerArray );
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();
		Local::WriteUint( &CsvOut, "GetState", (unsigned int)pPlayer->GetState() );
		Local::WriteUint( &CsvOut, "GetRequestedState", (unsigned int)pPlayer->GetRequestedState() );
		Local::WriteUint( &CsvOut, "GetPlayerNumber", (unsigned int)pPlayer->GetPlayerNumber() );
		Local::WriteUint( &CsvOut, "GetMoney", (unsigned int)pPlayer->GetMoney() );
		Local::WriteUint( &CsvOut, "IsFirstMoveOfPlayersTurn", (unsigned int)pPlayer->IsFirstMoveOfPlayersTurn() );
	}


	const char* cpFullPath = SoUtils::AddRelativePathToSaveData( pacFilename );
	CsvOut.WriteToFile( cpFullPath );
	
}



/******************************************************************************
 FUNCTION:  CaptureUndoState
******************************************************************************/
void SoGameStateMcts::CaptureUndoState()
{
	// Intentionally left blank.
	// We don't want to capture an undo state while simulating moves 
	// during mcts processing.
	// mdomowicz 2013_10_12
}



/******************************************************************************
 FUNCTION:  PrintToConsole
******************************************************************************/
void SoGameStateMcts::PrintToConsole( const std::string& s )
{
	/*
	char acText[ 256 ] = "";
	sprintf( acText, "MCTS: %s", s.c_str() );
	CCLOG( acText );
	*/
}
