/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "SoGuiManager.h"
#include "SoGuiMessage.h"
#include "SoGuiAnim.h"
#include "SoUtils.h"

#include "mmgrChooser.h"


USING_NS_CC;



/******************************************************************************
 FUNCTION:  SoGuiAnim
******************************************************************************/
SoGuiAnim::SoGuiAnim() :
_pSpriteBackground( NULL ),
_eTransitionStyle( TRANSITION_STYLE_TYPE_LAST ),
_eTransitionState( TRANSITION_STATE_LAST ),
_fTransitionTimer( 0.0f ),
_fTransitionSpeed( 1.0f ),
_bDeleteOnTransitionOut( false )
{
}



/******************************************************************************
 FUNCTION:  ~SoGuiAnim
******************************************************************************/
SoGuiAnim::~SoGuiAnim()
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoGuiAnim::Init( CCNode* pParentNode, SoGuiManager* pGuiManager )
{
	SoGui::Init( pParentNode, pGuiManager );

	_eTransitionState = TRANSITION_STATE_INIT;
	_eTransitionStyle = TRANSITION_STYLE_POP_UP;
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoGuiAnim::Deinit()
{
	SoGui::Deinit();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoGuiAnim::Update( float dt )
{
	SoGui::Update( dt );


	if ( _pSpriteBackground )
	{
		ScaleSpriteToSizeOfGui( _pSpriteBackground );

		_pSpriteBackground->SetPosition( _vAbsolutePos );
	
		switch ( _eTransitionStyle )
		{
			case TRANSITION_STYLE_FADE_IN_OUT:
				UpdateStyleFadeInOut( dt );
				break;

			case TRANSITION_STYLE_POP_UP:
				UpdateStylePopup( dt );
				break;

			case TRANSITION_STYLE_NONE:
				UpdateStyleNone( dt );
				break;

			default:
				ASSERTF( 0, ( "\nUnknown transition style." ) )
		}
	}
	else
	{
		// Update style "none" anyway, so that this GuiAnim still goes through
		// state changes into TRANSITION_STATE_SHOW and TRANSITION_STATE_HIDE, etc
		// which allows IsShowing() and IsHiding() to still work.
		// mdomowicz 2014_08_04
		UpdateStyleNone( dt );
	}
}



/******************************************************************************
 FUNCTION:  HandleMessage
******************************************************************************/
void SoGuiAnim::HandleMessage( SoGuiMessage* pMessage )
{
	switch ( pMessage->GetType() )
	{
		case SoGuiMessage::TYPE_TRANSITION_IN:
		{
			TransitionIn();
			SoGui::HandleMessage( pMessage );
			break;
		}
		case SoGuiMessage::TYPE_TRANSITION_OUT:
		{
			TransitionOut();
			SoGui::HandleMessage( pMessage );
			break;
		}
		default:
		{
			SoGui::HandleMessage( pMessage );
			break;
		}
	}
}



/******************************************************************************
 FUNCTION:  CreateBackgroundSprite
******************************************************************************/
void SoGuiAnim::CreateBackgroundSprite( const std::string& sPathAndFilename )
{
	_pSpriteBackground = CreateSprite( sPathAndFilename, BASE_RENDER_ORDER_GUI );
	SetWidthHeightFromSprite( _pSpriteBackground );
}



/******************************************************************************
 FUNCTION:  UpdateStyleFadeInOut
******************************************************************************/
void SoGuiAnim::UpdateStyleFadeInOut( float dt )
{
	_fTransitionTimer += dt;

	switch ( _eTransitionState )
	{
		case TRANSITION_STATE_INIT:
		{
			_pSpriteBackground->setOpacity( 0 );
			_pSpriteBackground->setVisible( false );
			break;
		}
		case TRANSITION_STATE_ANIM_IN:
		{
			_pSpriteBackground->setVisible( true );

			float fPercent = _fTransitionTimer * _fTransitionSpeed;
			fPercent = Clamp( fPercent, 0.0, 1.0f );
			_pSpriteBackground->setOpacity( fPercent * ( 255.0f * _fSpriteAlpha ) );
			if ( fPercent == 1.0f )
			{
				_fTransitionTimer = 0.0f;
				_eTransitionState = TRANSITION_STATE_SHOW;
			}

			break;
		}
		case TRANSITION_STATE_SHOW:
		{
			break;
		}
		case TRANSITION_STATE_ANIM_OUT:
		{
			float fPercent = ( 1.0f - ( _fTransitionTimer * _fTransitionSpeed ) );
			fPercent = Clamp( fPercent, 0.0, 1.0f );
			_pSpriteBackground->setOpacity( fPercent * ( 255.0f * _fSpriteAlpha ) );
			if ( fPercent == 0.0f )
			{
				_fTransitionTimer = 0.0f;
				_eTransitionState = TRANSITION_STATE_HIDE;
			}
			break;
		}
		case TRANSITION_STATE_HIDE:
		{
			_pSpriteBackground->setVisible( false );
			break;
		}
		default:
			ASSERTF( 0, ( "\nUnknown transition state." ) );
	}
}



/******************************************************************************
 FUNCTION:  UpdateStylePopup
******************************************************************************/
void SoGuiAnim::UpdateStylePopup( float dt )
{
	_fTransitionTimer += dt;

	switch ( _eTransitionState )
	{
		case TRANSITION_STATE_INIT:
		{
			_pSpriteBackground->setScale( 0.0f );
			SetSpriteAlphaToPercentOfTarget( 0.0f );
			SetTextAlphaToPercentOfTarget( 0.0f );

			break;
		}
		case TRANSITION_STATE_ANIM_IN:
		{
			_pSpriteBackground->setVisible( true );

			float fPercent = _fTransitionTimer * _fTransitionSpeed * 6.0f;
			fPercent = Clamp( fPercent, 0.0, 1.0f );
			SetSpriteAlphaToPercentOfTarget( fPercent );
			SetTextAlphaToPercentOfTarget( fPercent );

			float fTargetScale = 1.0f;
			float fMarginTimer = Clamp( _fTransitionTimer * _fTransitionSpeed * 15.0f, 0.0f, JpUtils::_fPi );
			float fMarginSize = 15.0f * SoUtils::GetiPadRatioX() * SoUtils::GetSpriteZoom();
			float fMargin = sinf( fMarginTimer ) * fMarginSize;

			_fWidth = _fTargetWidth + fMargin;
			_fHeight = _fTargetHeight + fMargin;

			if ( fMarginTimer == JpUtils::_fPi && fPercent == 1.0f )
			{
				_fTransitionTimer = 0.0f;
				_eTransitionState = TRANSITION_STATE_SHOW;
			}

			break;
		}
		case TRANSITION_STATE_SHOW:
		{
			break;
		}
		case TRANSITION_STATE_ANIM_OUT:
		{
			float fPercent = ( 1.0f - ( _fTransitionTimer * _fTransitionSpeed * 2.0f ) );
			fPercent = Clamp( fPercent, 0.0, 1.0f );
			SetSpriteAlphaToPercentOfTarget( fPercent );
			SetTextAlphaToPercentOfTarget( fPercent );

			float fTargetScale = 1.0f;
			float fMarginTimer = Clamp( _fTransitionTimer * _fTransitionSpeed * 10.0f, 0.0f, JpUtils::_fPi );
			float fMarginSize = 25.0f * SoUtils::GetiPadRatioX() * SoUtils::GetSpriteZoom();
			float fMargin = sinf( fMarginTimer ) * fMarginSize;

			_fWidth = _fTargetWidth + fMargin;
			_fHeight = _fTargetHeight + fMargin;

			if ( fPercent == 0.0f )
			{
				if ( _bDeleteOnTransitionOut )
				{
					_pGuiManager->RemoveAndDeleteGui( this );
				}

				_fTransitionTimer = 0.0f;
				_eTransitionState = TRANSITION_STATE_HIDE;
			}
			break;
		}
		case TRANSITION_STATE_HIDE:
		{
			_pSpriteBackground->setVisible( false );
			break;
		}
		default:
			ASSERTF( 0, ( "\nUnknown transition state." ) );
	}
}



/******************************************************************************
 FUNCTION:  UpdateStyleNone
******************************************************************************/
void SoGuiAnim::UpdateStyleNone( float dt )
{
	_fTransitionTimer += dt;

	switch ( _eTransitionState )
	{
		case TRANSITION_STATE_INIT:
		{
			if ( _pSpriteBackground )
			{
				_pSpriteBackground->setOpacity( 0 );
				_pSpriteBackground->setVisible( false );
			}
			break;
		}
		case TRANSITION_STATE_ANIM_IN:
		{
			if ( _pSpriteBackground )
			{
				_pSpriteBackground->setVisible( true );
			}

			_eTransitionState = TRANSITION_STATE_SHOW;
			break;
		}
		case TRANSITION_STATE_SHOW:
		{
			break;
		}
		case TRANSITION_STATE_ANIM_OUT:
		{
			if ( _pSpriteBackground )
			{
				_pSpriteBackground->setOpacity( 0.0f );
			}

			_eTransitionState = TRANSITION_STATE_HIDE;
			break;
		}
		case TRANSITION_STATE_HIDE:
		{
			if ( _pSpriteBackground )
			{
				_pSpriteBackground->setVisible( false );
			}

			break;
		}
		default:
			ASSERTF( 0, ( "\nUnknown transition state." ) );
	}
}



/******************************************************************************
 FUNCTION:  SetSpriteAlphaToPercentOfTarget
******************************************************************************/
void SoGuiAnim::SetSpriteAlphaToPercentOfTarget( float fPercent )
{
	_fSpriteAlpha = ( fPercent * _fTargetSpriteAlpha );

	SoSpriteList::Iterator Iter( &_SpriteList );
	while ( Iter.Next() )
	{
		SoSprite* pSprite = Iter.GetNodeData();
		pSprite->setOpacity( _fSpriteAlpha * 255.0f );
	}
}



/******************************************************************************
 FUNCTION:  SetTextAlphaToPercentOfTarget
******************************************************************************/
void SoGuiAnim::SetTextAlphaToPercentOfTarget( float fPercent )
{
	_fTextAlpha = ( fPercent * _fTargetTextAlpha );
	
	SoLabelBmFontList::Iterator Iter2( &_LabelList );
	while ( Iter2.Next() )
	{
		SoLabelBmFont* pLabel = Iter2.GetNodeData();
		pLabel->setOpacity( _fTextAlpha * 255.0f );
	}
}



/******************************************************************************
 FUNCTION:  TransitionIn
******************************************************************************/
void SoGuiAnim::TransitionIn()
{
	if ( _eTransitionState != TRANSITION_STATE_ANIM_IN && _eTransitionState != TRANSITION_STATE_SHOW )
	{
		_fTransitionTimer = 0.0f;
		_eTransitionState = TRANSITION_STATE_ANIM_IN;
	}
}



/******************************************************************************
 FUNCTION:  TransitionOut
******************************************************************************/
void SoGuiAnim::TransitionOut()
{
	if ( _eTransitionState != TRANSITION_STATE_ANIM_OUT && _eTransitionState != TRANSITION_STATE_HIDE )
	{
		_fTransitionTimer = 0.0f;
		_eTransitionState = TRANSITION_STATE_ANIM_OUT;
	}
}
