/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "JpDebug.h"
#include "SoGuiManager.h"

#include "mmgrChooser.h"


USING_NS_CC;


SoGuiManager* SoGuiManager::_pInstance = NULL;


/******************************************************************************
 FUNCTION:  SoGuiManager
******************************************************************************/
SoGuiManager::SoGuiManager() :
_pParentNode( NULL ),
_eInputSource( INPUT_SOURCE_INTERNAL )
{
}



/******************************************************************************
 FUNCTION:  ~SoGuiManager
******************************************************************************/
SoGuiManager::~SoGuiManager()
{
	Deinit();
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoGuiManager::Init( CCNode* pParentNode, SoGuiManager::INPUT_SOURCE eInputSource )
{
	_pParentNode = pParentNode;
	_eInputSource = eInputSource;

	if ( _eInputSource == INPUT_SOURCE_INTERNAL )
	{
		_InputTouchListener.SetOwner( this );
		SoInputTouch::GetInstance()->AddListener( &_InputTouchListener );

		_InputGestureListener.SetOwner( this );
		SoInputGesture::GetInstance()->AddListener( &_InputGestureListener );

		_InputKeypadListener.SetOwner( this );
		SoInputKeypad::GetInstance()->AddListener( &_InputKeypadListener );

		#if DCG_IS_TARGET_WIN32
			_InputMouseListener.SetOwner( this );
			SoInputMouse::GetInstance()->AddListener( &_InputMouseListener );
		#endif
	}
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoGuiManager::Deinit()
{
	RemoveAndDeleteAllGuisNow();

	if ( _eInputSource == INPUT_SOURCE_INTERNAL )
	{
		SoInputTouch::GetInstance()->RemoveListener( &_InputTouchListener );
		SoInputGesture::GetInstance()->RemoveListener( &_InputGestureListener );
		SoInputKeypad::GetInstance()->RemoveListener( &_InputKeypadListener );

		#if DCG_IS_TARGET_WIN32
			SoInputMouse::GetInstance()->RemoveListener( &_InputMouseListener );
		#endif
	}
}



/******************************************************************************
 FUNCTION:  Update
 NOTE:      The update is hierarchical.  This is needed for relative position
            updating, and likely other things.
******************************************************************************/
void SoGuiManager::Update( float dt )
{
	struct Local
	{
		static void Process( SoGuiManager* pGuiManager, SoGui* pParent, float dt )
		{
			if ( pParent )
			{	
				// update the parent
				if ( pParent != pGuiManager->GetHeadGui() )
					pParent->Update( dt );

				// update all the parents children
				SoChildGuiGuidList::Iterator Iter( pParent->GetChildGuiGuidList() );
				while ( Iter.Next() )
				{
					guid gChild = (guid)Iter.GetNodeData();
					SoGui* pChild = pGuiManager->GetGuiByGuid( gChild );
					Process( pGuiManager, pChild, dt );
				}
			}
		}
	};

	Local::Process( this, &_HeadGui, dt );

	DeleteAllGuisInPendingDeletionList();
}



/******************************************************************************
 FUNCTION:  AddGui
 NOTE:      If pParentGui is NULL, then the pGuiToAdd is made a child of
            the _HeadGui.
******************************************************************************/
guid SoGuiManager::AddGui( SoGui* pGuiToAdd, SoGui* pParentGui )
{
	// This new Gui needs a guid before calling its Init.
	// This is because the Init might create new child guis which
	// will need this new guis guid.
	pGuiToAdd->SetGuid( GenerateGuid() );

	pGuiToAdd->Init( _pParentNode, this );

	if ( pParentGui )
		pParentGui->AddChildGui( pGuiToAdd );
	else
		_HeadGui.AddChildGui( pGuiToAdd );

	_GuiList.AddLast( pGuiToAdd );
	_GuiGuidTable.Store( pGuiToAdd->GetGuid(), pGuiToAdd );

	return pGuiToAdd->GetGuid();
}



/******************************************************************************
 FUNCTION:  CreateAndAddAnimButton
******************************************************************************/
guid SoGuiManager::CreateAndAddAnimButton( const SoGuiAnimButton::SetupParams& Params, SoGui* pParentGui )
{
	SoGuiAnimButton* pButton = new SoGuiAnimButton;

	guid gButton = AddGui( pButton, pParentGui );

	pButton->Setup( Params );

	return gButton;
}



/******************************************************************************
 FUNCTION:  CreateAndAddAnimComboButton
******************************************************************************/
guid SoGuiManager::CreateAndAddAnimComboButton( const SoGuiAnimComboButton::SetupParams& Params, SoGui* pParentGui )
{
	SoGuiAnimComboButton* pComboButton = new SoGuiAnimComboButton;

	guid gComboButton = AddGui( pComboButton, pParentGui );

	pComboButton->Setup( Params );

	return gComboButton;
}



/******************************************************************************
 FUNCTION:  CreateAndAddAnimSlider
******************************************************************************/
guid SoGuiManager::CreateAndAddAnimSlider( const SoGuiAnimSlider::SetupParams& Params, SoGui* pParentGui )
{
	SoGuiAnimSlider* pSlider = new SoGuiAnimSlider;

	guid gSlider = AddGui( pSlider, pParentGui );

	pSlider->Setup( Params );

	return gSlider;
}



/******************************************************************************
 FUNCTION:  CreateAndAddAnimToggle
******************************************************************************/
guid SoGuiManager::CreateAndAddAnimToggle( const SoGuiAnimToggle::SetupParams& Params, SoGui* pParentGui )
{
	SoGuiAnimToggle* pToggle = new SoGuiAnimToggle;

	guid gToggle = AddGui( pToggle, pParentGui );

	pToggle->Setup( Params );

	return gToggle;
}



/******************************************************************************
 FUNCTION:  RemoveAndDeleteGui
******************************************************************************/
void SoGuiManager::RemoveAndDeleteGui( SoGui* pGui )
{
	if ( pGui )
	{
		_PendingDeletionList.AddUniquely( pGui, 0 );
	}
}



/******************************************************************************
 FUNCTION:  RemoveAndDeleteAllGuis
******************************************************************************/
void SoGuiManager::RemoveAndDeleteAllGuis()
{
	SoGuiList::Iterator ListIter( &_GuiList );
	while ( ListIter.Next() )
	{
		SoGui* pGui = ListIter.GetNodeData();

		// When we remove all the guis, we only add to the pending deletion list
		// the guis directly connected to the head gui.  We do not add any
		// child guis, as they get deleted when the parents are deleted.
		// mdomowicz 2013_10_26
		bool bParentIsHeadGui = ( pGui->GetParentGui() == NULL );

		if ( bParentIsHeadGui )
			_PendingDeletionList.AddUniquely( pGui, 0 );
	}

	_GuiList.RemoveAll();
}



/******************************************************************************
 FUNCTION:  RemoveAndDeleteAllGuisNow
******************************************************************************/
void SoGuiManager::RemoveAndDeleteAllGuisNow()
{
	RemoveAndDeleteAllGuis();
	DeleteAllGuisInPendingDeletionList();
}



/******************************************************************************
 FUNCTION:  DeleteAllGuisInPendingDeletionList
******************************************************************************/
void SoGuiManager::DeleteAllGuisInPendingDeletionList()
{
	//-----------------------
	// This was originally based on the implementation in SoActorManager.
	// However, a key difference requires this version to be much more complex.
	//
	// The difference is that Guis can have child Guis, and when a Parent Gui
	// gets deleted, it automatically deletes all of its child guis.
	//
	// The loop below takes this into account.  Basically it makes a copy of
	// the pending list, and deletes guis from the copy.  Thus when child
	// guis get deleted (ie, added to the pending list), it's not corrupting
	// the list we are currently working with (the copy).
	//
	// mdomowicz 2013_08_07

	while ( _PendingDeletionList.GetNumEntries() > 0 )
	{
		//-----------------------
		// Make a copy of the pending list
		SoGuiList ListCopy;
		SoGuiList::Iterator Iter( &_PendingDeletionList );

		while ( Iter.Next() )
		{
			SoGui* pGui = Iter.GetNodeData();
			ListCopy.AddLast( pGui );
		}


		//-----------------------
		// Then empty the pending list.  Child Gui's may be added to
		// this as we deinit/delete all the gui's in from the ListCopy.
		_PendingDeletionList.RemoveAll();


		//-----------------------
		// Now deinit and delete all the Guis in the Listcopy
		SoGuiList::Iterator Iter2( &ListCopy );
		while ( Iter2.Next() )
		{
			SoGui* pGui = Iter2.GetNodeData();
			pGui->Deinit();

			_GuiList.Remove( pGui );
			_GuiGuidTable.Remove( pGui->GetGuid() );

			delete pGui;
		}


		//-----------------------
		// Just in case!
		// As we deinited/deleted Guis above, child Guis were likely
		// added to the pending list.  Just to make absolutely sure we
		// don't delete a gui twice, we remove all Guis currently in
		// the ListCopy (ie, the guis we just deleted above) from the 
		// Pendinglist.  If things are working, nothing in ListCopy will
		// be found in Pending, and so no removals will happen.
		Iter2.Reset();
		while ( Iter2.Next() )
		{
			SoGui* pGui = Iter2.GetNodeData();

			ASSERTF( !_PendingDeletionList.Find( pGui ), ( "\nA gui that was just deleted, was added back to the pending delete list!  That's bad." ) );

			_PendingDeletionList.Remove( pGui );
		}


		ListCopy.RemoveAll();
	}
}



/******************************************************************************
 FUNCTION:  GenerateGuid
******************************************************************************/
guid SoGuiManager::GenerateGuid()
{
	unsigned int uiGuid = JpRandom::G.GetInt();

	while ( GetGuiByGuid( uiGuid ) != NULL )
		uiGuid = JpRandom::G.GetInt();

	return uiGuid;
}



/******************************************************************************
 FUNCTION:  GetGuiByGuid
******************************************************************************/
SoGui* SoGuiManager::GetGuiByGuid( guid g )
{
	return _GuiGuidTable.Find( g );
}



/******************************************************************************
 FUNCTION:  SetHeadGuiAbsolutionPositionToScreenCenter
******************************************************************************/
void SoGuiManager::SetHeadGuiAbsolutionPositionToScreenCenter()
{
	CCSize SizeInPixels = CCDirector::sharedDirector()->getWinSize();
	float fCenterX = SizeInPixels.width / 2.0f;
	float fCenterY = SizeInPixels.height / 2.0f;
	_HeadGui.SetAbsolutePos( JpVector2f( fCenterX, fCenterY ) );
}



/******************************************************************************
 FUNCTION:  SendMessageToGui
******************************************************************************/
void SoGuiManager::SendMessageToGui( guid gGui, SoGuiMessage* pMessage )
{
	SoGui* pGui = GetGuiByGuid( gGui );
	
	ASSERTF( pGui, ( "\nGui not found.  Cannot send message." ) );

	if ( pGui )
		pGui->HandleMessage( pMessage );
}



/******************************************************************************
 FUNCTION:  OnTouchesDown
******************************************************************************/
bool SoGuiManager::OnTouchesDown( const SoCcTouchList* pSoCcTouchList ) 
{
	bool bInputHandled = false;

	SoGuiList::Iterator Iter( &_GuiList );
	while ( Iter.Next() )
	{
		SoGui* pGui = Iter.GetNodeData();
		if ( pGui->OnTouchesDown( pSoCcTouchList ) )
			bInputHandled = true;
	}

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  OnTouchesUp
******************************************************************************/
bool SoGuiManager::OnTouchesUp( const SoCcTouchList* pSoCcTouchList )
{
	bool bInputHandled = false;

	SoGuiList::Iterator Iter( &_GuiList );
	while ( Iter.Next() )
	{
		SoGui* pGui = Iter.GetNodeData();
		if ( pGui->OnTouchesUp( pSoCcTouchList ) )
			bInputHandled = true;
	}

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  OnTouchesMove
******************************************************************************/
bool SoGuiManager::OnTouchesMove( const SoCcTouchList* pSoCcTouchList ) 
{
	bool bInputHandled = false;

	SoGuiList::Iterator Iter( &_GuiList );
	while ( Iter.Next() )
	{
		SoGui* pGui = Iter.GetNodeData();
		if ( pGui->OnTouchesMove( pSoCcTouchList ) )
			bInputHandled = true;
	}

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  OnMouseMoved
******************************************************************************/
#if DCG_IS_TARGET_WIN32
	bool SoGuiManager::OnMouseMoved( const OIS::MouseEvent &MouseEventArg )
	{
		bool bInputHandled = false;

		SoGuiList::Iterator Iter( &_GuiList );
		while ( Iter.Next() )
		{
			SoGui* pGui = Iter.GetNodeData();
			if ( pGui->OnMouseMoved( MouseEventArg ) )
				bInputHandled = true;
		}

		return bInputHandled;
	}
#endif



/******************************************************************************
 FUNCTION:  OnGestureDown
******************************************************************************/
bool SoGuiManager::OnGestureDown( SoGesture* pGesture ) 
{
	bool bInputHandled = false;

	SoGuiList::Iterator Iter( &_GuiList );
	while ( Iter.Next() )
	{
		SoGui* pGui = Iter.GetNodeData();
		if ( pGui->OnGestureDown( pGesture ) )
			bInputHandled = true;
	}

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  OnGestureUp
******************************************************************************/
bool SoGuiManager::OnGestureUp( SoGesture* pGesture ) 
{
	bool bInputHandled = false;

	SoGuiList::Iterator Iter( &_GuiList );
	while ( Iter.Next() )
	{
		SoGui* pGui = Iter.GetNodeData();
		if ( pGui->OnGestureUp( pGesture ) )
			bInputHandled = true;
	}

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  OnGestureMove
******************************************************************************/
bool SoGuiManager::OnGestureMove( SoGesture* pGesture ) 
{
	bool bInputHandled = false;

	SoGuiList::Iterator Iter( &_GuiList );
	while ( Iter.Next() )
	{
		SoGui* pGui = Iter.GetNodeData();
		if ( pGui->OnGestureMove( pGesture ) )
			bInputHandled = true;
	}

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  OnKeypadBackClicked
******************************************************************************/
bool SoGuiManager::OnKeypadBackClicked()
{
	bool bInputHandled = false;

	SoGuiList::Iterator Iter( &_GuiList );
	while ( Iter.Next() )
	{
		SoGui* pGui = Iter.GetNodeData();
		if ( pGui->OnKeypadBackClicked() )
			bInputHandled = true;
	}

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  OnKeypadMenuClicked
******************************************************************************/
bool SoGuiManager::OnKeypadMenuClicked()
{
	bool bInputHandled = false;

	SoGuiList::Iterator Iter( &_GuiList );
	while ( Iter.Next() )
	{
		SoGui* pGui = Iter.GetNodeData();
		if ( pGui->OnKeypadMenuClicked() )
			bInputHandled = true;
	}

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  OnInput
******************************************************************************/
bool SoGuiManager::OnInput( SoInputDispatcher::InputPacket* pInput )
{
	bool bInputHandled = false;

	if ( pInput->_pSoCcTouchesUpList )
		bInputHandled = OnTouchesUp( pInput->_pSoCcTouchesUpList );
	if ( pInput->_pSoCcTouchesDownList )
		bInputHandled = OnTouchesDown( pInput->_pSoCcTouchesDownList );
	if ( pInput->_pSoCcTouchesMoveList )
		bInputHandled = OnTouchesMove( pInput->_pSoCcTouchesMoveList );

	if ( pInput->_pGestureUp )
		bInputHandled = OnGestureUp( pInput->_pGestureUp );
	if ( pInput->_pGestureDown )
		bInputHandled = OnGestureDown( pInput->_pGestureDown );
	if ( pInput->_pGestureMove )
		bInputHandled = OnGestureMove( pInput->_pGestureMove );

	#if DCG_IS_TARGET_WIN32
		if ( pInput->_pMouseEvent )
			bInputHandled = OnMouseMoved( *pInput->_pMouseEvent );
	#endif

	if ( pInput->_bKeypadBackButtonPressed )
		bInputHandled = OnKeypadBackClicked();
	if ( pInput->_bKeypadMenuButtonPressed )
		bInputHandled = OnKeypadMenuClicked();

	return bInputHandled;
}



/******************************************************************************
 FUNCTION:  CreateInstance
******************************************************************************/
void SoGuiManager::CreateInstance()
{
	ASSERTF( _pInstance == NULL, ( "\nCan't create more than one static instance.  Not allowed." ) );

	if ( !_pInstance )
		_pInstance = new SoGuiManager;
}



/******************************************************************************
 FUNCTION:  DeleteInstance
******************************************************************************/
void SoGuiManager::DeleteInstance()
{
	if ( _pInstance )
	{
		delete _pInstance;
		_pInstance = NULL;
	}
}


