/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#if DCG_IS_TARGET_WIN32
	#include "direct.h"
#endif


#include "SoFontDefs.h"
#include "SoHudLog.h"
#include "SoLabelBmFont.h"
#include "SoUtils.h"

#include "mmgrChooser.h"



/******************************************************************************
 FUNCTION:  SoLogMessage
******************************************************************************/
SoLogMessage::SoLogMessage() :
_fTimer( 0.0f ),
_fMessageShowDuration( 2.0f ),
_eCurrentState( STATE_ANIMATE_IN ),
_pBmFontLabel( NULL )
{
}



/******************************************************************************
 FUNCTION:  ~SoLogMessage
******************************************************************************/
SoLogMessage::~SoLogMessage()
{
	Deinit();
}



/******************************************************************************
 FUNCTION:  CreateBmFontLabel
******************************************************************************/
SoLabelBmFont* SoLogMessage::CreateBmFontLabel( const std::string& sText, const std::string& sFontName )
{
	CCSize size = CCDirector::sharedDirector()->getWinSize();

	SoLabelBmFont* pLabel = SoLabelBmFont::create
	(
		sText.c_str(), 
		SoUtils::AddRelativePathToFont( sFontName.c_str() ), 
		size.width, 
		kCCTextAlignmentLeft
	);
	pLabel->setOpacity( 128.0f );
	pLabel->retain();
	pLabel->setAnchorPoint( CCPoint( 0, 0 ) );
	pLabel->setScaleX( SoUtils::GetiPadRatioX() );
	pLabel->setScaleY( SoUtils::GetiPadRatioY() );

	_pBmFontLabel = pLabel;

	SetPosition( 0.0f, 0.0f );

	return _pBmFontLabel;
}



/******************************************************************************
 FUNCTION:  DestroyBmFontLabel
******************************************************************************/
void SoLogMessage::DestroyBmFontLabel()
{
	if ( _pBmFontLabel )
	{
		_pBmFontLabel->release();
		_pBmFontLabel = NULL;
	}
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoLogMessage::Deinit()
{
	DestroyBmFontLabel();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoLogMessage::Update( float dt )
{
	_fTimer += dt;


	switch ( _eCurrentState )
	{
		case STATE_ANIMATE_IN:
		{
			float fAlpha = _fTimer * 5.0f;
			fAlpha = Clamp( fAlpha, 0.0f, 1.0f );

			_pBmFontLabel->setColor( _Color );

			int iAlpha = fAlpha * 255.0f;
			_pBmFontLabel->setOpacity( iAlpha );

			if ( fAlpha == 1.0f )
			{
				_fTimer = 0.0f;
				_eCurrentState = STATE_SHOW;
			}

			break;
		}
		case STATE_SHOW:
		{
			// negative duration means, "show forever"
			if ( _fMessageShowDuration >= 0.0f ) 
			{
				if ( _fTimer > _fMessageShowDuration )
				{
					_fTimer = 0.0f;
					_eCurrentState = STATE_ANIMATE_OUT;
				}
			}

			break;
		}
		case STATE_ANIMATE_OUT:
		{

			float fAlpha = 1.0f - _fTimer;
			fAlpha = Clamp( fAlpha, 0.25f, 1.0f );

			_pBmFontLabel->setColor( _Color );

			int iAlpha = fAlpha * 255.0f;
			_pBmFontLabel->setOpacity( iAlpha );

			if ( fAlpha == 0.25f )
			{
				_fTimer = 0.0f;
				_eCurrentState = STATE_DONE;
			}

			break;
		}
		case STATE_DONE:
		{
			float fTween = Clamp( _fTimer, 0.0f, 1.0f );
			ccColor3B Color = Interpolate( _Color, JpUtils::_White, fTween );
			_pBmFontLabel->setColor( Color );
			break;
		}
		default:
			ASSERTF( 0, ( "\nunknown state" ) );
	}
}



/******************************************************************************
 FUNCTION:  GetPosition
******************************************************************************/
JpVector2f SoLogMessage::GetPosition()
{
	if ( _pBmFontLabel )
		return _pBmFontLabel->GetPosition();

	return SoUtils::_v2fZero;
}



/******************************************************************************
 FUNCTION:  SetPosition
******************************************************************************/
void SoLogMessage::SetPosition( float fX, float fY )
{
	_pBmFontLabel->setPosition( ccp( fX, fY ) );
}









/******************************************************************************
 FUNCTION:  SoHudLog
******************************************************************************/
SoHudLog::SoHudLog() : 
_bIsVisible( true ),
_fX( 0.0f ),
_fY( 0.0f ),
_fWidth( 0.0f ),
_fHeight( 0.0f ),
_fMessageShowDuration( 2.0f ),
_iMaxMessagesBeforeFifoRemoval( -1 ),
_pParentNode( NULL ),
_fFillHeight( 0.0f )
{
}



/******************************************************************************
 FUNCTION:  ~SoHudLog
******************************************************************************/
SoHudLog::~SoHudLog()
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoHudLog::Init()
{
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoHudLog::Deinit()
{
	Clear();

	_pParentNode = NULL;
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoHudLog::Update( float dt )
{
	if ( _bIsVisible )
	{
		float fLineCount = 0.0f;

		SoLogMessageList::Iterator Iter( &_LogMessageList );
		while( Iter.Prev() )
		{
			SoLogMessage* pMessage = Iter.GetNodeData();
			float fLineHeight = pMessage->GetBmFontLabel()->GetLineHeight() * SoUtils::GetiPadRatioY();

			if ( pMessage->IsDone() && false )
			{
				_pParentNode->removeChild( pMessage->GetBmFontLabel(), DCG_CLEANUP_ALL_RUNNING_ACTIONS );
				_LogMessageList.Remove( pMessage );
				pMessage->Deinit();
				delete pMessage;
			}
			else
			{
				pMessage->Update( dt );

				float fY = _fY + ( fLineHeight * fLineCount );
				pMessage->SetPosition( _fX, fY );

				fLineCount += 1.0f;
			}
		}


		AdjustHeightsForFillHeight();


		if ( _iMaxMessagesBeforeFifoRemoval >= 0 )
		{
			while ( _LogMessageList.GetNumEntries() > _iMaxMessagesBeforeFifoRemoval )
			{
				SoLogMessageList::Node* pNode = _LogMessageList.GetFirst();

				SoLogMessage* pMessage = pNode->GetData();

				_pParentNode->removeChild( pMessage->GetBmFontLabel(), DCG_CLEANUP_ALL_RUNNING_ACTIONS );
				_LogMessageList.Remove( pMessage );
				pMessage->Deinit();
				delete pMessage;
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  GetHeightOfAllEntries
******************************************************************************/
float SoHudLog::GetHeightOfAllEntries()
{
	float fTotalHeight = 0.0f;

	SoLogMessageList::Iterator Iter( &_LogMessageList );
	while( Iter.Prev() )
	{
		SoLogMessage* pMessage = Iter.GetNodeData();
		float fLineHeight = pMessage->GetBmFontLabel()->GetLineHeight() * SoUtils::GetiPadRatioY();
		fTotalHeight += fLineHeight;
	}

	return fTotalHeight;
}



/******************************************************************************
 FUNCTION:  AdjustHeightsForFillHeight

		This function can make logs that are at the top edge of the screen
		slightly nicer.  Normally, the log behaves like this:

           ____________      ____________     ____________     ____________
                                               line1            line2
                       -->    line1      -->   line2      -->   line3
            line1             line2            line3            line4


		This looks slightly bad because the first few lines are just "floating"
		down from the top edge, until enough lines are added.  It would be
		better if the behavior was more like:

           ____________      ____________     ____________     ____________  _
            line1             line1            line1            line2         \
                       -->    line2      -->   line2      -->   line3          --- "fill Height"
                                               line3            line4        _/


		This function basically checks to see if the height of all the
		lines in the log are shorter than the "fill height", and if they
		are, they are all pushed upwards.  So for example, if the fill
		height is 100.0f, but the total log height is only 90.0f, we
		must push the log up by 10.0f.

		mdomowicz 2013_10_24
******************************************************************************/
void SoHudLog::AdjustHeightsForFillHeight()
{
	float fTotalHeight = GetHeightOfAllEntries();
	if ( fTotalHeight < _fFillHeight )
	{
		float fDelta = ( _fFillHeight - fTotalHeight );

		SoLogMessageList::Iterator Iter2( &_LogMessageList );
		while( Iter2.Next() )
		{
			SoLogMessage* pMessage = Iter2.GetNodeData();

			JpVector2f vPos = pMessage->GetPosition();
			vPos.Y += fDelta;
			pMessage->SetPosition( vPos.X, vPos.Y );
		}
	}
}



/******************************************************************************
 FUNCTION:  Clear
******************************************************************************/
void SoHudLog::Clear()
{
	SoLogMessageList::Iterator Iter( &_LogMessageList );
	while( Iter.Prev() )
	{
		SoLogMessage* pMessage = Iter.GetNodeData();
		_pParentNode->removeChild( pMessage->GetBmFontLabel(), DCG_CLEANUP_ALL_RUNNING_ACTIONS );
		pMessage->Deinit();
	}

	_LogMessageList.DeleteAll();

	ClearFullHistory();	
}



/******************************************************************************
 FUNCTION:  AddMessage
******************************************************************************/
void SoHudLog::AddMessage( const std::string& sMessage, const ccColor3B& Color )
{
	if ( _pParentNode )
	{
		SoLogMessage* pMessage = new SoLogMessage;
		
		float fStartX = _fX;
		float fStartY = _fY;

		if ( _sFontName == "" )
			_sFontName = SoFontDefs::GetTestLarge();

		SoLabelBmFont* pLabel = pMessage->CreateBmFontLabel( sMessage, _sFontName );
		pLabel->setVisible( _bIsVisible );
		_pParentNode->addChild(pLabel, 1);
		SoUtils::SetRenderZOrder( _pParentNode, pLabel, BASE_RENDER_ORDER_LOG );

		pMessage->SetColor( Color );
		pMessage->SetMessageShowDuration( _fMessageShowDuration );
		pMessage->SetPosition( fStartX, fStartY );


		_LogMessageList.AddLast( pMessage );

		_FullHistoryCsv.AddNewRow();
		_FullHistoryCsv.AppendStringToLastRow( sMessage.c_str() );
	}
}



/******************************************************************************
 FUNCTION:  AddMessageF
******************************************************************************/
void SoHudLog::AddMessageF( const ccColor3B& Color, const char* pFormat, ... )
{
	if ( strlen( pFormat ) < 2048 )
	{
		char buffer[2048];
		va_list args;
		va_start (args, pFormat);
		vsprintf (buffer,pFormat, args);
		va_end (args);

		AddMessage( buffer, Color );
	}
}



/******************************************************************************
 FUNCTION:  SetPosition
******************************************************************************/
void SoHudLog::SetPosition
( 
	float fBottomLeftX, 
	float fBottomLeftY
)
{
	_fX = fBottomLeftX;
	_fY = fBottomLeftY;
}



/******************************************************************************
 FUNCTION:  SetPositionAndSize
******************************************************************************/
void SoHudLog::SetPositionAndSize
( 
	float fBottomLeftX, 
	float fBottomLeftY, 
	float fWidth, 
	float fHeight 
)
{
	_fX = fBottomLeftX;
	_fY = fBottomLeftY;
	_fWidth = fWidth;
	_fHeight = fHeight;
}



/******************************************************************************
 FUNCTION:  SetVisible
******************************************************************************/
void SoHudLog::SetVisible( bool b )
{
	_bIsVisible = b;

	SoLogMessageList::Iterator Iter( &_LogMessageList );
	while( Iter.Prev() )
	{
		SoLogMessage* pMessage = Iter.GetNodeData();
		pMessage->GetBmFontLabel()->setVisible( _bIsVisible );
	}	
}



/******************************************************************************
 FUNCTION:  WriteFullHistoryToDisk
******************************************************************************/
void SoHudLog::WriteFullHistoryToDisk()
{
	if ( !_FullHistoryCsv.IsClear()  )
	{
		//-----------------------
		// get the time info
		time_t rawtime;
		time( &rawtime );

		tm* timeinfo;
		timeinfo = localtime( &rawtime );

		char acTimeString[ 32 ] = "";
		strftime( acTimeString, 128, "%Y_%m_%d__%H.%M.%S.txt", timeinfo );


		//-----------------------
		// Assemble the file name with the time info
		std::string sPsName = "Gamelog-";
		sPsName.append( acTimeString );


		//-----------------------
		// This makes sure the "save" folder exists
		#if DCG_IS_TARGET_WIN32
			_mkdir( SoUtils::AddRelativePathToSaveData( "" ) );
		#endif


		//-----------------------
		// write the file
		const char* cpFullPath = SoUtils::AddRelativePathToSaveData( sPsName.c_str() );
		_FullHistoryCsv.WriteToFile( cpFullPath );
	}
}