/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "NeCsvIn.h"

#include "SoDrawDebug.h"
#include "SoInputGesture.h"
#include "SoSprite.h"
#include "SoSoundManager.h"
#include "SoUtils.h"

#include "mmgrChooser.h"



SoInputGesture* SoInputGesture::_pInstance = NULL;


#define SoInputGestureList JpList< SoInputGesture >



/******************************************************************************
 FUNCTION:  SoInputGesture
 ******************************************************************************/
SoGesture::SoGesture()
{
	_eGestureType = GESTURE_TYPE_LAST;
	_iNumMovesSingle = 0;
	_iNumMovesMulti = 0;
	_iNumTouches = 0;
	_fTimer = 0.0f;
	
	memset( _vPositionDeltaOverall, 0, GESTURE_NUM_FINGERS * sizeof( JpVector2f ) );
	memset( _vPositionDeltaPrevious, 0, GESTURE_NUM_FINGERS * sizeof( JpVector2f ) );
	memset( _vPosition, 0, GESTURE_NUM_FINGERS * GESTURE_NUM_FRAMES_TO_STORE * sizeof( JpVector2f ) );
	memset( _vPositionArray, 0, GESTURE_NUM_FINGERS * GESTURE_NUM_FRAMES_TO_STORE * sizeof( JpVector2f ) );
}



/******************************************************************************
 FUNCTION:  GrowOrCreateNewGesture
 ******************************************************************************/
void SoInputGesture::GrowOrCreateNewGesture( const SoCcTouchList* pSoCcTouchList )
{
	SoCcTouchList* pNonConstTouchList = (SoCcTouchList*)pSoCcTouchList;
	
	if ( !_pGesture )
		_pGesture = new SoGesture;

	// We may have added a new finger to a previous gesture
	_pGesture->_iNumTouches = Clamp ( _pGesture->_iNumTouches + pNonConstTouchList->GetNumEntries(), 0, 4 );
	
	
	// Go through all fingers.  If we have more than 1, we could have a
	// multi-finger gesture, so fill in the position data.
	for ( int i = 0; i < pNonConstTouchList->GetNumEntries(); i ++ )
	{
		SoCcTouchList::Node* pNode = pNonConstTouchList->GetAtIndex( i );
		CCTouch* pTouch = pNode->GetData();
	
		_pGesture->_vPosition[ i ] = SoUtils::ToJpVector2f( pTouch->locationInView() );
		_pGesture->_vPositionArray[ i ][ 0 ] = _pGesture->_vPosition[ i ];
	}


	// 'Reset' the gesture since we are growing or just created.
	_pGesture->_iNumMovesSingle = 1;
	_pGesture->_iNumMovesMulti = 1;
	_pGesture->_fTimer = 0.0f;
}








/******************************************************************************
 FUNCTION:  SoInputGesture
 ******************************************************************************/
SoInputGesture::SoInputGesture() :
_bListenerListLocked( false ),
_pGesture( NULL )
{
}



/******************************************************************************
 FUNCTION:  Init
 ******************************************************************************/
void SoInputGesture::Init()
{
	_InputTouchListener.SetOwner( this );
	SoInputTouch::GetInstance()->AddListener( &_InputTouchListener );
}



/******************************************************************************
 FUNCTION:  Deinit
 ******************************************************************************/
void SoInputGesture::Deinit()
{
	SoInputTouch::GetInstance()->RemoveListener( &_InputTouchListener );
	RemoveAllListeners();
}



/******************************************************************************
 FUNCTION:  Update
 ******************************************************************************/
void SoInputGesture::Update( float dt )
{
	if ( _pGesture )
	{
		_pGesture->_fTimer += dt;

		if ( _pGesture->_iNumTouches == 1 )
			DetectHold( _pGesture );
	}
}



/******************************************************************************
 FUNCTION:  BroadcastGestureDown
 ******************************************************************************/
void SoInputGesture::BroadcastGestureDown( SoGesture* pGesture )
{
	//-----------------------
	// The event handler we are calling can do anything it likes...even attempt to
	// add or remove to the listener list.  That would be bad since we are in
	// the middle of walking through the list.  So we lock the list to prevent tampering.
	
	_bListenerListLocked = true;
	
	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnGestureDown( pGesture );
	}
	
	_bListenerListLocked = false;
}



/******************************************************************************
 FUNCTION:  BroadcastGestureUp
 ******************************************************************************/
void SoInputGesture::BroadcastGestureUp( SoGesture* pGesture )
{
	_bListenerListLocked = true;
	
	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnGestureUp( pGesture );
	}
	
	_bListenerListLocked = false;
}



/******************************************************************************
 FUNCTION:  BroadcastGestureMove
 ******************************************************************************/
void SoInputGesture::BroadcastGestureMove( SoGesture* pGesture )
{
	_bListenerListLocked = true;
	
	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnGestureMove( pGesture );
	}
	
	_bListenerListLocked = false;
}



/******************************************************************************
 FUNCTION:  CreateInstance
 ******************************************************************************/
void SoInputGesture::CreateInstance()
{
	ASSERTF( _pInstance == NULL, ( "\nCan't create more than one instance.  Not allowed." ) );
	
	if ( !_pInstance )
	{
		_pInstance = new SoInputGesture;
		_pInstance->Init();
	}
}



/******************************************************************************
 FUNCTION:  DeleteInstance
 ******************************************************************************/
void SoInputGesture::DeleteInstance()
{
	if ( _pInstance )
	{
		_pInstance->Deinit();
		delete _pInstance;
		_pInstance = NULL;
	}
}



/******************************************************************************
 FUNCTION:  OnTouchesDown
 ******************************************************************************/
bool SoInputGesture::OnTouchesDown( const SoCcTouchList* pSoCcTouchList )
{
	GrowOrCreateNewGesture( pSoCcTouchList );
	
	if ( _pGesture )
	{
		if ( _pGesture->_iNumTouches == 1 )
		{
			//------------------
			// Detect gestures
			SoCcTouchList* pNonConstTouchList = (SoCcTouchList*)pSoCcTouchList;
			_pGesture->_vPosition[ 0 ] = SoUtils::ToJpVector2f( pNonConstTouchList->GetFirst()->GetData()->locationInView() );
			_pGesture->_vPositionArray[ 0 ][ _pGesture->_iNumMovesSingle%GESTURE_NUM_FRAMES_TO_STORE ] = _pGesture->_vPosition[ 0 ];
			_pGesture->_vPositionDeltaPrevious[ 0 ] = _pGesture->_vPositionArray[ 0 ][ _pGesture->_iNumMovesSingle%GESTURE_NUM_FRAMES_TO_STORE ] - _pGesture->_vPositionArray[ 0 ][ (_pGesture->_iNumMovesSingle - 1)%GESTURE_NUM_FRAMES_TO_STORE ];
			_pGesture->_vPositionDeltaOverall[ 0 ] = _pGesture->_vPositionArray[ 0 ][ _pGesture->_iNumMovesSingle%GESTURE_NUM_FRAMES_TO_STORE ] - _pGesture->_vPositionArray[ 0 ][ 0 ];
			_pGesture->_eGestureType = SoGesture::GESTURE_TYPE_PAN;
			
			//------------------
			// Broadcast anything detected:
			if ( _pGesture->_eGestureType != SoGesture::GESTURE_TYPE_LAST )
				BroadcastGestureDown( _pGesture );
		}
	}

	return false;
}



/******************************************************************************
 FUNCTION:  OnTouchesUp
 ******************************************************************************/
bool SoInputGesture::OnTouchesUp( const SoCcTouchList* pSoCcTouchList )
{
	if ( _pGesture )
	{
		//CCLOG( "%i finger gesture deleted.", _pGesture->_iNumTouches );

		BroadcastGestureUp( _pGesture );

		delete _pGesture;
		
		_pGesture = NULL;
	}

	return false;
}



/******************************************************************************
 FUNCTION:  OnTouchesMove
 ******************************************************************************/
bool SoInputGesture::OnTouchesMove( const SoCcTouchList* pSoCcTouchList )
{
	if ( _pGesture )
	{
		SoCcTouchList* pNonConstTouchList = (SoCcTouchList*)pSoCcTouchList;
		
		if ( pNonConstTouchList->GetNumEntries() == 1 )
		{
			if ( _pGesture->_iNumTouches == 1 )
			{
				_pGesture->_vPosition[ 0 ] = SoUtils::ToJpVector2f( pNonConstTouchList->GetAtIndex( 0 )->GetData()->locationInView() );
				_pGesture->_vPositionArray[ 0 ][ _pGesture->_iNumMovesSingle%GESTURE_NUM_FRAMES_TO_STORE ] = _pGesture->_vPosition[ 0 ];
				_pGesture->_vPositionDeltaPrevious[ 0 ] = _pGesture->_vPositionArray[ 0 ][ _pGesture->_iNumMovesSingle%GESTURE_NUM_FRAMES_TO_STORE ] - _pGesture->_vPositionArray[ 0 ][ (_pGesture->_iNumMovesSingle - 1)%GESTURE_NUM_FRAMES_TO_STORE ];
				_pGesture->_vPositionDeltaOverall[ 0 ] = _pGesture->_vPositionArray[ 0 ][ _pGesture->_iNumMovesSingle%GESTURE_NUM_FRAMES_TO_STORE ] - _pGesture->_vPositionArray[ 0 ][ 0 ];

				
				//------------------
				// Detect gestures
				DetectPan( _pGesture );

				
				//------------------
				// Broadcast anything detected:
				if ( _pGesture->_eGestureType != SoGesture::GESTURE_TYPE_LAST )
					BroadcastGestureMove( _pGesture );

				
				_pGesture->_iNumMovesSingle++;
			}
		}
		else
		{
			if ( _pGesture->_iNumTouches > 1 )
			{
				//------------------
				// Show a log
				//if ( pGesture && pNonConstTouchList->GetNumEntries() == pGesture->_iTouchIndex + 1 )
				//	CCLOG("%i finger gesture move", pGesture->_iTouchIndex + 1 );
				
				
				//------------------
				// Fill in latest gesture/movement data
				for ( int i = 0; i < pNonConstTouchList->GetNumEntries(); i ++ )
				{
					_pGesture->_vPosition[ i ] = SoUtils::ToJpVector2f( pNonConstTouchList->GetAtIndex( i )->GetData()->locationInView() );
					_pGesture->_vPositionArray[ i ][ _pGesture->_iNumMovesMulti%GESTURE_NUM_FRAMES_TO_STORE ] = _pGesture->_vPosition[ i ];
					_pGesture->_vPositionDeltaPrevious[ i ] = _pGesture->_vPositionArray[ i ][ _pGesture->_iNumMovesMulti%GESTURE_NUM_FRAMES_TO_STORE ] - _pGesture->_vPositionArray[ i ][ (_pGesture->_iNumMovesMulti - 1)%GESTURE_NUM_FRAMES_TO_STORE ];
					_pGesture->_vPositionDeltaOverall[ i ] = _pGesture->_vPositionArray[ i ][ _pGesture->_iNumMovesMulti%GESTURE_NUM_FRAMES_TO_STORE ] - _pGesture->_vPositionArray[ i ][ 0 ];
				}
				
		
				//------------------
				// Detect gestures
				DetectPinchOrZoom( _pGesture );
				
				
				//------------------
				// Broadcast anything detected:
				if ( _pGesture->_eGestureType != SoGesture::GESTURE_TYPE_LAST )
					BroadcastGestureMove( _pGesture );

				_pGesture->_iNumMovesMulti++;
			}
		}
	}

	return false;
}



/******************************************************************************
 FUNCTION:  DetectPinchOrZoom
 ******************************************************************************/
void SoInputGesture::DetectPinchOrZoom( SoGesture* pGesture )
{
	if ( pGesture->_iNumMovesMulti > 1 )
	{
		JpVector2f vDelta1 = pGesture->_vPositionDeltaPrevious[ 0 ];
		JpVector2f vDelta2 = pGesture->_vPositionDeltaPrevious[ 1 ];

		JpVector2f vDifferenceBetweenFingersCurrent = pGesture->_vPositionArray[ 0 ][ (pGesture->_iNumMovesMulti    )%GESTURE_NUM_FRAMES_TO_STORE ] - pGesture->_vPositionArray[ 1 ][ (pGesture->_iNumMovesMulti    )%GESTURE_NUM_FRAMES_TO_STORE ];
		JpVector2f vDifferenceBetweenFingersPrev    = pGesture->_vPositionArray[ 0 ][ (pGesture->_iNumMovesMulti - 1)%GESTURE_NUM_FRAMES_TO_STORE ] - pGesture->_vPositionArray[ 1 ][ (pGesture->_iNumMovesMulti - 1)%GESTURE_NUM_FRAMES_TO_STORE ];

		if  ( vDelta1.Length() > 1.0f || vDelta2.Length() > 1.0f )
		{
			if ( vDifferenceBetweenFingersCurrent.LengthSquared() < vDifferenceBetweenFingersPrev.LengthSquared() )
				pGesture->_eGestureType = SoGesture::GESTURE_TYPE_PINCH;
			else
				pGesture->_eGestureType = SoGesture::GESTURE_TYPE_ZOOM;
		}
	}
}



/******************************************************************************
 FUNCTION:  DetectPan
 ******************************************************************************/
void SoInputGesture::DetectPan( SoGesture* pGesture )
{
	if ( pGesture->_fTimer >= GESTURE_TIME_TO_MULTITOUCH )
	{
		pGesture->_eGestureType = SoGesture::GESTURE_TYPE_PAN;
	}
}



/******************************************************************************
 FUNCTION:  DetectHold
 ******************************************************************************/
void SoInputGesture::DetectHold( SoGesture* pGesture )
{
	if ( pGesture->_iNumTouches == 1 && pGesture->_fTimer >= GESTURE_TIME_TO_HOLD )
	{
		JpVector2f vDelta1 = pGesture->_vPositionArray[ 0 ][ 0 ] - pGesture->_vPositionArray[ 0 ][ (pGesture->_iNumMovesSingle - 1)%GESTURE_NUM_FRAMES_TO_STORE ];

		if  ( vDelta1.Length() < SoUtils::GetiPadRatio() * 3.0f )
		{
			pGesture->_eGestureType = SoGesture::GESTURE_TYPE_HOLD;

			BroadcastGestureDown( _pGesture );
			//CCLOG( "HOLD DETECTED!!" );

		}
		//CCLOG( "Delta: %f", vDelta1.Length() );
	}
}

















