/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/

#pragma once

#include "stdafx.h"

#include <string>

#include "cocos2d.h"

#include "JpList.h"

#include "SoInputTouch.h"


USING_NS_CC;



#define SoGestureList JpList< SoGesture >

#define GESTURE_TIME_TO_MULTITOUCH  0.000f
#define GESTURE_TIME_TO_HOLD        0.750f
#define GESTURE_ZOOM_SPEED          0.005f

#define GESTURE_NUM_FINGERS          5
#define GESTURE_NUM_FRAMES_TO_STORE  60



/******************************************************************************
 *********************************************  CLASS:  SoGesture
 ******************************************************************************/
class SoGesture
{
	public:
		
		//-----------------------
		// public enum and classes
		enum GESTURE_TYPE
		{
			GESTURE_TYPE_PINCH,
			GESTURE_TYPE_ZOOM,
			GESTURE_TYPE_PAN,
			GESTURE_TYPE_ROTATE,
			GESTURE_TYPE_HOLD,
			GESTURE_TYPE_LAST
		};
	
		
		SoGesture();
		virtual ~SoGesture() {}
		
		GESTURE_TYPE	_eGestureType;

		JpVector2f _vPosition[ GESTURE_NUM_FINGERS ];
		JpVector2f _vPositionArray[ GESTURE_NUM_FINGERS ][ GESTURE_NUM_FRAMES_TO_STORE ];  // 5 Fingers, last 60 spots on screen (1 second)
		JpVector2f _vPositionDeltaOverall[ GESTURE_NUM_FINGERS ];
		JpVector2f _vPositionDeltaPrevious[ GESTURE_NUM_FINGERS ];
	
		float _fTimer;
		int _iNumTouches;
		int _iNumMovesSingle;
		int _iNumMovesMulti;
};




/****************************************************************************** 
 *********************************************  CLASS:  SoInputGesture
 ******************************************************************************/
class SoInputGesture
{
	public:
		
		//-----------------------
		// public enum and classes
		class Listener
		{
		public:
			Listener() {}
			virtual ~Listener() {}
			virtual bool OnGestureDown( SoGesture* pGesture ) { return false; }
			virtual bool OnGestureUp( SoGesture* pGesture )   { return false; }
			virtual bool OnGestureMove( SoGesture* pGesture ) { return false; }
		};
				
		
		//-----------------------
		// public functions
		
		SoInputGesture();
		virtual ~SoInputGesture() {};
		
		void Init();
		void Deinit();
		void Update( float dt );
		
		void AddListener( Listener* pListener )
		{
			ASSERTF( _bListenerListLocked == false, ("Can't add while list is locked.") )
			if ( _bListenerListLocked == false )
				_ListenerList.AddUniquely( pListener, 0 );
		}
		
		void RemoveListener( Listener* pListener )
		{
			ASSERTF( _bListenerListLocked == false, ("Can't remove while list is locked.") )
			if ( _bListenerListLocked == false )
				_ListenerList.Remove( pListener );
		}
		
		void RemoveAllListeners()
		{
			ASSERTF( _bListenerListLocked == false, ("Can't remove while list is locked.") )
			if ( _bListenerListLocked == false )
				_ListenerList.RemoveAll();
		}
		
		
		// SoInputTouch callbacks
		virtual bool OnTouchesDown( const SoCcTouchList* pSoCcTouchList );
		virtual bool OnTouchesUp( const SoCcTouchList* pSoCcTouchList );
		virtual bool OnTouchesMove( const SoCcTouchList* pSoCcTouchList );

	
		static SoInputGesture* GetInstance() { return _pInstance; }
		static void CreateInstance();
		static void DeleteInstance();
	
		void GrowOrCreateNewGesture( const SoCcTouchList* pSoCcTouchList );
	
		void DetectPinchOrZoom( SoGesture* pGesture );
		void DetectPan( SoGesture* pGesture );
		void DetectHold( SoGesture* pGesture );
	
		SoGesture* _pGesture;
	
	
	private:
		
		
		//-----------------------
		// private variables
		static SoInputGesture* _pInstance;
		
		
		JpList< Listener > _ListenerList;
		bool _bListenerListLocked;
		
		//-----------------------
		// private functions
		void BroadcastGestureDown( SoGesture* pGesture );
		void BroadcastGestureUp( SoGesture* pGesture );
		void BroadcastGestureMove( SoGesture* pGesture );
	
		SoInputTouchListener< SoInputGesture > _InputTouchListener;
};





template< class TYPE >
class SoInputGestureListener : public SoInputGesture::Listener
{
	public:
		//-----------------------
		// public functions
		
		SoInputGestureListener() : _pOwner( NULL ) {}
		~SoInputGestureListener() {}
		void SetOwner( TYPE* p ) { _pOwner = p; }
		
		virtual bool OnGestureDown( SoGesture* pGesture ) { return _pOwner->OnGestureDown( pGesture ); }
		virtual bool OnGestureUp( SoGesture* pGesture )   { return _pOwner->OnGestureUp( pGesture ); }
		virtual bool OnGestureMove( SoGesture* pGesture ) { return _pOwner->OnGestureMove( pGesture ); }
		
	private:
		
		//-----------------------
		// private variables.
		TYPE* _pOwner;
		
		
		//-----------------------
		// private functions.
		SoInputGestureListener( SoInputGestureListener& Copy )
		{
			ASSERTF( 0, ( "\nCannot copy instances of this class" ) );
		}
};


