/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#if DCG_IS_TARGET_WIN32

#include "JpDebug.h"
#include "SoInputKeyboard.h"
#include "SoInputOisManager.h"

#include "mmgrChooser.h"


USING_NS_CC;


SoInputKeyboard* SoInputKeyboard::_pInstance = NULL;



/******************************************************************************
 FUNCTION:  SoInputKeyboard
******************************************************************************/
SoInputKeyboard::SoInputKeyboard() :
_bListenerListLocked( false )
{
}



/******************************************************************************
 FUNCTION:  ~SoInputKeyboard
******************************************************************************/
SoInputKeyboard::~SoInputKeyboard()
{
	Deinit();
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoInputKeyboard::Init()
{
	SoInputOisManager* pOisInputManager = SoInputOisManager::GetInstance();

	ASSERTF( pOisInputManager, ( "\nNo OIS Input manager exists!  Create it before initializing this class." ) );

	if ( pOisInputManager )
		pOisInputManager->SetKeyboardListener( this );
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoInputKeyboard::Deinit()
{
	SoInputOisManager* pOisInputManager = SoInputOisManager::GetInstance();

	ASSERTF( pOisInputManager, ( "\nNo OIS Input manager exists!" ) );

	if ( pOisInputManager )
		pOisInputManager->SetKeyboardListener( NULL );
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoInputKeyboard::Update( float dt )
{
	_bListenerListLocked = true;

	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->Update( dt );
	}

	_bListenerListLocked = false;
}



/******************************************************************************
 FUNCTION:  AddListener
******************************************************************************/
void SoInputKeyboard::AddListener( Listener* pListener )
{
	ASSERTF( _bListenerListLocked == false, ("Can't add while list is locked.") )
	if ( _bListenerListLocked == false )
		_ListenerList.AddUniquely( pListener, 0 );
}



/******************************************************************************
 FUNCTION:  RemoveListener
******************************************************************************/
void SoInputKeyboard::RemoveListener( Listener* pListener )
{
	ASSERTF( _bListenerListLocked == false, ("Can't remove while list is locked.") )
	if ( _bListenerListLocked == false )
	{
		pListener->DestroyKeyRepeastList();
		_ListenerList.Remove( pListener );
	}
}



/******************************************************************************
 FUNCTION:  RemoveAllListeners
******************************************************************************/
void SoInputKeyboard::RemoveAllListeners()
{
	ASSERTF( _bListenerListLocked == false, ("Can't remove while list is locked.") )
	if ( _bListenerListLocked == false )
	{
		JpList< Listener >::Iterator Iter( &_ListenerList );

		while ( Iter.Next() )
		{
			Listener* pListener = Iter.GetNodeData();
			pListener->DestroyKeyRepeastList();
		}

		_ListenerList.RemoveAll();
	}
}



/******************************************************************************
 FUNCTION:  keyPressed
******************************************************************************/
bool SoInputKeyboard::keyPressed( const OIS::KeyEvent &arg )
{
	//-----------------------
	// The event handler we are calling can do anything it likes...even attempt to
	// add or remove to the listener list.  That would be bad since we are in 
	// the middle of walking through the list.  So we lock the list to prevent tampering.

	_bListenerListLocked = true;

	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnKeyPressed( arg.key );


		if ( pListener->UseRepeat( arg.key ) )
		{
			SoKeyRepeat* pRepeat = new SoKeyRepeat;

			pRepeat->SetRepeatTimer( pListener->GetKeyRepeatCooldown() );
			pRepeat->SetKeyCode( arg.key );

			pListener->GetRepeatList()->AddLast( pRepeat );
		}
	}

	_bListenerListLocked = false;


	return true;
}



/******************************************************************************
 FUNCTION:  keyReleased
******************************************************************************/
bool SoInputKeyboard::keyReleased( const OIS::KeyEvent &arg )
{
	_bListenerListLocked = true;

	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnKeyReleased( arg.key );

		SoKeyRepeatList::Iterator Iter2( pListener->GetRepeatList() );

		while ( Iter2.Next() )
		{
			SoKeyRepeat* pInput = Iter2.GetNodeData();

			if ( arg.key == pInput->GetKeyCode() )
			{
				delete pInput;

				pListener->GetRepeatList()->Remove( pInput );
			}
		}
	}

	_bListenerListLocked = false;

	return true;
}



/******************************************************************************
 FUNCTION:  CreateInstance
******************************************************************************/
void SoInputKeyboard::CreateInstance()
{
	ASSERTF( _pInstance == NULL, ( "\nCan't create more than one instance.  Not allowed." ) );

	if ( !_pInstance )
		_pInstance = new SoInputKeyboard;
}



/******************************************************************************
 FUNCTION:  DeleteInstance
******************************************************************************/
void SoInputKeyboard::DeleteInstance()
{
	if ( _pInstance )
	{
		delete _pInstance;
		_pInstance = NULL;
	}
}



/******************************************************************************
FUNCTION:  Listener
******************************************************************************/
SoInputKeyboard::Listener::~Listener()
{
	DestroyKeyRepeastList();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoInputKeyboard::Listener::Update( float dt )
{
	UpdateRepeats( dt );
	BroadcastRepeats();
}



/******************************************************************************
 FUNCTION:  UpdateRepeats
******************************************************************************/
void SoInputKeyboard::Listener::UpdateRepeats( float dt )
{
	SoKeyRepeatList::Iterator Iter( &_RepeatList );

	while ( Iter.Next() )
	{
		SoKeyRepeat* pInput = Iter.GetNodeData();

		float fNewRepeatCountdown = pInput->GetRepeatTimer() - dt;
		pInput->SetRepeatTimer( fNewRepeatCountdown );
	}
}


	
/******************************************************************************
FUNCTION:  BroadcastRepeats
******************************************************************************/
void SoInputKeyboard::Listener::BroadcastRepeats()
{
	JpList< SoKeyRepeat >::Node* pToBroadcast = _RepeatList.GetFirst();
	SoKeyRepeat* pRepeat = pToBroadcast->GetData();

	if ( pRepeat )
	{
		if ( pRepeat->GetRepeatTimer() <= 0.0f )
		{
			OnKeyPressed( pRepeat->GetKeyCode() );

			pRepeat->SetRepeatTimer( pRepeat->GetRepeatTimer() + GetKeyRepeatCooldown() );
		}
	}
}



/******************************************************************************
FUNCTION:  DestroyKeyRepeastList
******************************************************************************/
void SoInputKeyboard::Listener::DestroyKeyRepeastList()
{
	SoKeyRepeatList::Iterator Iter( GetRepeatList() );

	while ( Iter.Next() )
	{
		SoKeyRepeat* pRepeat = Iter.GetNodeData();
		delete pRepeat;
	}

	GetRepeatList()->RemoveAll();
}



/******************************************************************************
 FUNCTION:  UseRepeat
******************************************************************************/
bool SoInputKeyboard::Listener::UseRepeat( OIS::KeyCode Key )
{
	JpList< int >::Iterator Iter( &_KeysToRepeatList );

	while ( Iter.Next() )
	{
		int iCode = (int)Iter.GetNodeData();
		if ( Key == (OIS::KeyCode)iCode )
			return true;
	}

	return false;
}
	
	


#endif // DCG_IS_TARGET_WIN32



























