/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "NeCsvIn.h"

#include "SoDrawDebug.h"
#include "SoInputTouch.h"
#include "SoSprite.h"
#include "SoSoundManager.h"
#include "SoUtils.h"

#include "mmgrChooser.h"



SoInputTouch* SoInputTouch::_pInstance = NULL;



/******************************************************************************
 FUNCTION:  SoInputTouch
******************************************************************************/
SoInputTouch::SoInputTouch() :
_bListenerListLocked( false ),
_pParentNode( NULL )
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoInputTouch::Init( CCNode* pParentNode )
{
	_pParentNode = pParentNode;

	if ( pParentNode )
		pParentNode->addChild( this );

	setTouchEnabled( true );
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoInputTouch::Deinit()
{
	if ( _pParentNode )
	{
		_pParentNode->removeChild( this, DCG_CLEANUP_ALL_RUNNING_ACTIONS );
		_pParentNode = NULL;
	}

	RemoveAllListeners();


	// Note: we don't delete the members, because they are just copies
	// of CCTouch*, which are sent from and managed by cocos.
	// mdomowicz 2013_03_30
	_TouchList.RemoveAll();
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoInputTouch::Update( float dt )
{
	if ( !_pParentNode )
	{
		_pParentNode = CCDirector::sharedDirector()->getRunningScene();
		_pParentNode->addChild( this );
	}
}




/******************************************************************************
 FUNCTION:  BroadcastTouchesDown
******************************************************************************/
void SoInputTouch::BroadcastTouchesDown( const SoCcTouchList* pSoCcTouchList )
{
	//-----------------------
	// The event handler we are calling can do anything it likes...even attempt to
	// add or remove to the listener list.  That would be bad since we are in 
	// the middle of walking through the list.  So we lock the list to prevent tampering.

	_bListenerListLocked = true;

	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnTouchesDown( pSoCcTouchList );
	}

	_bListenerListLocked = false;
}



/******************************************************************************
 FUNCTION:  BroadcastTouchesUp
******************************************************************************/
void SoInputTouch::BroadcastTouchesUp( const SoCcTouchList* pSoCcTouchList )
{
	_bListenerListLocked = true;

	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnTouchesUp( pSoCcTouchList );
	}

	_bListenerListLocked = false;
}



/******************************************************************************
 FUNCTION:  BroadcastTouchesMove
******************************************************************************/
void SoInputTouch::BroadcastTouchesMove( const SoCcTouchList* pSoCcTouchList )
{
	_bListenerListLocked = true;

	JpList< Listener >::Iterator Iter( &_ListenerList );
	while ( Iter.Next() )
	{
		Listener* pListener = Iter.GetNodeData();
		pListener->OnTouchesMove( pSoCcTouchList );
	}

	_bListenerListLocked = false;
}



/******************************************************************************
 FUNCTION:  ccTouchesBegan
******************************************************************************/
void SoInputTouch::ccTouchesBegan(CCSet *pTouches, CCEvent *pEvent)
{
	CopyCcToucheSetToSoTouchList( &_TouchList, pTouches );
	BroadcastTouchesDown( &_TouchList );
}



/******************************************************************************
 FUNCTION:  ccTouchesMoved
******************************************************************************/
void SoInputTouch::ccTouchesMoved(CCSet *pTouches, CCEvent *pEvent)
{
	CopyCcToucheSetToSoTouchList( &_TouchList, pTouches );
	BroadcastTouchesMove( &_TouchList );
}



/******************************************************************************
 FUNCTION:  ccTouchesEnded
******************************************************************************/
void SoInputTouch::ccTouchesEnded(CCSet *pTouches, CCEvent *pEvent)
{
	CopyCcToucheSetToSoTouchList( &_TouchList, pTouches );
	BroadcastTouchesUp( &_TouchList );
}



/******************************************************************************
 FUNCTION:  ccTouchesCancelled
******************************************************************************/
void SoInputTouch::ccTouchesCancelled(CCSet *pTouches, CCEvent *pEvent)
{
	ccTouchesEnded(pTouches, pEvent);
}



/******************************************************************************
 FUNCTION:  CopyCcToucheSetToSoTouchList
******************************************************************************/
void SoInputTouch::CopyCcToucheSetToSoTouchList
( 
	SoCcTouchList* pSoCcTouchList, 
	cocos2d::CCSet* pTouches 
)
{
	pSoCcTouchList->RemoveAll();

	CCSetIterator iter = pTouches->begin();
	for ( ; iter != pTouches->end(); iter ++ )
	{
		CCTouch* pTouch = (CCTouch*)(*iter);
		pSoCcTouchList->AddLast( pTouch );
	}
}



/******************************************************************************
 FUNCTION:  TouchPosToScreenPos
 NOTE:      Not sure why, but all CCNode objects treat 0,0 as the bottom
            left of the object, while touch positions are given relative to
            a 0,0 that's at the top left of the screen/window.  This function
            does the conversion.
            mdomowicz 2013_03_30
******************************************************************************/
JpVector2f SoInputTouch::TouchPosToScreenPos( const CCPoint& v )
{
	CCSize SizeInPixels = CCDirector::sharedDirector()->getWinSize();
	return JpVector2f( v.x, SizeInPixels.height - v.y );
}



/******************************************************************************
 FUNCTION:  CreateInstance
******************************************************************************/
void SoInputTouch::CreateInstance()
{
	ASSERTF( _pInstance == NULL, ( "\nCan't create more than one instance.  Not allowed." ) );

	if ( !_pInstance )
		_pInstance = new SoInputTouch;
}



/******************************************************************************
 FUNCTION:  DeleteInstance
******************************************************************************/
void SoInputTouch::DeleteInstance()
{
	if ( _pInstance )
	{
		_pInstance->Deinit();
		delete _pInstance;
		_pInstance = NULL;
	}
}






