/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "JpDebug.h"
#include "JpHashTable.h"
#include "SoActor.h"
#include "SoActorManager.h"
#include "SoFsmStateMain.h"
#include "SoHudLog.h"
#include "SoLuaState.h"
#include "SoSaveSystem.h"

#include "mmgrChooser.h"


USING_NS_CC;


SoLuaState* SoLuaState::_pInstance = NULL;

SoSaveSystem* SoLuaState::_pSaveSystem = NULL;

SoLuaStateTable SoLuaState::_LuaStateTable;


/******************************************************************************
 FUNCTION:  SoLuaState
******************************************************************************/
SoLuaState::SoLuaState() :
_pL( NULL )
{
}



/******************************************************************************
 FUNCTION:  ~SoLuaState
******************************************************************************/
SoLuaState::~SoLuaState()
{
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoLuaState::Init()
{
	ASSERTF( _pL == NULL, ( "\nTrying to init a non-null lua state pointer." ) );

	if ( _pL == NULL )
	{
		// lua stack sizes are dynamic
		_pL = lua_open();
		lua_checkstack( _pL, 108 );

		// open libraries
		luaopen_base( _pL );
		luaopen_table( _pL );
		luaopen_string( _pL );

		// clean up garbage left on the stack after loading libraries
		// see (http://lua-users.org/lists/lua-l/2003-10/msg00027.html)
		lua_settop(_pL, 0);


		// this is so that the static Api functions can refer
		// to this SoLuaState from the lua_State that they get
		// passed to them.
		SetLuaStateToSoLuaState( _pL, this );

		RegisterApiFunctions();
		RegisterGlobals();
	}
}



/******************************************************************************
 FUNCTION:  Deinit
******************************************************************************/
void SoLuaState::Deinit()
{
	//-----------------------
	// close the lua state.

	if ( _pL != NULL )
	{
		UnsetLuaStateToSoLuaState( _pL );

		lua_close( _pL );
		_pL = NULL;
	}
}



/******************************************************************************
 FUNCTION:  Update
******************************************************************************/
void SoLuaState::Update( float dt )
{
	if ( _pL != NULL )
	{
		//-----------------------
		// monitor stack integrity.
		const int iNumStackElementsOnExit = 0;
		SoLuaState::AutoStackCheck Check( _pL, iNumStackElementsOnExit );


		//-----------------------
		// call the update function.
		lua_getglobal( _pL, "ScriptUpdate" );

		if ( lua_isfunction( _pL, -1 ) )
		{
			lua_pushnumber( _pL, dt );
			const int iNumArgs = 1;
			SoPcall( _pL, iNumArgs, 0, "SoLuaState::Update" );
		}
		else
		{
			// the result of getglobal was not a function,
			// so remove whatever it was off the stack.
			lua_pop( _pL, 1 );
		}
	}
}



/******************************************************************************
 FUNCTION:  OpenScript
 PURPOSE:   Open a script an run the entry point function (defined by 
            convention)
******************************************************************************/
bool SoLuaState::OpenScript( const char acFileName[] )
{
	#if DCG_IS_TARGET_WIN32
		int iResult = luaL_dofile( _pL, acFileName );
	#else
		const char* pacFullPath = CCFileUtils::sharedFileUtils()->fullPathFromRelativePath( SoUtils::AddRelativePathToIni( acFileName ) );	
		int iResult = luaL_dofile( _pL, pacFullPath );
	#endif
	
	bool bSuccess = ( iResult == 0 );

	InvokeLuaFunction( "ScriptInit" );

	return bSuccess;
}



/******************************************************************************
 FUNCTION:  RegisterApiFunctions
 PURPOSE:   makes C functions callable from within lua scripts
******************************************************************************/
void SoLuaState::RegisterApiFunctions()
{
	lua_register( _pL, "Debug_Print", Debug_Print );
	lua_register( _pL, "Debug_Break", Debug_Break );
	lua_register( _pL, "Debug_Assert", Debug_Assert );
	lua_register( _pL, "Debug_GetIniSetting", Debug_GetIniSetting );
	lua_register( _pL, "Util_GetDistanceSq", Util_GetDistanceSq );
	lua_register( _pL, "Util_GetRandomFloatRanged", Util_GetRandomFloatRanged );
	lua_register( _pL, "Util_GetRandomInt", Util_GetRandomInt );
	lua_register( _pL, "Util_SetSpecialRandomSeed", Util_SetSpecialRandomSeed );
	lua_register( _pL, "Util_GetSpecialRandomFloatRanged", Util_GetSpecialRandomFloatRanged );
	lua_register( _pL, "Util_GetSpecialRandomInt", Util_GetSpecialRandomInt );
	lua_register( _pL, "Util_LoadScript", Util_LoadScript );
	lua_register( _pL, "Hud_PrintToLog", Hud_PrintToLog );
}




/******************************************************************************
 FUNCTION:  RegisterGlobals
 PURPOSE:   setup app specific globals that must be shared between lua and C.
******************************************************************************/
void SoLuaState::RegisterGlobals()
{
	/*
	for ( int i = 0; i < (int)SoMessage::TYPE_LAST; i ++ )
	{
		lua_pushnumber( _pL, i );
		lua_setglobal( _pL, SoMessage::GetStringFromType( (SoMessage::TYPE)i ) );
	};
	*/
}



/******************************************************************************
 FUNCTION:  InvokeLuaFunction
 PURPOSE:   For calling a function written in Lua from C
******************************************************************************/
void SoLuaState::InvokeLuaFunction( const char * szFunctionName )
{
	if ( szFunctionName != NULL && ::strlen(szFunctionName) > 0)
	{
		if ( _pL != NULL )
		{
			//-----------------------
			// monitor stack integrity.
			const int iNumStackElementsOnExit = 0;
			SoLuaState::AutoStackCheck Check( _pL, iNumStackElementsOnExit );


			//-----------------------
			// call the function.
			lua_getglobal( _pL, szFunctionName );
			if ( lua_isfunction( _pL, lua_gettop( _pL ) ) )
			{
				SoLuaState::SoPcall( _pL, 0, 0, "SoLuaState::InvokeLuaFunction" );
			}
			else
			{
				// the result of getglobal was not a function,
				// so remove whatever it was off the stack.
				lua_pop( _pL, 1 );
			}
		}
	}
}



/******************************************************************************
 FUNCTION:  SaveToFile
 NOTES:     This works by calling a function with a hard coded, pre-determined
            name.  This pre-determined function will be called on every 
            lua_state in the _LuaStateTable.  Each lua_state will then make 
            API calls as needed to interact with a shared common data base 
            file.
******************************************************************************/
void SoLuaState::SaveToFile( SoSaveSystem* pSaveSystem )
{
	// We have to hold on to the save system, because a lot of API functions 
	// that the lua scripts will call from here will need it.
	_pSaveSystem = pSaveSystem;

	SoLuaStateTable::Iterator Iter( &_LuaStateTable );
	while ( Iter.Next() )
	{
		SoLuaState* pL = Iter.GetData();
		pL->InvokeLuaFunction( "ScriptSaveGame" );
	}
}




/******************************************************************************
 FUNCTION:  LoadFromFile
 NOTES:     This works just like SaveToFile (see comments there)
******************************************************************************/
void SoLuaState::LoadFromFile( SoSaveSystem* pSaveSystem )
{
	// We have to hold on to the save system, because a lot of API functions 
	// that the lua scripts will call from here will need it.
	_pSaveSystem = pSaveSystem;

	SoLuaStateTable::Iterator Iter( &_LuaStateTable );
	while ( Iter.Next() )
	{
		SoLuaState* pL = Iter.GetData();
		pL->InvokeLuaFunction( "ScriptLoadGame" );
	}
}



/******************************************************************************
 FUNCTION:  SoPcall
 PURPOSE:   Does the actual work for InvokeLuaFunction - calls the lua
            function.
******************************************************************************/
void SoLuaState::SoPcall
( 
	lua_State* pL, 
	int nargs, 
	int nresults, 
	const char* sDesc 
)
{
	int err = lua_pcall( pL, nargs, nresults, 0 );
	if ( err != 0 )
	{
		PrintLuaCallstack( pL );
		const char* pacMessage = lua_tostring( pL, -1 );
		ASSERTF( 0, ( "\nLua error %d (%s)\n  %s\n", err, sDesc, pacMessage ) );
		lua_pop( pL, 1 );
	}
}



/******************************************************************************
 FUNCTION:  GetLuaNumber
 PURPOSE:   wrapper for getting a number from the lua stack
******************************************************************************/
float SoLuaState::GetLuaNumber
(
	lua_State* pL,
	int iStackIndex
)
{
	float fNumber = 0.0f;

	if ( lua_isnumber( pL, iStackIndex ) )
		fNumber = (float)lua_tonumber( pL, iStackIndex );
	else
		luaL_error( pL, "\nNumber not found.  Bad argument to Lua API." );

	return fNumber;
}



/******************************************************************************
 FUNCTION:  GetLuaBool
 PURPOSE:   wrapper for getting a bool from the lua stack
******************************************************************************/
bool SoLuaState::GetLuaBool( lua_State* pL, int iStackIndex )
{
	return ( lua_toboolean( pL, iStackIndex ) == 1 );
}



/******************************************************************************
 FUNCTION:  GetLuaString
 PURPOSE:   Wrapper for getting a string from the lua stack
******************************************************************************/
const char* SoLuaState::GetLuaString
(
	lua_State* pL,
	int iStackIndex
)
{
	const char* pcString = NULL;

	if ( lua_isstring( pL, iStackIndex ) )
		pcString = lua_tostring( pL, iStackIndex );
	else
		luaL_error( pL, "\nString not found.  Bad argument to Lua API." );

	return pcString;
}



/******************************************************************************
 FUNCTION:  GetLuaVector2
 PURPOSE:   Wrapper for getting a 2d vector from the lua stack
******************************************************************************/
JpVector2f SoLuaState::GetLuaVector2( lua_State* pL, int iStackIndex )
{
	float fX = GetLuaNumber( pL, iStackIndex );
	float fY = GetLuaNumber( pL, iStackIndex + 1 );
	JpVector2f vResult( fX, fY );
	return vResult;
}



/******************************************************************************
 FUNCTION:  GetLuaActor
 PURPOSE:   Wrapper for getting an actor guid from the lua stack, and
            converting it to an actor pointer
******************************************************************************/
SoActor* SoLuaState::GetLuaActor( lua_State* pL, int iStackIndex )
{
	if ( lua_islightuserdata ( pL, iStackIndex ) )
	{
		guid g = (guid)lua_touserdata( pL, iStackIndex );
		SoActor* pActor = SoActorManager::GetInstance()->GetActorByGuid( g );
		return pActor;
	}
	
	return NULL;
}



/******************************************************************************
 FUNCTION:  PushLuaNumber
 PURPOSE:   Wrapper for pushing a Number onto the lua stack
******************************************************************************/
void SoLuaState::PushLuaNumber
(
	lua_State* pL,
	float fNumber
)
{
	lua_pushnumber( pL, fNumber );
}



/******************************************************************************
 FUNCTION:  PushLuaBool
 PURPOSE:   wrapper for pushing a bool onto the lua stack
******************************************************************************/
void SoLuaState::PushLuaBool( lua_State* pL, bool bFlag )
{
	lua_pushboolean( pL, ( int )bFlag );
}



/******************************************************************************
 FUNCTION:  PushLuaString
 PURPOSE:   Wrapper for pushing a string onto the lua stack
******************************************************************************/
void SoLuaState::PushLuaString
(
	lua_State* pL,
	const char* pacString
)
{
	lua_pushstring( pL, pacString );
}



/******************************************************************************
 FUNCTION:  PushLuaActor
 PURPOSE:   Get the guid of the given actor and push it onto the lua stack
******************************************************************************/
void SoLuaState::PushLuaActor( lua_State* pL, SoActor* pActor )
{
	if ( pActor )
	{
		guid g = pActor->GetGuid();
		lua_pushlightuserdata ( pL, (void *)g );
	}
}



/******************************************************************************
 FUNCTION:  GetAbsoluteStackIndex
 PURPOSE:   Returns an absolute stack index when given either an absolute
            or relative stack index.
 NOTES:     The lua<->C stack can be indexed using both absolute and relative
            stack indices.
            Absoluate stack indices are positive starting at 1, where 1 is
            the first item pushed on the stack, 2 is the second, etc.
            Relative indices are counted back from the top of the stack and are
            specified using negative index values.  So -1 is the very top
            of the stack, -2 is the next item down, etc.
******************************************************************************/
int SoLuaState::GetAbsoluteStackIndex
( 
	lua_State* pL, 
	int iIndex 
)
{
	if ( iIndex < 0 )
		iIndex = ( lua_gettop( pL ) + iIndex ) + 1;

	return iIndex;
}



/******************************************************************************
 FUNCTION:  GetLuaArrayString
 PURPOSE:   Wrapper for getting a string from a lua table that is indexed
            specifically as an array (ie, consecutive numeric indices 
            starting at 1), 
 INPUT:     iTableStackIndex - the index of the lua table on the stack
            iTableKey - the index into the table
******************************************************************************/
const char* SoLuaState::GetLuaArrayString
( 
	lua_State* pL, 
	int iTableStackIndex, 
	int iTableIndex,
	const char* pacReturnOnFail 
)
{
	iTableStackIndex = GetAbsoluteStackIndex( pL, iTableStackIndex );

	lua_pushnumber( pL, (float)iTableIndex );
	lua_gettable( pL, iTableStackIndex );

	if ( lua_isstring( pL, -1 ) )
	{
		const char* pacResult = GetLuaString( pL, -1 );

		lua_pop( pL, 1 );
		return pacResult;
	}

	lua_pop( pL, 1 );
	return pacReturnOnFail;
}



/******************************************************************************
 FUNCTION:  GetLuaTableNumber
 PURPOSE:   Wrapper for getting a value from a lua table, given a key.
 INPUT:     iTableStackIndex - the index of the lua table on the stack
            pacKey - the identifier name of the key
******************************************************************************/
float SoLuaState::GetLuaTableNumber
(
	lua_State* pL,
	int iTableStackIndex,
	const char* pacKey,
	bool* pbSuccess
)
{
	iTableStackIndex = GetAbsoluteStackIndex( pL, iTableStackIndex );

	lua_pushstring( pL, pacKey );
	lua_gettable( pL, iTableStackIndex );

	*pbSuccess = false;
	if ( lua_isnumber( pL, -1 ) )
	{
		*pbSuccess = true;
		float fResult = GetLuaNumber( pL, -1 );

		lua_pop( pL, 1 );
		return fResult;
	}

	lua_pop( pL, 1 );
	return 0.0f;
}



/******************************************************************************
 FUNCTION:  GetLuaTableIntIntoArg
 PURPOSE:   does the same thing as GetLuaTableNumberIntoArg, but for integers
 INPUT:     piResult - this gets the result.  Unchanged if no result if found.
            iTableStackIndex - the index of the lua table on the stack
            pacKey - the identifier name of the key
******************************************************************************/
bool SoLuaState::GetLuaTableIntIntoArg
(
	lua_State* pL,
	int* piResult,
	int iTableStackIndex,
	const char* pacKey
)
{
	bool bSuccess;
	int iNumber = (int)GetLuaTableNumber( pL, iTableStackIndex, pacKey, &bSuccess );
	if ( bSuccess )
		*piResult = iNumber;

	return bSuccess;
}



/******************************************************************************
 FUNCTION:  GetLuaTableNumberIntoArg
 PURPOSE:   An alternative wrapper for getting a float from a lua table.
            This version gets the result back as a argument.  It leaves the
            argument unchanged if the table value cannot be found.  This is
            useful for when the user wants to use their own default value
            when the table value is not found.
 INPUT:     pfResult - this gets the result.  Unchanged if no result if found.
            iTableStackIndex - the index of the lua table on the stack
            pacKey - the identifier name of the key
******************************************************************************/
bool SoLuaState::GetLuaTableNumberIntoArg
(
	lua_State* pL,
	float* pfResult,
	int iTableStackIndex,
	const char* pacKey
)
{
	iTableStackIndex = GetAbsoluteStackIndex( pL, iTableStackIndex );


	lua_pushstring( pL, pacKey );
	lua_gettable( pL, iTableStackIndex );

	if ( lua_isnumber( pL, -1 ) )
	{
		*pfResult = GetLuaNumber( pL, -1 );
		lua_pop( pL, 1 );
		return true;
	}

	lua_pop( pL, 1 );
	return false;
}



/******************************************************************************
 FUNCTION:  GetLuaTableString
 PURPOSE:   Wrapper for getting a value from a lua table, given a key.
 INPUT:     iTableStackIndex - the index of the lua table on the stack
            pacKey - the identifier name of the key
******************************************************************************/
const char* SoLuaState::GetLuaTableString
(
	lua_State* pL,
	int iTableStackIndex,
	const char* pacKey,
	bool* pbSuccess
)
{
	iTableStackIndex = GetAbsoluteStackIndex( pL, iTableStackIndex );

	lua_pushstring( pL, pacKey );
	lua_gettable( pL, iTableStackIndex );

	*pbSuccess = false;
	if ( lua_isstring( pL, -1 ) )
	{
		*pbSuccess = true;
		const char* pacResult = GetLuaString( pL, -1 );

		lua_pop( pL, 1 );
		return pacResult;
	}

	lua_pop( pL, 1 );
	return NULL;
}



/******************************************************************************
 FUNCTION:  GetLuaTableStringIntoArg
 PURPOSE:   An alternative wrapper for getting a string from a lua table.
            This version gets the result back as a argument.  It leaves the
            argument unchanged if the table value cannot be found.  This is
            useful for when the user wants to use their own default value
            when the table value is not found.
 INPUT:     pacResult - this gets the result.  Unchanged if no result if found.
            iTableStackIndex - the index of the lua table on the stack
            pacKey - the identifier name of the key
******************************************************************************/
bool SoLuaState::GetLuaTableStringIntoArg
(
	lua_State* pL,
	const char** ppacResult,
	int iTableStackIndex,
	const char* pacKey
)
{
	iTableStackIndex = GetAbsoluteStackIndex( pL, iTableStackIndex );

	lua_pushstring( pL, pacKey );
	lua_gettable( pL, iTableStackIndex );

	if ( lua_isstring( pL, -1 ) )
	{
		*ppacResult = GetLuaString( pL, -1 );
		lua_pop( pL, 1 );
		return true;
	}

	lua_pop( pL, 1 );
	return false;
}



/******************************************************************************
 FUNCTION:  GetLuaTableActorIntoArg
 PURPOSE:   An alternative wrapper for getting an actor from a lua table.
            This version gets the result back as a argument.  It leaves the
            argument unchanged if the table value cannot be found.  This is
            useful for when the user wants to use their own default value
            when the table value is not found.
 INPUT:     ppActorResult - this gets the result.  Unchanged if no result is 
            found.
            iTableStackIndex - the index of the lua table on the stack
            pacKey - the identifier name of the key
******************************************************************************/
bool SoLuaState::GetLuaTableActorIntoArg
(
	lua_State* pL,
	SoActor** ppActorResult,
	int iTableStackIndex,
	const char* pacKey
)
{
	iTableStackIndex = GetAbsoluteStackIndex( pL, iTableStackIndex );

	lua_pushstring( pL, pacKey );
	lua_gettable( pL, iTableStackIndex );

	if ( lua_islightuserdata( pL, -1 ) )
	{
		*ppActorResult = GetLuaActor( pL, -1 );
		lua_pop( pL, 1 );
		return true;
	}

	lua_pop( pL, 1 );
	return false;
}



/******************************************************************************
 FUNCTION:  GetLuaTableBool
 PURPOSE:   Wrapper for getting a value from a lua table, given a key.
 INPUT:     iTableStackIndex - the index of the lua table on the stack
            pacKey - the identifier name of the key
******************************************************************************/
bool SoLuaState::GetLuaTableBool
(
	lua_State* pL,
	int iTableStackIndex,
	const char* pacKey,
	bool* pbSuccess
)
{
	iTableStackIndex = GetAbsoluteStackIndex( pL, iTableStackIndex );

	lua_pushstring( pL, pacKey );
	lua_gettable( pL, iTableStackIndex );

	*pbSuccess = false;
	if ( lua_isboolean( pL, -1 ) )
	{
		*pbSuccess = true;
		bool bResult = GetLuaBool( pL, -1 );

		lua_pop( pL, 1 );
		return bResult;
	}

	lua_pop( pL, 1 );
	return false;
}



/******************************************************************************
 FUNCTION:  GetLuaTableBoolIntoArg
 PURPOSE:   An alternative wrapper for getting a bool from a lua table.
            This version gets the result back as a argument.  It leaves the
            argument unchanged if the table value cannot be found.  This is
            useful for when the user wants to use their own default value
            when the table value is not found.
 INPUT:     pbResult - this gets the result.  Unchanged if no result if found.
            iTableStackIndex - the index of the lua table on the stack
            pacKey - the identifier name of the key
******************************************************************************/
bool SoLuaState::GetLuaTableBoolIntoArg
(
	lua_State* pL,
	bool* pbResult,
	int iTableStackIndex,
	const char* pacKey
)
{
	iTableStackIndex = GetAbsoluteStackIndex( pL, iTableStackIndex );

	lua_pushstring( pL, pacKey );
	lua_gettable( pL, iTableStackIndex );

	if ( lua_isboolean( pL, -1 ) )
	{
		*pbResult = GetLuaBool( pL, -1 );

		lua_pop( pL, 1 );
		return true;
	}

	lua_pop( pL, 1 );
	return false;
}



/******************************************************************************
 FUNCTION:  PushLuaTable
******************************************************************************/
void SoLuaState::PushLuaTable( lua_State* pL )
{
	lua_newtable( pL );
}



/******************************************************************************
 FUNCTION:  PushLuaTableNumber
******************************************************************************/
void SoLuaState::PushLuaTableNumber
( 
	lua_State* pL, 
	const char* pacKey, 
	float fValue 
)
{
	bool bTopOfStackIsTable = lua_istable( pL, -1 );

	ASSERTF( bTopOfStackIsTable, ( "\nPushLuaTable wrappers require target table at the top of the stack." ) );

	if ( bTopOfStackIsTable )
	{
		int iTableStackIndex = GetAbsoluteStackIndex( pL, -1 );
		lua_pushstring( pL, pacKey );          // push the key
		lua_pushnumber( pL, fValue );          // push the value

		// Tell lua to put the key and value into the table,
		// lua_settable pops both the key and the value off the stack when done
		lua_settable ( pL, iTableStackIndex ); 
	}
}



/******************************************************************************
 FUNCTION:  PrintLuaCallstack
 PURPOSE:   Prints the current lua callstack
******************************************************************************/
void SoLuaState::PrintLuaCallstack( lua_State* pL )
{
	lua_Debug DebugData;

	int iLevel = 0;
	int iStatus = lua_getstack(	pL,	iLevel,	&DebugData );

	CCLOG( "\n-----LUA CALLSTACK START-----" );

	while ( iStatus == 1 )
	{
		lua_getinfo( pL, "lnuS", &DebugData );
		CCLOG( "\n-->lvl:%i  line:%i fn:%s", iLevel, DebugData.currentline, DebugData.name );
		iLevel ++;
		iStatus = lua_getstack( pL, iLevel, &DebugData );
	}

	CCLOG(  "\n-------------END-------------\n" );
}



/******************************************************************************
 FUNCTION:  CreateInstance
******************************************************************************/
void SoLuaState::CreateInstance()
{
	ASSERTF( _pInstance == NULL, ( "\nCan't create more than one instance.  Not allowed." ) );

	if ( !_pInstance )
	{
		_pInstance = new SoLuaState;
		_pInstance->Init();
	}
}



/******************************************************************************
 FUNCTION:  DeleteInstance
******************************************************************************/
void SoLuaState::DeleteInstance()
{
	if ( _pInstance )
	{
		_pInstance->Deinit();
		delete _pInstance;
		_pInstance = NULL;
	}
}








/******************************************************************************
/******************************************************************************
/******************************************************************************

                                 LUA API 

/******************************************************************************
/******************************************************************************
******************************************************************************/



/******************************************************************************
 FUNCTION:  Debug_Break
 PURPOSE:   A poor mans lua break point.  To use: call this from lua where you
            want a break point, then put an actual break point in this C code.
 INPUT:     ARG 1: String: optional text.
 OUPUT:     none
 NOTES:
******************************************************************************/
int SoLuaState::Debug_Break( lua_State* pL )
{
	PrintLuaCallstack( pL );

	if ( lua_isstring( pL, 1 ) )
	{
		const char* pacText = NULL;
		pacText = lua_tostring( pL, 1 );
		CCLOG( pacText );	
	}

	return 0;
}




/******************************************************************************
 FUNCTION:  Debug_Assert
 PURPOSE:   
 INPUT:     ARG 1: The result of a condition check (number/boolean/nil).
            ARG 2: String: Optional error message.
 OUPUT:     none
 NOTES:
******************************************************************************/
int SoLuaState::Debug_Assert( lua_State* pL )
{
	bool bBool = true;
	if ( lua_isboolean( pL, 1 ) )
		bBool = GetLuaBool( pL, 1 );

	bool bIsZero = false;
	if ( lua_isnumber( pL, 1 ) )
		bIsZero = ( GetLuaNumber( pL, 1 ) == 0 );

	bool bIsNil = lua_isnil( pL, 1 );

	if ( bBool == false || bIsZero || bIsNil )
	{
		const char* pacText = NULL;
		if ( lua_isstring( pL, 2 ) )
			pacText = GetLuaString( pL, 2 );

		CCLOG( "\nLUA ASSERTION FAILED: " );
		if ( pacText )
			CCLOG( ( "%s", pacText ) );

		PrintLuaCallstack( pL );
		ASSERTF( 0, ( "\nStopping at lua assert" ) );
	}

	return 0;	
}




/******************************************************************************
 FUNCTION:  Debug_Print
 PURPOSE:   Lua API function: prints a string to the debug output window.
 ARGS:      ARG 1: String: the string to be printed.
 RETURN:    None
 NOTES:     
******************************************************************************/
int SoLuaState::Debug_Print( lua_State* pL )
{
	if ( lua_isstring( pL, 1 ) )
	{
		const char* pcString = NULL;
		pcString = lua_tostring( pL, 1 );

		CCLOG( pcString );
	}

	return 0;
}




/******************************************************************************
 FUNCTION:  Debug_GetIniSetting
 PURPOSE:   Lua API function: A way to get first.ini and local.ini settings
 ARGS:      ARG 1: String: the name of the desired ini file setting
 RETURN:    depends! - the number and type of returns varries depending on
            what ini option is being requested.  See impelmentation below for
            exact return types.
 NOTES:     Yeah, this is pretty quick and dirty, but whatever.
            mdomowicz 2009_07_31
******************************************************************************/
int SoLuaState::Debug_GetIniSetting( lua_State* pL )
{
	#if 0
		const char* pacSetting = GetLuaString( pL, 1 );
		unsigned int uiSettingHash = JpHash( pacSetting );

		SoGlobalIniSettings* pSettings = SoGlobalIniSettings::GetInstance();

		switch ( uiSettingHash )
		{
			case 0x87790aa5 /* "DebugLevelSkip" */:
			{
				int iLevel = pSettings->GetDebugLevelSkip();
				PushLuaNumber( pL, (float)iLevel );
				return 1;
			}
			case 0x21838186 /* "DebugModeOn" */:
			{
				PushLuaBool( pL, pSettings->GetDebugModeOn() );
				return 1;
			}
			default:
			{
				ASSERTF( 0, ( "\nIni setting not yet exposted to lua api" ) );
				break;
			}
		}
	#endif

	return 0;
}



/******************************************************************************
 FUNCTION:  Util_GetDistanceSq
 PURPOSE:   LUA API Function: Get the squared distance between two sets of 
            coordinates.
 INPUT:     ARG 1: Number: X1
            ARG 2: Number: Y1
            ARG 3: Number: X2
            ARG 4: Number: Y2
 OUPUT:     RET 1: Number: the squared distance between the given two points
 NOTES:
******************************************************************************/
int SoLuaState::Util_GetDistanceSq( lua_State* pL )
{
	float fX1 = GetLuaNumber( pL, 1 );
	float fY1 = GetLuaNumber( pL, 2 );
	float fX2 = GetLuaNumber( pL, 3 );
	float fY2 = GetLuaNumber( pL, 4 );

	JpVector2f vP1( fX1, fY1 );
	JpVector2f vP2( fX2, fY2 );

	float fDistSq = LengthSquared( vP1 - vP2 );

	PushLuaNumber( pL, fDistSq );
	return 1;
}



/******************************************************************************
 FUNCTION:  Util_GetRandomFloatRanged
 PURPOSE:   LUA API Function: Get a random float between fMin and fMax.
            fMin and fMax are possible return values.
 INPUT:     ARG 1: Number: The minimum value.
            ARG 2: Number: The maximum value.
 OUPUT:     RET 1: Number: a random float between the min and max
 NOTES:
******************************************************************************/
int SoLuaState::Util_GetRandomFloatRanged( lua_State* pL )
{
	float fMin = GetLuaNumber( pL, 1 );
	float fMax = GetLuaNumber( pL, 2 );
	float fValue = Interpolate( fMax, fMin, JpRandom::G.GetFloat() );
	PushLuaNumber( pL, fValue );

	return 1;
}



/******************************************************************************
 FUNCTION:  Util_GetRandomInt
 PURPOSE:   LUA API Function: Get a random integer.
 INPUT:     ARG 1: Number: The maximum value.
 OUPUT:     RET 1: Number: a whole number between and including 0 and 
            "maximum value - 1"
 NOTES:
******************************************************************************/
int SoLuaState::Util_GetRandomInt( lua_State* pL )
{
	float fMaxValue = GetLuaNumber( pL, 1 );

	int iMaxValue = (int)fMaxValue;
	int iValue = JpRandom::G.GetInt() % iMaxValue;

	PushLuaNumber( pL, (float)iValue );

	return 1;
}



/******************************************************************************
 FUNCTION:  Util_SetSpecialRandomSeed
 PURPOSE:   LUA API Function: Seed the Special Rng.
 INPUT:     ARG 1: String: text to be used as the rng seed.
 OUPUT:     
 NOTES:     

The Special Rng is a trick that allows the following properties:

1. If Seeded before use, it simulates "local" Rngs.
2. If Seeded before use, it can be used at any time to generate a consistent 
   number sequence (ie, it does not have to generate the sequence at startup, 
   or in a particular "known deterministic" part of the code).
3. With the exact same seed it will generate a *different* consistent number
   sequence for each different world MasterSeed.

Why is this useful?

a. You can get consistent random numbers when and where you need them.  Just
   use different seeds.  eg. Seed( "Alpha" ), Seed( "HolaCola" )
b. No need to change the seed to get a different set of numbers for a
   different world MasterSeed.  Say you have a boss that has to live in a
   random sector.  You can pick his sector using Special Random and the seed,
   "llama", and even if the script is shut down and restarted (perhaps on 
   game save/restore?) you can easily repick the same sector.  However, if the 
   world MasterSeed changes (on start of a new game) you want to pick a new
   random sector that can also be easily repicked on script restart.  With
   special Rngs, this all happens automatically.
******************************************************************************/
int SoLuaState::Util_SetSpecialRandomSeed( lua_State* pL )
{
	/*
	const char* pacSeed = GetLuaString( pL, 1 );
	hash hSeed = JpHash( pacSeed );
	
	SoLuaState* pSoLuaState = GetSoLuaStateFromLuaState( pL );
	JpRandom* pRng = pSoLuaState->GetSpecialRng();

	SoGameDungeon* pDungeon = SoGameDungeon::GetInstance();
	hash hMasterSeed = pDungeon->GetMasterRngSeed();

	hash hFinalSeed = ( hSeed ^ hMasterSeed );
	pRng->Seed( hFinalSeed );
	*/

	return 0;
}



/******************************************************************************
 FUNCTION:  Util_GetSpecialRandomFloatRanged
 PURPOSE:   LUA API Function: Get a random float between fMin and fMax.
            fMin and fMax are possible return values.
 INPUT:     ARG 1: Number: The minimum value.
            ARG 2: Number: The maximum value.
 OUPUT:     RET 1: Number: a random float between the min and max
 NOTES:     See Util_SetSpecialRandomSeed for more on the purpose of this
******************************************************************************/
int SoLuaState::Util_GetSpecialRandomFloatRanged( lua_State* pL )
{
	float fMin = GetLuaNumber( pL, 1 );
	float fMax = GetLuaNumber( pL, 2 );

	SoLuaState* pSoLuaState = GetSoLuaStateFromLuaState( pL );
	JpRandom* pRng = pSoLuaState->GetSpecialRng();

	float fValue = Interpolate( fMax, fMin, pRng->GetFloat() );
	PushLuaNumber( pL, fValue );

	return 1;
}



/******************************************************************************
 FUNCTION:  Util_GetSpecialRandomInt
 PURPOSE:   LUA API Function: Get a random integer.
 INPUT:     ARG 1: Number: The maximum value.
 OUPUT:     RET 1: Number: a whole number between and including 0 and 
            "maximum value - 1"
 NOTES:     See Util_SetSpecialRandomSeed for more on the purpose of this
******************************************************************************/
int SoLuaState::Util_GetSpecialRandomInt( lua_State* pL )
{
	float fMaxValue = GetLuaNumber( pL, 1 );

	SoLuaState* pSoLuaState = GetSoLuaStateFromLuaState( pL );
	JpRandom* pRng = pSoLuaState->GetSpecialRng();

	int iMaxValue = (int)fMaxValue;
	int iValue = pRng->GetInt() % iMaxValue;

	PushLuaNumber( pL, (float)iValue );

	return 1;
}



/******************************************************************************
 FUNCTION:  Util_LoadScript
 PURPOSE:   LUA API Function: load an external lua module into the current
            lua state environment.
 INPUT:     ARG 1: String: script name, without path.  eg, "GmEventManager.lua"
 OUPUT:
 NOTES:
******************************************************************************/
int SoLuaState::Util_LoadScript( lua_State* pL )
{
	const char* pacFilename = GetLuaString( pL, 1 );

	#if DCG_IS_TARGET_WIN32
		int iResult = luaL_dofile( pL, SoUtils::AddRelativePathToScript( pacFilename ) );	
	#else
		const char* pacFullPath = CCFileUtils::sharedFileUtils()->fullPathFromRelativePath( SoUtils::AddRelativePathToIni( pacFilename ) );
		int iResult = luaL_dofile( pL, SoUtils::AddRelativePathToScript( pacFullPath ) );	
	#endif


	bool bSuccess = ( iResult == 0 );

	if ( !bSuccess )
		PrintLuaCallstack( pL );

	ASSERTF( bSuccess, ( "\nFailed to load lua script." ) );

	PushLuaBool( pL, bSuccess );
	return 1;
}



/******************************************************************************
 FUNCTION:  Hud_PrintToLog
 PURPOSE:   Lua API function: prints a string to the HudLog
 ARGS:      ARG 1: String: the string to be printed.
 RETURN:    None
 NOTES:     
******************************************************************************/
int SoLuaState::Hud_PrintToLog( lua_State* pL )
{
	SoHudLog* pLog = SoFsmMainLayer::GetSharedHudLog();

	if ( lua_isstring( pL, 1 ) )
	{
		const char* pcString = NULL;
		pcString = lua_tostring( pL, 1 );

		CCLOG( pcString );
		pLog->AddMessageF( JpUtils::_White, pcString );
	}

	return 0;
}



