/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/



#pragma once

#include "stdafx.h"
#include "cocos2d.h"

extern "C" {
#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"
}

#include "JpRandom.h"
#include "JpVector.h"


USING_NS_CC;


#define SoLuaStateTable  JpHashTable< SoLuaState, 16 >

class SoActor;
class SoSaveSystem;

class SoLuaState
{
	public:

		//-----------------------
		// This class is used to watch the integrity of the lua stack.
		// To use, just make an instance of this class at the start
		// of any function that uses the lua stack.  The destructor
		// of this class will make sure the Lua stack has the expected
		// number of elements when the function exits.
		//-----------------------
		class AutoStackCheck
		{
			public:
				AutoStackCheck( lua_State *s, int iNumStackElementsOnExit ) : _pL( s )
				{
					_iFinalStackTop = lua_gettop( _pL ) + iNumStackElementsOnExit;
				}

				~AutoStackCheck()
				{
					int iCurrentStackTop = lua_gettop( _pL );
					ASSERTF( _iFinalStackTop == iCurrentStackTop, ( "\nLua stack check fail." ) );
				}

			private:
				lua_State* _pL;
				int _iFinalStackTop;
		};


		//-----------------------
		// public functions

		SoLuaState();

		virtual ~SoLuaState();

		void Init();

		void Deinit();

		void Update( float dt );

		bool OpenScript( const char acFileName[] );

		void RegisterApiFunctions();
		void RegisterGlobals();

		void InvokeLuaFunction( const char* szFunctionName );

		JpRandom* GetSpecialRng() { return &_SpecialRng; }

		void SaveToFile( SoSaveSystem* pSaveSystem );
		void LoadFromFile( SoSaveSystem* pSaveSystem );


		static SoLuaState* GetInstance() { return _pInstance; }
		static void CreateInstance();
		static void DeleteInstance();



		//-----------------------
		// API
		static int Debug_Print( lua_State* pL );
		static int Debug_Break( lua_State* pL );
		static int Debug_Assert( lua_State* pL );
		static int Debug_GetIniSetting( lua_State* pL );

		static int Util_GetDistanceSq( lua_State* pL );
		static int Util_GetRandomFloatRanged( lua_State* pL );
		static int Util_GetRandomInt( lua_State* pL );
		static int Util_SetSpecialRandomSeed( lua_State* pL );
		static int Util_GetSpecialRandomFloatRanged( lua_State* pL );
		static int Util_GetSpecialRandomInt( lua_State* pL );
		static int Util_LoadScript( lua_State* pL );

		static int Hud_PrintToLog( lua_State* pL );

	private:

		//-----------------------
		// private variables
		static SoLuaState* _pInstance;

		lua_State* _pL;
		JpRandom _SpecialRng;

		static SoSaveSystem* _pSaveSystem;
		static SoLuaStateTable _LuaStateTable;


		//-----------------------
		// private functions
		void SoPcall( lua_State* pL, int nargs, int nresults, const char* sDesc );

		static float GetLuaNumber( lua_State* pL, int iStackIndex );
		static bool GetLuaBool( lua_State* pL, int iStackIndex );
		static const char* GetLuaString( lua_State* pL, int iStackIndex );
		static JpVector2f GetLuaVector2( lua_State* pL, int iStackIndex );
		static SoActor* GetLuaActor( lua_State* pL, int iStackIndex );

		static void PushLuaNumber( lua_State* pL, float fNumber );
		static void PushLuaString( lua_State* pL, const char* pacString );
		static void PushLuaBool( lua_State* pL, bool bFlag );
		static void PushLuaActor( lua_State* pL, SoActor* pActor );

		static int GetAbsoluteStackIndex( lua_State* pL, int iIndex );

		static const char* GetLuaArrayString( lua_State* pL, int iTableStackIndex, int iTableIndex, const char* pacReturnOnFail );

		static float GetLuaTableNumber( lua_State* pL, int iTableStackIndex, const char* pacKey, bool* pbSuccess );
		static const char* GetLuaTableString( lua_State* pL, int iTableStackIndex, const char* pacKey, bool* pbSuccess );
		static bool GetLuaTableBool( lua_State* pL, int iTableStackIndex, const char* pacKey, bool* pbSuccess );

		static bool GetLuaTableNumberIntoArg( lua_State* pL, float* pfResult, int iTableStackIndex, const char* pacKey );
		static bool GetLuaTableIntIntoArg( lua_State* pL, int* piResult, int iTableStackIndex, const char* pacKey );
		static bool GetLuaTableStringIntoArg( lua_State* pL, const char** ppacResult, int iTableStackIndex, const char* pacKey);
		static bool GetLuaTableActorIntoArg( lua_State* pL, SoActor** ppActorResult, int iTableStackIndex, const char* pacKey );
		static bool GetLuaTableBoolIntoArg( lua_State* pL, bool* pbResult, int iTableStackIndex, const char* pacKey );


		static void PushLuaTable( lua_State* pL );
		static void PushLuaTableNumber( lua_State* pL, const char* pacKey, float fValue );

		static void PrintLuaCallstack( lua_State* pL );



		//-----------------------
		// this is so that the static Api functions can refer
		// to this SoLuaState from the lua_State that they get
		// passed to them.
		// mdomowicz 2009_03_10
		static SoLuaState* GetSoLuaStateFromLuaState( lua_State* pL ) { return _LuaStateTable.Find( (unsigned int)pL ); }

		static void SetLuaStateToSoLuaState( lua_State* pL, SoLuaState* pSoLuaState ) 
		{ 
			_LuaStateTable.Store( (unsigned int)pL, pSoLuaState ); 
		}

		static void UnsetLuaStateToSoLuaState( lua_State* pL ) 
		{ 
			_LuaStateTable.Remove( (unsigned int)pL ); 
		}
};




