/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#include "SoGameBitPlayerMcts.h"
#include "SoMctsTree.h"
#include "SoSaveSystem.h"

#include "mmgrChooser.h"



bool SoMctsTreeNode::_bDoChecksumTesting = false;


/******************************************************************************
 FUNCTION:  SoMctsTree
******************************************************************************/
SoMctsTree::SoMctsTree() :
_iMaxSimulations( 0 ),
_iNumSimulations( 0 ),
_iNumMovesPerSimulation( 100 )
{
	Init();
}



/******************************************************************************
 FUNCTION:  ~SoMctsTree
******************************************************************************/
SoMctsTree::~SoMctsTree()
{
	RemoveAndDeleteAllNodes();
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoMctsTree::Init()
{
	_iNumSimulations = 0;
	_HeadNode.SetNumVisits( 1 );
}



/******************************************************************************
 FUNCTION:  SearchForMove
******************************************************************************/
SoGameMessage* SoMctsTree::SearchForMove( SoGameBitPlayer* pPlayer, SoGameState* pGame, SoGameState::STATE eServersSoGameState )
{
	if ( _iNumSimulations < _iMaxSimulations )
	{
		_iNumSimulations ++;

		//-----------------------
		// copy the SoGame into an SoGameMcts - we only alter the copy.
		SoGameStateMcts GameMcts;
		GameMcts.MakeCopy( pGame, pPlayer->GetAiPlayerStrength() );


		//-----------------------
		// Both SoGameServer and SoGameClient have their own instances of SoGame
		// that are kept in-sync with each other.  The copy made above is of
		// the Client's SoGame.  But, the client's SoGame is missing some Server
		// specific information - most importantly the Server SoGames's current
		// state.  And so we (rather hackily) manually set it here.
		// Why do we even need the server SoGames's state?  Because when we simulate 
		// the game we share this one copy of SoGame between both the simulated
		// client and simulated server.  Thus this copy needs to have the data
		// relevant to both.
		GameMcts.SetCurrentState( eServersSoGameState );


		//-----------------------
		// This helps guarantee that simulation process is using a fresh
		// (but still deterministic) rng seed.  The GetInt() alone guarantees
		// the rng is not the same as last time, and the hNoise also helps
		// prevent any rng output sequence aliasing.
		int iRandom = pGame->GetMasterRng()->GetInt();
		hash hNoise = 0x904416d1 /* "Noise" */;
		hash hNewSeed = ( iRandom ^ hNoise );
		GameMcts.SeedMasterRng( hNewSeed );


		//-----------------------
		// Get a copy of the current player.
		// Because the GameMcts is a copy of the actual SoGame,
		// getting the player from it, is getting a copy.
		SoGameBitPlayer* pCopiedPlayer = GameMcts.GetPlayerByPlayerNumber( pPlayer->GetPlayerNumber() );
		SoGameBitPlayerMcts* pPlayerMcts = pCopiedPlayer->GetAsPlayerMcts();


		//-----------------------
		// Begin the core mcts processing!
		_HeadNode.SetPlayerNumber( pPlayerMcts->GetPlayerNumber() );
		_HeadNode.EvaluateNode( pPlayerMcts, &GameMcts );
		_HeadNode.AdjustNumVisits( 1 );
	}


	if ( _iNumSimulations == _iMaxSimulations )
	{
		DebugPrintDiagnostics( pGame );

		SoMctsTreeNode* pResult = _HeadNode.GetChildWithBestScore();
		if ( pResult )
		{
			return pResult->GetGameMessage();
		}
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  RemoveAndDeleteAllNodes
******************************************************************************/
void SoMctsTree::RemoveAndDeleteAllNodes()
{
	_HeadNode.RemoveAndDeleteAllChildNodes();
}



/******************************************************************************
 FUNCTION:  DebugPrintDiagnostics
******************************************************************************/
void SoMctsTree::DebugPrintDiagnostics( SoGameState* pGame )
{
	CCLOG( "Num Unique Moves Evaluated  =========~~~~~~~~~~~ %i", _HeadNode.GetChildNodeList()->GetNumEntries() );
	CCLOG( "mctsTree Max Depth   ================~~~~~~~~~~~ %i", _HeadNode.CalculateMaxDepth() );
	CCLOG( "mctsTree Max Breadth ================~~~~~~~~~~~ %i", _HeadNode.CalculateMaxBreadth() );
	CCLOG( "mctsTree Total Nodes ================~~~~~~~~~~~ %i", _HeadNode.CalculateTotalSubNodeCount() );


	SoMctsTreeNode* pBest = _HeadNode.GetChildWithBestScore();
	CCLOG( "~~~~~~~~~~~" );
	CCLOG( "Best Node Num Visits ================~~~~~~~~~~~ %i", pBest->GetNumVisits() );
	CCLOG( "mctsTree P1 Eval     ================~~~~~~~~~~~ %.2f", pBest->GetPlayerEvaluation( 0 ) / pBest->GetNumVisits() );
	CCLOG( "mctsTree P2 Eval     ================~~~~~~~~~~~ %.2f", pBest->GetPlayerEvaluation( 1 ) / pBest->GetNumVisits()  );
	CCLOG( "mctsTree P3 Eval     ================~~~~~~~~~~~ %.2f", pBest->GetPlayerEvaluation( 2 ) / pBest->GetNumVisits()  );
	CCLOG( "mctsTree P4 Eval     ================~~~~~~~~~~~ %.2f", pBest->GetPlayerEvaluation( 3 ) / pBest->GetNumVisits()  );
	CCLOG( "mctsTree P5 Eval     ================~~~~~~~~~~~ %.2f", pBest->GetPlayerEvaluation( 4 ) / pBest->GetNumVisits()  );
	CCLOG( "mctsTree P6 Eval     ================~~~~~~~~~~~ %.2f", pBest->GetPlayerEvaluation( 5 ) / pBest->GetNumVisits()  );
	CCLOG( "~~~~~~~~~~~" );
}











/******************************************************************************
 FUNCTION:  SoMctsTreeNode
******************************************************************************/
SoMctsTreeNode::SoMctsTreeNode() :
_pParentNode( NULL ),
_hRngSeed( 0 ),
_iNumVisits( 0 ),
_iPlayerNumber( PLAYER_UNDEFINED_PLAYER_NUMBER ),
_iNumMovesPerSimulation( 100 ),
_GameChecksum( 0 )
{
	Init();
}



/******************************************************************************
 FUNCTION:  ~SoMctsTreeNode
******************************************************************************/
SoMctsTreeNode::~SoMctsTreeNode()
{
	RemoveAndDeleteAllChildNodes();
}



/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoMctsTreeNode::Init()
{
	memset( _afPlayerEvaluations, 0.0f, GAMERULES_MAX_PLAYERS * sizeof( float ) );
}



/******************************************************************************
 FUNCTION:  EvaluateNode
******************************************************************************/
void SoMctsTreeNode::EvaluateNode( SoGameBitPlayerMcts* pPlayerMcts, SoGameStateMcts* pGameMcts )
{ 
	SoMctsTreeNode* pChildNode = CreateOrSelectNodeToEvaluate( pPlayerMcts, pGameMcts );

	if ( pChildNode->GetNumVisits() == 0 )
	{
		pChildNode->SetNumVisits( 1 );


		//-----------------------
		// "apply" the move to the SoGame
		pGameMcts->UpdateFromServer( pChildNode->GetGameMessage() );

		if ( _bDoChecksumTesting )
			pChildNode->SetChecksum( pGameMcts->CalculateChecksumForMcts() );


		//-----------------------
		// playout the rest of the game
		for ( int i = 0; i < _iNumMovesPerSimulation; i ++ )
		{
			pGameMcts->Simulate();

			if ( pGameMcts->IsGameOver() )
				break;
		}

		pGameMcts->DeleteMessageLists();


		//-----------------------
		// Set the node's evaluations
		pGameMcts->Evaluate();
		for ( int i = 0; i < GAMERULES_MAX_PLAYERS; i ++ )
		{
			pChildNode->SetPlayerEvaluation( i, pGameMcts->GetPlayerEvaluation( i ) );
		}


		//-----------------------
		// Add the node to the child list
		_ChildNodeList.AddLast( pChildNode );
	}
	else
	{
		pChildNode->AdjustNumVisits( 1 );


		//-----------------------
		// Some error checking.
		ASSERTF( pPlayerMcts->GetPlayerNumber() == pChildNode->GetPlayerNumber(), ( "\nMctsTreeNode has wrong player." ) );


		//-----------------------
		// Normally, when a player makes a move, it goes into a "wait for server" state.
		// Because we are making a move for the player, we have to manually push the player
		// into the correct state.
		if ( !pPlayerMcts->IsInWaitState() )
			pPlayerMcts->AdvanceToWaitState();


		//-----------------------
		// "Apply" the move represented by the node
		pGameMcts->UpdateFromServer( pChildNode->GetGameMessage() );

		if ( _bDoChecksumTesting )
		{
			ChecksumUint32 uiNodeChecksum = pChildNode->GetChecksum();
			ChecksumUint32 uiGameChecksum = pGameMcts->CalculateChecksumForMcts();
			ASSERTF( uiNodeChecksum == uiGameChecksum, ( "\nBad SoGame checksum!" ) );
		}


		//-----------------------
		// Step the game forward, until it sends a command to a player, or
		// the game is ended/won.
		pGameMcts->SimulateUntilNextPlayerMoveCommand();


		//-----------------------
		// ** IF ** the applied move ends the game
		// then we update the evaluation of the associated tree node, and
		// stop.  We make no more recursive calls to EvaluateNode().
		if ( pGameMcts->IsGameOver() )
		{
			pGameMcts->Evaluate();
			for ( int i = 0; i < GAMERULES_MAX_PLAYERS; i ++ )
			{
				pChildNode->AdjustPlayerEvaluation( i, pGameMcts->GetPlayerEvaluation( i ) );
			}
		}


		//-----------------------
		// ** ELSE ** if the applied move doesn't end the game, we continue
		// walking the tree.
		else
		{
			// Get the player that the next command is for
			SoGameBitPlayerMcts* pNextPlayerToMove = FindPlayerCommandedToMove( pGameMcts );

			// put the new player in the proper state.
			pNextPlayerToMove->AdvanceToMoveSelectionState();

			// "visit" the tree node 
			pChildNode->EvaluateNode( pNextPlayerToMove, pGameMcts );
		}
	}


	//-----------------------
	// Back propagation
	// According to Chaslot_thesis.pdf, "Monte Carlo Tree Search", ISBN 978-90-8559-099-6, page 23, section 3.3.4, 
	// It is best to use the average of the results of all games simulated through a node, as the node's value.
	// (An alternate strategy is to consider a node's value to be the maximum value of a node's children, but
	// the thesis noted this is not ideal when simulation counts are low).
	for ( int i = 0; i < GAMERULES_MAX_PLAYERS; i ++ )
	{
		_afPlayerEvaluations[ i ] += pGameMcts->GetPlayerEvaluation( i );
	}
}



/******************************************************************************
 FUNCTION:  CreateOrSelectNodeToEvaluate
******************************************************************************/
SoMctsTreeNode* SoMctsTreeNode::CreateOrSelectNodeToEvaluate( SoGameBitPlayerMcts* pPlayerMcts, SoGameStateMcts* pGameMcts )
{
	//-----------------------
	// Pick (or don't) a move from the list of already
	// analyzed moves.
	SoMctsTreeNode* pChosenTreeNode = SelectNodeToEvaluate( pPlayerMcts->GetPlayerNumber() );


	//-----------------------
	// If the above decided not to pick an existing move, 
	// generate a new one.
	if ( !pChosenTreeNode )
	{
		ASSERTF( pPlayerMcts->IsInMoveSelectionState(), ( "\nPassed in player must be in the right state." ) );

		//-----------------------
		// Pick a random move - do this by putting the PlayerMcts 
		// in a special mode, then force updates until the player
		// has a message for the server.
		pPlayerMcts->SetMessageTarget( SoGameBitPlayerMcts::MESSAGE_TARGET_MCTS_TREE );
		pPlayerMcts->ClearLastSentMessage();

		while ( pPlayerMcts->GetLastSentMessage()->IsClear() )
			pPlayerMcts->Update( 0.0f );
	
		pPlayerMcts->SetMessageTarget( SoGameBitPlayerMcts::MESSAGE_TARGET_SERVER );


		//-----------------------
		// The randomly generated move above might be a duplicate of 
		// a move we've already analyzed.  If, so we return the already
		// existing move.  We do not keep looping until we get a new move
		// because it could be that there are no new moves!
		pChosenTreeNode = FindChildNodeWithDuplicateMove( pPlayerMcts->GetLastSentMessage(), pGameMcts );


		//-----------------------
		if ( !pChosenTreeNode )
		{
			// Create a new tree node for the new move.
			// Note - we have to do this before we Simulate.  Else
			// the "last sent message" will be wrong.
			pChosenTreeNode = new SoMctsTreeNode;
			pChosenTreeNode->SetPlayerNumber( pPlayerMcts->GetPlayerNumber() );
			pChosenTreeNode->CopyGameMessage( pPlayerMcts->GetLastSentMessage() );
		}
	}

	return pChosenTreeNode;
}



/******************************************************************************
 FUNCTION:  SelectNodeToEvaluate
 NOTES:     This uses Upper Confidence Bounds for Trees, aka UCT, as described
            in mcts-survey-master.pdf (page 7), aka:

            "A Survey of Monte Carlo Tree Search Methods" 
            IEEE TRANSACTIONS ON COMPUTATIONAL INTELLIGENCE AND AI IN GAMES, 
            VOL. 4, NO. 1, MARCH 2012 

            A node's UCT value is defined by:

                      UCT = X + 2C * sqrt( 2ln( N ) / Nj )

            where:
                      X  = the value of the node, normalized to [0, 1]
                      C  = a tweakable constant, suggested to be 1/sqrt(2)
                      N  = number of visits of the node's parent
                      Nj = number of visits of the node

            Notice that X, the "exploit term" grows if the node and its
            children have good evaluations.  The rest, the "explore term"
            grows as the parent gets more visits than the node, thus
            encouraging selection of lesser valued nodes sometimes.
******************************************************************************/
SoMctsTreeNode* SoMctsTreeNode::SelectNodeToEvaluate( int iPlayerNumber )
{
	//-----------------------
	// Determine the highest evaluation.
	// This is needed for normalizing the evaluation.
	// (See above.  X must be between [0, 1])
	float fHighestEval = 0.0f;
	SoMctsTreeNodeList::Iterator Iter( &_ChildNodeList );
	while ( Iter.Next() )
	{
		SoMctsTreeNode* pTreeNode = Iter.GetNodeData();
		float fEval = pTreeNode->GetPlayerEvaluation( iPlayerNumber );
		if ( fEval > fHighestEval )
			fHighestEval = fEval;
	}


	//-----------------------
	// Calculate the UCT value for each node.
	// Find the node with the highest UCT value.
	SoMctsTreeNode* pChosenTreeNode = NULL;
	float fHighestUcb = -1.0f;

	Iter.Reset();
	while ( Iter.Next() )
	{
		SoMctsTreeNode* pTreeNode = Iter.GetNodeData();

		float _2_C = 2.0f * 0.707106f;

		float _2_LogN = ( 2.0f * logf( _iNumVisits ) );

		float Nj = pTreeNode->GetNumVisits();

		float fExploreTerm = ( _2_C * sqrtf( _2_LogN / Nj ) );

		float fExploitTerm = ( pTreeNode->GetPlayerEvaluation( iPlayerNumber ) / fHighestEval );

		float fUcb = ( fExploitTerm + fExploreTerm );

		if ( fUcb > fHighestUcb )
		{
			fHighestUcb = fUcb;
			pChosenTreeNode = pTreeNode;
		}
	}


	if ( fHighestUcb > 2.25f )
		return pChosenTreeNode;
	
	return NULL;
}



/******************************************************************************
 FUNCTION:  FindChildNodeWithDuplicateMove
******************************************************************************/
SoMctsTreeNode* SoMctsTreeNode::FindChildNodeWithDuplicateMove( SoGameMessage* pMessage, SoGameStateMcts* pGameMcts )
{
	SoMctsTreeNodeList::Iterator Iter( &_ChildNodeList );
	while ( Iter.Next() )
	{
		SoMctsTreeNode* pTreeNode = Iter.GetNodeData();

		if ( AreTheseMessagesDuplicates( pMessage, pTreeNode->GetGameMessage(), pGameMcts ) )
			return pTreeNode;
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  AreTheseMessagesDuplicates
******************************************************************************/
bool SoMctsTreeNode::AreTheseMessagesDuplicates
( 
	SoGameMessage* pMessage1, 
	SoGameMessage* pMessage2, 
	SoGameStateMcts* pGameMcts 
)
{
	if ( pMessage1->GetType() != pMessage2->GetType() )
		return false;

	switch ( pMessage1->GetType() )
	{
		case Msg::TYPE_PLYR_MOVE_PLACE_TILE:
		{
			SoCodecPlyrMovePlaceTile Decoder1;
			SoCodecPlyrMovePlaceTile Decoder2;

			Decoder1.Decode( pMessage1 );
			Decoder2.Decode( pMessage2 );

			if ( Decoder1._iPlacedTileId != Decoder2._iPlacedTileId )
				return false;

			if ( Decoder1._bSkippingTilePlacement != Decoder2._bSkippingTilePlacement )
				return false;

			break;
		}
		case Msg::TYPE_PLYR_MOVE_STOCK_PURCHASES:
		{
			SoCodecPlyrMoveStockPurchases Decoder1;
			SoCodecPlyrMoveStockPurchases Decoder2;

			Decoder1.Decode( pMessage1 );
			Decoder2.Decode( pMessage2 );

			if ( Decoder1._iAmountOfHotelWithId0 != Decoder2._iAmountOfHotelWithId0 )
				return false;
			if ( Decoder1._iAmountOfHotelWithId1 != Decoder2._iAmountOfHotelWithId1 )
				return false;
			if ( Decoder1._iAmountOfHotelWithId2 != Decoder2._iAmountOfHotelWithId2 )
				return false;
			if ( Decoder1._iAmountOfHotelWithId3 != Decoder2._iAmountOfHotelWithId3 )
				return false;
			if ( Decoder1._iAmountOfHotelWithId4 != Decoder2._iAmountOfHotelWithId4 )
				return false;
			if ( Decoder1._iAmountOfHotelWithId5 != Decoder2._iAmountOfHotelWithId5 )
				return false;
			if ( Decoder1._iAmountOfHotelWithId6 != Decoder2._iAmountOfHotelWithId6 )
				return false;

			break;
		}
		case Msg::TYPE_PLYR_MOVE_NEW_HOTEL_PICK:
		{
			SoCodecPlyrMoveNewHotelPick Decoder1;
			SoCodecPlyrMoveNewHotelPick Decoder2;

			Decoder1.Decode( pMessage1 );
			Decoder2.Decode( pMessage2 );

			if ( Decoder1._iHotelId != Decoder2._iHotelId )
				return false;

			break;
		}
		case Msg::TYPE_PLYR_MOVE_MERGE_HOTELS:
		{
			SoCodecPlyrMoveMergeHotels Decoder1;
			SoCodecPlyrMoveMergeHotels Decoder2;

			Decoder1.Decode( pMessage1 );
			Decoder2.Decode( pMessage2 );

			if ( Decoder1._iSurvivorHotelId != Decoder2._iSurvivorHotelId )
				return false;
			if ( Decoder1._iVictimHotelId != Decoder2._iVictimHotelId )
				return false;

			break;
		}
		case Msg::TYPE_PLYR_MOVE_MANAGE_OLD_STOCK:
		{
			SoCodecPlyrManageOldStock Decoder1;
			SoCodecPlyrManageOldStock Decoder2;

			Decoder1.Decode( pMessage1 );
			Decoder2.Decode( pMessage2 );

			if ( Decoder1._iNumStocksToKeep != Decoder2._iNumStocksToKeep )
				return false;
			if ( Decoder1._iNumStocksToSell != Decoder2._iNumStocksToSell )
				return false;
			if ( Decoder1._iNumStocksToTrade != Decoder2._iNumStocksToTrade )
				return false;

			break;
		}

		case Msg::TYPE_PLYR_MOVE_DECLARE_LAST_TURN:
		{
			break;
		}

		case Msg::TYPE_PLYR_MOVE_END_TURN:
		{
			break;
		}

		default:
			ASSERTF( 0, ( "\nUnhandled message type." ) );
	}


	return true;
}



/******************************************************************************
 FUNCTION:  FindPlayerCommandedToMove
******************************************************************************/
SoGameBitPlayerMcts* SoMctsTreeNode::FindPlayerCommandedToMove( SoGameStateMcts* pGameMcts )
{
	SoGameBitPlayerMcts* pNextPlayerToMove = NULL;
	SoGameBitPlayerArray::Iterator PlayerIter( pGameMcts->GetPlayerArray() );
	while ( PlayerIter.Next() )
	{
		SoGameBitPlayer* pPlayer = PlayerIter.GetData();
		if ( pPlayer->GetRequestedState() != SoGameBitPlayer::STATE_LAST )
		{
			return pPlayer->GetAsPlayerMcts();
		}
	}

	return NULL;
}



/******************************************************************************
 FUNCTION:  CalculateNodesValueToPlayer
******************************************************************************/
float SoMctsTreeNode::CalculateNodesValueToPlayer( int iPlayerNumber )
{
	return ( GetPlayerEvaluation( iPlayerNumber ) / _iNumVisits );
}



/******************************************************************************
 FUNCTION:  GetChildWithBestScore
 NOTE:      This function was changed so that it searches for the highest 
            rated child node.  It must do this now because the child list is
            no longer sorted/maintained during evaluation (which makes
            evaluation faster).

            IMPORTANT: Because of this change, the returned child is not
            always the same as it was before, making the results of an all-Ai
            play-through inconsistent compared to before this change.  From what
            I can tell, this is due to where CalculateNodesValueToPlayer()
            is called from.  Before, when sorting happened, it was called
            at the time of the sort.  But now it is only called here.  This
            leads to a slight difference in floating point division results
            (ie, eval/numVisits) and when two nodes are essentially the same, 
            the slight difference can lead to one node being chosen over
            another, differently than before.  As far as I can tell, this
            does not affect the overall quality of the results.  But keep
            an eye on this.  mdomowicz 2015_11_01
******************************************************************************/
SoMctsTreeNode* SoMctsTreeNode::GetChildWithBestScore()
{
	//-----------------------
	// Loop through all the children to find
	// the child with the highest score.
	SoMctsTreeNode* pChildNodeWithBestScore = NULL;
	float fBestScore = -FLT_MAX;

	SoMctsTreeNodeList::Iterator Iter( &_ChildNodeList );
	while ( Iter.Next() )
	{
		SoMctsTreeNode* pChildNode = Iter.GetNodeData();
		float fScore = pChildNode->CalculateNodesValueToPlayer( GetPlayerNumber() );

		if ( fScore > fBestScore )
		{
			fBestScore = fScore;
			pChildNodeWithBestScore = pChildNode;
		}
	}

	return pChildNodeWithBestScore;
}



/******************************************************************************
 FUNCTION:  GetPlayerEvaluation
******************************************************************************/
float SoMctsTreeNode::GetPlayerEvaluation( int iPlayerNumber )
{	
	ASSERTF( iPlayerNumber >= 0 && iPlayerNumber < GAMERULES_MAX_PLAYERS, ( "\nBad PlayerNumber" ) );

	if ( iPlayerNumber >= 0 && iPlayerNumber < GAMERULES_MAX_PLAYERS )
		return _afPlayerEvaluations[ iPlayerNumber ];

	return 0.0f;
}



/******************************************************************************
 FUNCTION:  SetPlayerEvaluation
******************************************************************************/
void SoMctsTreeNode::SetPlayerEvaluation( int iPlayerNumber, float fScore )
{
	ASSERTF( iPlayerNumber >= 0 && iPlayerNumber < GAMERULES_MAX_PLAYERS, ( "\nBad PlayerNumber" ) );

	if ( iPlayerNumber >= 0 && iPlayerNumber < GAMERULES_MAX_PLAYERS )
		_afPlayerEvaluations[ iPlayerNumber ] = fScore;
}



/******************************************************************************
 FUNCTION:  AdjustPlayerEvaluation
******************************************************************************/
void SoMctsTreeNode::AdjustPlayerEvaluation( int iPlayerNumber, float fScore )
{
	ASSERTF( iPlayerNumber >= 0 && iPlayerNumber < GAMERULES_MAX_PLAYERS, ( "\nBad PlayerNumber" ) );

	if ( iPlayerNumber >= 0 && iPlayerNumber < GAMERULES_MAX_PLAYERS )
		_afPlayerEvaluations[ iPlayerNumber ] += fScore;
}



/******************************************************************************
 FUNCTION:  RemoveAndDeleteAllChildNodes
******************************************************************************/
void SoMctsTreeNode::RemoveAndDeleteAllChildNodes()
{
	SoMctsTreeNodeList::Iterator Iter( &_ChildNodeList );
	while ( Iter.Next() )
	{
		SoMctsTreeNode* pTreeNode = Iter.GetNodeData();

		pTreeNode->RemoveAndDeleteAllChildNodes();

		delete pTreeNode;
	}

	_ChildNodeList.RemoveAll();
}



/******************************************************************************
 FUNCTION:  CopyGameMessage
******************************************************************************/
void SoMctsTreeNode::CopyGameMessage( SoGameMessage* pMessage ) 
{ 
	ASSERTF( _GameMessage.GetDataList()->GetNumEntries() == 0, ( "\nThis message already has data!" ) );
	_GameMessage.CopyMessage( *pMessage ); 
}



/******************************************************************************
 FUNCTION:  CalculateTotalSubNodeCount
******************************************************************************/
int SoMctsTreeNode::CalculateTotalSubNodeCount()
{
	int iSubNodes = _ChildNodeList.GetNumEntries();

	SoMctsTreeNodeList::Iterator Iter( &_ChildNodeList );
	while ( Iter.Next() )
	{
		SoMctsTreeNode* pTreeNode = Iter.GetNodeData();

		iSubNodes += pTreeNode->CalculateTotalSubNodeCount();
	}

	return iSubNodes;
}



/******************************************************************************
 FUNCTION:  CalculateMaxDepth
******************************************************************************/
int SoMctsTreeNode::CalculateMaxDepth() 
{ 
	int iDepthToFarthestChild = 0;	

	SoMctsTreeNodeList::Iterator Iter( &_ChildNodeList );
	while ( Iter.Next() )
	{
		SoMctsTreeNode* pTreeNode = Iter.GetNodeData();

		int iDepth = pTreeNode->CalculateMaxDepth();

		if ( iDepth > iDepthToFarthestChild )
			iDepthToFarthestChild = iDepth;

	}

	return ( 1 + iDepthToFarthestChild );
}



/******************************************************************************
 FUNCTION:  CalculateMaxBreadth
******************************************************************************/
int SoMctsTreeNode::CalculateMaxBreadth() 
{ 
	int iMaxChildBreadth = _ChildNodeList.GetNumEntries();	


	SoMctsTreeNodeList::Iterator Iter( &_ChildNodeList );
	while ( Iter.Next() )
	{
		SoMctsTreeNode* pTreeNode = Iter.GetNodeData();

		int iBreadth = pTreeNode->CalculateMaxBreadth();

		if ( iBreadth > iMaxChildBreadth )
			iMaxChildBreadth = iBreadth;

	}

	return iMaxChildBreadth;
}
