/*-----------------------------------------------------------------------------
 |  Copyright (c) 2011 Domowicz Creative Group, LLC    All Rights Reserved
 +----------------------------------------------------------------------------*/


#include "stdafx.h"

#if DCG_IS_TARGET_WIN32
	#include "direct.h"
#endif

#include <stdio.h>
#include <time.h>


#include "JpFileUtils.h"
#include "JpUtils.h"
#include "NeCsvOut.h"
#include "SoActorManager.h"
#include "SoSaveSystem.h"
#include "SoScript.h"
#include "mmgrChooser.h"


SoSaveSystem* SoSaveSystem::_pInstance = NULL;


/******************************************************************************
 FUNCTION:  Init
******************************************************************************/
void SoSaveSystem::Init()
{
	_acSavePath[ 0 ] = 0;
	_acConstructedPathAndFilename[ 0 ] = 0;

	_iSharedCsvReaderIndex = 0;
}




/******************************************************************************
 FUNCTION:  Save
******************************************************************************/
void SoSaveSystem::Save()
{
	//-----------------------
	// generate save path
	GenerateSavePath();

	// This makes sure the "save" folder exists
	SoUtils::CreateSaveDirectory();

	// and this makes the "save/SubDirectoryName" folder
	JpFileUtils::CreateDirectory( GetSavePath() ); 
	

	//-----------------------
	// saved the most recently used Dungeon
	_SharedCsvWriter.Clear();
	_SharedCsvWriter.AddNewRow();
	_SharedCsvWriter.AppendStringToLastRow( "sLastDungeonPath" );
	_SharedCsvWriter.AppendStringToLastRow( "iLastLevelId" );
	_SharedCsvWriter.AddNewRow();
	_SharedCsvWriter.AppendStringToLastRow( _acSavePath );
	_SharedCsvWriter.AppendIntToLastRow( GetCurrentLevelId() );
	_SharedCsvWriter.WriteToFile( SoUtils::AddRelativePathToSaveData( "LastDungeon.csv" ) );
	_SharedCsvWriter.Clear();


	//-----------------------
	// save the SoActors
	SoActorManager* pActorManager = SoActorManager::GetInstance();
	pActorManager->SaveToFile( this );


	//-----------------------
	// save the scripts
	SoScriptManager* pScriptManager = SoScriptManager::GetInstance();
	pScriptManager->SaveToFile( this );
}




/******************************************************************************
 FUNCTION:  LoadLevel
 NOTES:     This loads the given level.  Note that this is similar to
            LoadGame(), except it is not as comprehensive.  This can be thought
            of as a "partial load", only suitable for level switching.
******************************************************************************/
void SoSaveSystem::LoadLevel( int iLevelId )
{
	//-----------------------
	// load the path to the most recently used dungeon files
	_SharedCsvReader.Clear();
	_SharedCsvReader.ReadCsvFile( (char*)SoUtils::AddRelativePathToSaveData( "LastDungeon.csv" ) );
	_SharedCsvReader.MakeHashIndicesFromColValuesAtRow( 0 );
	char* pacSavedPath = _SharedCsvReader.GetStringByIntAndHash( 1, 0x134b3083 /* "sLastDungeonPath" */ );
	SetSavePath( pacSavedPath );


	//-----------------------
	// set the level we are going to load
	SetCurrentLevelId( iLevelId );


	//-----------------------
	// load the SoActors
	SoActorManager* pActorManager = SoActorManager::GetInstance();
	pActorManager->LoadFromFile( this );


	//-----------------------
	// load the scripts
	// Note: it could be argued that we don't need to do this on level
	// switches - only new game switches.  However, by forcing
	// a reload of the scripts on every level load, we are better
	// testing the save/load of every script.  Given how easy it
	// is to incorrectly save/load something or even forget to do it
	// all together, this extra testing is a good idea I think.
	// mdomowicz 2013_06_20
	SoScriptManager* pScriptManager = SoScriptManager::GetInstance();
	pScriptManager->LoadFromFile( this );
}



/******************************************************************************
 FUNCTION:  LoadGame
 NOTES:     This loads the last saved level.  Conceptually this does the same
            work as LoadLevel(), except: 1. the level id to load is not
            provided as a parameter, but rather it is read from the
            LastDungeon.csv.  and 2. Additional things get loaded such as
            the Hero, and so this can be thought of as a "full load".
******************************************************************************/
void SoSaveSystem::LoadGame()
{
	_SharedCsvReader.Clear();
	_SharedCsvReader.ReadCsvFile( (char*)SoUtils::AddRelativePathToSaveData( "LastDungeon.csv" ) );
	_SharedCsvReader.MakeHashIndicesFromColValuesAtRow( 0 );

	//-----------------------
	// load the path to the most recently used dungeon files
	char* pacSavedPath = _SharedCsvReader.GetStringByIntAndHash( 1, 0x134b3083 /* "sLastDungeonPath" */ );
	SetSavePath( pacSavedPath );


	//-----------------------
	// set the level id
	int iLastLevelId = _SharedCsvReader.GetIntByIntAndHash( 1, 0xe8c99777 /* "iLastLevelId" */, -1 );
	SetCurrentLevelId( iLastLevelId );


	//-----------------------
	// The hero is a special case - a hero actor must exist for the rest of
	// the dungeon load to work.
	SoActorManager* pActorManager = SoActorManager::GetInstance();
	pActorManager->LoadHeroFromFile( this );


	//-----------------------
	// load the SoActors
	pActorManager->LoadFromFile( this );


	//-----------------------
	// load the scripts
	SoScriptManager* pScriptManager = SoScriptManager::GetInstance();
	pScriptManager->LoadFromFile( this );
}



/******************************************************************************
 FUNCTION:  GenerateSavePath
******************************************************************************/
void SoSaveSystem::GenerateSavePath()
{
	sprintf( _acSavePath, SoUtils::AddRelativePathToSaveData( "%s" ), "TempPathName" );
}



/******************************************************************************
 FUNCTION:  CreateInstance
******************************************************************************/
void SoSaveSystem::CreateInstance()
{
	ASSERTF( _pInstance == NULL, ( "\nCan't create more than one instance.  Not allowed." ) );

	if ( !_pInstance )
		_pInstance = new SoSaveSystem;
}



/******************************************************************************
 FUNCTION:  DeleteInstance
******************************************************************************/
void SoSaveSystem::DeleteInstance()
{
	if ( _pInstance )
	{
		delete _pInstance;
		_pInstance = NULL;
	}
}



/******************************************************************************
 FUNCTION:  ConstructPathAndFileName
******************************************************************************/
char* SoSaveSystem::ConstructPathAndFileName( char* pacFileNameNoExt, int iLevelId )
{
	if ( iLevelId < 0 )
		iLevelId = _iCurrentLevelId;

	sprintf( _acConstructedPathAndFilename, "%s/%i_%s.csv", _acSavePath, iLevelId, pacFileNameNoExt );
	return _acConstructedPathAndFilename;
}



/******************************************************************************
 FUNCTION:  DoesSaveDataExistForLastDungeon
******************************************************************************/
bool SoSaveSystem::DoesSaveDataExistForLastDungeon()
{
	if ( JpFileUtils::FileExists( (char*)SoUtils::AddRelativePathToSaveData( "LastDungeon.csv" ) ) )
	{
		//-----------------------
		// get the save game path from the LastDungeon.csv, the top-level file.
		_SharedCsvReader.Clear();
		_SharedCsvReader.ReadCsvFile( (char*)SoUtils::AddRelativePathToSaveData( "LastDungeon.csv" ) );
		_SharedCsvReader.MakeHashIndicesFromColValuesAtRow( 0 );
		char* pacSavedPath = _SharedCsvReader.GetStringByIntAndHash( 1, 0x134b3083 /* "sLastDungeonPath" */ );
		SetSavePath( pacSavedPath );

		//-----------------------
		// see if the SoGameDungeon.csv file exists.  If it does, then we can
		// assume that save data exists.
		return JpFileUtils::FileExists( ConstructPathAndFileName( "SoGameDungeon", 0 ) );
	}

	return false;
}
